self.__BUILD_MANIFEST=function(s,c,a,e,t,u,n,b,o,i,r,j,k){return{__rewrites:{afterFiles:[],beforeFiles:[],fallback:[]},"/":["static/chunks/pages/index-bbfee9b059a805b6.js"],"/_error":["static/chunks/pages/_error-6ae619510b1539d6.js"],"/clusters":["static/chunks/pages/clusters-4d7539b652042b77.js"],"/clusters/[cluster]":[s,n,a,b,t,c,e,o,i,u,r,j,"static/chunks/871-9ae3562e676ccff7.js","static/chunks/pages/clusters/[cluster]-e2de6688998a4f8a.js"],"/clusters/[cluster]/[job]":[s,c,k,"static/chunks/pages/clusters/[cluster]/[job]-b3b89666d7a9b76a.js"],"/config":["static/chunks/pages/config-84a73b45c1311178.js"],"/infra":["static/chunks/pages/infra-b6d62a628e75f76e.js"],"/infra/[context]":["static/chunks/pages/infra/[context]-21d2fbd90066d95d.js"],"/jobs":["static/chunks/pages/jobs-714e53cea6a69592.js"],"/jobs/pools/[pool]":[s,a,t,"static/chunks/821-781977113952dbf3.js",c,e,u,"static/chunks/pages/jobs/pools/[pool]-2dab3a17dbce7912.js"],"/jobs/[job]":[s,a,t,c,e,k,"static/chunks/pages/jobs/[job]-a4989f80b40bac5a.js"],"/plugins/[...slug]":[s,"static/chunks/pages/plugins/[...slug]-b4e71b9f3d1d4985.js"],"/users":["static/chunks/pages/users-f90d17bb464b80cb.js"],"/volumes":["static/chunks/pages/volumes-ea066ac89a0c151c.js"],"/workspace/new":["static/chunks/pages/workspace/new-ea137ebf17f314b1.js"],"/workspaces":["static/chunks/pages/workspaces-01d3738858512c7c.js"],"/workspaces/[name]":[s,n,a,b,c,e,o,i,u,r,j,"static/chunks/50-2b1a60bdce926b03.js","static/chunks/pages/workspaces/[name]-6d5c35034ee45ecc.js"],sortedPages:["/","/_app","/_error","/clusters","/clusters/[cluster]","/clusters/[cluster]/[job]","/config","/infra","/infra/[context]","/jobs","/jobs/pools/[pool]","/jobs/[job]","/plugins/[...slug]","/users","/volumes","/workspace/new","/workspaces","/workspaces/[name]"]}}("static/chunks/739-ce131cf092b994c8.js","static/chunks/850-0b979d5ca596d640.js","static/chunks/798-c0525dc3f21e488d.js","static/chunks/969-f964172cb17e6c4a.js","static/chunks/754-cfc5d4ad1b843d29.js","static/chunks/248-886c8e51053f6e54.js","static/chunks/82-c5786ea3f4496d60.js","static/chunks/835-6242063df6081b97.js","static/chunks/990-91b4b5bea46084f1.js","static/chunks/56-9c314358cd0df8aa.js","static/chunks/353-d57fd280bfc5e7c9.js","static/chunks/109-0bc891b5e95953ad.js","static/chunks/83-ad357c443c0779c4.js"),self.__BUILD_MANIFEST_CB&&self.__BUILD_MANIFEST_CB();