(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[490],{94483:function(n,e,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/config",function(){return t(99649)}])},99649:function(n,e,t){"use strict";t.r(e),t.d(e,{default:function(){return c}});var r=t(85893);t(67294);var i=t(9008),o=t.n(i),u=t(5152);let s=t.n(u)()(()=>Promise.all([t.e(739),t.e(798),t.e(990),t.e(401)]).then(t.bind(t,90699)).then(n=>n.Config),{loadableGenerated:{webpack:()=>[90699]},ssr:!1});function c(){return(0,r.jsxs)(r.Fragment,{children:[(0,r.jsx)(o(),{children:(0,r.jsx)("title",{children:"SkyPilot API Server | SkyPilot Dashboard"})}),(0,r.jsx)(s,{})]})}},9008:function(n,e,t){n.exports=t(37219)}},function(n){n.O(0,[888,774,179],function(){return n(n.s=94483)}),_N_E=n.O()}]);