(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[590],{92592:function(e,n,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/workspaces",function(){return r(4980)}])},4980:function(e,n,r){"use strict";r.r(n),r.d(n,{default:function(){return u}});var t=r(85893);r(67294);var s=r(9008),c=r.n(s),i=r(5152);let o=r.n(i)()(()=>Promise.all([r.e(739),r.e(82),r.e(798),r.e(850),r.e(969),r.e(990),r.e(56),r.e(248),r.e(109),r.e(981)]).then(r.bind(r,82077)).then(e=>e.Workspaces),{loadableGenerated:{webpack:()=>[82077]},ssr:!1});function u(){return(0,t.jsxs)(t.Fragment,{children:[(0,t.jsx)(c(),{children:(0,t.jsx)("title",{children:"Workspaces | SkyPilot Dashboard"})}),(0,t.jsx)(o,{})]})}},9008:function(e,n,r){e.exports=r(37219)}},function(e){e.O(0,[888,774,179],function(){return e(e.s=92592)}),_N_E=e.O()}]);