(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[142],{41446:function(e,n,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/jobs",function(){return t(48929)}])},48929:function(e,n,t){"use strict";t.r(n),t.d(n,{default:function(){return a}});var r=t(85893);t(67294);var s=t(9008),i=t.n(s),o=t(5152);let u=t.n(o)()(()=>Promise.all([t.e(739),t.e(82),t.e(850),t.e(969),t.e(990),t.e(56),t.e(248),t.e(109),t.e(555)]).then(t.bind(t,2109)).then(e=>e.ManagedJobs),{loadableGenerated:{webpack:()=>[2109]},ssr:!1});function a(){return(0,r.jsxs)(r.Fragment,{children:[(0,r.jsx)(i(),{children:(0,r.jsx)("title",{children:"Managed Jobs | SkyPilot Dashboard"})}),(0,r.jsx)(u,{})]})}},9008:function(e,n,t){e.exports=t(37219)}},function(e){e.O(0,[888,774,179],function(){return e(e.s=41446)}),_N_E=e.O()}]);