(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[963],{35113:function(n,e,r){(window.__NEXT_P=window.__NEXT_P||[]).push(["/workspace/new",function(){return r(4250)}])},4250:function(n,e,r){"use strict";r.r(e),r.d(e,{default:function(){return u}});var t=r(85893);r(67294);var s=r(9008),c=r.n(s),i=r(5152);let o=r.n(i)()(()=>Promise.all([r.e(990),r.e(785)]).then(r.bind(r,33785)).then(n=>n.NewWorkspace),{loadableGenerated:{webpack:()=>[33785]},ssr:!1});function u(){return(0,t.jsxs)(t.Fragment,{children:[(0,t.jsx)(c(),{children:(0,t.jsx)("title",{children:"New Workspace | SkyPilot Dashboard"})}),(0,t.jsx)(o,{})]})}},9008:function(n,e,r){n.exports=r(37219)}},function(n){n.O(0,[888,774,179],function(){return n(n.s=35113)}),_N_E=n.O()}]);