(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[810],{45833:function(e,n,t){(window.__NEXT_P=window.__NEXT_P||[]).push(["/clusters",function(){return t(23450)}])},23450:function(e,n,t){"use strict";t.r(n),t.d(n,{default:function(){return l}});var r=t(85893);t(67294);var s=t(9008),u=t.n(s),i=t(5152);let c=t.n(i)()(()=>Promise.all([t.e(739),t.e(82),t.e(798),t.e(835),t.e(850),t.e(969),t.e(990),t.e(56),t.e(248),t.e(353),t.e(871),t.e(540)]).then(t.bind(t,39037)).then(e=>e.Clusters),{loadableGenerated:{webpack:()=>[39037]},ssr:!1});function l(){return(0,r.jsxs)(r.Fragment,{children:[(0,r.jsx)(u(),{children:(0,r.jsx)("title",{children:"Clusters | SkyPilot Dashboard"})}),(0,r.jsx)(c,{})]})}},9008:function(e,n,t){e.exports=t(37219)}},function(e){e.O(0,[888,774,179],function(){return e(e.s=45833)}),_N_E=e.O()}]);