(self.webpackChunk_N_E=self.webpackChunk_N_E||[]).push([[888],{23541:function(e,t,n){"use strict";n.d(t,{Ry:function(){return u}});var r=new WeakMap,o=new WeakMap,i={},l=0,a=function(e){return e&&(e.host||a(e.parentNode))},s=function(e,t,n,s){var u=(Array.isArray(e)?e:[e]).map(function(e){if(t.contains(e))return e;var n=a(e);return n&&t.contains(n)?n:(console.error("aria-hidden",e,"in not contained inside",t,". Doing nothing"),null)}).filter(function(e){return!!e});i[n]||(i[n]=new WeakMap);var c=i[n],d=[],f=new Set,p=new Set(u),h=function(e){!e||f.has(e)||(f.add(e),h(e.parentNode))};u.forEach(h);var m=function(e){!e||p.has(e)||Array.prototype.forEach.call(e.children,function(e){if(f.has(e))m(e);else try{var t=e.getAttribute(s),i=null!==t&&"false"!==t,l=(r.get(e)||0)+1,a=(c.get(e)||0)+1;r.set(e,l),c.set(e,a),d.push(e),1===l&&i&&o.set(e,!0),1===a&&e.setAttribute(n,"true"),i||e.setAttribute(s,"true")}catch(t){console.error("aria-hidden: cannot operate on ",e,t)}})};return m(t),f.clear(),l++,function(){d.forEach(function(e){var t=r.get(e)-1,i=c.get(e)-1;r.set(e,t),c.set(e,i),t||(o.has(e)||e.removeAttribute(s),o.delete(e)),i||e.removeAttribute(n)}),--l||(r=new WeakMap,r=new WeakMap,o=new WeakMap,i={})}},u=function(e,t,n){void 0===n&&(n="data-aria-hidden");var r=Array.from(Array.isArray(e)?e:[e]),o=t||("undefined"==typeof document?null:(Array.isArray(e)?e[0]:e).ownerDocument.body);return o?(r.push.apply(r,Array.from(o.querySelectorAll("[aria-live], script"))),s(r,o,n,"aria-hidden")):function(){return null}}},83454:function(e,t,n){"use strict";var r,o;e.exports=(null==(r=n.g.process)?void 0:r.env)&&"object"==typeof(null==(o=n.g.process)?void 0:o.env)?n.g.process:n(77663)},91118:function(e,t,n){(window.__NEXT_P=window.__NEXT_P||[]).push(["/_app",function(){return n(32586)}])},60998:function(e,t,n){"use strict";n.d(t,{Z:function(){return s}});var r=n(67294);/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let o=e=>e.replace(/([a-z0-9])([A-Z])/g,"$1-$2").toLowerCase(),i=function(){for(var e=arguments.length,t=Array(e),n=0;n<e;n++)t[n]=arguments[n];return t.filter((e,t,n)=>!!e&&n.indexOf(e)===t).join(" ")};/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */var l={xmlns:"http://www.w3.org/2000/svg",width:24,height:24,viewBox:"0 0 24 24",fill:"none",stroke:"currentColor",strokeWidth:2,strokeLinecap:"round",strokeLinejoin:"round"};/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let a=(0,r.forwardRef)((e,t)=>{let{color:n="currentColor",size:o=24,strokeWidth:a=2,absoluteStrokeWidth:s,className:u="",children:c,iconNode:d,...f}=e;return(0,r.createElement)("svg",{ref:t,...l,width:o,height:o,stroke:n,strokeWidth:s?24*Number(a)/Number(o):a,className:i("lucide",u),...f},[...d.map(e=>{let[t,n]=e;return(0,r.createElement)(t,n)}),...Array.isArray(c)?c:[c]])}),s=(e,t)=>{let n=(0,r.forwardRef)((n,l)=>{let{className:s,...u}=n;return(0,r.createElement)(a,{ref:l,iconNode:t,className:i("lucide-".concat(o(e)),s),...u})});return n.displayName="".concat(e),n}},70282:function(e,t,n){"use strict";n.d(t,{Z:function(){return r}});/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let r=(0,n(60998).Z)("Check",[["path",{d:"M20 6 9 17l-5-5",key:"1gmf2c"}]])},45895:function(e,t,n){"use strict";n.d(t,{Z:function(){return r}});/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let r=(0,n(60998).Z)("ChevronDown",[["path",{d:"m6 9 6 6 6-6",key:"qrunsl"}]])},56101:function(e,t){"use strict";var n,r,o,i;Object.defineProperty(t,"__esModule",{value:!0}),function(e,t){for(var n in t)Object.defineProperty(e,n,{enumerable:!0,get:t[n]})}(t,{ACTION_FAST_REFRESH:function(){return d},ACTION_NAVIGATE:function(){return a},ACTION_PREFETCH:function(){return c},ACTION_REFRESH:function(){return l},ACTION_RESTORE:function(){return s},ACTION_SERVER_ACTION:function(){return f},ACTION_SERVER_PATCH:function(){return u},PrefetchCacheEntryStatus:function(){return r},PrefetchKind:function(){return n},isThenable:function(){return p}});let l="refresh",a="navigate",s="restore",u="server-patch",c="prefetch",d="fast-refresh",f="server-action";function p(e){return e&&("object"==typeof e||"function"==typeof e)&&"function"==typeof e.then}(o=n||(n={})).AUTO="auto",o.FULL="full",o.TEMPORARY="temporary",(i=r||(r={})).fresh="fresh",i.reusable="reusable",i.expired="expired",i.stale="stale",("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},27670:function(e,t,n){"use strict";function r(e,t,n,r){return!1}Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"getDomainLocale",{enumerable:!0,get:function(){return r}}),n(51297),("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},24116:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"default",{enumerable:!0,get:function(){return w}});let r=n(10260),o=n(85893),i=r._(n(67294)),l=n(69975),a=n(92712),s=n(28547),u=n(84350),c=n(38109),d=n(54494),f=n(55716),p=n(388),h=n(27670),m=n(6220),g=n(56101),b=new Set;function v(e,t,n,r,o,i){if(i||(0,a.isLocalURL)(t)){if(!r.bypassPrefetchedCheck){let o=t+"%"+n+"%"+(void 0!==r.locale?r.locale:"locale"in e?e.locale:void 0);if(b.has(o))return;b.add(o)}(async()=>i?e.prefetch(t,o):e.prefetch(t,n,r))().catch(e=>{})}}function y(e){return"string"==typeof e?e:(0,s.formatUrl)(e)}let w=i.default.forwardRef(function(e,t){let n,r;let{href:s,as:b,children:w,prefetch:x=null,passHref:E,replace:S,shallow:k,scroll:T,locale:C,onClick:P,onMouseEnter:M,onTouchStart:R,legacyBehavior:j=!1,...O}=e;n=w,j&&("string"==typeof n||"number"==typeof n)&&(n=(0,o.jsx)("a",{children:n}));let L=i.default.useContext(d.RouterContext),A=i.default.useContext(f.AppRouterContext),_=null!=L?L:A,N=!L,$=!1!==x,I=null===x?g.PrefetchKind.AUTO:g.PrefetchKind.FULL,{href:D,as:W}=i.default.useMemo(()=>{if(!L){let e=y(s);return{href:e,as:b?y(b):e}}let[e,t]=(0,l.resolveHref)(L,s,!0);return{href:e,as:b?(0,l.resolveHref)(L,b):t||e}},[L,s,b]),H=i.default.useRef(D),F=i.default.useRef(W);j&&(r=i.default.Children.only(n));let z=j?r&&"object"==typeof r&&r.ref:t,[B,q,V]=(0,p.useIntersection)({rootMargin:"200px"}),U=i.default.useCallback(e=>{(F.current!==W||H.current!==D)&&(V(),F.current=W,H.current=D),B(e),z&&("function"==typeof z?z(e):"object"==typeof z&&(z.current=e))},[W,z,D,V,B]);i.default.useEffect(()=>{_&&q&&$&&v(_,D,W,{locale:C},{kind:I},N)},[W,D,q,C,$,null==L?void 0:L.locale,_,N,I]);let Y={ref:U,onClick(e){j||"function"!=typeof P||P(e),j&&r.props&&"function"==typeof r.props.onClick&&r.props.onClick(e),_&&!e.defaultPrevented&&function(e,t,n,r,o,l,s,u,c){let{nodeName:d}=e.currentTarget;if("A"===d.toUpperCase()&&(function(e){let t=e.currentTarget.getAttribute("target");return t&&"_self"!==t||e.metaKey||e.ctrlKey||e.shiftKey||e.altKey||e.nativeEvent&&2===e.nativeEvent.which}(e)||!c&&!(0,a.isLocalURL)(n)))return;e.preventDefault();let f=()=>{let e=null==s||s;"beforePopState"in t?t[o?"replace":"push"](n,r,{shallow:l,locale:u,scroll:e}):t[o?"replace":"push"](r||n,{scroll:e})};c?i.default.startTransition(f):f()}(e,_,D,W,S,k,T,C,N)},onMouseEnter(e){j||"function"!=typeof M||M(e),j&&r.props&&"function"==typeof r.props.onMouseEnter&&r.props.onMouseEnter(e),_&&($||!N)&&v(_,D,W,{locale:C,priority:!0,bypassPrefetchedCheck:!0},{kind:I},N)},onTouchStart:function(e){j||"function"!=typeof R||R(e),j&&r.props&&"function"==typeof r.props.onTouchStart&&r.props.onTouchStart(e),_&&($||!N)&&v(_,D,W,{locale:C,priority:!0,bypassPrefetchedCheck:!0},{kind:I},N)}};if((0,u.isAbsoluteUrl)(W))Y.href=W;else if(!j||E||"a"===r.type&&!("href"in r.props)){let e=void 0!==C?C:null==L?void 0:L.locale,t=(null==L?void 0:L.isLocaleDomain)&&(0,h.getDomainLocale)(W,e,null==L?void 0:L.locales,null==L?void 0:L.domainLocales);Y.href=t||(0,m.addBasePath)((0,c.addLocale)(W,e,null==L?void 0:L.defaultLocale))}return j?i.default.cloneElement(r,Y):(0,o.jsx)("a",{...O,...Y,children:n})});("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},388:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"useIntersection",{enumerable:!0,get:function(){return s}});let r=n(67294),o=n(40460),i="function"==typeof IntersectionObserver,l=new Map,a=[];function s(e){let{rootRef:t,rootMargin:n,disabled:s}=e,u=s||!i,[c,d]=(0,r.useState)(!1),f=(0,r.useRef)(null),p=(0,r.useCallback)(e=>{f.current=e},[]);return(0,r.useEffect)(()=>{if(i){if(u||c)return;let e=f.current;if(e&&e.tagName)return function(e,t,n){let{id:r,observer:o,elements:i}=function(e){let t;let n={root:e.root||null,margin:e.rootMargin||""},r=a.find(e=>e.root===n.root&&e.margin===n.margin);if(r&&(t=l.get(r)))return t;let o=new Map;return t={id:n,observer:new IntersectionObserver(e=>{e.forEach(e=>{let t=o.get(e.target),n=e.isIntersecting||e.intersectionRatio>0;t&&n&&t(n)})},e),elements:o},a.push(n),l.set(n,t),t}(n);return i.set(e,t),o.observe(e),function(){if(i.delete(e),o.unobserve(e),0===i.size){o.disconnect(),l.delete(r);let e=a.findIndex(e=>e.root===r.root&&e.margin===r.margin);e>-1&&a.splice(e,1)}}}(e,e=>e&&d(e),{root:null==t?void 0:t.current,rootMargin:n})}else if(!c){let e=(0,o.requestIdleCallback)(()=>d(!0));return()=>(0,o.cancelIdleCallback)(e)}},[u,n,t,c,f.current]),[p,c,(0,r.useCallback)(()=>{d(!1)},[])]}("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},19606:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),function(e,t){for(var n in t)Object.defineProperty(e,n,{enumerable:!0,get:t[n]})}(t,{default:function(){return a},noSSR:function(){return l}});let r=n(10260);n(85893),n(67294);let o=r._(n(6119));function i(e){return{default:(null==e?void 0:e.default)||e}}function l(e,t){return delete t.webpack,delete t.modules,e(t)}function a(e,t){let n=o.default,r={loading:e=>{let{error:t,isLoading:n,pastDelay:r}=e;return null}};e instanceof Promise?r.loader=()=>e:"function"==typeof e?r.loader=e:"object"==typeof e&&(r={...r,...e});let a=(r={...r,...t}).loader;return(r.loadableGenerated&&(r={...r,...r.loadableGenerated},delete r.loadableGenerated),"boolean"!=typeof r.ssr||r.ssr)?n({...r,loader:()=>null!=a?a().then(i):Promise.resolve(i(()=>null))}):(delete r.webpack,delete r.modules,l(n,r))}("function"==typeof t.default||"object"==typeof t.default&&null!==t.default)&&void 0===t.default.__esModule&&(Object.defineProperty(t.default,"__esModule",{value:!0}),Object.assign(t.default,t),e.exports=t.default)},16725:function(e,t,n){"use strict";Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"LoadableContext",{enumerable:!0,get:function(){return r}});let r=n(10260)._(n(67294)).default.createContext(null)},6119:function(e,t,n){"use strict";/**
@copyright (c) 2017-present James Kyle <me@thejameskyle.com>
 MIT License
 Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:
 The above copyright notice and this permission notice shall be
included in all copies or substantial portions of the Software.
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE
*/Object.defineProperty(t,"__esModule",{value:!0}),Object.defineProperty(t,"default",{enumerable:!0,get:function(){return f}});let r=n(10260)._(n(67294)),o=n(16725),i=[],l=[],a=!1;function s(e){let t=e(),n={loading:!0,loaded:null,error:null};return n.promise=t.then(e=>(n.loading=!1,n.loaded=e,e)).catch(e=>{throw n.loading=!1,n.error=e,e}),n}class u{promise(){return this._res.promise}retry(){this._clearTimeouts(),this._res=this._loadFn(this._opts.loader),this._state={pastDelay:!1,timedOut:!1};let{_res:e,_opts:t}=this;e.loading&&("number"==typeof t.delay&&(0===t.delay?this._state.pastDelay=!0:this._delay=setTimeout(()=>{this._update({pastDelay:!0})},t.delay)),"number"==typeof t.timeout&&(this._timeout=setTimeout(()=>{this._update({timedOut:!0})},t.timeout))),this._res.promise.then(()=>{this._update({}),this._clearTimeouts()}).catch(e=>{this._update({}),this._clearTimeouts()}),this._update({})}_update(e){this._state={...this._state,error:this._res.error,loaded:this._res.loaded,loading:this._res.loading,...e},this._callbacks.forEach(e=>e())}_clearTimeouts(){clearTimeout(this._delay),clearTimeout(this._timeout)}getCurrentValue(){return this._state}subscribe(e){return this._callbacks.add(e),()=>{this._callbacks.delete(e)}}constructor(e,t){this._loadFn=e,this._opts=t,this._callbacks=new Set,this._delay=null,this._timeout=null,this.retry()}}function c(e){return function(e,t){let n=Object.assign({loader:null,loading:null,delay:200,timeout:null,webpack:null,modules:null},t),i=null;function s(){if(!i){let t=new u(e,n);i={getCurrentValue:t.getCurrentValue.bind(t),subscribe:t.subscribe.bind(t),retry:t.retry.bind(t),promise:t.promise.bind(t)}}return i.promise()}if(!a){let e=n.webpack?n.webpack():n.modules;e&&l.push(t=>{for(let n of e)if(t.includes(n))return s()})}function c(e,t){!function(){s();let e=r.default.useContext(o.LoadableContext);e&&Array.isArray(n.modules)&&n.modules.forEach(t=>{e(t)})}();let l=r.default.useSyncExternalStore(i.subscribe,i.getCurrentValue,i.getCurrentValue);return r.default.useImperativeHandle(t,()=>({retry:i.retry}),[]),r.default.useMemo(()=>{var t;return l.loading||l.error?r.default.createElement(n.loading,{isLoading:l.loading,pastDelay:l.pastDelay,timedOut:l.timedOut,error:l.error,retry:i.retry}):l.loaded?r.default.createElement((t=l.loaded)&&t.default?t.default:t,e):null},[e,l])}return c.preload=()=>s(),c.displayName="LoadableComponent",r.default.forwardRef(c)}(s,e)}function d(e,t){let n=[];for(;e.length;){let r=e.pop();n.push(r(t))}return Promise.all(n).then(()=>{if(e.length)return d(e,t)})}c.preloadAll=()=>new Promise((e,t)=>{d(i).then(e,t)}),c.preloadReady=e=>(void 0===e&&(e=[]),new Promise(t=>{let n=()=>(a=!0,t());d(l,e).then(n,n)})),window.__NEXT_PRELOADREADY=c.preloadReady;let f=c},91623:function(e,t,n){"use strict";n.d(t,{$h:function(){return m},on:function(){return h},Bx:function(){return g},pU:function(){return d}});var r=n(85893),o=n(67294),i=n(60998);/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let l=(0,i.Z)("Bell",[["path",{d:"M6 8a6 6 0 0 1 12 0c0 7 3 9 3 9H3s3-2 3-9",key:"1qo2s2"}],["path",{d:"M10.3 21a1.94 1.94 0 0 0 3.4 0",key:"qgo35s"}]]),a=(0,i.Z)("CircleArrowUp",[["circle",{cx:"12",cy:"12",r:"10",key:"1mglay"}],["path",{d:"m16 12-4-4-4 4",key:"177agl"}],["path",{d:"M12 16V8",key:"1sbj14"}]]);var s=n(470),u=n(47145);let c=(0,o.createContext)({version:null,latestVersion:null,commit:null,plugins:[]});function d(e){let{children:t}=e,[n,i]=(0,o.useState)(null),[l,a]=(0,o.useState)(null),[s,d]=(0,o.useState)(null),[f,p]=(0,o.useState)([]),h=async()=>{let[e,t]=await Promise.all([u.x.get("/api/health"),u.x.get("/api/plugins")]);if(e.ok){let t=await e.json();t.version&&i(t.version),t.commit&&d(t.commit),t.latest_version&&a(t.latest_version)}else console.error("API request /api/health failed with status ".concat(e.status));if(t.ok){let e=await t.json();e.plugins&&e.plugins.length>0&&p(e.plugins)}else console.error("API request /api/plugins failed with status ".concat(t.status))};return(0,o.useEffect)(()=>{h()},[]),(0,r.jsx)(c.Provider,{value:{version:n,latestVersion:l,commit:s,plugins:f},children:t})}function f(){return(0,o.useContext)(c)}function p(e){let{children:t,version:n,latestVersion:o,commit:i,plugins:l,showUpdateInfo:a=!0,showCommit:u=!0}=e,c=(0,r.jsxs)("div",{className:"flex flex-col gap-0.5",children:[a&&o&&(0,r.jsxs)("div",{className:"mb-1",children:[(0,r.jsx)("div",{className:"font-bold",children:"Update Available"}),(0,r.jsxs)("div",{children:["Current version: ",n]}),(0,r.jsxs)("div",{children:["New version available: ",o]})]}),u&&i&&(0,r.jsxs)("div",{children:[l.length>0?"Core commit":"Commit",": ",i]}),l.map((e,t)=>{let n=e.name||"Unknown Plugin",o=[];return e.version&&o.push(e.version),u&&e.commit&&o.push(e.commit),o.length>0?(0,r.jsxs)("div",{children:[n,": ",o.join(" - ")]},t):null}),!i&&0===l.length&&(!o||!a)&&(0,r.jsx)("div",{children:"Version information not available"})]});return(0,r.jsx)(s.Md,{content:c,className:"text-sm text-muted-foreground",children:t})}function h(){let{version:e,latestVersion:t,commit:n,plugins:o}=f();return e&&t?(0,r.jsx)(p,{version:e,latestVersion:t,commit:n,plugins:o,showCommit:!1,children:(0,r.jsx)("div",{className:"inline-flex items-center justify-center transition-colors duration-150 cursor-help",children:(0,r.jsx)("div",{className:"p-2 rounded-full text-gray-600 hover:bg-gray-100 hover:text-blue-600",children:(0,r.jsx)(l,{className:"w-5 h-5"})})})}):null}function m(){let{latestVersion:e}=f();return e?(0,r.jsxs)("div",{className:"flex items-center mr-4 text-amber-600 animate-pulse",children:[(0,r.jsx)(a,{className:"w-4 h-4 mr-1.5"}),(0,r.jsxs)("span",{className:"text-sm font-medium",children:["New version available: ",e]})]}):null}function g(){let{version:e,latestVersion:t,commit:n,plugins:o}=f();return e?(0,r.jsx)(p,{version:e,latestVersion:t,commit:n,plugins:o,showUpdateInfo:!1,children:(0,r.jsx)("div",{className:"inline-flex items-center justify-center transition-colors duration-150 cursor-help",children:(0,r.jsxs)("div",{className:"text-sm text-gray-500 border-b border-dotted border-gray-400 hover:text-blue-600 hover:border-blue-600",children:["Version: ",e]})})}):null}},62935:function(e,t,n){"use strict";n.d(t,{Ph:function(){return t4},Bw:function(){return t8},Ql:function(){return t7},i4:function(){return t6},ki:function(){return t5}});var r=n(85893),o=n(67294),i=n(73935);function l(e,[t,n]){return Math.min(n,Math.max(t,e))}var a=n(36206),s=n(25360),u=n(28771),c=n(88426),d=o.createContext(void 0),f=n(46063),p=n(27552),h=n(95420),m=n(91276);let g=["top","right","bottom","left"],b=Math.min,v=Math.max,y=Math.round,w=Math.floor,x=e=>({x:e,y:e}),E={left:"right",right:"left",bottom:"top",top:"bottom"},S={start:"end",end:"start"};function k(e,t){return"function"==typeof e?e(t):e}function T(e){return e.split("-")[0]}function C(e){return e.split("-")[1]}function P(e){return"x"===e?"y":"x"}function M(e){return"y"===e?"height":"width"}function R(e){return["top","bottom"].includes(T(e))?"y":"x"}function j(e){return e.replace(/start|end/g,e=>S[e])}function O(e){return e.replace(/left|right|bottom|top/g,e=>E[e])}function L(e){return"number"!=typeof e?{top:0,right:0,bottom:0,left:0,...e}:{top:e,right:e,bottom:e,left:e}}function A(e){let{x:t,y:n,width:r,height:o}=e;return{width:r,height:o,top:n,left:t,right:t+r,bottom:n+o,x:t,y:n}}function _(e,t,n){let r,{reference:o,floating:i}=e,l=R(t),a=P(R(t)),s=M(a),u=T(t),c="y"===l,d=o.x+o.width/2-i.width/2,f=o.y+o.height/2-i.height/2,p=o[s]/2-i[s]/2;switch(u){case"top":r={x:d,y:o.y-i.height};break;case"bottom":r={x:d,y:o.y+o.height};break;case"right":r={x:o.x+o.width,y:f};break;case"left":r={x:o.x-i.width,y:f};break;default:r={x:o.x,y:o.y}}switch(C(t)){case"start":r[a]-=p*(n&&c?-1:1);break;case"end":r[a]+=p*(n&&c?-1:1)}return r}let N=async(e,t,n)=>{let{placement:r="bottom",strategy:o="absolute",middleware:i=[],platform:l}=n,a=i.filter(Boolean),s=await (null==l.isRTL?void 0:l.isRTL(t)),u=await l.getElementRects({reference:e,floating:t,strategy:o}),{x:c,y:d}=_(u,r,s),f=r,p={},h=0;for(let n=0;n<a.length;n++){let{name:i,fn:m}=a[n],{x:g,y:b,data:v,reset:y}=await m({x:c,y:d,initialPlacement:r,placement:f,strategy:o,middlewareData:p,rects:u,platform:l,elements:{reference:e,floating:t}});c=null!=g?g:c,d=null!=b?b:d,p={...p,[i]:{...p[i],...v}},y&&h<=50&&(h++,"object"==typeof y&&(y.placement&&(f=y.placement),y.rects&&(u=!0===y.rects?await l.getElementRects({reference:e,floating:t,strategy:o}):y.rects),{x:c,y:d}=_(u,f,s)),n=-1)}return{x:c,y:d,placement:f,strategy:o,middlewareData:p}};async function $(e,t){var n;void 0===t&&(t={});let{x:r,y:o,platform:i,rects:l,elements:a,strategy:s}=e,{boundary:u="clippingAncestors",rootBoundary:c="viewport",elementContext:d="floating",altBoundary:f=!1,padding:p=0}=k(t,e),h=L(p),m=a[f?"floating"===d?"reference":"floating":d],g=A(await i.getClippingRect({element:null==(n=await (null==i.isElement?void 0:i.isElement(m)))||n?m:m.contextElement||await (null==i.getDocumentElement?void 0:i.getDocumentElement(a.floating)),boundary:u,rootBoundary:c,strategy:s})),b="floating"===d?{x:r,y:o,width:l.floating.width,height:l.floating.height}:l.reference,v=await (null==i.getOffsetParent?void 0:i.getOffsetParent(a.floating)),y=await (null==i.isElement?void 0:i.isElement(v))&&await (null==i.getScale?void 0:i.getScale(v))||{x:1,y:1},w=A(i.convertOffsetParentRelativeRectToViewportRelativeRect?await i.convertOffsetParentRelativeRectToViewportRelativeRect({elements:a,rect:b,offsetParent:v,strategy:s}):b);return{top:(g.top-w.top+h.top)/y.y,bottom:(w.bottom-g.bottom+h.bottom)/y.y,left:(g.left-w.left+h.left)/y.x,right:(w.right-g.right+h.right)/y.x}}function I(e,t){return{top:e.top-t.height,right:e.right-t.width,bottom:e.bottom-t.height,left:e.left-t.width}}function D(e){return g.some(t=>e[t]>=0)}async function W(e,t){let{placement:n,platform:r,elements:o}=e,i=await (null==r.isRTL?void 0:r.isRTL(o.floating)),l=T(n),a=C(n),s="y"===R(n),u=["left","top"].includes(l)?-1:1,c=i&&s?-1:1,d=k(t,e),{mainAxis:f,crossAxis:p,alignmentAxis:h}="number"==typeof d?{mainAxis:d,crossAxis:0,alignmentAxis:null}:{mainAxis:d.mainAxis||0,crossAxis:d.crossAxis||0,alignmentAxis:d.alignmentAxis};return a&&"number"==typeof h&&(p="end"===a?-1*h:h),s?{x:p*c,y:f*u}:{x:f*u,y:p*c}}function H(){return"undefined"!=typeof window}function F(e){return q(e)?(e.nodeName||"").toLowerCase():"#document"}function z(e){var t;return(null==e||null==(t=e.ownerDocument)?void 0:t.defaultView)||window}function B(e){var t;return null==(t=(q(e)?e.ownerDocument:e.document)||window.document)?void 0:t.documentElement}function q(e){return!!H()&&(e instanceof Node||e instanceof z(e).Node)}function V(e){return!!H()&&(e instanceof Element||e instanceof z(e).Element)}function U(e){return!!H()&&(e instanceof HTMLElement||e instanceof z(e).HTMLElement)}function Y(e){return!!H()&&"undefined"!=typeof ShadowRoot&&(e instanceof ShadowRoot||e instanceof z(e).ShadowRoot)}function G(e){let{overflow:t,overflowX:n,overflowY:r,display:o}=Q(e);return/auto|scroll|overlay|hidden|clip/.test(t+r+n)&&!["inline","contents"].includes(o)}function X(e){return[":popover-open",":modal"].some(t=>{try{return e.matches(t)}catch(e){return!1}})}function K(e){let t=Z(),n=V(e)?Q(e):e;return["transform","translate","scale","rotate","perspective"].some(e=>!!n[e]&&"none"!==n[e])||!!n.containerType&&"normal"!==n.containerType||!t&&!!n.backdropFilter&&"none"!==n.backdropFilter||!t&&!!n.filter&&"none"!==n.filter||["transform","translate","scale","rotate","perspective","filter"].some(e=>(n.willChange||"").includes(e))||["paint","layout","strict","content"].some(e=>(n.contain||"").includes(e))}function Z(){return"undefined"!=typeof CSS&&!!CSS.supports&&CSS.supports("-webkit-backdrop-filter","none")}function J(e){return["html","body","#document"].includes(F(e))}function Q(e){return z(e).getComputedStyle(e)}function ee(e){return V(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function et(e){if("html"===F(e))return e;let t=e.assignedSlot||e.parentNode||Y(e)&&e.host||B(e);return Y(t)?t.host:t}function en(e,t,n){var r;void 0===t&&(t=[]),void 0===n&&(n=!0);let o=function e(t){let n=et(t);return J(n)?t.ownerDocument?t.ownerDocument.body:t.body:U(n)&&G(n)?n:e(n)}(e),i=o===(null==(r=e.ownerDocument)?void 0:r.body),l=z(o);if(i){let e=er(l);return t.concat(l,l.visualViewport||[],G(o)?o:[],e&&n?en(e):[])}return t.concat(o,en(o,[],n))}function er(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function eo(e){let t=Q(e),n=parseFloat(t.width)||0,r=parseFloat(t.height)||0,o=U(e),i=o?e.offsetWidth:n,l=o?e.offsetHeight:r,a=y(n)!==i||y(r)!==l;return a&&(n=i,r=l),{width:n,height:r,$:a}}function ei(e){return V(e)?e:e.contextElement}function el(e){let t=ei(e);if(!U(t))return x(1);let n=t.getBoundingClientRect(),{width:r,height:o,$:i}=eo(t),l=(i?y(n.width):n.width)/r,a=(i?y(n.height):n.height)/o;return l&&Number.isFinite(l)||(l=1),a&&Number.isFinite(a)||(a=1),{x:l,y:a}}let ea=x(0);function es(e){let t=z(e);return Z()&&t.visualViewport?{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}:ea}function eu(e,t,n,r){var o;void 0===t&&(t=!1),void 0===n&&(n=!1);let i=e.getBoundingClientRect(),l=ei(e),a=x(1);t&&(r?V(r)&&(a=el(r)):a=el(e));let s=(void 0===(o=n)&&(o=!1),r&&(!o||r===z(l))&&o)?es(l):x(0),u=(i.left+s.x)/a.x,c=(i.top+s.y)/a.y,d=i.width/a.x,f=i.height/a.y;if(l){let e=z(l),t=r&&V(r)?z(r):r,n=e,o=er(n);for(;o&&r&&t!==n;){let e=el(o),t=o.getBoundingClientRect(),r=Q(o),i=t.left+(o.clientLeft+parseFloat(r.paddingLeft))*e.x,l=t.top+(o.clientTop+parseFloat(r.paddingTop))*e.y;u*=e.x,c*=e.y,d*=e.x,f*=e.y,u+=i,c+=l,o=er(n=z(o))}}return A({width:d,height:f,x:u,y:c})}function ec(e,t){let n=ee(e).scrollLeft;return t?t.left+n:eu(B(e)).left+n}function ed(e,t,n){void 0===n&&(n=!1);let r=e.getBoundingClientRect();return{x:r.left+t.scrollLeft-(n?0:ec(e,r)),y:r.top+t.scrollTop}}function ef(e,t,n){let r;if("viewport"===t)r=function(e,t){let n=z(e),r=B(e),o=n.visualViewport,i=r.clientWidth,l=r.clientHeight,a=0,s=0;if(o){i=o.width,l=o.height;let e=Z();(!e||e&&"fixed"===t)&&(a=o.offsetLeft,s=o.offsetTop)}return{width:i,height:l,x:a,y:s}}(e,n);else if("document"===t)r=function(e){let t=B(e),n=ee(e),r=e.ownerDocument.body,o=v(t.scrollWidth,t.clientWidth,r.scrollWidth,r.clientWidth),i=v(t.scrollHeight,t.clientHeight,r.scrollHeight,r.clientHeight),l=-n.scrollLeft+ec(e),a=-n.scrollTop;return"rtl"===Q(r).direction&&(l+=v(t.clientWidth,r.clientWidth)-o),{width:o,height:i,x:l,y:a}}(B(e));else if(V(t))r=function(e,t){let n=eu(e,!0,"fixed"===t),r=n.top+e.clientTop,o=n.left+e.clientLeft,i=U(e)?el(e):x(1),l=e.clientWidth*i.x;return{width:l,height:e.clientHeight*i.y,x:o*i.x,y:r*i.y}}(t,n);else{let n=es(e);r={x:t.x-n.x,y:t.y-n.y,width:t.width,height:t.height}}return A(r)}function ep(e){return"static"===Q(e).position}function eh(e,t){if(!U(e)||"fixed"===Q(e).position)return null;if(t)return t(e);let n=e.offsetParent;return B(e)===n&&(n=n.ownerDocument.body),n}function em(e,t){let n=z(e);if(X(e))return n;if(!U(e)){let t=et(e);for(;t&&!J(t);){if(V(t)&&!ep(t))return t;t=et(t)}return n}let r=eh(e,t);for(;r&&["table","td","th"].includes(F(r))&&ep(r);)r=eh(r,t);return r&&J(r)&&ep(r)&&!K(r)?n:r||function(e){let t=et(e);for(;U(t)&&!J(t);){if(K(t))return t;if(X(t))break;t=et(t)}return null}(e)||n}let eg=async function(e){let t=this.getOffsetParent||em,n=this.getDimensions,r=await n(e.floating);return{reference:function(e,t,n){let r=U(t),o=B(t),i="fixed"===n,l=eu(e,!0,i,t),a={scrollLeft:0,scrollTop:0},s=x(0);if(r||!r&&!i){if(("body"!==F(t)||G(o))&&(a=ee(t)),r){let e=eu(t,!0,i,t);s.x=e.x+t.clientLeft,s.y=e.y+t.clientTop}else o&&(s.x=ec(o))}i&&!r&&o&&(s.x=ec(o));let u=!o||r||i?x(0):ed(o,a);return{x:l.left+a.scrollLeft-s.x-u.x,y:l.top+a.scrollTop-s.y-u.y,width:l.width,height:l.height}}(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:r.width,height:r.height}}},eb={convertOffsetParentRelativeRectToViewportRelativeRect:function(e){let{elements:t,rect:n,offsetParent:r,strategy:o}=e,i="fixed"===o,l=B(r),a=!!t&&X(t.floating);if(r===l||a&&i)return n;let s={scrollLeft:0,scrollTop:0},u=x(1),c=x(0),d=U(r);if((d||!d&&!i)&&(("body"!==F(r)||G(l))&&(s=ee(r)),U(r))){let e=eu(r);u=el(r),c.x=e.x+r.clientLeft,c.y=e.y+r.clientTop}let f=!l||d||i?x(0):ed(l,s,!0);return{width:n.width*u.x,height:n.height*u.y,x:n.x*u.x-s.scrollLeft*u.x+c.x+f.x,y:n.y*u.y-s.scrollTop*u.y+c.y+f.y}},getDocumentElement:B,getClippingRect:function(e){let{element:t,boundary:n,rootBoundary:r,strategy:o}=e,i=[..."clippingAncestors"===n?X(t)?[]:function(e,t){let n=t.get(e);if(n)return n;let r=en(e,[],!1).filter(e=>V(e)&&"body"!==F(e)),o=null,i="fixed"===Q(e).position,l=i?et(e):e;for(;V(l)&&!J(l);){let t=Q(l),n=K(l);n||"fixed"!==t.position||(o=null),(i?!n&&!o:!n&&"static"===t.position&&!!o&&["absolute","fixed"].includes(o.position)||G(l)&&!n&&function e(t,n){let r=et(t);return!(r===n||!V(r)||J(r))&&("fixed"===Q(r).position||e(r,n))}(e,l))?r=r.filter(e=>e!==l):o=t,l=et(l)}return t.set(e,r),r}(t,this._c):[].concat(n),r],l=i[0],a=i.reduce((e,n)=>{let r=ef(t,n,o);return e.top=v(r.top,e.top),e.right=b(r.right,e.right),e.bottom=b(r.bottom,e.bottom),e.left=v(r.left,e.left),e},ef(t,l,o));return{width:a.right-a.left,height:a.bottom-a.top,x:a.left,y:a.top}},getOffsetParent:em,getElementRects:eg,getClientRects:function(e){return Array.from(e.getClientRects())},getDimensions:function(e){let{width:t,height:n}=eo(e);return{width:t,height:n}},getScale:el,isElement:V,isRTL:function(e){return"rtl"===Q(e).direction}};function ev(e,t){return e.x===t.x&&e.y===t.y&&e.width===t.width&&e.height===t.height}let ey=e=>({name:"arrow",options:e,async fn(t){let{x:n,y:r,placement:o,rects:i,platform:l,elements:a,middlewareData:s}=t,{element:u,padding:c=0}=k(e,t)||{};if(null==u)return{};let d=L(c),f={x:n,y:r},p=P(R(o)),h=M(p),m=await l.getDimensions(u),g="y"===p,y=g?"clientHeight":"clientWidth",w=i.reference[h]+i.reference[p]-f[p]-i.floating[h],x=f[p]-i.reference[p],E=await (null==l.getOffsetParent?void 0:l.getOffsetParent(u)),S=E?E[y]:0;S&&await (null==l.isElement?void 0:l.isElement(E))||(S=a.floating[y]||i.floating[h]);let T=S/2-m[h]/2-1,j=b(d[g?"top":"left"],T),O=b(d[g?"bottom":"right"],T),A=S-m[h]-O,_=S/2-m[h]/2+(w/2-x/2),N=v(j,b(_,A)),$=!s.arrow&&null!=C(o)&&_!==N&&i.reference[h]/2-(_<j?j:O)-m[h]/2<0,I=$?_<j?_-j:_-A:0;return{[p]:f[p]+I,data:{[p]:N,centerOffset:_-N-I,...$&&{alignmentOffset:I}},reset:$}}}),ew=(e,t,n)=>{let r=new Map,o={platform:eb,...n},i={...o.platform,_c:r};return N(e,t,{...o,platform:i})};var ex="undefined"!=typeof document?o.useLayoutEffect:function(){};function eE(e,t){let n,r,o;if(e===t)return!0;if(typeof e!=typeof t)return!1;if("function"==typeof e&&e.toString()===t.toString())return!0;if(e&&t&&"object"==typeof e){if(Array.isArray(e)){if((n=e.length)!==t.length)return!1;for(r=n;0!=r--;)if(!eE(e[r],t[r]))return!1;return!0}if((n=(o=Object.keys(e)).length)!==Object.keys(t).length)return!1;for(r=n;0!=r--;)if(!({}).hasOwnProperty.call(t,o[r]))return!1;for(r=n;0!=r--;){let n=o[r];if(("_owner"!==n||!e.$$typeof)&&!eE(e[n],t[n]))return!1}return!0}return e!=e&&t!=t}function eS(e){return"undefined"==typeof window?1:(e.ownerDocument.defaultView||window).devicePixelRatio||1}function ek(e,t){let n=eS(e);return Math.round(t*n)/n}function eT(e){let t=o.useRef(e);return ex(()=>{t.current=e}),t}let eC=e=>({name:"arrow",options:e,fn(t){let{element:n,padding:r}="function"==typeof e?e(t):e;return n&&({}).hasOwnProperty.call(n,"current")?null!=n.current?ey({element:n.current,padding:r}).fn(t):{}:n?ey({element:n,padding:r}).fn(t):{}}}),eP=(e,t)=>{var n;return{...(void 0===(n=e)&&(n=0),{name:"offset",options:n,async fn(e){var t,r;let{x:o,y:i,placement:l,middlewareData:a}=e,s=await W(e,n);return l===(null==(t=a.offset)?void 0:t.placement)&&null!=(r=a.arrow)&&r.alignmentOffset?{}:{x:o+s.x,y:i+s.y,data:{...s,placement:l}}}}),options:[e,t]}},eM=(e,t)=>{var n;return{...(void 0===(n=e)&&(n={}),{name:"shift",options:n,async fn(e){let{x:t,y:r,placement:o}=e,{mainAxis:i=!0,crossAxis:l=!1,limiter:a={fn:e=>{let{x:t,y:n}=e;return{x:t,y:n}}},...s}=k(n,e),u={x:t,y:r},c=await $(e,s),d=R(T(o)),f=P(d),p=u[f],h=u[d];if(i){let e="y"===f?"top":"left",t="y"===f?"bottom":"right",n=p+c[e],r=p-c[t];p=v(n,b(p,r))}if(l){let e="y"===d?"top":"left",t="y"===d?"bottom":"right",n=h+c[e],r=h-c[t];h=v(n,b(h,r))}let m=a.fn({...e,[f]:p,[d]:h});return{...m,data:{x:m.x-t,y:m.y-r,enabled:{[f]:i,[d]:l}}}}}),options:[e,t]}},eR=(e,t)=>{var n;return{...(void 0===(n=e)&&(n={}),{options:n,fn(e){let{x:t,y:r,placement:o,rects:i,middlewareData:l}=e,{offset:a=0,mainAxis:s=!0,crossAxis:u=!0}=k(n,e),c={x:t,y:r},d=R(o),f=P(d),p=c[f],h=c[d],m=k(a,e),g="number"==typeof m?{mainAxis:m,crossAxis:0}:{mainAxis:0,crossAxis:0,...m};if(s){let e="y"===f?"height":"width",t=i.reference[f]-i.floating[e]+g.mainAxis,n=i.reference[f]+i.reference[e]-g.mainAxis;p<t?p=t:p>n&&(p=n)}if(u){var b,v;let e="y"===f?"width":"height",t=["top","left"].includes(T(o)),n=i.reference[d]-i.floating[e]+(t&&(null==(b=l.offset)?void 0:b[d])||0)+(t?0:g.crossAxis),r=i.reference[d]+i.reference[e]+(t?0:(null==(v=l.offset)?void 0:v[d])||0)-(t?g.crossAxis:0);h<n?h=n:h>r&&(h=r)}return{[f]:p,[d]:h}}}),options:[e,t]}},ej=(e,t)=>{var n;return{...(void 0===(n=e)&&(n={}),{name:"flip",options:n,async fn(e){var t,r,o,i,l;let{placement:a,middlewareData:s,rects:u,initialPlacement:c,platform:d,elements:f}=e,{mainAxis:p=!0,crossAxis:h=!0,fallbackPlacements:m,fallbackStrategy:g="bestFit",fallbackAxisSideDirection:b="none",flipAlignment:v=!0,...y}=k(n,e);if(null!=(t=s.arrow)&&t.alignmentOffset)return{};let w=T(a),x=R(c),E=T(c)===c,S=await (null==d.isRTL?void 0:d.isRTL(f.floating)),L=m||(E||!v?[O(c)]:function(e){let t=O(e);return[j(e),t,j(t)]}(c)),A="none"!==b;!m&&A&&L.push(...function(e,t,n,r){let o=C(e),i=function(e,t,n){let r=["left","right"],o=["right","left"];switch(e){case"top":case"bottom":if(n)return t?o:r;return t?r:o;case"left":case"right":return t?["top","bottom"]:["bottom","top"];default:return[]}}(T(e),"start"===n,r);return o&&(i=i.map(e=>e+"-"+o),t&&(i=i.concat(i.map(j)))),i}(c,v,b,S));let _=[c,...L],N=await $(e,y),I=[],D=(null==(r=s.flip)?void 0:r.overflows)||[];if(p&&I.push(N[w]),h){let e=function(e,t,n){void 0===n&&(n=!1);let r=C(e),o=P(R(e)),i=M(o),l="x"===o?r===(n?"end":"start")?"right":"left":"start"===r?"bottom":"top";return t.reference[i]>t.floating[i]&&(l=O(l)),[l,O(l)]}(a,u,S);I.push(N[e[0]],N[e[1]])}if(D=[...D,{placement:a,overflows:I}],!I.every(e=>e<=0)){let e=((null==(o=s.flip)?void 0:o.index)||0)+1,t=_[e];if(t&&(!("alignment"===h&&x!==R(t))||D.every(e=>e.overflows[0]>0&&R(e.placement)===x)))return{data:{index:e,overflows:D},reset:{placement:t}};let n=null==(i=D.filter(e=>e.overflows[0]<=0).sort((e,t)=>e.overflows[1]-t.overflows[1])[0])?void 0:i.placement;if(!n)switch(g){case"bestFit":{let e=null==(l=D.filter(e=>{if(A){let t=R(e.placement);return t===x||"y"===t}return!0}).map(e=>[e.placement,e.overflows.filter(e=>e>0).reduce((e,t)=>e+t,0)]).sort((e,t)=>e[1]-t[1])[0])?void 0:l[0];e&&(n=e);break}case"initialPlacement":n=c}if(a!==n)return{reset:{placement:n}}}return{}}}),options:[e,t]}},eO=(e,t)=>{var n;return{...(void 0===(n=e)&&(n={}),{name:"size",options:n,async fn(e){var t,r;let o,i;let{placement:l,rects:a,platform:s,elements:u}=e,{apply:c=()=>{},...d}=k(n,e),f=await $(e,d),p=T(l),h=C(l),m="y"===R(l),{width:g,height:y}=a.floating;"top"===p||"bottom"===p?(o=p,i=h===(await (null==s.isRTL?void 0:s.isRTL(u.floating))?"start":"end")?"left":"right"):(i=p,o="end"===h?"top":"bottom");let w=y-f.top-f.bottom,x=g-f.left-f.right,E=b(y-f[o],w),S=b(g-f[i],x),P=!e.middlewareData.shift,M=E,j=S;if(null!=(t=e.middlewareData.shift)&&t.enabled.x&&(j=x),null!=(r=e.middlewareData.shift)&&r.enabled.y&&(M=w),P&&!h){let e=v(f.left,0),t=v(f.right,0),n=v(f.top,0),r=v(f.bottom,0);m?j=g-2*(0!==e||0!==t?e+t:v(f.left,f.right)):M=y-2*(0!==n||0!==r?n+r:v(f.top,f.bottom))}await c({...e,availableWidth:j,availableHeight:M});let O=await s.getDimensions(u.floating);return g!==O.width||y!==O.height?{reset:{rects:!0}}:{}}}),options:[e,t]}},eL=(e,t)=>{var n;return{...(void 0===(n=e)&&(n={}),{name:"hide",options:n,async fn(e){let{rects:t}=e,{strategy:r="referenceHidden",...o}=k(n,e);switch(r){case"referenceHidden":{let n=I(await $(e,{...o,elementContext:"reference"}),t.reference);return{data:{referenceHiddenOffsets:n,referenceHidden:D(n)}}}case"escaped":{let n=I(await $(e,{...o,altBoundary:!0}),t.floating);return{data:{escapedOffsets:n,escaped:D(n)}}}default:return{}}}}),options:[e,t]}},eA=(e,t)=>({...eC(e),options:[e,t]});var e_=n(75320),eN=o.forwardRef((e,t)=>{let{children:n,width:o=10,height:i=5,...l}=e;return(0,r.jsx)(e_.WV.svg,{...l,ref:t,width:o,height:i,viewBox:"0 0 30 10",preserveAspectRatio:"none",children:e.asChild?n:(0,r.jsx)("polygon",{points:"0,0 30,0 15,10"})})});eN.displayName="Arrow";var e$=n(79698),eI=n(9981),eD="Popper",[eW,eH]=(0,s.b)(eD),[eF,ez]=eW(eD),eB=e=>{let{__scopePopper:t,children:n}=e,[i,l]=o.useState(null);return(0,r.jsx)(eF,{scope:t,anchor:i,onAnchorChange:l,children:n})};eB.displayName=eD;var eq="PopperAnchor",eV=o.forwardRef((e,t)=>{let{__scopePopper:n,virtualRef:i,...l}=e,a=ez(eq,n),s=o.useRef(null),c=(0,u.e)(t,s);return o.useEffect(()=>{a.onAnchorChange(i?.current||s.current)}),i?null:(0,r.jsx)(e_.WV.div,{...l,ref:c})});eV.displayName=eq;var eU="PopperContent",[eY,eG]=eW(eU),eX=o.forwardRef((e,t)=>{let{__scopePopper:n,side:l="bottom",sideOffset:a=0,align:s="center",alignOffset:c=0,arrowPadding:d=0,avoidCollisions:f=!0,collisionBoundary:p=[],collisionPadding:h=0,sticky:m="partial",hideWhenDetached:g=!1,updatePositionStrategy:y="optimized",onPlaced:x,...E}=e,S=ez(eU,n),[k,T]=o.useState(null),C=(0,u.e)(t,e=>T(e)),[P,M]=o.useState(null),R=function(e){let[t,n]=o.useState(void 0);return(0,eI.b)(()=>{if(e){n({width:e.offsetWidth,height:e.offsetHeight});let t=new ResizeObserver(t=>{let r,o;if(!Array.isArray(t)||!t.length)return;let i=t[0];if("borderBoxSize"in i){let e=i.borderBoxSize,t=Array.isArray(e)?e[0]:e;r=t.inlineSize,o=t.blockSize}else r=e.offsetWidth,o=e.offsetHeight;n({width:r,height:o})});return t.observe(e,{box:"border-box"}),()=>t.unobserve(e)}n(void 0)},[e]),t}(P),j=R?.width??0,O=R?.height??0,L="number"==typeof h?h:{top:0,right:0,bottom:0,left:0,...h},A=Array.isArray(p)?p:[p],_=A.length>0,N={padding:L,boundary:A.filter(eQ),altBoundary:_},{refs:$,floatingStyles:I,placement:D,isPositioned:W,middlewareData:H}=function(e){void 0===e&&(e={});let{placement:t="bottom",strategy:n="absolute",middleware:r=[],platform:l,elements:{reference:a,floating:s}={},transform:u=!0,whileElementsMounted:c,open:d}=e,[f,p]=o.useState({x:0,y:0,strategy:n,placement:t,middlewareData:{},isPositioned:!1}),[h,m]=o.useState(r);eE(h,r)||m(r);let[g,b]=o.useState(null),[v,y]=o.useState(null),w=o.useCallback(e=>{e!==k.current&&(k.current=e,b(e))},[]),x=o.useCallback(e=>{e!==T.current&&(T.current=e,y(e))},[]),E=a||g,S=s||v,k=o.useRef(null),T=o.useRef(null),C=o.useRef(f),P=null!=c,M=eT(c),R=eT(l),j=eT(d),O=o.useCallback(()=>{if(!k.current||!T.current)return;let e={placement:t,strategy:n,middleware:h};R.current&&(e.platform=R.current),ew(k.current,T.current,e).then(e=>{let t={...e,isPositioned:!1!==j.current};L.current&&!eE(C.current,t)&&(C.current=t,i.flushSync(()=>{p(t)}))})},[h,t,n,R,j]);ex(()=>{!1===d&&C.current.isPositioned&&(C.current.isPositioned=!1,p(e=>({...e,isPositioned:!1})))},[d]);let L=o.useRef(!1);ex(()=>(L.current=!0,()=>{L.current=!1}),[]),ex(()=>{if(E&&(k.current=E),S&&(T.current=S),E&&S){if(M.current)return M.current(E,S,O);O()}},[E,S,O,M,P]);let A=o.useMemo(()=>({reference:k,floating:T,setReference:w,setFloating:x}),[w,x]),_=o.useMemo(()=>({reference:E,floating:S}),[E,S]),N=o.useMemo(()=>{let e={position:n,left:0,top:0};if(!_.floating)return e;let t=ek(_.floating,f.x),r=ek(_.floating,f.y);return u?{...e,transform:"translate("+t+"px, "+r+"px)",...eS(_.floating)>=1.5&&{willChange:"transform"}}:{position:n,left:t,top:r}},[n,u,_.floating,f.x,f.y]);return o.useMemo(()=>({...f,update:O,refs:A,elements:_,floatingStyles:N}),[f,O,A,_,N])}({strategy:"fixed",placement:l+("center"!==s?"-"+s:""),whileElementsMounted:(...e)=>(function(e,t,n,r){let o;void 0===r&&(r={});let{ancestorScroll:i=!0,ancestorResize:l=!0,elementResize:a="function"==typeof ResizeObserver,layoutShift:s="function"==typeof IntersectionObserver,animationFrame:u=!1}=r,c=ei(e),d=i||l?[...c?en(c):[],...en(t)]:[];d.forEach(e=>{i&&e.addEventListener("scroll",n,{passive:!0}),l&&e.addEventListener("resize",n)});let f=c&&s?function(e,t){let n,r=null,o=B(e);function i(){var e;clearTimeout(n),null==(e=r)||e.disconnect(),r=null}return!function l(a,s){void 0===a&&(a=!1),void 0===s&&(s=1),i();let u=e.getBoundingClientRect(),{left:c,top:d,width:f,height:p}=u;if(a||t(),!f||!p)return;let h=w(d),m=w(o.clientWidth-(c+f)),g={rootMargin:-h+"px "+-m+"px "+-w(o.clientHeight-(d+p))+"px "+-w(c)+"px",threshold:v(0,b(1,s))||1},y=!0;function x(t){let r=t[0].intersectionRatio;if(r!==s){if(!y)return l();r?l(!1,r):n=setTimeout(()=>{l(!1,1e-7)},1e3)}1!==r||ev(u,e.getBoundingClientRect())||l(),y=!1}try{r=new IntersectionObserver(x,{...g,root:o.ownerDocument})}catch(e){r=new IntersectionObserver(x,g)}r.observe(e)}(!0),i}(c,n):null,p=-1,h=null;a&&(h=new ResizeObserver(e=>{let[r]=e;r&&r.target===c&&h&&(h.unobserve(t),cancelAnimationFrame(p),p=requestAnimationFrame(()=>{var e;null==(e=h)||e.observe(t)})),n()}),c&&!u&&h.observe(c),h.observe(t));let m=u?eu(e):null;return u&&function t(){let r=eu(e);m&&!ev(m,r)&&n(),m=r,o=requestAnimationFrame(t)}(),n(),()=>{var e;d.forEach(e=>{i&&e.removeEventListener("scroll",n),l&&e.removeEventListener("resize",n)}),null==f||f(),null==(e=h)||e.disconnect(),h=null,u&&cancelAnimationFrame(o)}})(...e,{animationFrame:"always"===y}),elements:{reference:S.anchor},middleware:[eP({mainAxis:a+O,alignmentAxis:c}),f&&eM({mainAxis:!0,crossAxis:!1,limiter:"partial"===m?eR():void 0,...N}),f&&ej({...N}),eO({...N,apply:({elements:e,rects:t,availableWidth:n,availableHeight:r})=>{let{width:o,height:i}=t.reference,l=e.floating.style;l.setProperty("--radix-popper-available-width",`${n}px`),l.setProperty("--radix-popper-available-height",`${r}px`),l.setProperty("--radix-popper-anchor-width",`${o}px`),l.setProperty("--radix-popper-anchor-height",`${i}px`)}}),P&&eA({element:P,padding:d}),e0({arrowWidth:j,arrowHeight:O}),g&&eL({strategy:"referenceHidden",...N})]}),[F,z]=e1(D),q=(0,e$.W)(x);(0,eI.b)(()=>{W&&q?.()},[W,q]);let V=H.arrow?.x,U=H.arrow?.y,Y=H.arrow?.centerOffset!==0,[G,X]=o.useState();return(0,eI.b)(()=>{k&&X(window.getComputedStyle(k).zIndex)},[k]),(0,r.jsx)("div",{ref:$.setFloating,"data-radix-popper-content-wrapper":"",style:{...I,transform:W?I.transform:"translate(0, -200%)",minWidth:"max-content",zIndex:G,"--radix-popper-transform-origin":[H.transformOrigin?.x,H.transformOrigin?.y].join(" "),...H.hide?.referenceHidden&&{visibility:"hidden",pointerEvents:"none"}},dir:e.dir,children:(0,r.jsx)(eY,{scope:n,placedSide:F,onArrowChange:M,arrowX:V,arrowY:U,shouldHideArrow:Y,children:(0,r.jsx)(e_.WV.div,{"data-side":F,"data-align":z,...E,ref:C,style:{...E.style,animation:W?void 0:"none"}})})})});eX.displayName=eU;var eK="PopperArrow",eZ={top:"bottom",right:"left",bottom:"top",left:"right"},eJ=o.forwardRef(function(e,t){let{__scopePopper:n,...o}=e,i=eG(eK,n),l=eZ[i.placedSide];return(0,r.jsx)("span",{ref:i.onArrowChange,style:{position:"absolute",left:i.arrowX,top:i.arrowY,[l]:0,transformOrigin:{top:"",right:"0 0",bottom:"center 0",left:"100% 0"}[i.placedSide],transform:{top:"translateY(100%)",right:"translateY(50%) rotate(90deg) translateX(-50%)",bottom:"rotate(180deg)",left:"translateY(50%) rotate(-90deg) translateX(50%)"}[i.placedSide],visibility:i.shouldHideArrow?"hidden":void 0},children:(0,r.jsx)(eN,{...o,ref:t,style:{...o.style,display:"block"}})})});function eQ(e){return null!==e}eJ.displayName=eK;var e0=e=>({name:"transformOrigin",options:e,fn(t){let{placement:n,rects:r,middlewareData:o}=t,i=o.arrow?.centerOffset!==0,l=i?0:e.arrowWidth,a=i?0:e.arrowHeight,[s,u]=e1(n),c={start:"0%",center:"50%",end:"100%"}[u],d=(o.arrow?.x??0)+l/2,f=(o.arrow?.y??0)+a/2,p="",h="";return"bottom"===s?(p=i?c:`${d}px`,h=`${-a}px`):"top"===s?(p=i?c:`${d}px`,h=`${r.floating.height+a}px`):"right"===s?(p=`${-a}px`,h=i?c:`${f}px`):"left"===s&&(p=`${r.floating.width+a}px`,h=i?c:`${f}px`),{data:{x:p,y:h}}}});function e1(e){let[t,n="center"]=e.split("-");return[t,n]}var e2=n(42651),e4=n(77342),e5=Object.freeze({position:"absolute",border:0,width:1,height:1,padding:0,margin:-1,overflow:"hidden",clip:"rect(0, 0, 0, 0)",whiteSpace:"nowrap",wordWrap:"normal"});o.forwardRef((e,t)=>(0,r.jsx)(e_.WV.span,{...e,ref:t,style:{...e5,...e.style}})).displayName="VisuallyHidden";var e6=n(23541),e3=n(6223),e9=[" ","Enter","ArrowUp","ArrowDown"],e8=[" ","Enter"],e7="Select",[te,tt,tn]=function(e){let t=e+"CollectionProvider",[n,i]=(0,s.b)(t),[l,a]=n(t,{collectionRef:{current:null},itemMap:new Map}),d=e=>{let{scope:t,children:n}=e,i=o.useRef(null),a=o.useRef(new Map).current;return(0,r.jsx)(l,{scope:t,itemMap:a,collectionRef:i,children:n})};d.displayName=t;let f=e+"CollectionSlot",p=(0,c.Z8)(f),h=o.forwardRef((e,t)=>{let{scope:n,children:o}=e,i=a(f,n),l=(0,u.e)(t,i.collectionRef);return(0,r.jsx)(p,{ref:l,children:o})});h.displayName=f;let m=e+"CollectionItemSlot",g="data-radix-collection-item",b=(0,c.Z8)(m),v=o.forwardRef((e,t)=>{let{scope:n,children:i,...l}=e,s=o.useRef(null),c=(0,u.e)(t,s),d=a(m,n);return o.useEffect(()=>(d.itemMap.set(s,{ref:s,...l}),()=>void d.itemMap.delete(s))),(0,r.jsx)(b,{[g]:"",ref:c,children:i})});return v.displayName=m,[{Provider:d,Slot:h,ItemSlot:v},function(t){let n=a(e+"CollectionConsumer",t);return o.useCallback(()=>{let e=n.collectionRef.current;if(!e)return[];let t=Array.from(e.querySelectorAll(`[${g}]`));return Array.from(n.itemMap.values()).sort((e,n)=>t.indexOf(e.ref.current)-t.indexOf(n.ref.current))},[n.collectionRef,n.itemMap])},i]}(e7),[tr,to]=(0,s.b)(e7,[tn,eH]),ti=eH(),[tl,ta]=tr(e7),[ts,tu]=tr(e7),tc=e=>{let{__scopeSelect:t,children:n,open:i,defaultOpen:l,onOpenChange:a,value:s,defaultValue:u,onValueChange:c,dir:f,name:p,autoComplete:h,disabled:g,required:b,form:v}=e,y=ti(t),[w,x]=o.useState(null),[E,S]=o.useState(null),[k,T]=o.useState(!1),C=function(e){let t=o.useContext(d);return e||t||"ltr"}(f),[P,M]=(0,e4.T)({prop:i,defaultProp:l??!1,onChange:a,caller:e7}),[R,j]=(0,e4.T)({prop:s,defaultProp:u,onChange:c,caller:e7}),O=o.useRef(null),L=!w||v||!!w.closest("form"),[A,_]=o.useState(new Set),N=Array.from(A).map(e=>e.props.value).join(";");return(0,r.jsx)(eB,{...y,children:(0,r.jsxs)(tl,{required:b,scope:t,trigger:w,onTriggerChange:x,valueNode:E,onValueNodeChange:S,valueNodeHasChildren:k,onValueNodeHasChildrenChange:T,contentId:(0,m.M)(),value:R,onValueChange:j,open:P,onOpenChange:M,dir:C,triggerPointerDownPosRef:O,disabled:g,children:[(0,r.jsx)(te.Provider,{scope:t,children:(0,r.jsx)(ts,{scope:e.__scopeSelect,onNativeOptionAdd:o.useCallback(e=>{_(t=>new Set(t).add(e))},[]),onNativeOptionRemove:o.useCallback(e=>{_(t=>{let n=new Set(t);return n.delete(e),n})},[]),children:n})}),L?(0,r.jsxs)(tX,{"aria-hidden":!0,required:b,tabIndex:-1,name:p,autoComplete:h,value:R,onChange:e=>j(e.target.value),disabled:g,form:v,children:[void 0===R?(0,r.jsx)("option",{value:""}):null,Array.from(A)]},N):null]})})};tc.displayName=e7;var td="SelectTrigger",tf=o.forwardRef((e,t)=>{let{__scopeSelect:n,disabled:i=!1,...l}=e,s=ti(n),c=ta(td,n),d=c.disabled||i,f=(0,u.e)(t,c.onTriggerChange),p=tt(n),h=o.useRef("touch"),[m,g,b]=tZ(e=>{let t=p().filter(e=>!e.disabled),n=t.find(e=>e.value===c.value),r=tJ(t,e,n);void 0!==r&&c.onValueChange(r.value)}),v=e=>{d||(c.onOpenChange(!0),b()),e&&(c.triggerPointerDownPosRef.current={x:Math.round(e.pageX),y:Math.round(e.pageY)})};return(0,r.jsx)(eV,{asChild:!0,...s,children:(0,r.jsx)(e_.WV.button,{type:"button",role:"combobox","aria-controls":c.contentId,"aria-expanded":c.open,"aria-required":c.required,"aria-autocomplete":"none",dir:c.dir,"data-state":c.open?"open":"closed",disabled:d,"data-disabled":d?"":void 0,"data-placeholder":tK(c.value)?"":void 0,...l,ref:f,onClick:(0,a.M)(l.onClick,e=>{e.currentTarget.focus(),"mouse"!==h.current&&v(e)}),onPointerDown:(0,a.M)(l.onPointerDown,e=>{h.current=e.pointerType;let t=e.target;t.hasPointerCapture(e.pointerId)&&t.releasePointerCapture(e.pointerId),0===e.button&&!1===e.ctrlKey&&"mouse"===e.pointerType&&(v(e),e.preventDefault())}),onKeyDown:(0,a.M)(l.onKeyDown,e=>{let t=""!==m.current;e.ctrlKey||e.altKey||e.metaKey||1!==e.key.length||g(e.key),(!t||" "!==e.key)&&e9.includes(e.key)&&(v(),e.preventDefault())})})})});tf.displayName=td;var tp="SelectValue",th=o.forwardRef((e,t)=>{let{__scopeSelect:n,className:o,style:i,children:l,placeholder:a="",...s}=e,c=ta(tp,n),{onValueNodeHasChildrenChange:d}=c,f=void 0!==l,p=(0,u.e)(t,c.onValueNodeChange);return(0,eI.b)(()=>{d(f)},[d,f]),(0,r.jsx)(e_.WV.span,{...s,ref:p,style:{pointerEvents:"none"},children:tK(c.value)?(0,r.jsx)(r.Fragment,{children:a}):l})});th.displayName=tp;var tm=o.forwardRef((e,t)=>{let{__scopeSelect:n,children:o,...i}=e;return(0,r.jsx)(e_.WV.span,{"aria-hidden":!0,...i,ref:t,children:o||"▼"})});tm.displayName="SelectIcon";var tg=e=>(0,r.jsx)(e2.h,{asChild:!0,...e});tg.displayName="SelectPortal";var tb="SelectContent",tv=o.forwardRef((e,t)=>{let n=ta(tb,e.__scopeSelect),[l,a]=o.useState();return((0,eI.b)(()=>{a(new DocumentFragment)},[]),n.open)?(0,r.jsx)(tE,{...e,ref:t}):l?i.createPortal((0,r.jsx)(ty,{scope:e.__scopeSelect,children:(0,r.jsx)(te.Slot,{scope:e.__scopeSelect,children:(0,r.jsx)("div",{children:e.children})})}),l):null});tv.displayName=tb;var[ty,tw]=tr(tb),tx=(0,c.Z8)("SelectContent.RemoveScroll"),tE=o.forwardRef((e,t)=>{let{__scopeSelect:n,position:i="item-aligned",onCloseAutoFocus:l,onEscapeKeyDown:s,onPointerDownOutside:c,side:d,sideOffset:m,align:g,alignOffset:b,arrowPadding:v,collisionBoundary:y,collisionPadding:w,sticky:x,hideWhenDetached:E,avoidCollisions:S,...k}=e,T=ta(tb,n),[C,P]=o.useState(null),[M,R]=o.useState(null),j=(0,u.e)(t,e=>P(e)),[O,L]=o.useState(null),[A,_]=o.useState(null),N=tt(n),[$,I]=o.useState(!1),D=o.useRef(!1);o.useEffect(()=>{if(C)return(0,e6.Ry)(C)},[C]),(0,p.EW)();let W=o.useCallback(e=>{let[t,...n]=N().map(e=>e.ref.current),[r]=n.slice(-1),o=document.activeElement;for(let n of e)if(n===o||(n?.scrollIntoView({block:"nearest"}),n===t&&M&&(M.scrollTop=0),n===r&&M&&(M.scrollTop=M.scrollHeight),n?.focus(),document.activeElement!==o))return},[N,M]),H=o.useCallback(()=>W([O,C]),[W,O,C]);o.useEffect(()=>{$&&H()},[$,H]);let{onOpenChange:F,triggerPointerDownPosRef:z}=T;o.useEffect(()=>{if(C){let e={x:0,y:0},t=t=>{e={x:Math.abs(Math.round(t.pageX)-(z.current?.x??0)),y:Math.abs(Math.round(t.pageY)-(z.current?.y??0))}},n=n=>{e.x<=10&&e.y<=10?n.preventDefault():C.contains(n.target)||F(!1),document.removeEventListener("pointermove",t),z.current=null};return null!==z.current&&(document.addEventListener("pointermove",t),document.addEventListener("pointerup",n,{capture:!0,once:!0})),()=>{document.removeEventListener("pointermove",t),document.removeEventListener("pointerup",n,{capture:!0})}}},[C,F,z]),o.useEffect(()=>{let e=()=>F(!1);return window.addEventListener("blur",e),window.addEventListener("resize",e),()=>{window.removeEventListener("blur",e),window.removeEventListener("resize",e)}},[F]);let[B,q]=tZ(e=>{let t=N().filter(e=>!e.disabled),n=t.find(e=>e.ref.current===document.activeElement),r=tJ(t,e,n);r&&setTimeout(()=>r.ref.current.focus())}),V=o.useCallback((e,t,n)=>{let r=!D.current&&!n;(void 0!==T.value&&T.value===t||r)&&(L(e),r&&(D.current=!0))},[T.value]),U=o.useCallback(()=>C?.focus(),[C]),Y=o.useCallback((e,t,n)=>{let r=!D.current&&!n;(void 0!==T.value&&T.value===t||r)&&_(e)},[T.value]),G="popper"===i?tk:tS,X=G===tk?{side:d,sideOffset:m,align:g,alignOffset:b,arrowPadding:v,collisionBoundary:y,collisionPadding:w,sticky:x,hideWhenDetached:E,avoidCollisions:S}:{};return(0,r.jsx)(ty,{scope:n,content:C,viewport:M,onViewportChange:R,itemRefCallback:V,selectedItem:O,onItemLeave:U,itemTextRefCallback:Y,focusSelectedItem:H,selectedItemText:A,position:i,isPositioned:$,searchRef:B,children:(0,r.jsx)(e3.Z,{as:tx,allowPinchZoom:!0,children:(0,r.jsx)(h.M,{asChild:!0,trapped:T.open,onMountAutoFocus:e=>{e.preventDefault()},onUnmountAutoFocus:(0,a.M)(l,e=>{T.trigger?.focus({preventScroll:!0}),e.preventDefault()}),children:(0,r.jsx)(f.XB,{asChild:!0,disableOutsidePointerEvents:!0,onEscapeKeyDown:s,onPointerDownOutside:c,onFocusOutside:e=>e.preventDefault(),onDismiss:()=>T.onOpenChange(!1),children:(0,r.jsx)(G,{role:"listbox",id:T.contentId,"data-state":T.open?"open":"closed",dir:T.dir,onContextMenu:e=>e.preventDefault(),...k,...X,onPlaced:()=>I(!0),ref:j,style:{display:"flex",flexDirection:"column",outline:"none",...k.style},onKeyDown:(0,a.M)(k.onKeyDown,e=>{let t=e.ctrlKey||e.altKey||e.metaKey;if("Tab"===e.key&&e.preventDefault(),t||1!==e.key.length||q(e.key),["ArrowUp","ArrowDown","Home","End"].includes(e.key)){let t=N().filter(e=>!e.disabled).map(e=>e.ref.current);if(["ArrowUp","End"].includes(e.key)&&(t=t.slice().reverse()),["ArrowUp","ArrowDown"].includes(e.key)){let n=e.target,r=t.indexOf(n);t=t.slice(r+1)}setTimeout(()=>W(t)),e.preventDefault()}})})})})})})});tE.displayName="SelectContentImpl";var tS=o.forwardRef((e,t)=>{let{__scopeSelect:n,onPlaced:i,...a}=e,s=ta(tb,n),c=tw(tb,n),[d,f]=o.useState(null),[p,h]=o.useState(null),m=(0,u.e)(t,e=>h(e)),g=tt(n),b=o.useRef(!1),v=o.useRef(!0),{viewport:y,selectedItem:w,selectedItemText:x,focusSelectedItem:E}=c,S=o.useCallback(()=>{if(s.trigger&&s.valueNode&&d&&p&&y&&w&&x){let e=s.trigger.getBoundingClientRect(),t=p.getBoundingClientRect(),n=s.valueNode.getBoundingClientRect(),r=x.getBoundingClientRect();if("rtl"!==s.dir){let o=r.left-t.left,i=n.left-o,a=e.left-i,s=e.width+a,u=Math.max(s,t.width),c=l(i,[10,Math.max(10,window.innerWidth-10-u)]);d.style.minWidth=s+"px",d.style.left=c+"px"}else{let o=t.right-r.right,i=window.innerWidth-n.right-o,a=window.innerWidth-e.right-i,s=e.width+a,u=Math.max(s,t.width),c=l(i,[10,Math.max(10,window.innerWidth-10-u)]);d.style.minWidth=s+"px",d.style.right=c+"px"}let o=g(),a=window.innerHeight-20,u=y.scrollHeight,c=window.getComputedStyle(p),f=parseInt(c.borderTopWidth,10),h=parseInt(c.paddingTop,10),m=parseInt(c.borderBottomWidth,10),v=f+h+u+parseInt(c.paddingBottom,10)+m,E=Math.min(5*w.offsetHeight,v),S=window.getComputedStyle(y),k=parseInt(S.paddingTop,10),T=parseInt(S.paddingBottom,10),C=e.top+e.height/2-10,P=w.offsetHeight/2,M=f+h+(w.offsetTop+P);if(M<=C){let e=o.length>0&&w===o[o.length-1].ref.current;d.style.bottom="0px";let t=p.clientHeight-y.offsetTop-y.offsetHeight;d.style.height=M+Math.max(a-C,P+(e?T:0)+t+m)+"px"}else{let e=o.length>0&&w===o[0].ref.current;d.style.top="0px";let t=Math.max(C,f+y.offsetTop+(e?k:0)+P);d.style.height=t+(v-M)+"px",y.scrollTop=M-C+y.offsetTop}d.style.margin="10px 0",d.style.minHeight=E+"px",d.style.maxHeight=a+"px",i?.(),requestAnimationFrame(()=>b.current=!0)}},[g,s.trigger,s.valueNode,d,p,y,w,x,s.dir,i]);(0,eI.b)(()=>S(),[S]);let[k,T]=o.useState();(0,eI.b)(()=>{p&&T(window.getComputedStyle(p).zIndex)},[p]);let C=o.useCallback(e=>{e&&!0===v.current&&(S(),E?.(),v.current=!1)},[S,E]);return(0,r.jsx)(tT,{scope:n,contentWrapper:d,shouldExpandOnScrollRef:b,onScrollButtonChange:C,children:(0,r.jsx)("div",{ref:f,style:{display:"flex",flexDirection:"column",position:"fixed",zIndex:k},children:(0,r.jsx)(e_.WV.div,{...a,ref:m,style:{boxSizing:"border-box",maxHeight:"100%",...a.style}})})})});tS.displayName="SelectItemAlignedPosition";var tk=o.forwardRef((e,t)=>{let{__scopeSelect:n,align:o="start",collisionPadding:i=10,...l}=e,a=ti(n);return(0,r.jsx)(eX,{...a,...l,ref:t,align:o,collisionPadding:i,style:{boxSizing:"border-box",...l.style,"--radix-select-content-transform-origin":"var(--radix-popper-transform-origin)","--radix-select-content-available-width":"var(--radix-popper-available-width)","--radix-select-content-available-height":"var(--radix-popper-available-height)","--radix-select-trigger-width":"var(--radix-popper-anchor-width)","--radix-select-trigger-height":"var(--radix-popper-anchor-height)"}})});tk.displayName="SelectPopperPosition";var[tT,tC]=tr(tb,{}),tP="SelectViewport",tM=o.forwardRef((e,t)=>{let{__scopeSelect:n,nonce:i,...l}=e,s=tw(tP,n),c=tC(tP,n),d=(0,u.e)(t,s.onViewportChange),f=o.useRef(0);return(0,r.jsxs)(r.Fragment,{children:[(0,r.jsx)("style",{dangerouslySetInnerHTML:{__html:"[data-radix-select-viewport]{scrollbar-width:none;-ms-overflow-style:none;-webkit-overflow-scrolling:touch;}[data-radix-select-viewport]::-webkit-scrollbar{display:none}"},nonce:i}),(0,r.jsx)(te.Slot,{scope:n,children:(0,r.jsx)(e_.WV.div,{"data-radix-select-viewport":"",role:"presentation",...l,ref:d,style:{position:"relative",flex:1,overflow:"hidden auto",...l.style},onScroll:(0,a.M)(l.onScroll,e=>{let t=e.currentTarget,{contentWrapper:n,shouldExpandOnScrollRef:r}=c;if(r?.current&&n){let e=Math.abs(f.current-t.scrollTop);if(e>0){let r=window.innerHeight-20,o=Math.max(parseFloat(n.style.minHeight),parseFloat(n.style.height));if(o<r){let i=o+e,l=Math.min(r,i),a=i-l;n.style.height=l+"px","0px"===n.style.bottom&&(t.scrollTop=a>0?a:0,n.style.justifyContent="flex-end")}}}f.current=t.scrollTop})})})]})});tM.displayName=tP;var tR="SelectGroup",[tj,tO]=tr(tR);o.forwardRef((e,t)=>{let{__scopeSelect:n,...o}=e,i=(0,m.M)();return(0,r.jsx)(tj,{scope:n,id:i,children:(0,r.jsx)(e_.WV.div,{role:"group","aria-labelledby":i,...o,ref:t})})}).displayName=tR;var tL="SelectLabel",tA=o.forwardRef((e,t)=>{let{__scopeSelect:n,...o}=e,i=tO(tL,n);return(0,r.jsx)(e_.WV.div,{id:i.id,...o,ref:t})});tA.displayName=tL;var t_="SelectItem",[tN,t$]=tr(t_),tI=o.forwardRef((e,t)=>{let{__scopeSelect:n,value:i,disabled:l=!1,textValue:s,...c}=e,d=ta(t_,n),f=tw(t_,n),p=d.value===i,[h,g]=o.useState(s??""),[b,v]=o.useState(!1),y=(0,u.e)(t,e=>f.itemRefCallback?.(e,i,l)),w=(0,m.M)(),x=o.useRef("touch"),E=()=>{l||(d.onValueChange(i),d.onOpenChange(!1))};if(""===i)throw Error("A <Select.Item /> must have a value prop that is not an empty string. This is because the Select value can be set to an empty string to clear the selection and show the placeholder.");return(0,r.jsx)(tN,{scope:n,value:i,disabled:l,textId:w,isSelected:p,onItemTextChange:o.useCallback(e=>{g(t=>t||(e?.textContent??"").trim())},[]),children:(0,r.jsx)(te.ItemSlot,{scope:n,value:i,disabled:l,textValue:h,children:(0,r.jsx)(e_.WV.div,{role:"option","aria-labelledby":w,"data-highlighted":b?"":void 0,"aria-selected":p&&b,"data-state":p?"checked":"unchecked","aria-disabled":l||void 0,"data-disabled":l?"":void 0,tabIndex:l?void 0:-1,...c,ref:y,onFocus:(0,a.M)(c.onFocus,()=>v(!0)),onBlur:(0,a.M)(c.onBlur,()=>v(!1)),onClick:(0,a.M)(c.onClick,()=>{"mouse"!==x.current&&E()}),onPointerUp:(0,a.M)(c.onPointerUp,()=>{"mouse"===x.current&&E()}),onPointerDown:(0,a.M)(c.onPointerDown,e=>{x.current=e.pointerType}),onPointerMove:(0,a.M)(c.onPointerMove,e=>{x.current=e.pointerType,l?f.onItemLeave?.():"mouse"===x.current&&e.currentTarget.focus({preventScroll:!0})}),onPointerLeave:(0,a.M)(c.onPointerLeave,e=>{e.currentTarget===document.activeElement&&f.onItemLeave?.()}),onKeyDown:(0,a.M)(c.onKeyDown,e=>{f.searchRef?.current!==""&&" "===e.key||(e8.includes(e.key)&&E()," "===e.key&&e.preventDefault())})})})})});tI.displayName=t_;var tD="SelectItemText",tW=o.forwardRef((e,t)=>{let{__scopeSelect:n,className:l,style:a,...s}=e,c=ta(tD,n),d=tw(tD,n),f=t$(tD,n),p=tu(tD,n),[h,m]=o.useState(null),g=(0,u.e)(t,e=>m(e),f.onItemTextChange,e=>d.itemTextRefCallback?.(e,f.value,f.disabled)),b=h?.textContent,v=o.useMemo(()=>(0,r.jsx)("option",{value:f.value,disabled:f.disabled,children:b},f.value),[f.disabled,f.value,b]),{onNativeOptionAdd:y,onNativeOptionRemove:w}=p;return(0,eI.b)(()=>(y(v),()=>w(v)),[y,w,v]),(0,r.jsxs)(r.Fragment,{children:[(0,r.jsx)(e_.WV.span,{id:f.textId,...s,ref:g}),f.isSelected&&c.valueNode&&!c.valueNodeHasChildren?i.createPortal(s.children,c.valueNode):null]})});tW.displayName=tD;var tH="SelectItemIndicator",tF=o.forwardRef((e,t)=>{let{__scopeSelect:n,...o}=e;return t$(tH,n).isSelected?(0,r.jsx)(e_.WV.span,{"aria-hidden":!0,...o,ref:t}):null});tF.displayName=tH;var tz="SelectScrollUpButton",tB=o.forwardRef((e,t)=>{let n=tw(tz,e.__scopeSelect),i=tC(tz,e.__scopeSelect),[l,a]=o.useState(!1),s=(0,u.e)(t,i.onScrollButtonChange);return(0,eI.b)(()=>{if(n.viewport&&n.isPositioned){let e=function(){a(t.scrollTop>0)},t=n.viewport;return e(),t.addEventListener("scroll",e),()=>t.removeEventListener("scroll",e)}},[n.viewport,n.isPositioned]),l?(0,r.jsx)(tU,{...e,ref:s,onAutoScroll:()=>{let{viewport:e,selectedItem:t}=n;e&&t&&(e.scrollTop=e.scrollTop-t.offsetHeight)}}):null});tB.displayName=tz;var tq="SelectScrollDownButton",tV=o.forwardRef((e,t)=>{let n=tw(tq,e.__scopeSelect),i=tC(tq,e.__scopeSelect),[l,a]=o.useState(!1),s=(0,u.e)(t,i.onScrollButtonChange);return(0,eI.b)(()=>{if(n.viewport&&n.isPositioned){let e=function(){let e=t.scrollHeight-t.clientHeight;a(Math.ceil(t.scrollTop)<e)},t=n.viewport;return e(),t.addEventListener("scroll",e),()=>t.removeEventListener("scroll",e)}},[n.viewport,n.isPositioned]),l?(0,r.jsx)(tU,{...e,ref:s,onAutoScroll:()=>{let{viewport:e,selectedItem:t}=n;e&&t&&(e.scrollTop=e.scrollTop+t.offsetHeight)}}):null});tV.displayName=tq;var tU=o.forwardRef((e,t)=>{let{__scopeSelect:n,onAutoScroll:i,...l}=e,s=tw("SelectScrollButton",n),u=o.useRef(null),c=tt(n),d=o.useCallback(()=>{null!==u.current&&(window.clearInterval(u.current),u.current=null)},[]);return o.useEffect(()=>()=>d(),[d]),(0,eI.b)(()=>{let e=c().find(e=>e.ref.current===document.activeElement);e?.ref.current?.scrollIntoView({block:"nearest"})},[c]),(0,r.jsx)(e_.WV.div,{"aria-hidden":!0,...l,ref:t,style:{flexShrink:0,...l.style},onPointerDown:(0,a.M)(l.onPointerDown,()=>{null===u.current&&(u.current=window.setInterval(i,50))}),onPointerMove:(0,a.M)(l.onPointerMove,()=>{s.onItemLeave?.(),null===u.current&&(u.current=window.setInterval(i,50))}),onPointerLeave:(0,a.M)(l.onPointerLeave,()=>{d()})})}),tY=o.forwardRef((e,t)=>{let{__scopeSelect:n,...o}=e;return(0,r.jsx)(e_.WV.div,{"aria-hidden":!0,...o,ref:t})});tY.displayName="SelectSeparator";var tG="SelectArrow";o.forwardRef((e,t)=>{let{__scopeSelect:n,...o}=e,i=ti(n),l=ta(tG,n),a=tw(tG,n);return l.open&&"popper"===a.position?(0,r.jsx)(eJ,{...i,...o,ref:t}):null}).displayName=tG;var tX=o.forwardRef(({__scopeSelect:e,value:t,...n},i)=>{let l=o.useRef(null),a=(0,u.e)(i,l),s=function(e){let t=o.useRef({value:e,previous:e});return o.useMemo(()=>(t.current.value!==e&&(t.current.previous=t.current.value,t.current.value=e),t.current.previous),[e])}(t);return o.useEffect(()=>{let e=l.current;if(!e)return;let n=Object.getOwnPropertyDescriptor(window.HTMLSelectElement.prototype,"value").set;if(s!==t&&n){let r=new Event("change",{bubbles:!0});n.call(e,t),e.dispatchEvent(r)}},[s,t]),(0,r.jsx)(e_.WV.select,{...n,style:{...e5,...n.style},ref:a,defaultValue:t})});function tK(e){return""===e||void 0===e}function tZ(e){let t=(0,e$.W)(e),n=o.useRef(""),r=o.useRef(0),i=o.useCallback(e=>{let o=n.current+e;t(o),function e(t){n.current=t,window.clearTimeout(r.current),""!==t&&(r.current=window.setTimeout(()=>e(""),1e3))}(o)},[t]),l=o.useCallback(()=>{n.current="",window.clearTimeout(r.current)},[]);return o.useEffect(()=>()=>window.clearTimeout(r.current),[]),[n,i,l]}function tJ(e,t,n){var r;let o=t.length>1&&Array.from(t).every(e=>e===t[0])?t[0]:t,i=(r=Math.max(n?e.indexOf(n):-1,0),e.map((t,n)=>e[(r+n)%e.length]));1===o.length&&(i=i.filter(e=>e!==n));let l=i.find(e=>e.textValue.toLowerCase().startsWith(o.toLowerCase()));return l!==n?l:void 0}tX.displayName="SelectBubbleInput";var tQ=n(45895);/**
 * @license lucide-react v0.407.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */let t0=(0,n(60998).Z)("ChevronUp",[["path",{d:"m18 15-6-6-6 6",key:"153udz"}]]);var t1=n(70282),t2=n(46990);let t4=tc,t5=th,t6=o.forwardRef((e,t)=>{let{className:n,children:o,...i}=e;return(0,r.jsxs)(tf,{ref:t,className:(0,t2.cn)("flex h-10 w-full items-center justify-between rounded-md border border-input bg-background px-3 py-2 text-sm ring-offset-background placeholder:text-muted-foreground focus:outline-none focus:ring-2 focus:ring-ring focus:ring-offset-2 disabled:cursor-not-allowed disabled:opacity-50 [&>span]:line-clamp-1",n),...i,children:[o,(0,r.jsx)(tm,{asChild:!0,children:(0,r.jsx)(tQ.Z,{className:"h-4 w-4 opacity-50"})})]})});t6.displayName=tf.displayName;let t3=o.forwardRef((e,t)=>{let{className:n,...o}=e;return(0,r.jsx)(tB,{ref:t,className:(0,t2.cn)("flex cursor-default items-center justify-center py-1",n),...o,children:(0,r.jsx)(t0,{className:"h-4 w-4"})})});t3.displayName=tB.displayName;let t9=o.forwardRef((e,t)=>{let{className:n,...o}=e;return(0,r.jsx)(tV,{ref:t,className:(0,t2.cn)("flex cursor-default items-center justify-center py-1",n),...o,children:(0,r.jsx)(tQ.Z,{className:"h-4 w-4"})})});t9.displayName=tV.displayName;let t8=o.forwardRef((e,t)=>{let{className:n,children:o,position:i="popper",...l}=e;return(0,r.jsx)(tg,{children:(0,r.jsxs)(tv,{ref:t,className:(0,t2.cn)("relative z-50 max-h-96 min-w-[8rem] overflow-hidden rounded-md border bg-popover text-popover-foreground shadow-md data-[state=open]:animate-in data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=open]:fade-in-0 data-[state=closed]:zoom-out-95 data-[state=open]:zoom-in-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2","popper"===i&&"data-[side=bottom]:translate-y-1 data-[side=left]:-translate-x-1 data-[side=right]:translate-x-1 data-[side=top]:-translate-y-1",n),position:i,...l,children:[(0,r.jsx)(t3,{}),(0,r.jsx)(tM,{className:(0,t2.cn)("p-1","popper"===i&&"h-[var(--radix-select-trigger-height)] w-full min-w-[var(--radix-select-trigger-width)]"),children:o}),(0,r.jsx)(t9,{})]})})});t8.displayName=tv.displayName,o.forwardRef((e,t)=>{let{className:n,...o}=e;return(0,r.jsx)(tA,{ref:t,className:(0,t2.cn)("py-1.5 pl-8 pr-2 text-sm font-semibold",n),...o})}).displayName=tA.displayName;let t7=o.forwardRef((e,t)=>{let{className:n,children:o,...i}=e;return(0,r.jsxs)(tI,{ref:t,className:(0,t2.cn)("relative flex w-full cursor-default select-none items-center rounded-sm py-1.5 pl-8 pr-2 text-sm outline-none focus:bg-accent focus:text-accent-foreground data-[disabled]:pointer-events-none data-[disabled]:opacity-50",n),...i,children:[(0,r.jsx)("span",{className:"absolute left-2 flex h-3.5 w-3.5 items-center justify-center",children:(0,r.jsx)(tF,{children:(0,r.jsx)(t1.Z,{className:"h-4 w-4"})})}),(0,r.jsx)(tW,{children:o})]})});t7.displayName=tI.displayName,o.forwardRef((e,t)=>{let{className:n,...o}=e;return(0,r.jsx)(tY,{ref:t,className:(0,t2.cn)("-mx-1 my-1 h-px bg-muted",n),...o})}).displayName=tY.displayName},470:function(e,t,n){"use strict";let r,o,i;n.d(t,{WH:function(){return r$},Kl:function(){return rG},x9:function(){return rY},$3:function(){return rD},$B:function(){return rz},Md:function(){return rI},yc:function(){return r_},Zg:function(){return rB},LU:function(){return rW},xn:function(){return rq},sY:function(){return rV},os:function(){return rX},Zt:function(){return rF},Zn:function(){return rH}});var l,a,s,u,c,d,f,p=n(85893),h=n(67294);let m={},g=0,b=!1,v=null,y=null,w="undefined"!=typeof document?h.useLayoutEffect:()=>{},x={prefix:String(Math.round(1e10*Math.random())),current:0},E=h.createContext(x),S=h.createContext(!1);"undefined"!=typeof window&&window.document&&window.document.createElement;let k=new WeakMap,T="function"==typeof h.useId?function(e){let t=h.useId(),[n]=(0,h.useState)(R()),r=n?"react-aria":`react-aria${x.prefix}`;return e||`${r}-${t}`}:function(e){let t=(0,h.useContext)(E),n=function(e=!1){let t=(0,h.useContext)(E),n=(0,h.useRef)(null);if(null===n.current&&!e){var r,o;let e=null===(o=h.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED)||void 0===o?void 0:null===(r=o.ReactCurrentOwner)||void 0===r?void 0:r.current;if(e){let n=k.get(e);null==n?k.set(e,{id:t.current,state:e.memoizedState}):e.memoizedState!==n.state&&(t.current=n.id,k.delete(e))}n.current=++t.current}return n.current}(!!e),r=`react-aria${t.prefix}`;return e||`${r}-${n}`};function C(){return!1}function P(){return!0}function M(e){return()=>{}}function R(){return"function"==typeof h.useSyncExternalStore?h.useSyncExternalStore(M,C,P):(0,h.useContext)(S)}let j=!!("undefined"!=typeof window&&window.document&&window.document.createElement),O=new Map;var L=n(90512);function A(...e){let t={...e[0]};for(let n=1;n<e.length;n++){let r=e[n];for(let e in r){let n=t[e],o=r[e];"function"==typeof n&&"function"==typeof o&&"o"===e[0]&&"n"===e[1]&&e.charCodeAt(2)>=65&&90>=e.charCodeAt(2)?t[e]=function(...e){return(...t)=>{for(let n of e)"function"==typeof n&&n(...t)}}(n,o):("className"===e||"UNSAFE_className"===e)&&"string"==typeof n&&"string"==typeof o?t[e]=(0,L.Z)(n,o):"id"===e&&n&&o?t.id=function(e,t){if(e===t)return e;let n=O.get(e);if(n)return n.forEach(e=>e(t)),t;let r=O.get(t);return r?(r.forEach(t=>t(e)),e):t}(n,o):t[e]=void 0!==o?o:n}}return t}function _(e){var t;return"undefined"!=typeof window&&null!=window.navigator&&((null===(t=window.navigator.userAgentData)||void 0===t?void 0:t.brands.some(t=>e.test(t.brand)))||e.test(window.navigator.userAgent))}function N(e){var t;return"undefined"!=typeof window&&null!=window.navigator&&e.test((null===(t=window.navigator.userAgentData)||void 0===t?void 0:t.platform)||window.navigator.platform)}function $(e){let t=null;return()=>(null==t&&(t=e()),t)}let I=$(function(){return N(/^Mac/i)}),D=$(function(){return N(/^iPhone/i)}),W=$(function(){return N(/^iPad/i)||I()&&navigator.maxTouchPoints>1}),H=$(function(){return D()||W()});$(function(){return I()||H()});let F=$(function(){return _(/AppleWebKit/i)&&!z()}),z=$(function(){return _(/Chrome/i)}),B=$(function(){return _(/Android/i)});$(function(){return _(/Firefox/i)});let q=e=>{var t;return null!==(t=null==e?void 0:e.ownerDocument)&&void 0!==t?t:document},V=e=>e&&"window"in e&&e.window===e?e:q(e).defaultView||window,U=null,Y=new Set,G=new Map,X=!1,K=!1;function Z(e,t){for(let n of Y)n(e,t)}function J(e){X=!0,e.metaKey||!I()&&e.altKey||e.ctrlKey||"Control"===e.key||"Shift"===e.key||"Meta"===e.key||(U="keyboard",Z("keyboard",e))}function Q(e){U="pointer",("mousedown"===e.type||"pointerdown"===e.type)&&(X=!0,Z("pointer",e))}function ee(e){(0===e.mozInputSource&&e.isTrusted||(B()&&e.pointerType?"click"===e.type&&1===e.buttons:0===e.detail&&!e.pointerType))&&(X=!0,U="virtual")}function et(e){e.target!==window&&e.target!==document&&(X||K||(U="virtual",Z("virtual",e)),X=!1,K=!1)}function en(){X=!1,K=!0}function er(e){if("undefined"==typeof window||G.get(V(e)))return;let t=V(e),n=q(e),r=t.HTMLElement.prototype.focus;t.HTMLElement.prototype.focus=function(){X=!0,r.apply(this,arguments)},n.addEventListener("keydown",J,!0),n.addEventListener("keyup",J,!0),n.addEventListener("click",ee,!0),t.addEventListener("focus",et,!0),t.addEventListener("blur",en,!1),"undefined"!=typeof PointerEvent?(n.addEventListener("pointerdown",Q,!0),n.addEventListener("pointermove",Q,!0),n.addEventListener("pointerup",Q,!0)):(n.addEventListener("mousedown",Q,!0),n.addEventListener("mousemove",Q,!0),n.addEventListener("mouseup",Q,!0)),t.addEventListener("beforeunload",()=>{eo(e)},{once:!0}),G.set(t,{focus:r})}let eo=(e,t)=>{let n=V(e),r=q(e);t&&r.removeEventListener("DOMContentLoaded",t),G.has(n)&&(n.HTMLElement.prototype.focus=G.get(n).focus,r.removeEventListener("keydown",J,!0),r.removeEventListener("keyup",J,!0),r.removeEventListener("click",ee,!0),n.removeEventListener("focus",et,!0),n.removeEventListener("blur",en,!1),"undefined"!=typeof PointerEvent?(r.removeEventListener("pointerdown",Q,!0),r.removeEventListener("pointermove",Q,!0),r.removeEventListener("pointerup",Q,!0)):(r.removeEventListener("mousedown",Q,!0),r.removeEventListener("mousemove",Q,!0),r.removeEventListener("mouseup",Q,!0)),G.delete(n))};"undefined"!=typeof document&&function(e){let t;let n=q(void 0);"loading"!==n.readyState?er(void 0):(t=()=>{er(void 0)},n.addEventListener("DOMContentLoaded",t)),()=>eo(e,t)}();let ei=!1,el=0;function ea(){ei=!0,setTimeout(()=>{ei=!1},50)}function es(e){"touch"===e.pointerType&&ea()}function eu(){if("undefined"!=typeof document)return"undefined"!=typeof PointerEvent?document.addEventListener("pointerup",es):document.addEventListener("touchend",ea),el++,()=>{--el>0||("undefined"!=typeof PointerEvent?document.removeEventListener("pointerup",es):document.removeEventListener("touchend",ea))}}function ec(e){let{onHoverStart:t,onHoverChange:n,onHoverEnd:r,isDisabled:o}=e,[i,l]=(0,h.useState)(!1),a=(0,h.useRef)({isHovered:!1,ignoreEmulatedMouseEvents:!1,pointerType:"",target:null}).current;(0,h.useEffect)(eu,[]);let{hoverProps:s,triggerHoverEnd:u}=(0,h.useMemo)(()=>{let e=(e,r)=>{if(a.pointerType=r,o||"touch"===r||a.isHovered||!e.currentTarget.contains(e.target))return;a.isHovered=!0;let i=e.currentTarget;a.target=i,t&&t({type:"hoverstart",target:i,pointerType:r}),n&&n(!0),l(!0)},i=(e,t)=>{if(a.pointerType="",a.target=null,"touch"===t||!a.isHovered)return;a.isHovered=!1;let o=e.currentTarget;r&&r({type:"hoverend",target:o,pointerType:t}),n&&n(!1),l(!1)},s={};return"undefined"!=typeof PointerEvent?(s.onPointerEnter=t=>{ei&&"mouse"===t.pointerType||e(t,t.pointerType)},s.onPointerLeave=e=>{!o&&e.currentTarget.contains(e.target)&&i(e,e.pointerType)}):(s.onTouchStart=()=>{a.ignoreEmulatedMouseEvents=!0},s.onMouseEnter=t=>{a.ignoreEmulatedMouseEvents||ei||e(t,"mouse"),a.ignoreEmulatedMouseEvents=!1},s.onMouseLeave=e=>{!o&&e.currentTarget.contains(e.target)&&i(e,"mouse")}),{hoverProps:s,triggerHoverEnd:i}},[t,n,r,o,a]);return(0,h.useEffect)(()=>{o&&u({currentTarget:a.target},a.pointerType)},[o]),{hoverProps:s,isHovered:i}}let ed="undefined"!=typeof document?h.useLayoutEffect:()=>{};function ef(e){var t;return"undefined"!=typeof window&&null!=window.navigator&&((null===(t=window.navigator.userAgentData)||void 0===t?void 0:t.brands.some(t=>e.test(t.brand)))||e.test(window.navigator.userAgent))}function ep(e){var t;return"undefined"!=typeof window&&null!=window.navigator&&e.test((null===(t=window.navigator.userAgentData)||void 0===t?void 0:t.platform)||window.navigator.platform)}function eh(e){let t=null;return()=>(null==t&&(t=e()),t)}let em=eh(function(){return ep(/^Mac/i)}),eg=eh(function(){return ep(/^iPhone/i)}),eb=eh(function(){return ep(/^iPad/i)||em()&&navigator.maxTouchPoints>1}),ev=eh(function(){return eg()||eb()});eh(function(){return em()||ev()}),eh(function(){return ef(/AppleWebKit/i)&&!ey()});let ey=eh(function(){return ef(/Chrome/i)}),ew=eh(function(){return ef(/Android/i)});eh(function(){return ef(/Firefox/i)});let ex=e=>{var t;return null!==(t=null==e?void 0:e.ownerDocument)&&void 0!==t?t:document},eE=e=>e&&"window"in e&&e.window===e?e:ex(e).defaultView||window,eS=null,ek=new Set,eT=new Map,eC=!1,eP=!1;function eM(e,t){for(let n of ek)n(e,t)}function eR(e){eC=!0,e.metaKey||!em()&&e.altKey||e.ctrlKey||"Control"===e.key||"Shift"===e.key||"Meta"===e.key||(eS="keyboard",eM("keyboard",e))}function ej(e){eS="pointer",("mousedown"===e.type||"pointerdown"===e.type)&&(eC=!0,eM("pointer",e))}function eO(e){(0===e.mozInputSource&&e.isTrusted||(ew()&&e.pointerType?"click"===e.type&&1===e.buttons:0===e.detail&&!e.pointerType))&&(eC=!0,eS="virtual")}function eL(e){e.target!==window&&e.target!==document&&e.isTrusted&&(eC||eP||(eS="virtual",eM("virtual",e)),eC=!1,eP=!1)}function eA(){eC=!1,eP=!0}function e_(e){if("undefined"==typeof window||"undefined"==typeof document||eT.get(eE(e)))return;let t=eE(e),n=ex(e),r=t.HTMLElement.prototype.focus;t.HTMLElement.prototype.focus=function(){eC=!0,r.apply(this,arguments)},n.addEventListener("keydown",eR,!0),n.addEventListener("keyup",eR,!0),n.addEventListener("click",eO,!0),t.addEventListener("focus",eL,!0),t.addEventListener("blur",eA,!1),"undefined"!=typeof PointerEvent&&(n.addEventListener("pointerdown",ej,!0),n.addEventListener("pointermove",ej,!0),n.addEventListener("pointerup",ej,!0)),t.addEventListener("beforeunload",()=>{eN(e)},{once:!0}),eT.set(t,{focus:r})}let eN=(e,t)=>{let n=eE(e),r=ex(e);t&&r.removeEventListener("DOMContentLoaded",t),eT.has(n)&&(n.HTMLElement.prototype.focus=eT.get(n).focus,r.removeEventListener("keydown",eR,!0),r.removeEventListener("keyup",eR,!0),r.removeEventListener("click",eO,!0),n.removeEventListener("focus",eL,!0),n.removeEventListener("blur",eA,!1),"undefined"!=typeof PointerEvent&&(r.removeEventListener("pointerdown",ej,!0),r.removeEventListener("pointermove",ej,!0),r.removeEventListener("pointerup",ej,!0)),eT.delete(n))};"undefined"!=typeof document&&function(e){let t;let n=ex(void 0);"loading"!==n.readyState?e_(void 0):(t=()=>{e_(void 0)},n.addEventListener("DOMContentLoaded",t)),()=>eN(e,t)}();let e$=(e=document)=>e.activeElement,eI=new Map,eD=new Set;function eW(){if("undefined"==typeof window)return;function e(e){return"propertyName"in e}let t=n=>{if(!e(n)||!n.target)return;let r=eI.get(n.target);if(r&&(r.delete(n.propertyName),0===r.size&&(n.target.removeEventListener("transitioncancel",t),eI.delete(n.target)),0===eI.size)){for(let e of eD)e();eD.clear()}};document.body.addEventListener("transitionrun",n=>{if(!e(n)||!n.target)return;let r=eI.get(n.target);r||(r=new Set,eI.set(n.target,r),n.target.addEventListener("transitioncancel",t,{once:!0})),r.add(n.propertyName)}),document.body.addEventListener("transitionend",t)}function eH(e){if(function(){if(null==eF){eF=!1;try{document.createElement("div").focus({get preventScroll(){return eF=!0,!0}})}catch{}}return eF}())e.focus({preventScroll:!0});else{let t=function(e){let t=e.parentNode,n=[],r=document.scrollingElement||document.documentElement;for(;t instanceof HTMLElement&&t!==r;)(t.offsetHeight<t.scrollHeight||t.offsetWidth<t.scrollWidth)&&n.push({element:t,scrollTop:t.scrollTop,scrollLeft:t.scrollLeft}),t=t.parentNode;return r instanceof HTMLElement&&n.push({element:r,scrollTop:r.scrollTop,scrollLeft:r.scrollLeft}),n}(e);e.focus(),function(e){for(let{element:t,scrollTop:n,scrollLeft:r}of e)t.scrollTop=n,t.scrollLeft=r}(t)}}"undefined"!=typeof document&&("loading"!==document.readyState?eW():document.addEventListener("DOMContentLoaded",eW));let eF=null;function ez(e){if(!e)return;let t=!0;return n=>{e({...n,preventDefault(){n.preventDefault()},isDefaultPrevented:()=>n.isDefaultPrevented(),stopPropagation(){t=!0},continuePropagation(){t=!1},isPropagationStopped:()=>t}),t&&n.stopPropagation()}}"undefined"!=typeof window&&window.document&&window.document.createElement;let eB=new Map;function eq(...e){let t={...e[0]};for(let n=1;n<e.length;n++){let r=e[n];for(let e in r){let n=t[e],o=r[e];"function"==typeof n&&"function"==typeof o&&"o"===e[0]&&"n"===e[1]&&e.charCodeAt(2)>=65&&90>=e.charCodeAt(2)?t[e]=function(...e){return(...t)=>{for(let n of e)"function"==typeof n&&n(...t)}}(n,o):("className"===e||"UNSAFE_className"===e)&&"string"==typeof n&&"string"==typeof o?t[e]=(0,L.Z)(n,o):"id"===e&&n&&o?t.id=function(e,t){if(e===t)return e;let n=eB.get(e);if(n)return n.forEach(e=>e.current=t),t;let r=eB.get(t);return r?(r.forEach(t=>t.current=e),e):t}(n,o):t[e]=void 0!==o?o:n}}return t}"undefined"!=typeof FinalizationRegistry&&new FinalizationRegistry(e=>{eB.delete(e)});let eV=h.createContext(null),eU=new Set(["id"]),eY=new Set(["aria-label","aria-labelledby","aria-describedby","aria-details"]),eG=new Set(["href","hrefLang","target","rel","download","ping","referrerPolicy"]),eX=/^(data-.*)$/;function eK(e,t=-1/0,n=1/0){return Math.min(Math.max(e,t),n)}let eZ={top:"top",bottom:"top",left:"left",right:"left"},eJ={top:"bottom",bottom:"top",left:"right",right:"left"},eQ={top:"left",left:"top"},e0={top:"height",left:"width"},e1={width:"totalWidth",height:"totalHeight"},e2={},e4="undefined"!=typeof document?window.visualViewport:null;function e5(e){var t,n,r,o,i;let l=0,a=0,s=0,u=0,c=0,d=0,f={},p=(null!==(t=null==e4?void 0:e4.scale)&&void 0!==t?t:1)>1;if("BODY"===e.tagName){let t=document.documentElement;s=t.clientWidth,u=t.clientHeight,l=null!==(n=null==e4?void 0:e4.width)&&void 0!==n?n:s,a=null!==(r=null==e4?void 0:e4.height)&&void 0!==r?r:u,f.top=t.scrollTop||e.scrollTop,f.left=t.scrollLeft||e.scrollLeft,e4&&(c=e4.offsetTop,d=e4.offsetLeft)}else({width:l,height:a,top:c,left:d}=e7(e)),f.top=e.scrollTop,f.left=e.scrollLeft,s=l,u=a;return F()&&("BODY"===e.tagName||"HTML"===e.tagName)&&p&&(f.top=0,f.left=0,c=null!==(o=null==e4?void 0:e4.pageTop)&&void 0!==o?o:0,d=null!==(i=null==e4?void 0:e4.pageLeft)&&void 0!==i?i:0),{width:l,height:a,totalWidth:s,totalHeight:u,scroll:f,top:c,left:d}}function e6(e,t,n,r,o,i,l){var a;let s=null!==(a=o.scroll[e])&&void 0!==a?a:0,u=r[e0[e]],c=r.scroll[eZ[e]]+i,d=u+r.scroll[eZ[e]]-i,f=t-s+l[e]-r[eZ[e]],p=t-s+n+l[e]-r[eZ[e]];return f<c?c-f:p>d?Math.max(d-p,c-f):0}function e3(e){if(e2[e])return e2[e];let[t,n]=e.split(" "),r=eZ[t]||"right",o=eQ[r];eZ[n]||(n="center");let i=e0[r],l=e0[o];return e2[e]={placement:t,crossPlacement:n,axis:r,crossAxis:o,size:i,crossSize:l},e2[e]}function e9(e,t,n,r,o,i,l,a,s,u){var c,d,f,p,h;let{placement:m,crossPlacement:g,axis:b,crossAxis:v,size:y,crossSize:w}=r,x={};x[v]=null!==(c=e[v])&&void 0!==c?c:0,"center"===g?x[v]+=((null!==(d=e[w])&&void 0!==d?d:0)-(null!==(f=n[w])&&void 0!==f?f:0))/2:g!==v&&(x[v]+=(null!==(p=e[w])&&void 0!==p?p:0)-(null!==(h=n[w])&&void 0!==h?h:0)),x[v]+=i;let E=e[v]-n[w]+s+u,S=e[v]+e[w]-s-u;if(x[v]=eK(x[v],E,S),m===b){let n=a?l[y]:t[e1[y]];x[eJ[b]]=Math.floor(n-e[b]+o)}else x[b]=Math.floor(e[b]+e[y]+o);return x}function e8(e,t,n,r,o,i){var l,a,s;let{placement:u,axis:c,size:d}=i;return u===c?Math.max(0,n[c]-e[c]-(null!==(l=e.scroll[c])&&void 0!==l?l:0)+t[c]-(null!==(a=r[c])&&void 0!==a?a:0)-r[eJ[c]]-o):Math.max(0,e[d]+e[c]+e.scroll[c]-t[c]-n[c]-n[d]-(null!==(s=r[c])&&void 0!==s?s:0)-r[eJ[c]]-o)}function e7(e){let{top:t,left:n,width:r,height:o}=e.getBoundingClientRect(),{scrollTop:i,scrollLeft:l,clientTop:a,clientLeft:s}=document.documentElement;return{top:t+i-a,left:n+l-s,width:r,height:o}}function te(e,t){let n,r=window.getComputedStyle(e);if("fixed"===r.position){let{top:t,left:r,width:o,height:i}=e.getBoundingClientRect();n={top:t,left:r,width:o,height:i}}else{n=e7(e);let r=e7(t),o=window.getComputedStyle(t);r.top+=(parseInt(o.borderTopWidth,10)||0)-t.scrollTop,r.left+=(parseInt(o.borderLeftWidth,10)||0)-t.scrollLeft,n.top-=r.top,n.left-=r.left}return n.top-=parseInt(r.marginTop,10)||0,n.left-=parseInt(r.marginLeft,10)||0,n}function tt(e){let t=window.getComputedStyle(e);return"none"!==t.transform||/transform|perspective/.test(t.willChange)||"none"!==t.filter||"paint"===t.contain||"backdropFilter"in t&&"none"!==t.backdropFilter||"WebkitBackdropFilter"in t&&"none"!==t.WebkitBackdropFilter}let tn=new WeakMap;function tr(e){let{ref:t,box:n,onResize:r}=e;(0,h.useEffect)(()=>{let e=null==t?void 0:t.current;if(e){if(!(void 0!==window.ResizeObserver))return window.addEventListener("resize",r,!1),()=>{window.removeEventListener("resize",r,!1)};{let t=new window.ResizeObserver(e=>{e.length&&r()});return t.observe(e,{box:n}),()=>{e&&t.unobserve(e)}}}},[r,t,n])}let to=new Set(["Arab","Syrc","Samr","Mand","Thaa","Mend","Nkoo","Adlm","Rohg","Hebr"]),ti=new Set(["ae","ar","arc","bcc","bqi","ckb","dv","fa","glk","he","ku","mzn","nqo","pnb","ps","sd","ug","ur","yi"]),tl=Symbol.for("react-aria.i18n.locale");function ta(){let e="undefined"!=typeof window&&window[tl]||"undefined"!=typeof navigator&&(navigator.language||navigator.userLanguage)||"en-US";try{Intl.DateTimeFormat.supportedLocalesOf([e])}catch{e="en-US"}return{locale:e,direction:!function(e){if(Intl.Locale){let t=new Intl.Locale(e).maximize(),n="function"==typeof t.getTextInfo?t.getTextInfo():t.textInfo;if(n)return"rtl"===n.direction;if(t.script)return to.has(t.script)}let t=e.split("-")[0];return ti.has(t)}(e)?"ltr":"rtl"}}let ts=ta(),tu=new Set;function tc(){for(let e of(ts=ta(),tu))e(ts)}let td=h.createContext(null),tf="undefined"!=typeof document?window.visualViewport:null;function tp(e,t){return!!e&&!!t&&t.some(t=>t.contains(e))}class th{get size(){return this.fastMap.size}getTreeNode(e){return this.fastMap.get(e)}addTreeNode(e,t,n){let r=this.fastMap.get(null!=t?t:null);if(!r)return;let o=new tm({scopeRef:e});r.addChild(o),o.parent=r,this.fastMap.set(e,o),n&&(o.nodeToRestore=n)}addNode(e){this.fastMap.set(e.scopeRef,e)}removeTreeNode(e){if(null===e)return;let t=this.fastMap.get(e);if(!t)return;let n=t.parent;for(let e of this.traverse())e!==t&&t.nodeToRestore&&e.nodeToRestore&&t.scopeRef&&t.scopeRef.current&&tp(e.nodeToRestore,t.scopeRef.current)&&(e.nodeToRestore=t.nodeToRestore);let r=t.children;n&&(n.removeChild(t),r.size>0&&r.forEach(e=>n&&n.addChild(e))),this.fastMap.delete(t.scopeRef)}*traverse(e=this.root){if(null!=e.scopeRef&&(yield e),e.children.size>0)for(let t of e.children)yield*this.traverse(t)}clone(){var e,t;let n=new th;for(let r of this.traverse())n.addTreeNode(r.scopeRef,null!==(t=null===(e=r.parent)||void 0===e?void 0:e.scopeRef)&&void 0!==t?t:null,r.nodeToRestore);return n}constructor(){this.fastMap=new Map,this.root=new tm({scopeRef:null}),this.fastMap.set(null,this.root)}}class tm{addChild(e){this.children.add(e),e.parent=this}removeChild(e){this.children.delete(e),e.parent=void 0}constructor(e){this.children=new Set,this.contain=!1,this.scopeRef=e.scopeRef}}let tg=new th;function tb(e){let t=(0,h.useRef)(null);return w(()=>{t.current=e},[e]),(0,h.useCallback)((...e)=>{let n=t.current;return null==n?void 0:n(...e)},[])}function tv(e,t){if(e.button>0)return!1;if(e.target){let t=e.target.ownerDocument;if(!t||!t.documentElement.contains(e.target)||e.target.closest("[data-react-aria-top-layer]"))return!1}return t.current&&!t.current.contains(e.target)}class ty{isDefaultPrevented(){return this.nativeEvent.defaultPrevented}preventDefault(){this.defaultPrevented=!0,this.nativeEvent.preventDefault()}stopPropagation(){this.nativeEvent.stopPropagation(),this.isPropagationStopped=()=>!0}isPropagationStopped(){return!1}persist(){}constructor(e,t){this.nativeEvent=t,this.target=t.target,this.currentTarget=t.currentTarget,this.relatedTarget=t.relatedTarget,this.bubbles=t.bubbles,this.cancelable=t.cancelable,this.defaultPrevented=t.defaultPrevented,this.eventPhase=t.eventPhase,this.isTrusted=t.isTrusted,this.timeStamp=t.timeStamp,this.type=e}}let tw=[];var[tx,tE]=function(e={}){let{strict:t=!0,errorMessage:n="useContext: `context` is undefined. Seems you forgot to wrap component within the Provider",name:r}=e,o=h.createContext(void 0);return o.displayName=r,[o.Provider,function e(){var r;let i=h.useContext(o);if(!i&&t){let t=Error(n);throw t.name="ContextError",null==(r=Error.captureStackTrace)||r.call(Error,t,e),t}return i},o]}({name:"ProviderContext",strict:!1}),tS=(e,t,n=!0)=>{if(!t)return[e,{}];let r=t.reduce((t,n)=>n in e?{...t,[n]:e[n]}:t,{});return n?[Object.keys(e).filter(e=>!t.includes(e)).reduce((t,n)=>({...t,[n]:e[n]}),{}),r]:[e,r]},tk=["small","medium","large"],tT={theme:{opacity:["disabled"],spacing:["divider"],borderWidth:tk,borderRadius:tk},classGroups:{shadow:[{shadow:tk}],"font-size":[{text:["tiny",...tk]}],"bg-image":["bg-stripe-gradient-default","bg-stripe-gradient-primary","bg-stripe-gradient-secondary","bg-stripe-gradient-success","bg-stripe-gradient-warning","bg-stripe-gradient-danger"]}},tC=e=>"boolean"==typeof e?`${e}`:0===e?"0":e,tP=e=>!e||"object"!=typeof e||0===Object.keys(e).length,tM=(e,t)=>JSON.stringify(e)===JSON.stringify(t);function tR(e){let t=[];return function e(t,n){t.forEach(function(t){Array.isArray(t)?e(t,n):n.push(t)})}(e,t),t}var tj=(...e)=>tR(e).filter(Boolean),tO=(e,t)=>{let n={},r=Object.keys(e),o=Object.keys(t);for(let i of r)if(o.includes(i)){let r=e[i],o=t[i];"object"==typeof r&&"object"==typeof o?n[i]=tO(r,o):Array.isArray(r)||Array.isArray(o)?n[i]=tj(o,r):n[i]=o+" "+r}else n[i]=e[i];for(let e of o)r.includes(e)||(n[e]=t[e]);return n},tL=e=>e&&"string"==typeof e?e.replace(/\s+/g," ").trim():e,tA=/^\[(.+)\]$/;function t_(e,t){var n=e;return t.split("-").forEach(function(e){n.nextPart.has(e)||n.nextPart.set(e,{nextPart:new Map,validators:[]}),n=n.nextPart.get(e)}),n}var tN=/\s+/;function t$(){for(var e,t,n=0,r="";n<arguments.length;)(e=arguments[n++])&&(t=function e(t){if("string"==typeof t)return t;for(var n,r="",o=0;o<t.length;o++)t[o]&&(n=e(t[o]))&&(r&&(r+=" "),r+=n);return r}(e))&&(r&&(r+=" "),r+=t);return r}function tI(){for(var e,t,n,r=arguments.length,o=Array(r),i=0;i<r;i++)o[i]=arguments[i];var l=function(r){var i=o[0];return t=(e=function(e){var t,n,r,o,i,l,a,s,u,c,d;return{cache:function(e){if(e<1)return{get:function(){},set:function(){}};var t=0,n=new Map,r=new Map;function o(o,i){n.set(o,i),++t>e&&(t=0,r=n,n=new Map)}return{get:function(e){var t=n.get(e);return void 0!==t?t:void 0!==(t=r.get(e))?(o(e,t),t):void 0},set:function(e,t){n.has(e)?n.set(e,t):o(e,t)}}}(e.cacheSize),splitModifiers:(n=1===(t=e.separator||":").length,r=t[0],o=t.length,function(e){for(var i,l=[],a=0,s=0,u=0;u<e.length;u++){var c=e[u];if(0===a){if(c===r&&(n||e.slice(u,u+o)===t)){l.push(e.slice(s,u)),s=u+o;continue}if("/"===c){i=u;continue}}"["===c?a++:"]"===c&&a--}var d=0===l.length?e:e.substring(s),f=d.startsWith("!"),p=f?d.substring(1):d;return{modifiers:l,hasImportantModifier:f,baseClassName:p,maybePostfixModifierPosition:i&&i>s?i-s:void 0}}),...(s=e.theme,u=e.prefix,c={nextPart:new Map,validators:[]},(d=Object.entries(e.classGroups),u?d.map(function(e){return[e[0],e[1].map(function(e){return"string"==typeof e?u+e:"object"==typeof e?Object.fromEntries(Object.entries(e).map(function(e){return[u+e[0],e[1]]})):e})]}):d).forEach(function(e){var t=e[0];(function e(t,n,r,o){t.forEach(function(t){if("string"==typeof t){(""===t?n:t_(n,t)).classGroupId=r;return}if("function"==typeof t){if(t.isThemeGetter){e(t(o),n,r,o);return}n.validators.push({validator:t,classGroupId:r});return}Object.entries(t).forEach(function(t){var i=t[0];e(t[1],t_(n,i),r,o)})})})(e[1],c,t,s)}),i=e.conflictingClassGroups,a=void 0===(l=e.conflictingClassGroupModifiers)?{}:l,{getClassGroupId:function(e){var t=e.split("-");return""===t[0]&&1!==t.length&&t.shift(),function e(t,n){if(0===t.length)return n.classGroupId;var r=t[0],o=n.nextPart.get(r),i=o?e(t.slice(1),o):void 0;if(i)return i;if(0!==n.validators.length){var l=t.join("-");return n.validators.find(function(e){return(0,e.validator)(l)})?.classGroupId}}(t,c)||function(e){if(tA.test(e)){var t=tA.exec(e)[1],n=t?.substring(0,t.indexOf(":"));if(n)return"arbitrary.."+n}}(e)},getConflictingClassGroupIds:function(e,t){var n=i[e]||[];return t&&a[e]?[].concat(n,a[e]):n}})}}(o.slice(1).reduce(function(e,t){return t(e)},i()))).cache.get,n=e.cache.set,l=a,a(r)};function a(r){var o,i,l,a,s,u=t(r);if(u)return u;var c=(i=(o=e).splitModifiers,l=o.getClassGroupId,a=o.getConflictingClassGroupIds,s=new Set,r.trim().split(tN).map(function(e){var t=i(e),n=t.modifiers,r=t.hasImportantModifier,o=t.baseClassName,a=t.maybePostfixModifierPosition,s=l(a?o.substring(0,a):o),u=!!a;if(!s){if(!a||!(s=l(o)))return{isTailwindClass:!1,originalClassName:e};u=!1}var c=(function(e){if(e.length<=1)return e;var t=[],n=[];return e.forEach(function(e){"["===e[0]?(t.push.apply(t,n.sort().concat([e])),n=[]):n.push(e)}),t.push.apply(t,n.sort()),t})(n).join(":");return{isTailwindClass:!0,modifierId:r?c+"!":c,classGroupId:s,originalClassName:e,hasPostfixModifier:u}}).reverse().filter(function(e){if(!e.isTailwindClass)return!0;var t=e.modifierId,n=e.classGroupId,r=e.hasPostfixModifier,o=t+n;return!s.has(o)&&(s.add(o),a(n,r).forEach(function(e){return s.add(t+e)}),!0)}).reverse().map(function(e){return e.originalClassName}).join(" "));return n(r,c),c}return function(){return l(t$.apply(null,arguments))}}function tD(e){var t=function(t){return t[e]||[]};return t.isThemeGetter=!0,t}var tW=/^\[(?:([a-z-]+):)?(.+)\]$/i,tH=/^\d+\/\d+$/,tF=new Set(["px","full","screen"]),tz=/^(\d+(\.\d+)?)?(xs|sm|md|lg|xl)$/,tB=/\d+(%|px|r?em|[sdl]?v([hwib]|min|max)|pt|pc|in|cm|mm|cap|ch|ex|r?lh|cq(w|h|i|b|min|max))|\b(calc|min|max|clamp)\(.+\)|^0$/,tq=/^-?((\d+)?\.?(\d+)[a-z]+|0)_-?((\d+)?\.?(\d+)[a-z]+|0)/;function tV(e){return tZ(e)||tF.has(e)||tH.test(e)||tU(e)}function tU(e){return t5(e,"length",t6)}function tY(e){return t5(e,"size",t3)}function tG(e){return t5(e,"position",t3)}function tX(e){return t5(e,"url",t9)}function tK(e){return t5(e,"number",tZ)}function tZ(e){return!Number.isNaN(Number(e))}function tJ(e){return e.endsWith("%")&&tZ(e.slice(0,-1))}function tQ(e){return t8(e)||t5(e,"number",t8)}function t0(e){return tW.test(e)}function t1(){return!0}function t2(e){return tz.test(e)}function t4(e){return t5(e,"",t7)}function t5(e,t,n){var r=tW.exec(e);return!!r&&(r[1]?r[1]===t:n(r[2]))}function t6(e){return tB.test(e)}function t3(){return!1}function t9(e){return e.startsWith("url(")}function t8(e){return Number.isInteger(Number(e))}function t7(e){return tq.test(e)}function ne(){var e=tD("colors"),t=tD("spacing"),n=tD("blur"),r=tD("brightness"),o=tD("borderColor"),i=tD("borderRadius"),l=tD("borderSpacing"),a=tD("borderWidth"),s=tD("contrast"),u=tD("grayscale"),c=tD("hueRotate"),d=tD("invert"),f=tD("gap"),p=tD("gradientColorStops"),h=tD("gradientColorStopPositions"),m=tD("inset"),g=tD("margin"),b=tD("opacity"),v=tD("padding"),y=tD("saturate"),w=tD("scale"),x=tD("sepia"),E=tD("skew"),S=tD("space"),k=tD("translate"),T=function(){return["auto","contain","none"]},C=function(){return["auto","hidden","clip","visible","scroll"]},P=function(){return["auto",t0,t]},M=function(){return[t0,t]},R=function(){return["",tV]},j=function(){return["auto",tZ,t0]},O=function(){return["bottom","center","left","left-bottom","left-top","right","right-bottom","right-top","top"]},L=function(){return["solid","dashed","dotted","double","none"]},A=function(){return["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity","plus-lighter"]},_=function(){return["start","end","center","between","around","evenly","stretch"]},N=function(){return["","0",t0]},$=function(){return["auto","avoid","all","avoid-page","page","left","right","column"]},I=function(){return[tZ,tK]},D=function(){return[tZ,t0]};return{cacheSize:500,theme:{colors:[t1],spacing:[tV],blur:["none","",t2,t0],brightness:I(),borderColor:[e],borderRadius:["none","","full",t2,t0],borderSpacing:M(),borderWidth:R(),contrast:I(),grayscale:N(),hueRotate:D(),invert:N(),gap:M(),gradientColorStops:[e],gradientColorStopPositions:[tJ,tU],inset:P(),margin:P(),opacity:I(),padding:M(),saturate:I(),scale:I(),sepia:N(),skew:D(),space:M(),translate:M()},classGroups:{aspect:[{aspect:["auto","square","video",t0]}],container:["container"],columns:[{columns:[t2]}],"break-after":[{"break-after":$()}],"break-before":[{"break-before":$()}],"break-inside":[{"break-inside":["auto","avoid","avoid-page","avoid-column"]}],"box-decoration":[{"box-decoration":["slice","clone"]}],box:[{box:["border","content"]}],display:["block","inline-block","inline","flex","inline-flex","table","inline-table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row-group","table-row","flow-root","grid","inline-grid","contents","list-item","hidden"],float:[{float:["right","left","none"]}],clear:[{clear:["left","right","both","none"]}],isolation:["isolate","isolation-auto"],"object-fit":[{object:["contain","cover","fill","none","scale-down"]}],"object-position":[{object:[].concat(O(),[t0])}],overflow:[{overflow:C()}],"overflow-x":[{"overflow-x":C()}],"overflow-y":[{"overflow-y":C()}],overscroll:[{overscroll:T()}],"overscroll-x":[{"overscroll-x":T()}],"overscroll-y":[{"overscroll-y":T()}],position:["static","fixed","absolute","relative","sticky"],inset:[{inset:[m]}],"inset-x":[{"inset-x":[m]}],"inset-y":[{"inset-y":[m]}],start:[{start:[m]}],end:[{end:[m]}],top:[{top:[m]}],right:[{right:[m]}],bottom:[{bottom:[m]}],left:[{left:[m]}],visibility:["visible","invisible","collapse"],z:[{z:["auto",tQ]}],basis:[{basis:P()}],"flex-direction":[{flex:["row","row-reverse","col","col-reverse"]}],"flex-wrap":[{flex:["wrap","wrap-reverse","nowrap"]}],flex:[{flex:["1","auto","initial","none",t0]}],grow:[{grow:N()}],shrink:[{shrink:N()}],order:[{order:["first","last","none",tQ]}],"grid-cols":[{"grid-cols":[t1]}],"col-start-end":[{col:["auto",{span:["full",tQ]},t0]}],"col-start":[{"col-start":j()}],"col-end":[{"col-end":j()}],"grid-rows":[{"grid-rows":[t1]}],"row-start-end":[{row:["auto",{span:[tQ]},t0]}],"row-start":[{"row-start":j()}],"row-end":[{"row-end":j()}],"grid-flow":[{"grid-flow":["row","col","dense","row-dense","col-dense"]}],"auto-cols":[{"auto-cols":["auto","min","max","fr",t0]}],"auto-rows":[{"auto-rows":["auto","min","max","fr",t0]}],gap:[{gap:[f]}],"gap-x":[{"gap-x":[f]}],"gap-y":[{"gap-y":[f]}],"justify-content":[{justify:["normal"].concat(_())}],"justify-items":[{"justify-items":["start","end","center","stretch"]}],"justify-self":[{"justify-self":["auto","start","end","center","stretch"]}],"align-content":[{content:["normal"].concat(_(),["baseline"])}],"align-items":[{items:["start","end","center","baseline","stretch"]}],"align-self":[{self:["auto","start","end","center","stretch","baseline"]}],"place-content":[{"place-content":[].concat(_(),["baseline"])}],"place-items":[{"place-items":["start","end","center","baseline","stretch"]}],"place-self":[{"place-self":["auto","start","end","center","stretch"]}],p:[{p:[v]}],px:[{px:[v]}],py:[{py:[v]}],ps:[{ps:[v]}],pe:[{pe:[v]}],pt:[{pt:[v]}],pr:[{pr:[v]}],pb:[{pb:[v]}],pl:[{pl:[v]}],m:[{m:[g]}],mx:[{mx:[g]}],my:[{my:[g]}],ms:[{ms:[g]}],me:[{me:[g]}],mt:[{mt:[g]}],mr:[{mr:[g]}],mb:[{mb:[g]}],ml:[{ml:[g]}],"space-x":[{"space-x":[S]}],"space-x-reverse":["space-x-reverse"],"space-y":[{"space-y":[S]}],"space-y-reverse":["space-y-reverse"],w:[{w:["auto","min","max","fit",t0,t]}],"min-w":[{"min-w":["min","max","fit",t0,tV]}],"max-w":[{"max-w":["0","none","full","min","max","fit","prose",{screen:[t2]},t2,t0]}],h:[{h:[t0,t,"auto","min","max","fit"]}],"min-h":[{"min-h":["min","max","fit",t0,tV]}],"max-h":[{"max-h":[t0,t,"min","max","fit"]}],"font-size":[{text:["base",t2,tU]}],"font-smoothing":["antialiased","subpixel-antialiased"],"font-style":["italic","not-italic"],"font-weight":[{font:["thin","extralight","light","normal","medium","semibold","bold","extrabold","black",tK]}],"font-family":[{font:[t1]}],"fvn-normal":["normal-nums"],"fvn-ordinal":["ordinal"],"fvn-slashed-zero":["slashed-zero"],"fvn-figure":["lining-nums","oldstyle-nums"],"fvn-spacing":["proportional-nums","tabular-nums"],"fvn-fraction":["diagonal-fractions","stacked-fractons"],tracking:[{tracking:["tighter","tight","normal","wide","wider","widest",t0]}],"line-clamp":[{"line-clamp":["none",tZ,tK]}],leading:[{leading:["none","tight","snug","normal","relaxed","loose",t0,tV]}],"list-image":[{"list-image":["none",t0]}],"list-style-type":[{list:["none","disc","decimal",t0]}],"list-style-position":[{list:["inside","outside"]}],"placeholder-color":[{placeholder:[e]}],"placeholder-opacity":[{"placeholder-opacity":[b]}],"text-alignment":[{text:["left","center","right","justify","start","end"]}],"text-color":[{text:[e]}],"text-opacity":[{"text-opacity":[b]}],"text-decoration":["underline","overline","line-through","no-underline"],"text-decoration-style":[{decoration:[].concat(L(),["wavy"])}],"text-decoration-thickness":[{decoration:["auto","from-font",tV]}],"underline-offset":[{"underline-offset":["auto",t0,tV]}],"text-decoration-color":[{decoration:[e]}],"text-transform":["uppercase","lowercase","capitalize","normal-case"],"text-overflow":["truncate","text-ellipsis","text-clip"],indent:[{indent:M()}],"vertical-align":[{align:["baseline","top","middle","bottom","text-top","text-bottom","sub","super",t0]}],whitespace:[{whitespace:["normal","nowrap","pre","pre-line","pre-wrap","break-spaces"]}],break:[{break:["normal","words","all","keep"]}],hyphens:[{hyphens:["none","manual","auto"]}],content:[{content:["none",t0]}],"bg-attachment":[{bg:["fixed","local","scroll"]}],"bg-clip":[{"bg-clip":["border","padding","content","text"]}],"bg-opacity":[{"bg-opacity":[b]}],"bg-origin":[{"bg-origin":["border","padding","content"]}],"bg-position":[{bg:[].concat(O(),[tG])}],"bg-repeat":[{bg:["no-repeat",{repeat:["","x","y","round","space"]}]}],"bg-size":[{bg:["auto","cover","contain",tY]}],"bg-image":[{bg:["none",{"gradient-to":["t","tr","r","br","b","bl","l","tl"]},tX]}],"bg-color":[{bg:[e]}],"gradient-from-pos":[{from:[h]}],"gradient-via-pos":[{via:[h]}],"gradient-to-pos":[{to:[h]}],"gradient-from":[{from:[p]}],"gradient-via":[{via:[p]}],"gradient-to":[{to:[p]}],rounded:[{rounded:[i]}],"rounded-s":[{"rounded-s":[i]}],"rounded-e":[{"rounded-e":[i]}],"rounded-t":[{"rounded-t":[i]}],"rounded-r":[{"rounded-r":[i]}],"rounded-b":[{"rounded-b":[i]}],"rounded-l":[{"rounded-l":[i]}],"rounded-ss":[{"rounded-ss":[i]}],"rounded-se":[{"rounded-se":[i]}],"rounded-ee":[{"rounded-ee":[i]}],"rounded-es":[{"rounded-es":[i]}],"rounded-tl":[{"rounded-tl":[i]}],"rounded-tr":[{"rounded-tr":[i]}],"rounded-br":[{"rounded-br":[i]}],"rounded-bl":[{"rounded-bl":[i]}],"border-w":[{border:[a]}],"border-w-x":[{"border-x":[a]}],"border-w-y":[{"border-y":[a]}],"border-w-s":[{"border-s":[a]}],"border-w-e":[{"border-e":[a]}],"border-w-t":[{"border-t":[a]}],"border-w-r":[{"border-r":[a]}],"border-w-b":[{"border-b":[a]}],"border-w-l":[{"border-l":[a]}],"border-opacity":[{"border-opacity":[b]}],"border-style":[{border:[].concat(L(),["hidden"])}],"divide-x":[{"divide-x":[a]}],"divide-x-reverse":["divide-x-reverse"],"divide-y":[{"divide-y":[a]}],"divide-y-reverse":["divide-y-reverse"],"divide-opacity":[{"divide-opacity":[b]}],"divide-style":[{divide:L()}],"border-color":[{border:[o]}],"border-color-x":[{"border-x":[o]}],"border-color-y":[{"border-y":[o]}],"border-color-t":[{"border-t":[o]}],"border-color-r":[{"border-r":[o]}],"border-color-b":[{"border-b":[o]}],"border-color-l":[{"border-l":[o]}],"divide-color":[{divide:[o]}],"outline-style":[{outline:[""].concat(L())}],"outline-offset":[{"outline-offset":[t0,tV]}],"outline-w":[{outline:[tV]}],"outline-color":[{outline:[e]}],"ring-w":[{ring:R()}],"ring-w-inset":["ring-inset"],"ring-color":[{ring:[e]}],"ring-opacity":[{"ring-opacity":[b]}],"ring-offset-w":[{"ring-offset":[tV]}],"ring-offset-color":[{"ring-offset":[e]}],shadow:[{shadow:["","inner","none",t2,t4]}],"shadow-color":[{shadow:[t1]}],opacity:[{opacity:[b]}],"mix-blend":[{"mix-blend":A()}],"bg-blend":[{"bg-blend":A()}],filter:[{filter:["","none"]}],blur:[{blur:[n]}],brightness:[{brightness:[r]}],contrast:[{contrast:[s]}],"drop-shadow":[{"drop-shadow":["","none",t2,t0]}],grayscale:[{grayscale:[u]}],"hue-rotate":[{"hue-rotate":[c]}],invert:[{invert:[d]}],saturate:[{saturate:[y]}],sepia:[{sepia:[x]}],"backdrop-filter":[{"backdrop-filter":["","none"]}],"backdrop-blur":[{"backdrop-blur":[n]}],"backdrop-brightness":[{"backdrop-brightness":[r]}],"backdrop-contrast":[{"backdrop-contrast":[s]}],"backdrop-grayscale":[{"backdrop-grayscale":[u]}],"backdrop-hue-rotate":[{"backdrop-hue-rotate":[c]}],"backdrop-invert":[{"backdrop-invert":[d]}],"backdrop-opacity":[{"backdrop-opacity":[b]}],"backdrop-saturate":[{"backdrop-saturate":[y]}],"backdrop-sepia":[{"backdrop-sepia":[x]}],"border-collapse":[{border:["collapse","separate"]}],"border-spacing":[{"border-spacing":[l]}],"border-spacing-x":[{"border-spacing-x":[l]}],"border-spacing-y":[{"border-spacing-y":[l]}],"table-layout":[{table:["auto","fixed"]}],caption:[{caption:["top","bottom"]}],transition:[{transition:["none","all","","colors","opacity","shadow","transform",t0]}],duration:[{duration:D()}],ease:[{ease:["linear","in","out","in-out",t0]}],delay:[{delay:D()}],animate:[{animate:["none","spin","ping","pulse","bounce",t0]}],transform:[{transform:["","gpu","none"]}],scale:[{scale:[w]}],"scale-x":[{"scale-x":[w]}],"scale-y":[{"scale-y":[w]}],rotate:[{rotate:[tQ,t0]}],"translate-x":[{"translate-x":[k]}],"translate-y":[{"translate-y":[k]}],"skew-x":[{"skew-x":[E]}],"skew-y":[{"skew-y":[E]}],"transform-origin":[{origin:["center","top","top-right","right","bottom-right","bottom","bottom-left","left","top-left",t0]}],accent:[{accent:["auto",e]}],appearance:["appearance-none"],cursor:[{cursor:["auto","default","pointer","wait","text","move","help","not-allowed","none","context-menu","progress","cell","crosshair","vertical-text","alias","copy","no-drop","grab","grabbing","all-scroll","col-resize","row-resize","n-resize","e-resize","s-resize","w-resize","ne-resize","nw-resize","se-resize","sw-resize","ew-resize","ns-resize","nesw-resize","nwse-resize","zoom-in","zoom-out",t0]}],"caret-color":[{caret:[e]}],"pointer-events":[{"pointer-events":["none","auto"]}],resize:[{resize:["none","y","x",""]}],"scroll-behavior":[{scroll:["auto","smooth"]}],"scroll-m":[{"scroll-m":M()}],"scroll-mx":[{"scroll-mx":M()}],"scroll-my":[{"scroll-my":M()}],"scroll-ms":[{"scroll-ms":M()}],"scroll-me":[{"scroll-me":M()}],"scroll-mt":[{"scroll-mt":M()}],"scroll-mr":[{"scroll-mr":M()}],"scroll-mb":[{"scroll-mb":M()}],"scroll-ml":[{"scroll-ml":M()}],"scroll-p":[{"scroll-p":M()}],"scroll-px":[{"scroll-px":M()}],"scroll-py":[{"scroll-py":M()}],"scroll-ps":[{"scroll-ps":M()}],"scroll-pe":[{"scroll-pe":M()}],"scroll-pt":[{"scroll-pt":M()}],"scroll-pr":[{"scroll-pr":M()}],"scroll-pb":[{"scroll-pb":M()}],"scroll-pl":[{"scroll-pl":M()}],"snap-align":[{snap:["start","end","center","align-none"]}],"snap-stop":[{snap:["normal","always"]}],"snap-type":[{snap:["none","x","y","both"]}],"snap-strictness":[{snap:["mandatory","proximity"]}],touch:[{touch:["auto","none","pinch-zoom","manipulation",{pan:["x","left","right","y","up","down"]}]}],select:[{select:["none","text","all","auto"]}],"will-change":[{"will-change":["auto","scroll","contents","transform",t0]}],fill:[{fill:[e,"none"]}],"stroke-w":[{stroke:[tV,tK]}],stroke:[{stroke:[e,"none"]}],sr:["sr-only","not-sr-only"]},conflictingClassGroups:{overflow:["overflow-x","overflow-y"],overscroll:["overscroll-x","overscroll-y"],inset:["inset-x","inset-y","start","end","top","right","bottom","left"],"inset-x":["right","left"],"inset-y":["top","bottom"],flex:["basis","grow","shrink"],gap:["gap-x","gap-y"],p:["px","py","ps","pe","pt","pr","pb","pl"],px:["pr","pl"],py:["pt","pb"],m:["mx","my","ms","me","mt","mr","mb","ml"],mx:["mr","ml"],my:["mt","mb"],"font-size":["leading"],"fvn-normal":["fvn-ordinal","fvn-slashed-zero","fvn-figure","fvn-spacing","fvn-fraction"],"fvn-ordinal":["fvn-normal"],"fvn-slashed-zero":["fvn-normal"],"fvn-figure":["fvn-normal"],"fvn-spacing":["fvn-normal"],"fvn-fraction":["fvn-normal"],rounded:["rounded-s","rounded-e","rounded-t","rounded-r","rounded-b","rounded-l","rounded-ss","rounded-se","rounded-ee","rounded-es","rounded-tl","rounded-tr","rounded-br","rounded-bl"],"rounded-s":["rounded-ss","rounded-es"],"rounded-e":["rounded-se","rounded-ee"],"rounded-t":["rounded-tl","rounded-tr"],"rounded-r":["rounded-tr","rounded-br"],"rounded-b":["rounded-br","rounded-bl"],"rounded-l":["rounded-tl","rounded-bl"],"border-spacing":["border-spacing-x","border-spacing-y"],"border-w":["border-w-s","border-w-e","border-w-t","border-w-r","border-w-b","border-w-l"],"border-w-x":["border-w-r","border-w-l"],"border-w-y":["border-w-t","border-w-b"],"border-color":["border-color-t","border-color-r","border-color-b","border-color-l"],"border-color-x":["border-color-r","border-color-l"],"border-color-y":["border-color-t","border-color-b"],"scroll-m":["scroll-mx","scroll-my","scroll-ms","scroll-me","scroll-mt","scroll-mr","scroll-mb","scroll-ml"],"scroll-mx":["scroll-mr","scroll-ml"],"scroll-my":["scroll-mt","scroll-mb"],"scroll-p":["scroll-px","scroll-py","scroll-ps","scroll-pe","scroll-pt","scroll-pr","scroll-pb","scroll-pl"],"scroll-px":["scroll-pr","scroll-pl"],"scroll-py":["scroll-pt","scroll-pb"]},conflictingClassGroupModifiers:{"font-size":["leading"]}}}var nt=tI(ne),nn=Object.prototype.hasOwnProperty,nr=new Set(["string","number","boolean"]),no={twMerge:!0,twMergeConfig:{},responsiveVariants:!1},ni=e=>e||void 0,nl=(...e)=>ni(tR(e).filter(Boolean).join(" ")),na=null,ns={},nu=!1,nc=(...e)=>t=>t.twMerge?((!na||nu)&&(nu=!1,na=tP(ns)?nt:function(e){for(var t=arguments.length,n=Array(t>1?t-1:0),r=1;r<t;r++)n[r-1]=arguments[r];return"function"==typeof e?tI.apply(void 0,[ne,e].concat(n)):tI.apply(void 0,[function(){return function(e,t){for(var n in t)(function e(t,n,r){if(!nn.call(t,n)||nr.has(typeof r)||null===r){t[n]=r;return}if(Array.isArray(r)&&Array.isArray(t[n])){t[n]=t[n].concat(r);return}if("object"==typeof r&&"object"==typeof t[n]){if(null===t[n]){t[n]=r;return}for(var o in r)e(t[n],o,r[o])}})(e,n,t[n]);return e}(ne(),e)}].concat(n))}(ns)),ni(na(nl(e)))):nl(e),nd=(e,t)=>{for(let n in t)e.hasOwnProperty(n)?e[n]=nl(e[n],t[n]):e[n]=t[n];return e},nf=((e,t)=>{let{extend:n=null,slots:r={},variants:o={},compoundVariants:i=[],compoundSlots:l=[],defaultVariants:a={}}=e,s={...no,...t},u=null!=n&&n.base?nl(n.base,null==e?void 0:e.base):null==e?void 0:e.base,c=null!=n&&n.variants&&!tP(n.variants)?tO(o,n.variants):o,d=null!=n&&n.defaultVariants&&!tP(n.defaultVariants)?{...n.defaultVariants,...a}:a;tP(s.twMergeConfig)||tM(s.twMergeConfig,ns)||(nu=!0,ns=s.twMergeConfig);let f=tP(null==n?void 0:n.slots),p=tP(r)?{}:{base:nl(null==e?void 0:e.base,f&&(null==n?void 0:n.base)),...r},h=f?p:nd({...null==n?void 0:n.slots},tP(p)?{base:null==e?void 0:e.base}:p),m=e=>{if(tP(c)&&tP(r)&&f)return nc(u,null==e?void 0:e.class,null==e?void 0:e.className)(s);if(i&&!Array.isArray(i))throw TypeError(`The "compoundVariants" prop must be an array. Received: ${typeof i}`);if(l&&!Array.isArray(l))throw TypeError(`The "compoundSlots" prop must be an array. Received: ${typeof l}`);let t=(e,t,n=[],r)=>{let o=n;if("string"==typeof t)o=o.concat(tL(t).split(" ").map(t=>`${e}:${t}`));else if(Array.isArray(t))o=o.concat(t.reduce((t,n)=>t.concat(`${e}:${n}`),[]));else if("object"==typeof t&&"string"==typeof r){for(let n in t)if(t.hasOwnProperty(n)&&n===r){let i=t[n];if(i&&"string"==typeof i){let t=tL(i);o[r]?o[r]=o[r].concat(t.split(" ").map(t=>`${e}:${t}`)):o[r]=t.split(" ").map(t=>`${e}:${t}`)}else Array.isArray(i)&&i.length>0&&(o[r]=i.reduce((t,n)=>t.concat(`${e}:${n}`),[]))}}return o},o=(n,r=c,o=null,i=null)=>{var l;let a=r[n];if(!a||tP(a))return null;let u=null!=(l=null==i?void 0:i[n])?l:null==e?void 0:e[n];if(null===u)return null;let f=tC(u),p=Array.isArray(s.responsiveVariants)&&s.responsiveVariants.length>0||!0===s.responsiveVariants,h=null==d?void 0:d[n],m=[];if("object"==typeof f&&p)for(let[e,n]of Object.entries(f)){let r=a[n];if("initial"===e){h=n;continue}Array.isArray(s.responsiveVariants)&&!s.responsiveVariants.includes(e)||(m=t(e,r,m,o))}let g=a[f]||a[tC(h)];return"object"==typeof m&&"string"==typeof o&&m[o]?nd(m,g):m.length>0?(m.push(g),m):g},a=(e,t)=>{if(!c||"object"!=typeof c)return null;let n=[];for(let r in c){let i=o(r,c,e,t),l="base"===e&&"string"==typeof i?i:i&&i[e];l&&(n[n.length]=l)}return n},p={};for(let t in e)void 0!==e[t]&&(p[t]=e[t]);let m=(t,n)=>{var r;let o="object"==typeof(null==e?void 0:e[t])?{[t]:null==(r=e[t])?void 0:r.initial}:{};return{...d,...p,...o,...n}},g=(e=[],t)=>{let n=[];for(let{class:r,className:o,...i}of e){let e=!0;for(let[n,r]of Object.entries(i)){let o=m(n,t);if(Array.isArray(r)){if(!r.includes(o[n])){e=!1;break}}else if(o[n]!==r){e=!1;break}}e&&(r&&n.push(r),o&&n.push(o))}return n},b=e=>{let t=g(i,e);return tj(g(null==n?void 0:n.compoundVariants,e),t)},v=e=>{let t=b(e);if(!Array.isArray(t))return t;let n={};for(let e of t)if("string"==typeof e&&(n.base=nc(n.base,e)(s)),"object"==typeof e)for(let[t,r]of Object.entries(e))n[t]=nc(n[t],r)(s);return n},y=e=>{if(l.length<1)return null;let t={};for(let{slots:n=[],class:r,className:o,...i}of l){if(!tP(i)){let t=!0;for(let n of Object.keys(i)){let r=m(n,e)[n];if(void 0===r||(Array.isArray(i[n])?!i[n].includes(r):i[n]!==r)){t=!1;break}}if(!t)continue}for(let e of n)t[e]=t[e]||[],t[e].push([r,o])}return t};if(!tP(r)||!f){let e={};if("object"==typeof h&&!tP(h))for(let t of Object.keys(h))e[t]=e=>{var n,r;return nc(h[t],a(t,e),(null!=(n=v(e))?n:[])[t],(null!=(r=y(e))?r:[])[t],null==e?void 0:e.class,null==e?void 0:e.className)(s)};return e}return nc(u,c?Object.keys(c).map(e=>o(e,c)):null,b(),null==e?void 0:e.class,null==e?void 0:e.className)(s)};return m.variantKeys=(()=>{if(!(!c||"object"!=typeof c))return Object.keys(c)})(),m.extend=n,m.base=u,m.slots=h,m.variants=c,m.defaultVariants=d,m.compoundSlots=l,m.compoundVariants=i,m})({slots:{base:["z-0","relative","bg-transparent","before:content-['']","before:hidden","before:z-[-1]","before:absolute","before:rotate-45","before:w-2.5","before:h-2.5","before:rounded-sm","data-[arrow=true]:before:block","data-[placement=top]:before:-bottom-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=top]:before:left-1/2","data-[placement=top]:before:-translate-x-1/2","data-[placement=top-start]:before:-bottom-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=top-start]:before:left-3","data-[placement=top-end]:before:-bottom-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=top-end]:before:right-3","data-[placement=bottom]:before:-top-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=bottom]:before:left-1/2","data-[placement=bottom]:before:-translate-x-1/2","data-[placement=bottom-start]:before:-top-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=bottom-start]:before:left-3","data-[placement=bottom-end]:before:-top-[calc(theme(spacing.5)/4_-_1.5px)]","data-[placement=bottom-end]:before:right-3","data-[placement=left]:before:-right-[calc(theme(spacing.5)/4_-_2px)]","data-[placement=left]:before:top-1/2","data-[placement=left]:before:-translate-y-1/2","data-[placement=left-start]:before:-right-[calc(theme(spacing.5)/4_-_3px)]","data-[placement=left-start]:before:top-1/4","data-[placement=left-end]:before:-right-[calc(theme(spacing.5)/4_-_3px)]","data-[placement=left-end]:before:bottom-1/4","data-[placement=right]:before:-left-[calc(theme(spacing.5)/4_-_2px)]","data-[placement=right]:before:top-1/2","data-[placement=right]:before:-translate-y-1/2","data-[placement=right-start]:before:-left-[calc(theme(spacing.5)/4_-_3px)]","data-[placement=right-start]:before:top-1/4","data-[placement=right-end]:before:-left-[calc(theme(spacing.5)/4_-_3px)]","data-[placement=right-end]:before:bottom-1/4","outline-none","data-[focus-visible=true]:z-10","data-[focus-visible=true]:outline-2","data-[focus-visible=true]:outline-focus","data-[focus-visible=true]:outline-offset-2"],content:["z-10","px-2.5","py-1","w-full","inline-flex","flex-col","items-center","justify-center","box-border","subpixel-antialiased","outline-none","box-border"],trigger:["z-10"],backdrop:["hidden"],arrow:[]},variants:{size:{sm:{content:"text-tiny"},md:{content:"text-small"},lg:{content:"text-medium"}},color:{default:{base:"before:bg-content1 before:shadow-small",content:"bg-content1"},foreground:{base:"before:bg-foreground",content:"bg-foreground text-background"},primary:{base:"before:bg-primary",content:"bg-primary text-primary-foreground"},secondary:{base:"before:bg-secondary",content:"bg-secondary text-secondary-foreground"},success:{base:"before:bg-success",content:"bg-success text-success-foreground"},warning:{base:"before:bg-warning",content:"bg-warning text-warning-foreground"},danger:{base:"before:bg-danger",content:"bg-danger text-danger-foreground"}},radius:{none:{content:"rounded-none"},sm:{content:"rounded-small"},md:{content:"rounded-medium"},lg:{content:"rounded-large"},full:{content:"rounded-full"}},shadow:{sm:{content:"shadow-small"},md:{content:"shadow-medium"},lg:{content:"shadow-large"}},backdrop:{transparent:{},opaque:{backdrop:"bg-overlay/50 backdrop-opacity-disabled"},blur:{backdrop:"backdrop-blur-sm backdrop-saturate-150 bg-overlay/30"}},triggerScaleOnOpen:{true:{trigger:["aria-expanded:scale-[0.97]","aria-expanded:opacity-70","subpixel-antialiased"]},false:{}},disableAnimation:{true:{base:"animate-none"}},isTriggerDisabled:{true:{trigger:"opacity-disabled pointer-events-none"},false:{}}},defaultVariants:{color:"default",radius:"lg",size:"md",shadow:"md",backdrop:"transparent",triggerScaleOnOpen:!0},compoundVariants:[{backdrop:["opaque","blur"],class:{backdrop:"block w-full h-full fixed inset-0 -z-30"}}]},{...l,twMerge:null==(a=null==l?void 0:l.twMerge)||a,twMergeConfig:{...null==l?void 0:l.twMergeConfig,theme:{...null==(s=null==l?void 0:l.twMergeConfig)?void 0:s.theme,...tT.theme},classGroups:{...null==(u=null==l?void 0:l.twMergeConfig)?void 0:u.classGroups,...tT.classGroups}}}),np=Object.create,nh=Object.defineProperty,nm=Object.getOwnPropertyDescriptor,ng=Object.getOwnPropertyNames,nb=Object.getPrototypeOf,nv=Object.prototype.hasOwnProperty,ny=(e,t)=>function(){return t||(0,e[ng(e)[0]])((t={exports:{}}).exports,t),t.exports},nw=ny({"../../../node_modules/.pnpm/react@18.2.0/node_modules/react/cjs/react.production.min.js"(e){var t=Symbol.for("react.element"),n=Symbol.for("react.portal"),r=Symbol.for("react.fragment"),o=Symbol.for("react.strict_mode"),i=Symbol.for("react.profiler"),l=Symbol.for("react.provider"),a=Symbol.for("react.context"),s=Symbol.for("react.forward_ref"),u=Symbol.for("react.suspense"),c=Symbol.for("react.memo"),d=Symbol.for("react.lazy"),f=Symbol.iterator,p={isMounted:function(){return!1},enqueueForceUpdate:function(){},enqueueReplaceState:function(){},enqueueSetState:function(){}},h=Object.assign,m={};function g(e,t,n){this.props=e,this.context=t,this.refs=m,this.updater=n||p}function b(){}function v(e,t,n){this.props=e,this.context=t,this.refs=m,this.updater=n||p}g.prototype.isReactComponent={},g.prototype.setState=function(e,t){if("object"!=typeof e&&"function"!=typeof e&&null!=e)throw Error("setState(...): takes an object of state variables to update or a function which returns an object of state variables.");this.updater.enqueueSetState(this,e,t,"setState")},g.prototype.forceUpdate=function(e){this.updater.enqueueForceUpdate(this,e,"forceUpdate")},b.prototype=g.prototype;var y=v.prototype=new b;y.constructor=v,h(y,g.prototype),y.isPureReactComponent=!0;var w=Array.isArray,x=Object.prototype.hasOwnProperty,E={current:null},S={key:!0,ref:!0,__self:!0,__source:!0};function k(e,n,r){var o,i={},l=null,a=null;if(null!=n)for(o in void 0!==n.ref&&(a=n.ref),void 0!==n.key&&(l=""+n.key),n)x.call(n,o)&&!S.hasOwnProperty(o)&&(i[o]=n[o]);var s=arguments.length-2;if(1===s)i.children=r;else if(1<s){for(var u=Array(s),c=0;c<s;c++)u[c]=arguments[c+2];i.children=u}if(e&&e.defaultProps)for(o in s=e.defaultProps)void 0===i[o]&&(i[o]=s[o]);return{$$typeof:t,type:e,key:l,ref:a,props:i,_owner:E.current}}function T(e){return"object"==typeof e&&null!==e&&e.$$typeof===t}var C=/\/+/g;function P(e,t){var n,r;return"object"==typeof e&&null!==e&&null!=e.key?(n=""+e.key,r={"=":"=0",":":"=2"},"$"+n.replace(/[=:]/g,function(e){return r[e]})):t.toString(36)}function M(e,r,o){if(null==e)return e;var i=[],l=0;return!function e(r,o,i,l,a){var s,u,c,d=typeof r;("undefined"===d||"boolean"===d)&&(r=null);var p=!1;if(null===r)p=!0;else switch(d){case"string":case"number":p=!0;break;case"object":switch(r.$$typeof){case t:case n:p=!0}}if(p)return a=a(p=r),r=""===l?"."+P(p,0):l,w(a)?(i="",null!=r&&(i=r.replace(C,"$&/")+"/"),e(a,o,i,"",function(e){return e})):null!=a&&(T(a)&&(s=a,u=i+(!a.key||p&&p.key===a.key?"":(""+a.key).replace(C,"$&/")+"/")+r,a={$$typeof:t,type:s.type,key:u,ref:s.ref,props:s.props,_owner:s._owner}),o.push(a)),1;if(p=0,l=""===l?".":l+":",w(r))for(var h=0;h<r.length;h++){var m=l+P(d=r[h],h);p+=e(d,o,i,m,a)}else if("function"==typeof(m=null===(c=r)||"object"!=typeof c?null:"function"==typeof(c=f&&c[f]||c["@@iterator"])?c:null))for(r=m.call(r),h=0;!(d=r.next()).done;)m=l+P(d=d.value,h++),p+=e(d,o,i,m,a);else if("object"===d)throw Error("Objects are not valid as a React child (found: "+("[object Object]"===(o=String(r))?"object with keys {"+Object.keys(r).join(", ")+"}":o)+"). If you meant to render a collection of children, use an array instead.");return p}(e,i,"","",function(e){return r.call(o,e,l++)}),i}function R(e){if(-1===e._status){var t=e._result;(t=t()).then(function(t){(0===e._status||-1===e._status)&&(e._status=1,e._result=t)},function(t){(0===e._status||-1===e._status)&&(e._status=2,e._result=t)}),-1===e._status&&(e._status=0,e._result=t)}if(1===e._status)return e._result.default;throw e._result}var j={current:null},O={transition:null};e.Children={map:M,forEach:function(e,t,n){M(e,function(){t.apply(this,arguments)},n)},count:function(e){var t=0;return M(e,function(){t++}),t},toArray:function(e){return M(e,function(e){return e})||[]},only:function(e){if(!T(e))throw Error("React.Children.only expected to receive a single React element child.");return e}},e.Component=g,e.Fragment=r,e.Profiler=i,e.PureComponent=v,e.StrictMode=o,e.Suspense=u,e.__SECRET_INTERNALS_DO_NOT_USE_OR_YOU_WILL_BE_FIRED={ReactCurrentDispatcher:j,ReactCurrentBatchConfig:O,ReactCurrentOwner:E},e.cloneElement=function(e,n,r){if(null==e)throw Error("React.cloneElement(...): The argument must be a React element, but you passed "+e+".");var o=h({},e.props),i=e.key,l=e.ref,a=e._owner;if(null!=n){if(void 0!==n.ref&&(l=n.ref,a=E.current),void 0!==n.key&&(i=""+n.key),e.type&&e.type.defaultProps)var s=e.type.defaultProps;for(u in n)x.call(n,u)&&!S.hasOwnProperty(u)&&(o[u]=void 0===n[u]&&void 0!==s?s[u]:n[u])}var u=arguments.length-2;if(1===u)o.children=r;else if(1<u){s=Array(u);for(var c=0;c<u;c++)s[c]=arguments[c+2];o.children=s}return{$$typeof:t,type:e.type,key:i,ref:l,props:o,_owner:a}},e.createContext=function(e){return(e={$$typeof:a,_currentValue:e,_currentValue2:e,_threadCount:0,Provider:null,Consumer:null,_defaultValue:null,_globalName:null}).Provider={$$typeof:l,_context:e},e.Consumer=e},e.createElement=k,e.createFactory=function(e){var t=k.bind(null,e);return t.type=e,t},e.createRef=function(){return{current:null}},e.forwardRef=function(e){return{$$typeof:s,render:e}},e.isValidElement=T,e.lazy=function(e){return{$$typeof:d,_payload:{_status:-1,_result:e},_init:R}},e.memo=function(e,t){return{$$typeof:c,type:e,compare:void 0===t?null:t}},e.startTransition=function(e){var t=O.transition;O.transition={};try{e()}finally{O.transition=t}},e.unstable_act=function(){throw Error("act(...) is not supported in production builds of React.")},e.useCallback=function(e,t){return j.current.useCallback(e,t)},e.useContext=function(e){return j.current.useContext(e)},e.useDebugValue=function(){},e.useDeferredValue=function(e){return j.current.useDeferredValue(e)},e.useEffect=function(e,t){return j.current.useEffect(e,t)},e.useId=function(){return j.current.useId()},e.useImperativeHandle=function(e,t,n){return j.current.useImperativeHandle(e,t,n)},e.useInsertionEffect=function(e,t){return j.current.useInsertionEffect(e,t)},e.useLayoutEffect=function(e,t){return j.current.useLayoutEffect(e,t)},e.useMemo=function(e,t){return j.current.useMemo(e,t)},e.useReducer=function(e,t,n){return j.current.useReducer(e,t,n)},e.useRef=function(e){return j.current.useRef(e)},e.useState=function(e){return j.current.useState(e)},e.useSyncExternalStore=function(e,t,n){return j.current.useSyncExternalStore(e,t,n)},e.useTransition=function(){return j.current.useTransition()},e.version="18.2.0"}});ny({"../../../node_modules/.pnpm/react@18.2.0/node_modules/react/cjs/react.development.js"(e,t){}}),i=null!=(r=ny({"../../../node_modules/.pnpm/react@18.2.0/node_modules/react/index.js"(e,t){t.exports=nw()}})())?np(nb(r)):{},((e,t,n,r)=>{if(t&&"object"==typeof t||"function"==typeof t)for(let o of ng(t))nv.call(e,o)||o===n||nh(e,o,{get:()=>t[o],enumerable:!(r=nm(t,o))||r.enumerable})})(!o&&r&&r.__esModule?i:nh(i,"default",{value:r,enumerable:!0}),r);var nx=e=>e?"true":void 0;"undefined"!=typeof window&&window.document&&window.document.createElement;var nE=e=>{let t={top:{originY:1},bottom:{originY:0},left:{originX:1},right:{originX:0},"top-start":{originX:0,originY:1},"top-end":{originX:1,originY:1},"bottom-start":{originX:0,originY:0},"bottom-end":{originX:1,originY:0},"right-start":{originX:0,originY:0},"right-end":{originX:0,originY:1},"left-start":{originX:1,originY:0},"left-end":{originX:1,originY:1}};return(null==t?void 0:t[e])||{}},nS=e=>({top:"top",bottom:"bottom",left:"left",right:"right","top-start":"top start","top-end":"top end","bottom-start":"bottom start","bottom-end":"bottom end","left-start":"left top","left-end":"left bottom","right-start":"right top","right-end":"right bottom"})[e],nk=(e,t)=>{if(t.includes("-")){let[,n]=t.split("-");return`${e}-${n}`}return e},nT=(null==globalThis?void 0:globalThis.document)?h.useLayoutEffect:h.useEffect,nC=n(73935);let nP=h.createContext(null);function nM(e){let{children:t}=e,n=(0,h.useContext)(nP),[r,o]=(0,h.useState)(0),i=(0,h.useMemo)(()=>({parent:n,modalCount:r,addModal(){o(e=>e+1),n&&n.addModal()},removeModal(){o(e=>e-1),n&&n.removeModal()}}),[n,r]);return h.createElement(nP.Provider,{value:i},t)}function nR(e){let t;let{modalProviderProps:n}={modalProviderProps:{"aria-hidden":!!(t=(0,h.useContext)(nP))&&t.modalCount>0||void 0}};return h.createElement("div",{"data-overlay-container":!0,...e,...n})}function nj(e){return h.createElement(nM,null,h.createElement(nR,e))}function nO(e){let t=R(),{portalContainer:n=t?null:document.body,...r}=e;if(h.useEffect(()=>{if(null==n?void 0:n.closest("[data-overlay-container]"))throw Error("An OverlayContainer must not be inside another container. Please change the portalContainer prop.")},[n]),!n)return null;let o=h.createElement(nj,r);return nC.createPortal(o,n)}let nL=(0,h.createContext)({strict:!1});var nA=n(9442);function n_(e){for(let t in e)nA.featureDefinitions[t]={...nA.featureDefinitions[t],...e[t]}}function nN({children:e,features:t,strict:n=!1}){let[,r]=(0,h.useState)(!n$(t)),o=(0,h.useRef)(void 0);if(!n$(t)){let{renderer:e,...n}=t;o.current=e,n_(n)}return(0,h.useEffect)(()=>{n$(t)&&t().then(({renderer:e,...t})=>{n_(t),o.current=e,r(!0)})},[]),(0,p.jsx)(nL.Provider,{value:{renderer:o.current,strict:n},children:e})}function n$(e){return"function"==typeof e}let nI=(0,h.createContext)({}),nD=(0,h.createContext)({transformPagePoint:e=>e,isStatic:!1,reducedMotion:"never"}),nW=(0,h.createContext)({});var nH=n(7504),nF=n(97732);function nz(e){return Array.isArray(e)?e.join(" "):e}var nB=n(11741);let nq=Symbol.for("motionComponentSymbol");function nV(e){return e&&"object"==typeof e&&Object.prototype.hasOwnProperty.call(e,"current")}let{schedule:nU,cancel:nY}=(0,n(93674).Z)(queueMicrotask,!1);var nG=n(68588);let nX=(0,h.createContext)(null),nK=(0,h.createContext)({}),nZ=nB.j?h.useLayoutEffect:h.useEffect;var nJ=n(32558),nQ=n(96728),n0=n(41069);let n1=()=>({style:{},transform:{},transformOrigin:{},vars:{}});function n2(e,t,n){for(let r in t)(0,nJ.i)(t[r])||(0,nQ.j)(r,n)||(e[r]=t[r])}var n4=n(42617);let n5=()=>({...n1(),attrs:{}});var n6=n(79854);let n3=new Set(["animate","exit","variants","initial","style","values","variants","transition","transformTemplate","custom","inherit","onBeforeLayoutMeasure","onAnimationStart","onAnimationComplete","onUpdate","onDragStart","onDrag","onDragEnd","onMeasureDragConstraints","onDirectionLock","onDragTransitionEnd","_dragX","_dragY","onHoverStart","onHoverEnd","onViewportEnter","onViewportLeave","globalTapTarget","ignoreStrict","viewport"]);function n9(e){return e.startsWith("while")||e.startsWith("drag")&&"draggable"!==e||e.startsWith("layout")||e.startsWith("onTap")||e.startsWith("onPan")||e.startsWith("onLayout")||n3.has(e)}let n8=e=>!n9(e);try{c=require("@emotion/is-prop-valid").default,"function"==typeof c&&(n8=e=>e.startsWith("on")?!n9(e):c(e))}catch{}var n7=n(62627),re=n(2445),rt=n(79432);function rn(e){let t=(0,h.useRef)(null);return null===t.current&&(t.current=e()),t.current}let rr=e=>(t,n)=>{let r=(0,h.useContext)(nW),o=(0,h.useContext)(nX),i=()=>(function({scrapeMotionValuesFromProps:e,createRenderState:t},n,r,o){return{latestValues:function(e,t,n,r){let o={},i=r(e,{});for(let e in i){var l;o[e]=(l=i[e],(0,nJ.i)(l)?l.get():l)}let{initial:a,animate:s}=e,u=(0,nH.G)(e),c=(0,nH.M)(e);t&&c&&!u&&!1!==e.inherit&&(void 0===a&&(a=t.initial),void 0===s&&(s=t.animate));let d=!!n&&!1===n.initial,f=(d=d||!1===a)?s:a;if(f&&"boolean"!=typeof f&&!(0,re.H)(f)){let t=Array.isArray(f)?f:[f];for(let n=0;n<t.length;n++){let r=(0,rt.o)(e,t[n]);if(r){let{transitionEnd:e,transition:t,...n}=r;for(let e in n){let t=n[e];if(Array.isArray(t)){let e=d?t.length-1:0;t=t[e]}null!==t&&(o[e]=t)}for(let t in e)o[t]=e[t]}}}return o}(n,r,o,e),renderState:t()}})(e,t,r,o);return n?i():rn(i)},ro={useVisualState:rr({scrapeMotionValuesFromProps:n(50189).U,createRenderState:n1})},ri={useVisualState:rr({scrapeMotionValuesFromProps:n(16832).U,createRenderState:n5})},rl=function(e){if("undefined"==typeof Proxy)return e;let t=new Map;return new Proxy((...t)=>e(...t),{get:(n,r)=>"create"===r?e:(t.has(r)||t.set(r,e(r)),t.get(r))})}(function(e,{forwardMotionProps:t}={forwardMotionProps:!1}){return function({preloadedFeatures:e,createVisualElement:t,useRender:n,useVisualState:r,Component:o}){function i(e,i){var l;let a;let s={...(0,h.useContext)(nD),...e,layoutId:function({layoutId:e}){let t=(0,h.useContext)(nI).id;return t&&void 0!==e?t+"-"+e:e}(e)},{isStatic:u}=s,c=function(e){let{initial:t,animate:n}=function(e,t){if((0,nH.G)(e)){let{initial:t,animate:n}=e;return{initial:!1===t||(0,nF.$)(t)?t:void 0,animate:(0,nF.$)(n)?n:void 0}}return!1!==e.inherit?t:{}}(e,(0,h.useContext)(nW));return(0,h.useMemo)(()=>({initial:t,animate:n}),[nz(t),nz(n)])}(e),d=r(e,u);if(!u&&nB.j){(0,h.useContext)(nL).strict;let e=function(e){let{drag:t,layout:n}=nA.featureDefinitions;if(!t&&!n)return{};let r={...t,...n};return{MeasureLayout:t?.isEnabled(e)||n?.isEnabled(e)?r.MeasureLayout:void 0,ProjectionNode:r.ProjectionNode}}(s);a=e.MeasureLayout,c.visualElement=function(e,t,n,r,o){let{visualElement:i}=(0,h.useContext)(nW),l=(0,h.useContext)(nL),a=(0,h.useContext)(nX),s=(0,h.useContext)(nD).reducedMotion,u=(0,h.useRef)(null);r=r||l.renderer,!u.current&&r&&(u.current=r(e,{visualState:t,parent:i,props:n,presenceContext:a,blockInitialAnimation:!!a&&!1===a.initial,reducedMotionConfig:s}));let c=u.current,d=(0,h.useContext)(nK);c&&!c.projection&&o&&("html"===c.type||"svg"===c.type)&&function(e,t,n,r){let{layoutId:o,layout:i,drag:l,dragConstraints:a,layoutScroll:s,layoutRoot:u,layoutCrossfade:c}=t;e.projection=new n(e.latestValues,t["data-framer-portal-id"]?void 0:function e(t){if(t)return!1!==t.options.allowProjection?t.projection:e(t.parent)}(e.parent)),e.projection.setOptions({layoutId:o,layout:i,alwaysMeasureLayout:!!l||a&&nV(a),visualElement:e,animationType:"string"==typeof i?i:"both",initialPromotionConfig:r,crossfade:c,layoutScroll:s,layoutRoot:u})}(u.current,n,o,d);let f=(0,h.useRef)(!1);(0,h.useInsertionEffect)(()=>{c&&f.current&&c.update(n,a)});let p=n[nG.M],m=(0,h.useRef)(!!p&&!window.MotionHandoffIsComplete?.(p)&&window.MotionHasOptimisedAnimation?.(p));return nZ(()=>{c&&(f.current=!0,window.MotionIsMounted=!0,c.updateFeatures(),nU.render(c.render),m.current&&c.animationState&&c.animationState.animateChanges())}),(0,h.useEffect)(()=>{c&&(!m.current&&c.animationState&&c.animationState.animateChanges(),m.current&&(queueMicrotask(()=>{window.MotionHandoffMarkAsComplete?.(p)}),m.current=!1))}),c}(o,d,s,t,e.ProjectionNode)}return(0,p.jsxs)(nW.Provider,{value:c,children:[a&&c.visualElement?(0,p.jsx)(a,{visualElement:c.visualElement,...s}):null,n(o,e,(l=c.visualElement,(0,h.useCallback)(e=>{e&&d.onMount&&d.onMount(e),l&&(e?l.mount(e):l.unmount()),i&&("function"==typeof i?i(e):nV(i)&&(i.current=e))},[l])),d,u,c.visualElement)]})}e&&n_(e),i.displayName=`motion.${"string"==typeof o?o:`create(${o.displayName??o.name??""})`}`;let l=(0,h.forwardRef)(i);return l[nq]=o,l}({...(0,n7.q)(e)?ri:ro,preloadedFeatures:void 0,useRender:function(e=!1){return(t,n,r,{latestValues:o},i)=>{let l=((0,n7.q)(t)?function(e,t,n,r){let o=(0,h.useMemo)(()=>{let n=n5();return(0,n4.i)(n,t,(0,n6.a)(r),e.transformTemplate,e.style),{...n.attrs,style:{...n.style}}},[t]);if(e.style){let t={};n2(t,e.style,e),o.style={...t,...o.style}}return o}:function(e,t){let n={},r=function(e,t){let n=e.style||{},r={};return n2(r,n,e),Object.assign(r,function({transformTemplate:e},t){return(0,h.useMemo)(()=>{let n=n1();return(0,n0.r)(n,t,e),Object.assign({},n.vars,n.style)},[t])}(e,t)),r}(e,t);return e.drag&&!1!==e.dragListener&&(n.draggable=!1,r.userSelect=r.WebkitUserSelect=r.WebkitTouchCallout="none",r.touchAction=!0===e.drag?"none":`pan-${"x"===e.drag?"y":"x"}`),void 0===e.tabIndex&&(e.onTap||e.onTapStart||e.whileTap)&&(n.tabIndex=0),n.style=r,n})(n,o,i,t),a=function(e,t,n){let r={};for(let o in e)("values"!==o||"object"!=typeof e.values)&&(n8(o)||!0===n&&n9(o)||!t&&!n9(o)||e.draggable&&o.startsWith("onDrag"))&&(r[o]=e[o]);return r}(n,"string"==typeof t,e),s=t!==h.Fragment?{...a,...l,ref:r}:{},{children:u}=n,c=(0,h.useMemo)(()=>(0,nJ.i)(u)?u.get():u,[u]);return(0,h.createElement)(t,{...s,children:c})}}(t),createVisualElement:void 0,Component:e})});var ra=n(77596);class rs extends h.Component{getSnapshotBeforeUpdate(e){let t=this.props.childRef.current;if(t&&e.isPresent&&!this.props.isPresent){let e=t.offsetParent,n=(0,ra.R)(e)&&e.offsetWidth||0,r=this.props.sizeRef.current;r.height=t.offsetHeight||0,r.width=t.offsetWidth||0,r.top=t.offsetTop,r.left=t.offsetLeft,r.right=n-r.width-r.left}return null}componentDidUpdate(){}render(){return this.props.children}}function ru({children:e,isPresent:t,anchorX:n}){let r=(0,h.useId)(),o=(0,h.useRef)(null),i=(0,h.useRef)({width:0,height:0,top:0,left:0,right:0}),{nonce:l}=(0,h.useContext)(nD);return(0,h.useInsertionEffect)(()=>{let{width:e,height:a,top:s,left:u,right:c}=i.current;if(t||!o.current||!e||!a)return;let d="left"===n?`left: ${u}`:`right: ${c}`;o.current.dataset.motionPopId=r;let f=document.createElement("style");return l&&(f.nonce=l),document.head.appendChild(f),f.sheet&&f.sheet.insertRule(`
          [data-motion-pop-id="${r}"] {
            position: absolute !important;
            width: ${e}px !important;
            height: ${a}px !important;
            ${d}px !important;
            top: ${s}px !important;
          }
        `),()=>{document.head.contains(f)&&document.head.removeChild(f)}},[t]),(0,p.jsx)(rs,{isPresent:t,childRef:o,sizeRef:i,children:h.cloneElement(e,{ref:o})})}let rc=({children:e,initial:t,isPresent:n,onExitComplete:r,custom:o,presenceAffectsLayout:i,mode:l,anchorX:a})=>{let s=rn(rd),u=(0,h.useId)(),c=!0,d=(0,h.useMemo)(()=>(c=!1,{id:u,initial:t,isPresent:n,custom:o,onExitComplete:e=>{for(let t of(s.set(e,!0),s.values()))if(!t)return;r&&r()},register:e=>(s.set(e,!1),()=>s.delete(e))}),[n,s,r]);return i&&c&&(d={...d}),(0,h.useMemo)(()=>{s.forEach((e,t)=>s.set(t,!1))},[n]),h.useEffect(()=>{n||s.size||!r||r()},[n]),"popLayout"===l&&(e=(0,p.jsx)(ru,{isPresent:n,anchorX:a,children:e})),(0,p.jsx)(nX.Provider,{value:d,children:e})};function rd(){return new Map}let rf=e=>e.key||"";function rp(e){let t=[];return h.Children.forEach(e,e=>{(0,h.isValidElement)(e)&&t.push(e)}),t}let rh=({children:e,custom:t,initial:n=!0,onExitComplete:r,presenceAffectsLayout:o=!0,mode:i="sync",propagate:l=!1,anchorX:a="left"})=>{let[s,u]=function(e=!0){let t=(0,h.useContext)(nX);if(null===t)return[!0,null];let{isPresent:n,onExitComplete:r,register:o}=t,i=(0,h.useId)();(0,h.useEffect)(()=>{if(e)return o(i)},[e]);let l=(0,h.useCallback)(()=>e&&r&&r(i),[i,r,e]);return!n&&r?[!1,l]:[!0]}(l),c=(0,h.useMemo)(()=>rp(e),[e]),d=l&&!s?[]:c.map(rf),f=(0,h.useRef)(!0),m=(0,h.useRef)(c),g=rn(()=>new Map),[b,v]=(0,h.useState)(c),[y,w]=(0,h.useState)(c);nZ(()=>{f.current=!1,m.current=c;for(let e=0;e<y.length;e++){let t=rf(y[e]);d.includes(t)?g.delete(t):!0!==g.get(t)&&g.set(t,!1)}},[y,d.length,d.join("-")]);let x=[];if(c!==b){let e=[...c];for(let t=0;t<y.length;t++){let n=y[t],r=rf(n);d.includes(r)||(e.splice(t,0,n),x.push(n))}return"wait"===i&&x.length&&(e=x),w(rp(e)),v(c),null}let{forceRender:E}=(0,h.useContext)(nI);return(0,p.jsx)(p.Fragment,{children:y.map(e=>{let h=rf(e),b=(!l||!!s)&&(c===y||d.includes(h));return(0,p.jsx)(rc,{isPresent:b,initial:(!f.current||!!n)&&void 0,custom:t,presenceAffectsLayout:o,mode:i,onExitComplete:b?void 0:()=>{if(!g.has(h))return;g.set(h,!0);let e=!0;g.forEach(t=>{t||(e=!1)}),e&&(E?.(),w(m.current),l&&u?.(),r&&r())},anchorX:a,children:e},h)})})};var rm={enter:{transform:"scale(1)",opacity:1,transition:{type:"spring",bounce:0,duration:.2}},exit:{transform:"scale(0.85)",opacity:0,transition:{type:"easeOut",duration:.15}}},rg={},rb=()=>n.e(937).then(n.bind(n,83937)).then(e=>e.default),rv=(d=(e,t)=>{let n;let{Component:r,children:o,content:i,isOpen:l,portalContainer:a,placement:s,disableAnimation:u,motionProps:c,getTriggerProps:d,getTooltipProps:f,getTooltipContentProps:x}=function(e){var t,n;let r=tE(),[o,i]=tS(e,nf.variantKeys),{ref:l,as:a,isOpen:s,content:u,children:c,defaultOpen:d,onOpenChange:f,isDisabled:p,trigger:x,shouldFlip:E=!0,containerPadding:S=12,placement:k="top",delay:C=0,closeDelay:P=500,showArrow:M=!1,offset:L=7,crossOffset:_=0,isDismissable:N,shouldCloseOnBlur:$=!0,portalContainer:I,isKeyboardDismissDisabled:D=!1,updatePositionDeps:W=[],shouldCloseOnInteractOutside:H,className:F,onClose:z,motionProps:B,classNames:V,...Y}=o,G=null!=(n=null!=(t=null==e?void 0:e.disableAnimation)?t:null==r?void 0:r.disableAnimation)&&n,X=function(e={}){let{delay:t=1500,closeDelay:n=500}=e,{isOpen:r,open:o,close:i}=function(e){let[t,n]=function(e,t,n){let[r,o]=(0,h.useState)(e||t),i=(0,h.useRef)(void 0!==e),l=void 0!==e;(0,h.useEffect)(()=>{let e=i.current;e!==l&&console.warn(`WARN: A component changed from ${e?"controlled":"uncontrolled"} to ${l?"controlled":"uncontrolled"}.`),i.current=l},[l]);let a=l?e:r,s=(0,h.useCallback)((e,...t)=>{let r=(e,...t)=>{n&&!Object.is(a,e)&&n(e,...t),l||(a=e)};"function"==typeof e?(console.warn("We can not support a function callback. See Github Issues for details https://github.com/adobe/react-spectrum/issues/2320"),o((n,...o)=>{let i=e(l?a:n,...o);return(r(i,...t),l)?n:i})):(l||o(e),r(e,...t))},[l,a,n]);return[a,s]}(e.isOpen,e.defaultOpen||!1,e.onOpenChange),r=(0,h.useCallback)(()=>{n(!0)},[n]),o=(0,h.useCallback)(()=>{n(!1)},[n]),i=(0,h.useCallback)(()=>{n(!t)},[n,t]);return{isOpen:t,setOpen:n,open:r,close:o,toggle:i}}(e),l=(0,h.useMemo)(()=>`${++g}`,[]),a=(0,h.useRef)(null),s=(0,h.useRef)(i),u=()=>{m[l]=f},c=()=>{for(let e in m)e!==l&&(m[e](!0),delete m[e])},d=()=>{a.current&&clearTimeout(a.current),a.current=null,c(),u(),b=!0,o(),v&&(clearTimeout(v),v=null),y&&(clearTimeout(y),y=null)},f=e=>{e||n<=0?(a.current&&clearTimeout(a.current),a.current=null,s.current()):a.current||(a.current=setTimeout(()=>{a.current=null,s.current()},n)),v&&(clearTimeout(v),v=null),b&&(y&&clearTimeout(y),y=setTimeout(()=>{delete m[l],y=null,b=!1},Math.max(500,n)))},p=()=>{c(),u(),r||v||b?r||d():v=setTimeout(()=>{v=null,b=!0,d()},t)};return(0,h.useEffect)(()=>{s.current=i},[i]),(0,h.useEffect)(()=>()=>{a.current&&clearTimeout(a.current),m[l]&&delete m[l]},[l]),{isOpen:r,open:e=>{e||!(t>0)||a.current?d():p()},close:f}}({delay:C,closeDelay:P,isDisabled:p,defaultOpen:d,isOpen:s,onOpenChange:e=>{null==f||f(e),e||null==z||z()}}),K=(0,h.useRef)(null),Z=(0,h.useRef)(null),J=(0,h.useId)(),Q=X.isOpen&&!p;(0,h.useImperativeHandle)(l,()=>({UNSAFE_getDOMNode:()=>Z.current}));let{triggerProps:ee,tooltipProps:et}=function(e,t,n){let{isDisabled:r,trigger:o}=e,i=function(e){let[t,n]=(0,h.useState)(void 0),r=(0,h.useRef)(null),o=T(t),i=(0,h.useCallback)(e=>{r.current=e},[]);return j&&(O.has(o)&&!O.get(o).includes(i)?O.set(o,[...O.get(o),i]):O.set(o,[i])),w(()=>()=>{O.delete(o)},[o]),(0,h.useEffect)(()=>{let e=r.current;e&&(r.current=null,n(e))}),o}(),l=(0,h.useRef)(!1),a=(0,h.useRef)(!1),s=()=>{(l.current||a.current)&&t.open(a.current)},u=e=>{l.current||a.current||t.close(e)};(0,h.useEffect)(()=>{let e=e=>{n&&n.current&&"Escape"===e.key&&(e.stopPropagation(),t.close(!0))};if(t.isOpen)return document.addEventListener("keydown",e,!0),()=>{document.removeEventListener("keydown",e,!0)}},[n,t]);let c=()=>{a.current=!1,l.current=!1,u(!0)},{hoverProps:d}=ec({isDisabled:r,onHoverStart:()=>{"focus"!==o&&("pointer"===U?l.current=!0:l.current=!1,s())},onHoverEnd:()=>{"focus"!==o&&(a.current=!1,l.current=!1,u())}}),{focusableProps:f}=function(e,t){let{focusProps:n}=function(e){let t,n,{isDisabled:r,onFocus:o,onBlur:i,onFocusChange:l}=e,a=(0,h.useCallback)(e=>{if(e.target===e.currentTarget)return i&&i(e),l&&l(!1),!0},[i,l]),s=(t=(0,h.useRef)({isFocused:!1,observer:null}),ed(()=>{let e=t.current;return()=>{e.observer&&(e.observer.disconnect(),e.observer=null)}},[]),n=function(e){let t=(0,h.useRef)(null);return ed(()=>{t.current=e},[e]),(0,h.useCallback)((...e)=>{let n=t.current;return null==n?void 0:n(...e)},[])}(e=>{null==a||a(e)}),(0,h.useCallback)(e=>{if(e.target instanceof HTMLButtonElement||e.target instanceof HTMLInputElement||e.target instanceof HTMLTextAreaElement||e.target instanceof HTMLSelectElement){t.current.isFocused=!0;let r=e.target;r.addEventListener("focusout",e=>{t.current.isFocused=!1,r.disabled&&n((e.nativeEvent=e,e.isDefaultPrevented=()=>e.defaultPrevented,e.isPropagationStopped=()=>e.cancelBubble,e.persist=()=>{},e)),t.current.observer&&(t.current.observer.disconnect(),t.current.observer=null)},{once:!0}),t.current.observer=new MutationObserver(()=>{if(t.current.isFocused&&r.disabled){var e;null===(e=t.current.observer)||void 0===e||e.disconnect();let n=r===document.activeElement?null:document.activeElement;r.dispatchEvent(new FocusEvent("blur",{relatedTarget:n})),r.dispatchEvent(new FocusEvent("focusout",{bubbles:!0,relatedTarget:n}))}}),t.current.observer.observe(r,{attributes:!0,attributeFilter:["disabled"]})}},[n])),u=(0,h.useCallback)(e=>{let t=ex(e.target),n=t?e$(t):e$();e.target===e.currentTarget&&n===e.nativeEvent.target&&(o&&o(e),l&&l(!0),s(e))},[l,o,s]);return{focusProps:{onFocus:!r&&(o||l||i)?u:void 0,onBlur:!r&&(i||l)?a:void 0}}}(e),{keyboardProps:r}={keyboardProps:e.isDisabled?{}:{onKeyDown:ez(e.onKeyDown),onKeyUp:ez(e.onKeyUp)}},o=eq(n,r),i=function(e){let t=(0,h.useContext)(eV)||{};ed(()=>{if(t&&t.ref&&e)return t.ref.current=e.current,()=>{t.ref&&(t.ref.current=null)}});let{ref:n,...r}=t;return r}(t),l=e.isDisabled?{}:i,a=(0,h.useRef)(e.autoFocus);(0,h.useEffect)(()=>{a.current&&t.current&&function(e){let t=ex(e),n=e$(t);if("virtual"===eS){var r;r=()=>{e$(t)===n&&e.isConnected&&eH(e)},requestAnimationFrame(()=>{(function(){for(let[e]of eI)"isConnected"in e&&!e.isConnected&&eI.delete(e)})(),0===eI.size?r():eD.add(r)})}else eH(e)}(t.current),a.current=!1},[t]);let s=e.excludeFromTabOrder?-1:0;return e.isDisabled&&(s=void 0),{focusableProps:eq({...o,tabIndex:s},l)}}({isDisabled:r,onFocus:()=>{"pointer"!==U&&(a.current=!0,s())},onBlur:()=>{a.current=!1,l.current=!1,u(!0)}},n);return{triggerProps:{"aria-describedby":t.isOpen?i:void 0,...A(f,d,{onPointerDown:c,onKeyDown:c})},tooltipProps:{id:i}}}({isDisabled:p,trigger:x},X,K),{tooltipProps:en}=function(e,t){let n=function(e,t={}){let{labelable:n,isLink:r,propNames:o}=t,i={};for(let t in e)Object.prototype.hasOwnProperty.call(e,t)&&(eU.has(t)||n&&eY.has(t)||r&&eG.has(t)||(null==o?void 0:o.has(t))||eX.test(t))&&(i[t]=e[t]);return i}(e,{labelable:!0}),{hoverProps:r}=ec({onHoverStart:()=>null==t?void 0:t.open(!0),onHoverEnd:()=>null==t?void 0:t.close()});return{tooltipProps:A(n,r,{role:"tooltip"})}}({isOpen:Q,...A(o,et)},X),{overlayProps:er,placement:eo,updatePosition:ei}=function(e){var t,n,r;let o;let{direction:i}=(o=function(){let e=R(),[t,n]=(0,h.useState)(ts);return((0,h.useEffect)(()=>(0===tu.size&&window.addEventListener("languagechange",tc),tu.add(n),()=>{tu.delete(n),0===tu.size&&window.removeEventListener("languagechange",tc)}),[]),e)?{locale:"en-US",direction:"ltr"}:t}(),(0,h.useContext)(td)||o),{arrowSize:l=0,targetRef:a,overlayRef:s,scrollRef:u=s,placement:c="bottom",containerPadding:d=12,shouldFlip:f=!0,boundaryElement:p="undefined"!=typeof document?document.body:null,offset:m=0,crossOffset:g=0,shouldUpdatePosition:b=!0,isOpen:v=!0,onClose:y,maxHeight:x,arrowBoundaryOffset:E=0}=e,[S,k]=(0,h.useState)(null),T=[b,c,s.current,a.current,u.current,d,f,p,m,g,v,i,x,E,l],C=(0,h.useRef)(null==tf?void 0:tf.scale);(0,h.useEffect)(()=>{v&&(C.current=null==tf?void 0:tf.scale)},[v]);let P=(0,h.useCallback)(()=>{var e,t,n,r,o;if(!1===b||!v||!s.current||!a.current||!p||(null==tf?void 0:tf.scale)!==C.current)return;let h=null;if(u.current&&u.current.contains(document.activeElement)){let r=null===(e=document.activeElement)||void 0===e?void 0:e.getBoundingClientRect(),o=u.current.getBoundingClientRect();(h={type:"top",offset:(null!==(t=null==r?void 0:r.top)&&void 0!==t?t:0)-o.top}).offset>o.height/2&&(h.type="bottom",h.offset=(null!==(n=null==r?void 0:r.bottom)&&void 0!==n?n:0)-o.bottom)}let y=s.current;!x&&s.current&&(y.style.top="0px",y.style.bottom="",y.style.maxHeight=(null!==(o=null===(r=window.visualViewport)||void 0===r?void 0:r.height)&&void 0!==o?o:window.innerHeight)+"px");let w=function(e){var t,n,r,o;let i;let{placement:l,targetNode:a,overlayNode:s,scrollNode:u,padding:c,shouldFlip:d,boundaryElement:f,offset:p,crossOffset:h,maxHeight:m,arrowSize:g=0,arrowBoundaryOffset:b=0}=e,v=s instanceof HTMLElement?function(e){let t=e.offsetParent;if(t&&t===document.body&&"static"===window.getComputedStyle(t).position&&!tt(t)&&(t=document.documentElement),null==t)for(t=e.parentElement;t&&!tt(t);)t=t.parentElement;return t||document.documentElement}(s):document.documentElement,y=v===document.documentElement,w=window.getComputedStyle(v).position,x=y?e7(a):te(a,v);if(!y){let{marginTop:e,marginLeft:t}=window.getComputedStyle(a);x.top+=parseInt(e,10)||0,x.left+=parseInt(t,10)||0}let E=e7(s),S={top:parseInt((i=window.getComputedStyle(s)).marginTop,10)||0,bottom:parseInt(i.marginBottom,10)||0,left:parseInt(i.marginLeft,10)||0,right:parseInt(i.marginRight,10)||0};E.width+=(null!==(t=S.left)&&void 0!==t?t:0)+(null!==(n=S.right)&&void 0!==n?n:0),E.height+=(null!==(r=S.top)&&void 0!==r?r:0)+(null!==(o=S.bottom)&&void 0!==o?o:0);let k={top:u.scrollTop,left:u.scrollLeft,width:u.scrollWidth,height:u.scrollHeight},T=e5(f),C=e5(v),P="BODY"===f.tagName?e7(v):te(v,f);return"HTML"===v.tagName&&"BODY"===f.tagName&&(C.scroll.top=0,C.scroll.left=0),function(e,t,n,r,o,i,l,a,s,u,c,d,f,p,h,m){var g,b,v,y;let w=e3(e),{size:x,crossAxis:E,crossSize:S,placement:k,crossPlacement:T}=w,C=e9(t,a,n,w,c,d,u,f,h,m),P=c,M=e8(a,u,t,o,i+c,w);if(l&&r[x]>M){let e=e3(`${eJ[k]} ${T}`),r=e9(t,a,n,e,c,d,u,f,h,m);e8(a,u,t,o,i+c,e)>M&&(w=e,C=r,P=c)}let R="bottom";"top"===w.axis?"top"===w.placement?R="top":"bottom"===w.placement&&(R="bottom"):"top"===w.crossAxis&&("top"===w.crossPlacement?R="bottom":"bottom"===w.crossPlacement&&(R="top"));let j=e6(E,C[E],n[S],a,s,i,u);C[E]+=j;let O=function(e,t,n,r,o,i,l,a){var s,u,c,d,f,p,h;let m=r?n.height:t[e1.height],g=null!=e.top?n.top+e.top:n.top+(m-(null!==(s=e.bottom)&&void 0!==s?s:0)-l),b="top"!==a?Math.max(0,t.height+t.top+(null!==(u=t.scroll.top)&&void 0!==u?u:0)-g-((null!==(c=o.top)&&void 0!==c?c:0)+(null!==(d=o.bottom)&&void 0!==d?d:0)+i)):Math.max(0,g+l-(t.top+(null!==(f=t.scroll.top)&&void 0!==f?f:0))-((null!==(p=o.top)&&void 0!==p?p:0)+(null!==(h=o.bottom)&&void 0!==h?h:0)+i));return Math.min(t.height-2*i,b)}(C,a,u,f,o,i,n.height,R);p&&p<O&&(O=p),n.height=Math.min(n.height,O),j=e6(E,(C=e9(t,a,n,w,P,d,u,f,h,m))[E],n[S],a,s,i,u),C[E]+=j;let L={},A=t[E]+.5*t[S]-C[E]-o[eZ[E]],_=h/2+m,N="left"===eZ[E]?(null!==(g=o.left)&&void 0!==g?g:0)+(null!==(b=o.right)&&void 0!==b?b:0):(null!==(v=o.top)&&void 0!==v?v:0)+(null!==(y=o.bottom)&&void 0!==y?y:0),$=n[S]-N-h/2-m,I=eK(A,t[E]+h/2-(C[E]+o[eZ[E]]),t[E]+t[S]-h/2-(C[E]+o[eZ[E]]));return L[E]=eK(I,_,$),{position:C,maxHeight:O,arrowOffsetLeft:L.left,arrowOffsetTop:L.top,placement:w.placement}}(l,x,E,k,S,c,d,T,C,P,p,h,!!w&&"static"!==w,m,g,b)}({placement:"rtl"===i?c.replace("start","right").replace("end","left"):c.replace("start","left").replace("end","right"),overlayNode:s.current,targetNode:a.current,scrollNode:u.current||s.current,padding:d,shouldFlip:f,boundaryElement:p,offset:m,crossOffset:g,maxHeight:x,arrowSize:l,arrowBoundaryOffset:E});if(w.position){if(y.style.top="",y.style.bottom="",y.style.left="",y.style.right="",Object.keys(w.position).forEach(e=>y.style[e]=w.position[e]+"px"),y.style.maxHeight=null!=w.maxHeight?w.maxHeight+"px":"",h&&document.activeElement&&u.current){let e=document.activeElement.getBoundingClientRect(),t=u.current.getBoundingClientRect(),n=e[h.type]-t[h.type];u.current.scrollTop+=n-h.offset}k(w)}},T);w(P,T),w(()=>(window.addEventListener("resize",r,!1),()=>{window.removeEventListener("resize",r,!1)}),[r=P]),tr({ref:s,onResize:P}),tr({ref:a,onResize:P});let M=(0,h.useRef)(!1);w(()=>{let e;let t=()=>{M.current=!0,clearTimeout(e),e=setTimeout(()=>{M.current=!1},500),P()},n=()=>{M.current&&t()};return null==tf||tf.addEventListener("resize",t),null==tf||tf.addEventListener("scroll",n),()=>{null==tf||tf.removeEventListener("resize",t),null==tf||tf.removeEventListener("scroll",n)}},[P]);let j=(0,h.useCallback)(()=>{M.current||null==y||y()},[y,M]);return function(e){let{triggerRef:t,isOpen:n,onClose:r}=e;(0,h.useEffect)(()=>{if(!n||null===r)return;let e=e=>{let n=e.target;if(!t.current||n instanceof Node&&!n.contains(t.current)||e.target instanceof HTMLInputElement||e.target instanceof HTMLTextAreaElement)return;let o=r||tn.get(t.current);o&&o()};return window.addEventListener("scroll",e,!0),()=>{window.removeEventListener("scroll",e,!0)}},[n,r,t])}({triggerRef:a,isOpen:v,onClose:y&&j}),{overlayProps:{style:{position:"absolute",zIndex:1e5,...null==S?void 0:S.position,maxHeight:null!==(t=null==S?void 0:S.maxHeight)&&void 0!==t?t:"100vh"}},placement:null!==(n=null==S?void 0:S.placement)&&void 0!==n?n:null,arrowProps:{"aria-hidden":"true",role:"presentation",style:{left:null==S?void 0:S.arrowOffsetLeft,top:null==S?void 0:S.arrowOffsetTop}},updatePosition:P}}({isOpen:Q,targetRef:K,placement:nS(k),overlayRef:Z,offset:M?L+3:L,crossOffset:_,shouldFlip:E,containerPadding:S});nT(()=>{W.length&&ei()},W);let{overlayProps:el}=function(e,t){let{onClose:n,shouldCloseOnBlur:r,isOpen:o,isDismissable:i=!1,isKeyboardDismissDisabled:l=!1,shouldCloseOnInteractOutside:a}=e;(0,h.useEffect)(()=>(o&&tw.push(t),()=>{let e=tw.indexOf(t);e>=0&&tw.splice(e,1)}),[o,t]);let s=()=>{tw[tw.length-1]===t&&n&&n()};!function(e){let{ref:t,onInteractOutside:n,isDisabled:r,onInteractOutsideStart:o}=e,i=(0,h.useRef)({isPointerDown:!1,ignoreEmulatedMouseEvents:!1}),l=tb(e=>{n&&tv(e,t)&&(o&&o(e),i.current.isPointerDown=!0)}),a=tb(e=>{n&&n(e)});(0,h.useEffect)(()=>{let e=i.current;if(r)return;let n=q(t.current);if("undefined"!=typeof PointerEvent){let r=n=>{e.isPointerDown&&tv(n,t)&&a(n),e.isPointerDown=!1};return n.addEventListener("pointerdown",l,!0),n.addEventListener("pointerup",r,!0),()=>{n.removeEventListener("pointerdown",l,!0),n.removeEventListener("pointerup",r,!0)}}{let r=n=>{e.ignoreEmulatedMouseEvents?e.ignoreEmulatedMouseEvents=!1:e.isPointerDown&&tv(n,t)&&a(n),e.isPointerDown=!1},o=n=>{e.ignoreEmulatedMouseEvents=!0,e.isPointerDown&&tv(n,t)&&a(n),e.isPointerDown=!1};return n.addEventListener("mousedown",l,!0),n.addEventListener("mouseup",r,!0),n.addEventListener("touchstart",l,!0),n.addEventListener("touchend",o,!0),()=>{n.removeEventListener("mousedown",l,!0),n.removeEventListener("mouseup",r,!0),n.removeEventListener("touchstart",l,!0),n.removeEventListener("touchend",o,!0)}}},[t,r,l,a])}({ref:t,onInteractOutside:i&&o?e=>{(!a||a(e.target))&&(tw[tw.length-1]===t&&(e.stopPropagation(),e.preventDefault()),s())}:void 0,onInteractOutsideStart:e=>{(!a||a(e.target))&&tw[tw.length-1]===t&&(e.stopPropagation(),e.preventDefault())}});let{focusWithinProps:u}=function(e){let t,n,{isDisabled:r,onBlurWithin:o,onFocusWithin:i,onFocusWithinChange:l}=e,a=(0,h.useRef)({isFocusWithin:!1}),s=(0,h.useCallback)(e=>{a.current.isFocusWithin&&!e.currentTarget.contains(e.relatedTarget)&&(a.current.isFocusWithin=!1,o&&o(e),l&&l(!1))},[o,l,a]),u=(t=(0,h.useRef)({isFocused:!1,observer:null}),w(()=>{let e=t.current;return()=>{e.observer&&(e.observer.disconnect(),e.observer=null)}},[]),n=tb(e=>{null==s||s(e)}),(0,h.useCallback)(e=>{if(e.target instanceof HTMLButtonElement||e.target instanceof HTMLInputElement||e.target instanceof HTMLTextAreaElement||e.target instanceof HTMLSelectElement){t.current.isFocused=!0;let r=e.target;r.addEventListener("focusout",e=>{t.current.isFocused=!1,r.disabled&&n(new ty("blur",e)),t.current.observer&&(t.current.observer.disconnect(),t.current.observer=null)},{once:!0}),t.current.observer=new MutationObserver(()=>{if(t.current.isFocused&&r.disabled){var e;null===(e=t.current.observer)||void 0===e||e.disconnect();let n=r===document.activeElement?null:document.activeElement;r.dispatchEvent(new FocusEvent("blur",{relatedTarget:n})),r.dispatchEvent(new FocusEvent("focusout",{bubbles:!0,relatedTarget:n}))}}),t.current.observer.observe(r,{attributes:!0,attributeFilter:["disabled"]})}},[n])),c=(0,h.useCallback)(e=>{a.current.isFocusWithin||document.activeElement!==e.target||(i&&i(e),l&&l(!0),a.current.isFocusWithin=!0,u(e))},[i,l,u]);return r?{focusWithinProps:{onFocus:void 0,onBlur:void 0}}:{focusWithinProps:{onFocus:c,onBlur:s}}}({isDisabled:!r,onBlurWithin:e=>{!(!e.relatedTarget||function(e,t=null){if(e instanceof Element&&e.closest("[data-react-aria-top-layer]"))return!0;for(let{scopeRef:n}of tg.traverse(tg.getTreeNode(t)))if(n&&tp(e,n.current))return!0;return!1}(e.relatedTarget,null))&&(!a||a(e.relatedTarget))&&(null==n||n())}});return{overlayProps:{onKeyDown:e=>{"Escape"!==e.key||l||e.nativeEvent.isComposing||(e.stopPropagation(),e.preventDefault(),s())},...u},underlayProps:{onPointerDown:e=>{e.target===e.currentTarget&&e.preventDefault()}}}}({isOpen:Q,onClose:X.close,isDismissable:N,shouldCloseOnBlur:$,isKeyboardDismissDisabled:D,shouldCloseOnInteractOutside:H},Z),ea=(0,h.useMemo)(()=>{var t,n,r;return nf({...i,disableAnimation:G,radius:null!=(t=null==e?void 0:e.radius)?t:"md",size:null!=(n=null==e?void 0:e.size)?n:"md",shadow:null!=(r=null==e?void 0:e.shadow)?r:"sm"})},[function(e){if(!e||"object"!=typeof e)return"";try{return JSON.stringify(e)}catch(e){return""}}(i),G,null==e?void 0:e.radius,null==e?void 0:e.size,null==e?void 0:e.shadow]),es=(0,h.useCallback)((e={},t=null)=>({...A(ee,e),ref:function(...e){return t=>{e.forEach(e=>(function(e,t){if(null!=e){if("function"==typeof e){e(t);return}try{e.current=t}catch(n){throw Error(`Cannot assign value '${t}' to ref '${e}'`)}}})(e,t))}}(t,K),"aria-describedby":Q?J:void 0}),[ee,Q,J,X]),eu=(0,h.useCallback)(()=>({ref:Z,"data-slot":"base","data-open":nx(Q),"data-arrow":nx(M),"data-disabled":nx(p),"data-placement":nk(eo||"top",k),...A(en,el,Y),style:A(er.style,Y.style,o.style),className:ea.base({class:null==V?void 0:V.base}),id:J}),[ea,Q,M,p,eo,k,en,el,Y,er,o,J]),ef=(0,h.useCallback)(()=>({"data-slot":"content","data-open":nx(Q),"data-arrow":nx(M),"data-disabled":nx(p),"data-placement":nk(eo||"top",k),className:ea.content({class:function(...e){for(var t,n,r=0,o="";r<e.length;)(t=e[r++])&&(n=function e(t){var n,r,o="";if("string"==typeof t||"number"==typeof t)o+=t;else if("object"==typeof t){if(Array.isArray(t))for(n=0;n<t.length;n++)t[n]&&(r=e(t[n]))&&(o&&(o+=" "),o+=r);else for(n in t)t[n]&&(o&&(o+=" "),o+=n)}return o}(t))&&(o&&(o+=" "),o+=n);return o}(null==V?void 0:V.content,F)})}),[ea,Q,M,p,eo,k,V]);return{Component:a||"div",content:u,children:c,isOpen:Q,triggerRef:K,showArrow:M,portalContainer:I,placement:k,disableAnimation:G,isDisabled:p,motionProps:B,getTooltipContentProps:ef,getTriggerProps:es,getTooltipProps:eu}}({...e,ref:t});try{let e=h.Children.count(o);if(1!==e)throw Error();n=(0,h.isValidElement)(o)?(0,h.cloneElement)(o,d(o.props,o.ref)):(0,p.jsx)("p",{...d(),children:o})}catch(e){n=(0,p.jsx)("span",{}),function(e,t,...n){let r=`[Next UI] : ${e}`;"undefined"!=typeof console&&(rg[r]||(rg[r]=!0))}("Tooltip must have only one child node. Please, check your code.")}let{ref:E,id:S,style:k,...C}=f(),P=(0,p.jsx)("div",{ref:E,id:S,style:k,children:(0,p.jsx)(nN,{features:rb,children:(0,p.jsx)(rl.div,{animate:"enter",exit:"exit",initial:"exit",variants:rm,...A(c,C),style:{...nE(s)},children:(0,p.jsx)(r,{...x(),children:i})})})});return(0,p.jsxs)(p.Fragment,{children:[n,u&&l?(0,p.jsx)(nO,{portalContainer:a,children:(0,p.jsx)("div",{ref:E,id:S,style:k,...C,children:(0,p.jsx)(r,{...x(),children:i})})}):(0,p.jsx)(rh,{children:l?(0,p.jsx)(nO,{portalContainer:a,children:P}):null})]})},(0,h.forwardRef)(d));function ry(e){let t=Object.prototype.toString.call(e);return e instanceof Date||"object"==typeof e&&"[object Date]"===t?new e.constructor(+e):new Date("number"==typeof e||"[object Number]"===t||"string"==typeof e||"[object String]"===t?e:NaN)}function rw(e,t){let n=ry(e),r=ry(t),o=n.getTime()-r.getTime();return o<0?-1:o>0?1:o}rv.displayName="NextUI.Tooltip";let rx={lessThanXSeconds:{one:"less than a second",other:"less than {{count}} seconds"},xSeconds:{one:"1 second",other:"{{count}} seconds"},halfAMinute:"half a minute",lessThanXMinutes:{one:"less than a minute",other:"less than {{count}} minutes"},xMinutes:{one:"1 minute",other:"{{count}} minutes"},aboutXHours:{one:"about 1 hour",other:"about {{count}} hours"},xHours:{one:"1 hour",other:"{{count}} hours"},xDays:{one:"1 day",other:"{{count}} days"},aboutXWeeks:{one:"about 1 week",other:"about {{count}} weeks"},xWeeks:{one:"1 week",other:"{{count}} weeks"},aboutXMonths:{one:"about 1 month",other:"about {{count}} months"},xMonths:{one:"1 month",other:"{{count}} months"},aboutXYears:{one:"about 1 year",other:"about {{count}} years"},xYears:{one:"1 year",other:"{{count}} years"},overXYears:{one:"over 1 year",other:"over {{count}} years"},almostXYears:{one:"almost 1 year",other:"almost {{count}} years"}};function rE(e){return function(){let t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},n=t.width?String(t.width):e.defaultWidth;return e.formats[n]||e.formats[e.defaultWidth]}}let rS={date:rE({formats:{full:"EEEE, MMMM do, y",long:"MMMM do, y",medium:"MMM d, y",short:"MM/dd/yyyy"},defaultWidth:"full"}),time:rE({formats:{full:"h:mm:ss a zzzz",long:"h:mm:ss a z",medium:"h:mm:ss a",short:"h:mm a"},defaultWidth:"full"}),dateTime:rE({formats:{full:"{{date}} 'at' {{time}}",long:"{{date}} 'at' {{time}}",medium:"{{date}}, {{time}}",short:"{{date}}, {{time}}"},defaultWidth:"full"})},rk={lastWeek:"'last' eeee 'at' p",yesterday:"'yesterday at' p",today:"'today at' p",tomorrow:"'tomorrow at' p",nextWeek:"eeee 'at' p",other:"P"};function rT(e){return(t,n)=>{let r;if("formatting"===((null==n?void 0:n.context)?String(n.context):"standalone")&&e.formattingValues){let t=e.defaultFormattingWidth||e.defaultWidth,o=(null==n?void 0:n.width)?String(n.width):t;r=e.formattingValues[o]||e.formattingValues[t]}else{let t=e.defaultWidth,o=(null==n?void 0:n.width)?String(n.width):e.defaultWidth;r=e.values[o]||e.values[t]}return r[e.argumentCallback?e.argumentCallback(t):t]}}function rC(e){return function(t){let n,r=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},o=r.width,i=o&&e.matchPatterns[o]||e.matchPatterns[e.defaultMatchWidth],l=t.match(i);if(!l)return null;let a=l[0],s=o&&e.parsePatterns[o]||e.parsePatterns[e.defaultParseWidth],u=Array.isArray(s)?function(e,t){for(let n=0;n<e.length;n++)if(t(e[n]))return n}(s,e=>e.test(a)):function(e,t){for(let n in e)if(Object.prototype.hasOwnProperty.call(e,n)&&t(e[n]))return n}(s,e=>e.test(a));return n=e.valueCallback?e.valueCallback(u):u,{value:n=r.valueCallback?r.valueCallback(n):n,rest:t.slice(a.length)}}}let rP={code:"en-US",formatDistance:(e,t,n)=>{let r;let o=rx[e];return(r="string"==typeof o?o:1===t?o.one:o.other.replace("{{count}}",t.toString()),null==n?void 0:n.addSuffix)?n.comparison&&n.comparison>0?"in "+r:r+" ago":r},formatLong:rS,formatRelative:(e,t,n,r)=>rk[e],localize:{ordinalNumber:(e,t)=>{let n=Number(e),r=n%100;if(r>20||r<10)switch(r%10){case 1:return n+"st";case 2:return n+"nd";case 3:return n+"rd"}return n+"th"},era:rT({values:{narrow:["B","A"],abbreviated:["BC","AD"],wide:["Before Christ","Anno Domini"]},defaultWidth:"wide"}),quarter:rT({values:{narrow:["1","2","3","4"],abbreviated:["Q1","Q2","Q3","Q4"],wide:["1st quarter","2nd quarter","3rd quarter","4th quarter"]},defaultWidth:"wide",argumentCallback:e=>e-1}),month:rT({values:{narrow:["J","F","M","A","M","J","J","A","S","O","N","D"],abbreviated:["Jan","Feb","Mar","Apr","May","Jun","Jul","Aug","Sep","Oct","Nov","Dec"],wide:["January","February","March","April","May","June","July","August","September","October","November","December"]},defaultWidth:"wide"}),day:rT({values:{narrow:["S","M","T","W","T","F","S"],short:["Su","Mo","Tu","We","Th","Fr","Sa"],abbreviated:["Sun","Mon","Tue","Wed","Thu","Fri","Sat"],wide:["Sunday","Monday","Tuesday","Wednesday","Thursday","Friday","Saturday"]},defaultWidth:"wide"}),dayPeriod:rT({values:{narrow:{am:"a",pm:"p",midnight:"mi",noon:"n",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"},abbreviated:{am:"AM",pm:"PM",midnight:"midnight",noon:"noon",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"},wide:{am:"a.m.",pm:"p.m.",midnight:"midnight",noon:"noon",morning:"morning",afternoon:"afternoon",evening:"evening",night:"night"}},defaultWidth:"wide",formattingValues:{narrow:{am:"a",pm:"p",midnight:"mi",noon:"n",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"},abbreviated:{am:"AM",pm:"PM",midnight:"midnight",noon:"noon",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"},wide:{am:"a.m.",pm:"p.m.",midnight:"midnight",noon:"noon",morning:"in the morning",afternoon:"in the afternoon",evening:"in the evening",night:"at night"}},defaultFormattingWidth:"wide"})},match:{ordinalNumber:(f={matchPattern:/^(\d+)(th|st|nd|rd)?/i,parsePattern:/\d+/i,valueCallback:e=>parseInt(e,10)},function(e){let t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{},n=e.match(f.matchPattern);if(!n)return null;let r=n[0],o=e.match(f.parsePattern);if(!o)return null;let i=f.valueCallback?f.valueCallback(o[0]):o[0];return{value:i=t.valueCallback?t.valueCallback(i):i,rest:e.slice(r.length)}}),era:rC({matchPatterns:{narrow:/^(b|a)/i,abbreviated:/^(b\.?\s?c\.?|b\.?\s?c\.?\s?e\.?|a\.?\s?d\.?|c\.?\s?e\.?)/i,wide:/^(before christ|before common era|anno domini|common era)/i},defaultMatchWidth:"wide",parsePatterns:{any:[/^b/i,/^(a|c)/i]},defaultParseWidth:"any"}),quarter:rC({matchPatterns:{narrow:/^[1234]/i,abbreviated:/^q[1234]/i,wide:/^[1234](th|st|nd|rd)? quarter/i},defaultMatchWidth:"wide",parsePatterns:{any:[/1/i,/2/i,/3/i,/4/i]},defaultParseWidth:"any",valueCallback:e=>e+1}),month:rC({matchPatterns:{narrow:/^[jfmasond]/i,abbreviated:/^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)/i,wide:/^(january|february|march|april|may|june|july|august|september|october|november|december)/i},defaultMatchWidth:"wide",parsePatterns:{narrow:[/^j/i,/^f/i,/^m/i,/^a/i,/^m/i,/^j/i,/^j/i,/^a/i,/^s/i,/^o/i,/^n/i,/^d/i],any:[/^ja/i,/^f/i,/^mar/i,/^ap/i,/^may/i,/^jun/i,/^jul/i,/^au/i,/^s/i,/^o/i,/^n/i,/^d/i]},defaultParseWidth:"any"}),day:rC({matchPatterns:{narrow:/^[smtwf]/i,short:/^(su|mo|tu|we|th|fr|sa)/i,abbreviated:/^(sun|mon|tue|wed|thu|fri|sat)/i,wide:/^(sunday|monday|tuesday|wednesday|thursday|friday|saturday)/i},defaultMatchWidth:"wide",parsePatterns:{narrow:[/^s/i,/^m/i,/^t/i,/^w/i,/^t/i,/^f/i,/^s/i],any:[/^su/i,/^m/i,/^tu/i,/^w/i,/^th/i,/^f/i,/^sa/i]},defaultParseWidth:"any"}),dayPeriod:rC({matchPatterns:{narrow:/^(a|p|mi|n|(in the|at) (morning|afternoon|evening|night))/i,any:/^([ap]\.?\s?m\.?|midnight|noon|(in the|at) (morning|afternoon|evening|night))/i},defaultMatchWidth:"any",parsePatterns:{any:{am:/^a/i,pm:/^p/i,midnight:/^mi/i,noon:/^no/i,morning:/morning/i,afternoon:/afternoon/i,evening:/evening/i,night:/night/i}},defaultParseWidth:"any"})},options:{weekStartsOn:0,firstWeekContainsDate:1}},rM={};function rR(e){let t=ry(e),n=new Date(Date.UTC(t.getFullYear(),t.getMonth(),t.getDate(),t.getHours(),t.getMinutes(),t.getSeconds(),t.getMilliseconds()));return n.setUTCFullYear(t.getFullYear()),+e-+n}var rj=n(62935),rO=n(51214),rL=n(41664),rA=n.n(rL);let r_=rO.nb.REFRESH_INTERVAL,rN={placement:"bottom",color:"default"},r$=e=>{let{children:t,...n}=e,r=n.content;return n.content=void 0,(0,p.jsx)(rv,{...rN,...n,content:(0,p.jsx)("span",{className:"left-full w-max px-2 py-1 text-sm text-gray-100 bg-gray-500 text-sm capitalize rounded",children:r}),children:t})},rI=e=>{let{children:t,...n}=e,r=n.content;return n.content=void 0,(0,p.jsx)(rv,{...rN,...n,content:(0,p.jsx)("span",{className:"left-full w-max px-2 py-1 text-sm text-gray-100 bg-gray-500 text-sm rounded whitespace-pre-line",children:r}),children:t})},rD=e=>{let t,{timestamp:n,className:r=""}=e,[,o]=(0,h.useState)(0);if((0,h.useEffect)(()=>{if(!n)return;let e=setInterval(()=>{o(e=>e+1)},1e4);return()=>clearInterval(e)},[n]),!n)return null;let i=new Date-n;return t=i<5e3?"just now":i<6e4?"".concat(Math.floor(i/1e3),"s ago"):i<36e5?"".concat(Math.floor(i/6e4),"m ago"):i<864e5?"".concat(Math.floor(i/36e5),"h ago"):"".concat(Math.floor(i/864e5),"d ago"),(0,p.jsx)(rI,{content:"Last updated: ".concat(n.toLocaleString("en-CA",{year:"numeric",month:"2-digit",day:"2-digit",hour:"numeric",minute:"2-digit",second:"2-digit",hour12:!1,timeZoneName:"short"}).replace(",","")),className:"text-sm text-muted-foreground",children:(0,p.jsxs)("span",{className:"text-xs text-gray-500 ".concat(r),children:["Updated ",t]})})};function rW(e){if(!e&&0!==e)return"-";let t=e=Math.floor(e),n="",r=0;for(let e of[{value:31536e3,label:"y"},{value:2592e3,label:"mo"},{value:86400,label:"d"},{value:3600,label:"h"},{value:60,label:"m"},{value:1,label:"s"}])if(t>=e.value&&r<2){let o=Math.floor(t/e.value);n+="".concat(o).concat(e.label," "),t%=e.value,r++}return n.trim()||"0s"}function rH(e){return e.replace(/\x1b\[([0-9]{1,2}(;[0-9]{1,2})?)?[mGKH]/g,"")}function rF(e){return!!(!e||e.includes("<sky-payload")||/rich_(init|update|start|exit)/i.test(e)||/<rich_.*?>.*<\/rich_.*?>/i.test(e)||/^├──/.test(e)||/^└──/.test(e))}function rz(e){let{logs:t,controller:n=!1}=e,[r,o]=(0,h.useState)("all"),i=t?Array.isArray(t)?t:"string"==typeof t?t.split("\n").filter(e=>""!==e):[]:[],[l,a]=(0,h.useState)(i),[s,u]=(0,h.useState)([]);return(0,h.useEffect)(()=>{u(function(e){let t;let n=/\((head|worker\d+),/g,r=new Set;for(;null!==(t=n.exec(e));)r.add(t[1]);return Array.from(r).sort((e,t)=>"head"===e?-1:"head"===t?1:e.localeCompare(t,void 0,{numeric:!0,sensitivity:"base"}))}(i.join("\n")))},[i]),(0,h.useEffect)(()=>{"all"===r?a(i):a(i.filter(e=>e.includes("(".concat(r,","))))},[r,i]),(0,p.jsxs)("div",{children:[(0,p.jsx)("style",{children:'\n  .logs-container {\n    background-color: #f7f7f7;\n    padding: 16px;\n    max-height: calc(100vh - 300px);\n    overflow-y: auto;\n    overflow-x: hidden;\n    font-family: ui-monospace, SFMono-Regular, Menlo, Monaco, Consolas, "Liberation Mono", "Courier New", monospace;\n    line-height: 1.5;\n    border-radius: 6px;\n    min-height: fit-content;\n  }\n\n  .log-line {\n    display: block;\n    white-space: pre-wrap;\n    margin: 2px 0;\n  }\n\n  .log-line .level {\n    display: inline;\n    width: 1ch;\n    margin-right: 1ch;\n    font-weight: bold;\n  }\n\n  .log-line.INFO .level {\n    color: #2563eb;\n  }\n\n  .log-line.WARNING .level {\n    color: #d97706;\n  }\n\n  .log-line.ERROR .level {\n    color: #dc2626;\n  }\n\n  .log-line.DEBUG .level {\n    color: #6b7280;\n  }\n\n  .log-line .timestamp {\n    color: #059669;\n    margin-right: 1ch;\n    white-space: nowrap;\n  }\n\n  .log-line .location {\n    color: #6366f1;\n    margin-right: 1ch;\n    white-space: nowrap;\n  }\n\n  .log-line .message {\n    color: #111827;\n    word-break: break-word;\n    white-space: pre-wrap;\n  }\n\n  .log-line .log-prefix {\n    color: #6366f1;\n    font-weight: 500;\n  }\n\n  .log-line .log-rest {\n    color: #111827;\n    word-break: break-word;\n    white-space: pre-wrap;\n  }\n'}),!n&&(0,p.jsx)("div",{style:{marginBottom:"1rem"},children:(0,p.jsxs)(rj.Ph,{onValueChange:e=>o(e),value:r,children:[(0,p.jsx)(rj.i4,{"aria-label":"Node",className:"focus:ring-0 focus:ring-offset-0",children:(0,p.jsx)(rj.ki,{placeholder:"Select Node"})}),(0,p.jsxs)(rj.Bw,{children:[(0,p.jsx)(rj.Ql,{value:"all",children:"All Nodes"}),s.map(e=>(0,p.jsx)(rj.Ql,{value:e,children:e},e))]})]})}),(0,p.jsx)("div",{className:"logs-container whitespace-pre-wrap break-all font-mono text-sm text-gray-900","aria-label":"job-logs",children:l.join("\n")})]})}function rB(e){let t,{date:n}=e;if(!n)return"N/A";let r=new Date,o=n.getFullYear()+"-"+String(n.getMonth()+1).padStart(2,"0")+"-"+String(n.getDate()).padStart(2,"0"),i=n.toLocaleString("en-US",{hour:"numeric",minute:"2-digit",second:"2-digit",hour12:!0,timeZoneName:"short"});return t=function(e){if(!e||"string"!=typeof e)return e;if("just now"===e)return"now";if("less than a minute ago"===e.toLowerCase())return"Less than 1m ago";let t=e.match(/^about\s+(\d+)\s+(\w+)\s+ago$/i);if(t){let e=t[1],n=t[2],r={second:"s",seconds:"s",minute:"m",minutes:"m",hour:"h",hours:"h",day:"d",days:"d",month:"mo",months:"mo",year:"yr",years:"yr"};if(r[n])return"".concat(e).concat(r[n]," ago")}let n=e.match(/^a[n]?\s+(\w+)\s+ago$/i);if(n){let e=n[1],t={second:"s",minute:"m",hour:"h",day:"d",month:"mo",year:"yr"};if(t[e])return"1".concat(t[e]," ago")}let r=e.match(/^(\d+)\s+(\w+)\s+ago$/i);if(r){let e=r[1],t=r[2],n={second:"s",seconds:"s",minute:"m",minutes:"m",hour:"h",hours:"h",day:"d",days:"d",month:"mo",months:"mo",year:"yr",years:"yr"};if(n[t])return"".concat(e).concat(n[t]," ago")}return e}(function(e,t,n){var r,o,i,l,a,s;let u,c,d;let f=null!==(o=null!==(r=null==n?void 0:n.locale)&&void 0!==r?r:rM.locale)&&void 0!==o?o:rP,p=rw(e,t);if(isNaN(p))throw RangeError("Invalid time value");let h=Object.assign({},n,{addSuffix:null==n?void 0:n.addSuffix,comparison:p});p>0?(u=ry(t),c=ry(e)):(u=ry(e),c=ry(t));let m=(i=c,l=u,(s=null==void 0?void 0:(void 0).roundingMethod,e=>{let t=(s?Math[s]:Math.trunc)(e);return 0===t?0:t})((+ry(i)-+ry(l))/1e3)),g=Math.round((m-(rR(c)-rR(u))/1e3)/60);if(g<2){if(null==n?void 0:n.includeSeconds){if(m<5)return f.formatDistance("lessThanXSeconds",5,h);if(m<10)return f.formatDistance("lessThanXSeconds",10,h);if(m<20)return f.formatDistance("lessThanXSeconds",20,h);if(m<40)return f.formatDistance("halfAMinute",0,h);else if(m<60)return f.formatDistance("lessThanXMinutes",1,h);else return f.formatDistance("xMinutes",1,h)}return 0===g?f.formatDistance("lessThanXMinutes",1,h):f.formatDistance("xMinutes",g,h)}if(g<45)return f.formatDistance("xMinutes",g,h);if(g<90)return f.formatDistance("aboutXHours",1,h);if(g<1440)return f.formatDistance("aboutXHours",Math.round(g/60),h);if(g<2520)return f.formatDistance("xDays",1,h);if(g<43200)return f.formatDistance("xDays",Math.round(g/1440),h);if(g<86400)return d=Math.round(g/43200),f.formatDistance("aboutXMonths",d,h);if((d=function(e,t){let n;let r=ry(e),o=ry(t),i=rw(r,o),l=Math.abs(function(e,t){let n=ry(e),r=ry(t);return 12*(n.getFullYear()-r.getFullYear())+(n.getMonth()-r.getMonth())}(r,o));if(l<1)n=0;else{1===r.getMonth()&&r.getDate()>27&&r.setDate(30),r.setMonth(r.getMonth()-i*l);let t=rw(r,o)===-i;(function(e){let t=ry(e);return+function(e){let t=ry(e);return t.setHours(23,59,59,999),t}(t)==+function(e){let t=ry(e),n=t.getMonth();return t.setFullYear(t.getFullYear(),n+1,0),t.setHours(23,59,59,999),t}(t)})(ry(e))&&1===l&&1===rw(e,o)&&(t=!1),n=i*(l-Number(t))}return 0===n?0:n}(c,u))<12)return f.formatDistance("xMonths",Math.round(g/43200),h);{let e=d%12,t=Math.trunc(d/12);return e<3?f.formatDistance("aboutXYears",t,h):e<9?f.formatDistance("overXYears",t,h):f.formatDistance("almostXYears",t+1,h)}}(n,r,{addSuffix:!0})),(0,p.jsx)(r$,{content:o+" "+i,className:"text-sm text-muted-foreground",children:(0,p.jsx)("span",{className:"border-b border-dotted border-gray-400 cursor-help",children:t})})}function rq(e){return e?e.getFullYear()+"-"+String(e.getMonth()+1).padStart(2,"0")+"-"+String(e.getDate()).padStart(2,"0")+" "+e.toLocaleString("en-US",{hour:"numeric",minute:"2-digit",second:"2-digit",hour12:!0,timeZoneName:"short"}):"N/A"}let rV=e=>e&&e.jobCounts?e.jobCounts:{},rU=e=>{if(!e||0===e.length)return{};let t=e.filter(e=>"READY"===e.status),n={};t.forEach(e=>{try{let t=e.cloud&&""!==e.cloud.trim()&&"undefined"!==e.cloud,r=e.region&&"undefined"!==e.region&&null!==e.region&&""!==e.region.trim();if(!t&&!r)return;let o=t?e.cloud:"Unknown",i=r?e.region:null;n[o]||(n[o]={count:0,regions:new Set}),n[o].count+=1,i&&n[o].regions.add(i)}catch(e){n.Unknown||(n.Unknown={count:0,regions:new Set}),n.Unknown.count+=1}});let r={};return Object.entries(n).forEach(e=>{let[t,n]=e,o=n.regions.size,i=t.toLowerCase().includes("kubernetes")||t.toLowerCase().includes("k8s")?"context":"region",l=1===o?"1 ".concat(i):"".concat(o," ").concat(i,"s");r[o>0?"".concat(t," (").concat(l,")"):t]=n.count}),r},rY=e=>{let{jobCounts:t,getStatusStyle:n}=e;return t&&0!==Object.keys(t).length?(0,p.jsx)("div",{className:"flex flex-wrap gap-1",children:Object.entries(t).map(e=>{let[t,r]=e,o=n(t);return(0,p.jsxs)("span",{className:"px-2 py-1 rounded-full flex items-center space-x-2 text-xs font-medium ".concat(o),children:[(0,p.jsx)("span",{children:t}),(0,p.jsx)("span",{className:"text-xs bg-white/50 px-1.5 py-0.5 rounded",children:r})]},t)})}):(0,p.jsx)("span",{className:"text-gray-1000",children:"No active jobs"})},rG=e=>{let{replicaInfo:t}=e,n=rU(t);if(0===Object.keys(n).length)return(0,p.jsx)("span",{className:"text-gray-500 text-sm",children:"-"});let r=rO.MO.NAME_TRUNCATE_LENGTH,o=e=>{let t=e.indexOf("(");if(-1===t)return e;let n=e.substring(0,t).trim(),o=e.substring(t+1,e.length-1);if(o.length<=r)return e;let i="".concat(o.substring(0,Math.floor((r-3)/2)),"...").concat(o.substring(o.length-Math.ceil((r-3)/2)));return"".concat(n," (").concat(i,")")};return(0,p.jsx)("div",{className:"flex flex-wrap gap-1",children:Object.entries(n).map(e=>{let[t,n]=e,r=o(t),i=r!==t;return(0,p.jsxs)("span",{className:"px-2 py-1 rounded-full flex items-center space-x-2 text-xs font-medium bg-blue-50 text-blue-700",children:[i?(0,p.jsx)(rI,{content:t,className:"text-sm text-muted-foreground",children:(0,p.jsx)("span",{children:r})}):(0,p.jsx)("span",{children:t}),(0,p.jsx)("span",{className:"text-xs bg-white/50 px-1.5 py-0.5 rounded",children:n})]},t)})})},rX=(e,t,n)=>{if(!e)return"-";if(n.find(n=>n.name===e&&n.hash===t)&&t)return(0,p.jsxs)("div",{className:"flex items-center space-x-2",children:[(0,p.jsx)(rI,{content:"This pool is running",placement:"top",children:(0,p.jsx)("div",{className:"w-2 h-2 bg-green-700 rounded-full"})}),(0,p.jsx)(rA(),{href:"/jobs/pools/".concat(e),className:"text-gray-700 hover:text-blue-600 hover:underline",children:e})]});let r=t?t.substring(0,4):"";return(0,p.jsxs)("div",{className:"flex items-center space-x-2",children:[(0,p.jsx)(rI,{content:"This pool is terminated",placement:"top",children:(0,p.jsx)("div",{className:"w-2 h-2 bg-gray-800"})}),(0,p.jsxs)("span",{children:[e," (",r,")"]})]})}},47145:function(e,t,n){"use strict";n.d(t,{x:function(){return i}});var r=n(94545),o=n(93225);let i={fetchImmediate:async function(e,t){let n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:"POST",r=arguments.length>3&&void 0!==arguments[3]?arguments[3]:{},i="POST"===n?{"Content-Type":"application/json",...r.headers||{}}:{...r.headers||{}},l=window.location.origin,a="".concat(l).concat(o.f4).concat(e);return void 0!==t&&(t.env_vars={...t.env_vars||{},SKYPILOT_IS_FROM_DASHBOARD:"true",SKYPILOT_USER_ID:"dashboard",SKYPILOT_USER:"dashboard"}),await fetch(a,{method:n,headers:i,body:"POST"===n?JSON.stringify(t):void 0,signal:r.signal})},fetch:async function(e,t){let n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:"POST";try{let l=window.location.origin,a=await i.fetchImmediate(e,t,n);if(!a.ok){let t="Initial API request ".concat(e," failed with status ").concat(a.status);throw Error(t)}let s=a.headers.get("X-Skypilot-Request-ID")||a.headers.get("X-Request-ID");if(!s){let t="No request ID received from server for ".concat(e);throw Error(t)}let u=await fetch("".concat(l).concat(o.f4,"/api/get?request_id=").concat(s));if(!u.ok){let t=await (0,r.oN)(u),n="API request to get ".concat(e," result failed with status ").concat(u.status,", error: ").concat(t);throw Error(n)}let c=await u.json();return c.return_value?JSON.parse(c.return_value):[]}catch(n){let t="Error in apiClient.fetch for ".concat(e,": ").concat(n);throw console.error(t),Error(t)}},post:async function(e,t){let n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{},r={"Content-Type":"application/json",...n.headers||{}},i=window.location.origin,l="".concat(i).concat(o.f4).concat(e);return await fetch(l,{method:"POST",headers:r,body:JSON.stringify(t),signal:n.signal})},stream:async function(e,t,n){let r=arguments.length>3&&void 0!==arguments[3]?arguments[3]:{},o=await i.post(e,t,r);if(!o.ok){let t="API request ".concat(e," failed with status ").concat(o.status);throw console.error(t),Error(t)}let l=o.body.getReader();try{for(;;){let{done:e,value:t}=await l.read();if(e)break;let r=new TextDecoder().decode(t);n(r)}}catch(e){throw console.error("Error in stream:",e),e}},get:async e=>{let t=window.location.origin,n="".concat(t).concat(o.f4).concat(e);return await fetch(n)}}},93225:function(e,t,n){"use strict";n.d(t,{$m:function(){return s},Bo:function(){return l},GW:function(){return o},Z2:function(){return u},f4:function(){return r},iW:function(){return a},mF:function(){return i}}),n(83454);let r=function(){{let e=window.location.pathname,t=e.indexOf("/dashboard");if(-1!==t){let n=e.substring(0,t);return"".concat(n,"/internal/dashboard")}return"/internal/dashboard"}}(),o=function(){{let e=window.location.pathname,t=e.indexOf("/dashboard");return -1!==t?e.substring(0,t+10):"/dashboard"}}();"/api/v1".replace(/^http/,"ws");let i="ClusterDoesNotExist",l="NotSupportedError",a="ClusterNotUpError",s=["AWS","Azure","GCP","IBM","Lambda","SCP","OCI","RunPod","VAST","vSphere","Cudo","FluidStack","Paperspace","DO","Nebius"],u=Object.fromEntries([...s.map(e=>[e.toLowerCase(),e]),["kubernetes","Kubernetes"],["ssh","SSH Node Pool"]])},94545:function(e,t,n){"use strict";function r(e){return e.startsWith("sky-jobs-controller-")}function o(e,t,n){return null===t?e:[...e].sort((e,r)=>e[t]<r[t]?"ascending"===n?-1:1:e[t]>r[t]?"ascending"===n?1:-1:0)}async function i(e){let t=e.statusText;if(500===e.status)try{let n=await e.json();if(n.detail&&n.detail.error)try{t=JSON.parse(n.detail.error).message||String(n.detail.error)}catch(e){console.error("Error parsing JSON from data.detail.error:",e),t=String(n.detail.error)}}catch(e){console.error("Error parsing response JSON:",e),t=String(e)}return t}n.d(t,{R0:function(){return o},Ym:function(){return r},oN:function(){return i}})},6556:function(e,t,n){"use strict";n.d(t,{n:function(){return l}});var r=n(67294);let o="skypilot-dashboard-first-visit",i="skypilot-dashboard-tour-completed";function l(){let[e,t]=(0,r.useState)(!1),[n,l]=(0,r.useState)(!1),[a,s]=(0,r.useState)(!1);return(0,r.useEffect)(()=>{s(!0);let e=localStorage.getItem(o),n=localStorage.getItem(i);e||(t(!0),localStorage.setItem(o,"true")),n&&l(!0)},[]),{isFirstVisit:!!a&&e,tourCompleted:!!a&&n,shouldShowTourPrompt:a&&e&&!n,shouldPulseHelpButton:a&&e&&!n,markTourCompleted:()=>{localStorage.setItem(i,"true"),l(!0)},resetFirstVisit:()=>{localStorage.removeItem(o),localStorage.removeItem(i),t(!0),l(!1)},isClient:a}}},94849:function(e,t,n){"use strict";let r,o;n.d(t,{C:function(){return t3},r:function(){return t6}});var i,l,a=n(85893),s=n(67294),u=n(11163);function c(e){return e instanceof HTMLElement}function d(e){return"function"==typeof e}function f(e){return"string"==typeof e}function p(e){return void 0===e}class h{on(e,t,n,r=!1){var o;return p(this.bindings)&&(this.bindings={}),p(this.bindings[e])&&(this.bindings[e]=[]),null==(o=this.bindings[e])||o.push({handler:t,ctx:n,once:r}),this}once(e,t,n){return this.on(e,t,n,!0)}off(e,t){if(p(this.bindings)||p(this.bindings[e]))return this;if(p(t))delete this.bindings[e];else{var n;null==(n=this.bindings[e])||n.forEach((n,r)=>{if(n.handler===t){var o;null==(o=this.bindings[e])||o.splice(r,1)}})}return this}trigger(e,...t){if(!p(this.bindings)&&this.bindings[e]){var n;null==(n=this.bindings[e])||n.forEach((n,r)=>{let{ctx:o,handler:i,once:l}=n;if(i.apply(o||this,t),l){var a;null==(a=this.bindings[e])||a.splice(r,1)}})}return this}}function m(){return(m=Object.assign?Object.assign.bind():function(e){for(var t=1;t<arguments.length;t++){var n=arguments[t];for(var r in n)({}).hasOwnProperty.call(n,r)&&(e[r]=n[r])}return e}).apply(null,arguments)}function g(e,t){if(null==e)return{};var n={};for(var r in e)if(({}).hasOwnProperty.call(e,r)){if(t.includes(r))continue;n[r]=e[r]}return n}let b={defaultMerge:Symbol("deepmerge-ts: default merge"),skip:Symbol("deepmerge-ts: skip")};function v(e,t){return t}function y(e,t){return e.filter(e=>void 0!==e)}function w(e){return"object"!=typeof e||null===e?0:Array.isArray(e)?2:!function(e){if(!E.has(Object.prototype.toString.call(e)))return!1;let{constructor:t}=e;if(void 0===t)return!0;let n=t.prototype;return!!(null!==n&&"object"==typeof n&&E.has(Object.prototype.toString.call(n))&&n.hasOwnProperty("isPrototypeOf"))}(e)?e instanceof Set?3:e instanceof Map?4:5:1}function x(e){return{*[Symbol.iterator](){for(let t of e)for(let e of t)yield e}}}b.defaultMerge,(i=l||(l={}))[i.NOT=0]="NOT",i[i.RECORD=1]="RECORD",i[i.ARRAY=2]="ARRAY",i[i.SET=3]="SET",i[i.MAP=4]="MAP",i[i.OTHER=5]="OTHER";let E=new Set(["[object Object]","[object Module]"]),S={mergeRecords:function(e,t,n){let r={};for(let o of function(e){let t=new Set;for(let n of e)for(let e of[...Object.keys(n),...Object.getOwnPropertySymbols(n)])t.add(e);return t}(e)){let i=[];for(let t of e)"object"==typeof t&&Object.prototype.propertyIsEnumerable.call(t,o)&&i.push(t[o]);if(0===i.length)continue;let l=t.metaDataUpdater(n,{key:o,parents:e}),a=T(i,t,l);a!==b.skip&&("__proto__"===o?Object.defineProperty(r,o,{value:a,configurable:!0,enumerable:!0,writable:!0}):r[o]=a)}return r},mergeArrays:function(e){return e.flat()},mergeSets:function(e){return new Set(x(e))},mergeMaps:function(e){return new Map(x(e))},mergeOthers:function(e){return e.at(-1)}};function k(...e){return(function(e,t){var n,r,o;let i={defaultMergeFunctions:S,mergeFunctions:m({},S,Object.fromEntries(Object.entries(e).filter(([e,t])=>Object.hasOwn(S,e)).map(([e,t])=>!1===t?[e,S.mergeOthers]:[e,t]))),metaDataUpdater:null!=(n=e.metaDataUpdater)?n:v,deepmerge:l,useImplicitDefaultMerging:null!=(r=e.enableImplicitDefaultMerging)&&r,filterValues:!1===e.filterValues?void 0:null!=(o=e.filterValues)?o:y,actions:b};function l(...e){return T(e,i,void 0)}return l})({})(...e)}function T(e,t,n){var r;let o=null!=(r=null==t.filterValues?void 0:t.filterValues(e,n))?r:e;if(0===o.length)return;if(1===o.length)return C(o,t,n);let i=w(o[0]);if(0!==i&&5!==i){for(let e=1;e<o.length;e++)if(w(o[e])!==i)return C(o,t,n)}switch(i){case 1:return function(e,t,n){let r=t.mergeFunctions.mergeRecords(e,t,n);return r===b.defaultMerge||t.useImplicitDefaultMerging&&void 0===r&&t.mergeFunctions.mergeRecords!==t.defaultMergeFunctions.mergeRecords?t.defaultMergeFunctions.mergeRecords(e,t,n):r}(o,t,n);case 2:return function(e,t,n){let r=t.mergeFunctions.mergeArrays(e,t,n);return r===b.defaultMerge||t.useImplicitDefaultMerging&&void 0===r&&t.mergeFunctions.mergeArrays!==t.defaultMergeFunctions.mergeArrays?t.defaultMergeFunctions.mergeArrays(e):r}(o,t,n);case 3:return function(e,t,n){let r=t.mergeFunctions.mergeSets(e,t,n);return r===b.defaultMerge||t.useImplicitDefaultMerging&&void 0===r&&t.mergeFunctions.mergeSets!==t.defaultMergeFunctions.mergeSets?t.defaultMergeFunctions.mergeSets(e):r}(o,t,n);case 4:return function(e,t,n){let r=t.mergeFunctions.mergeMaps(e,t,n);return r===b.defaultMerge||t.useImplicitDefaultMerging&&void 0===r&&t.mergeFunctions.mergeMaps!==t.defaultMergeFunctions.mergeMaps?t.defaultMergeFunctions.mergeMaps(e):r}(o,t,n);default:return C(o,t,n)}}function C(e,t,n){let r=t.mergeFunctions.mergeOthers(e,t,n);return r===b.defaultMerge||t.useImplicitDefaultMerging&&void 0===r&&t.mergeFunctions.mergeOthers!==t.defaultMergeFunctions.mergeOthers?t.defaultMergeFunctions.mergeOthers(e):r}function P(e){let t=Object.getOwnPropertyNames(e.constructor.prototype);for(let n=0;n<t.length;n++){let r=t[n],o=e[r];"constructor"!==r&&"function"==typeof o&&(e[r]=o.bind(e))}return e}class M{constructor(e){}}class R{constructor(e,t){}}function j(e){return f(e)&&""!==e?"-"!==e.charAt(e.length-1)?`${e}-`:e:""}function O(e){return null==e||!e.element||!e.on}function L(){let e=Date.now();return"xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx".replace(/[xy]/g,t=>{let n=(e+16*Math.random())%16|0;return e=Math.floor(e/16),("x"==t?n:3&n|8).toString(16)})}let A=["top","right","bottom","left"].reduce((e,t)=>e.concat(t,t+"-start",t+"-end"),[]),_=Math.min,N=Math.max,$=Math.round,I=Math.floor,D=e=>({x:e,y:e}),W={left:"right",right:"left",bottom:"top",top:"bottom"},H={start:"end",end:"start"};function F(e,t){return"function"==typeof e?e(t):e}function z(e){return e.split("-")[0]}function B(e){return e.split("-")[1]}function q(e){return"x"===e?"y":"x"}function V(e){return"y"===e?"height":"width"}function U(e){return["top","bottom"].includes(z(e))?"y":"x"}function Y(e,t,n){void 0===n&&(n=!1);let r=B(e),o=q(U(e)),i=V(o),l="x"===o?r===(n?"end":"start")?"right":"left":"start"===r?"bottom":"top";return t.reference[i]>t.floating[i]&&(l=X(l)),[l,X(l)]}function G(e){return e.replace(/start|end/g,e=>H[e])}function X(e){return e.replace(/left|right|bottom|top/g,e=>W[e])}function K(e){return"number"!=typeof e?m({top:0,right:0,bottom:0,left:0},e):{top:e,right:e,bottom:e,left:e}}function Z(e){let{x:t,y:n,width:r,height:o}=e;return{width:r,height:o,top:n,left:t,right:t+r,bottom:n+o,x:t,y:n}}let J=["crossAxis","alignment","allowedPlacements","autoAlignment"],Q=["mainAxis","crossAxis","fallbackPlacements","fallbackStrategy","fallbackAxisSideDirection","flipAlignment"],ee=["mainAxis","crossAxis","limiter"];function et(e,t,n){let r,{reference:o,floating:i}=e,l=U(t),a=q(U(t)),s=V(a),u=z(t),c="y"===l,d=o.x+o.width/2-i.width/2,f=o.y+o.height/2-i.height/2,p=o[s]/2-i[s]/2;switch(u){case"top":r={x:d,y:o.y-i.height};break;case"bottom":r={x:d,y:o.y+o.height};break;case"right":r={x:o.x+o.width,y:f};break;case"left":r={x:o.x-i.width,y:f};break;default:r={x:o.x,y:o.y}}switch(B(t)){case"start":r[a]-=p*(n&&c?-1:1);break;case"end":r[a]+=p*(n&&c?-1:1)}return r}let en=async(e,t,n)=>{let{placement:r="bottom",strategy:o="absolute",middleware:i=[],platform:l}=n,a=i.filter(Boolean),s=await (null==l.isRTL?void 0:l.isRTL(t)),u=await l.getElementRects({reference:e,floating:t,strategy:o}),{x:c,y:d}=et(u,r,s),f=r,p={},h=0;for(let n=0;n<a.length;n++){let{name:i,fn:g}=a[n],{x:b,y:v,data:y,reset:w}=await g({x:c,y:d,initialPlacement:r,placement:f,strategy:o,middlewareData:p,rects:u,platform:l,elements:{reference:e,floating:t}});c=null!=b?b:c,d=null!=v?v:d,p=m({},p,{[i]:m({},p[i],y)}),w&&h<=50&&(h++,"object"==typeof w&&(w.placement&&(f=w.placement),w.rects&&(u=!0===w.rects?await l.getElementRects({reference:e,floating:t,strategy:o}):w.rects),{x:c,y:d}=et(u,f,s)),n=-1)}return{x:c,y:d,placement:f,strategy:o,middlewareData:p}};async function er(e,t){var n;void 0===t&&(t={});let{x:r,y:o,platform:i,rects:l,elements:a,strategy:s}=e,{boundary:u="clippingAncestors",rootBoundary:c="viewport",elementContext:d="floating",altBoundary:f=!1,padding:p=0}=F(t,e),h=K(p),m=a[f?"floating"===d?"reference":"floating":d],g=Z(await i.getClippingRect({element:null==(n=await (null==i.isElement?void 0:i.isElement(m)))||n?m:m.contextElement||await (null==i.getDocumentElement?void 0:i.getDocumentElement(a.floating)),boundary:u,rootBoundary:c,strategy:s})),b="floating"===d?{x:r,y:o,width:l.floating.width,height:l.floating.height}:l.reference,v=await (null==i.getOffsetParent?void 0:i.getOffsetParent(a.floating)),y=await (null==i.isElement?void 0:i.isElement(v))&&await (null==i.getScale?void 0:i.getScale(v))||{x:1,y:1},w=Z(i.convertOffsetParentRelativeRectToViewportRelativeRect?await i.convertOffsetParentRelativeRectToViewportRelativeRect({elements:a,rect:b,offsetParent:v,strategy:s}):b);return{top:(g.top-w.top+h.top)/y.y,bottom:(w.bottom-g.bottom+h.bottom)/y.y,left:(g.left-w.left+h.left)/y.x,right:(w.right-g.right+h.right)/y.x}}function eo(){return"undefined"!=typeof window}function ei(e){return es(e)?(e.nodeName||"").toLowerCase():"#document"}function el(e){var t;return(null==e||null==(t=e.ownerDocument)?void 0:t.defaultView)||window}function ea(e){var t;return null==(t=(es(e)?e.ownerDocument:e.document)||window.document)?void 0:t.documentElement}function es(e){return!!eo()&&(e instanceof Node||e instanceof el(e).Node)}function eu(e){return!!eo()&&(e instanceof Element||e instanceof el(e).Element)}function ec(e){return!!eo()&&(e instanceof HTMLElement||e instanceof el(e).HTMLElement)}function ed(e){return!!eo()&&"undefined"!=typeof ShadowRoot&&(e instanceof ShadowRoot||e instanceof el(e).ShadowRoot)}function ef(e){let{overflow:t,overflowX:n,overflowY:r,display:o}=eb(e);return/auto|scroll|overlay|hidden|clip/.test(t+r+n)&&!["inline","contents"].includes(o)}function ep(e){return[":popover-open",":modal"].some(t=>{try{return e.matches(t)}catch(e){return!1}})}function eh(e){let t=em(),n=eu(e)?eb(e):e;return"none"!==n.transform||"none"!==n.perspective||!!n.containerType&&"normal"!==n.containerType||!t&&!!n.backdropFilter&&"none"!==n.backdropFilter||!t&&!!n.filter&&"none"!==n.filter||["transform","perspective","filter"].some(e=>(n.willChange||"").includes(e))||["paint","layout","strict","content"].some(e=>(n.contain||"").includes(e))}function em(){return"undefined"!=typeof CSS&&!!CSS.supports&&CSS.supports("-webkit-backdrop-filter","none")}function eg(e){return["html","body","#document"].includes(ei(e))}function eb(e){return el(e).getComputedStyle(e)}function ev(e){return eu(e)?{scrollLeft:e.scrollLeft,scrollTop:e.scrollTop}:{scrollLeft:e.scrollX,scrollTop:e.scrollY}}function ey(e){if("html"===ei(e))return e;let t=e.assignedSlot||e.parentNode||ed(e)&&e.host||ea(e);return ed(t)?t.host:t}function ew(e,t,n){var r;void 0===t&&(t=[]),void 0===n&&(n=!0);let o=function e(t){let n=ey(t);return eg(n)?t.ownerDocument?t.ownerDocument.body:t.body:ec(n)&&ef(n)?n:e(n)}(e),i=o===(null==(r=e.ownerDocument)?void 0:r.body),l=el(o);if(i){let e=ex(l);return t.concat(l,l.visualViewport||[],ef(o)?o:[],e&&n?ew(e):[])}return t.concat(o,ew(o,[],n))}function ex(e){return e.parent&&Object.getPrototypeOf(e.parent)?e.frameElement:null}function eE(e){let t=eb(e),n=parseFloat(t.width)||0,r=parseFloat(t.height)||0,o=ec(e),i=o?e.offsetWidth:n,l=o?e.offsetHeight:r,a=$(n)!==i||$(r)!==l;return a&&(n=i,r=l),{width:n,height:r,$:a}}function eS(e){return eu(e)?e:e.contextElement}function ek(e){let t=eS(e);if(!ec(t))return D(1);let n=t.getBoundingClientRect(),{width:r,height:o,$:i}=eE(t),l=(i?$(n.width):n.width)/r,a=(i?$(n.height):n.height)/o;return l&&Number.isFinite(l)||(l=1),a&&Number.isFinite(a)||(a=1),{x:l,y:a}}let eT=D(0);function eC(e){let t=el(e);return em()&&t.visualViewport?{x:t.visualViewport.offsetLeft,y:t.visualViewport.offsetTop}:eT}function eP(e,t,n,r){var o;void 0===t&&(t=!1),void 0===n&&(n=!1);let i=e.getBoundingClientRect(),l=eS(e),a=D(1);t&&(r?eu(r)&&(a=ek(r)):a=ek(e));let s=(void 0===(o=n)&&(o=!1),r&&(!o||r===el(l))&&o)?eC(l):D(0),u=(i.left+s.x)/a.x,c=(i.top+s.y)/a.y,d=i.width/a.x,f=i.height/a.y;if(l){let e=el(l),t=r&&eu(r)?el(r):r,n=e,o=ex(n);for(;o&&r&&t!==n;){let e=ek(o),t=o.getBoundingClientRect(),r=eb(o),i=t.left+(o.clientLeft+parseFloat(r.paddingLeft))*e.x,l=t.top+(o.clientTop+parseFloat(r.paddingTop))*e.y;u*=e.x,c*=e.y,d*=e.x,f*=e.y,u+=i,c+=l,o=ex(n=el(o))}}return Z({width:d,height:f,x:u,y:c})}function eM(e,t){let n=ev(e).scrollLeft;return t?t.left+n:eP(ea(e)).left+n}function eR(e,t,n){void 0===n&&(n=!1);let r=e.getBoundingClientRect();return{x:r.left+t.scrollLeft-(n?0:eM(e,r)),y:r.top+t.scrollTop}}function ej(e,t,n){let r;if("viewport"===t)r=function(e,t){let n=el(e),r=ea(e),o=n.visualViewport,i=r.clientWidth,l=r.clientHeight,a=0,s=0;if(o){i=o.width,l=o.height;let e=em();(!e||e&&"fixed"===t)&&(a=o.offsetLeft,s=o.offsetTop)}return{width:i,height:l,x:a,y:s}}(e,n);else if("document"===t)r=function(e){let t=ea(e),n=ev(e),r=e.ownerDocument.body,o=N(t.scrollWidth,t.clientWidth,r.scrollWidth,r.clientWidth),i=N(t.scrollHeight,t.clientHeight,r.scrollHeight,r.clientHeight),l=-n.scrollLeft+eM(e),a=-n.scrollTop;return"rtl"===eb(r).direction&&(l+=N(t.clientWidth,r.clientWidth)-o),{width:o,height:i,x:l,y:a}}(ea(e));else if(eu(t))r=function(e,t){let n=eP(e,!0,"fixed"===t),r=n.top+e.clientTop,o=n.left+e.clientLeft,i=ec(e)?ek(e):D(1),l=e.clientWidth*i.x;return{width:l,height:e.clientHeight*i.y,x:o*i.x,y:r*i.y}}(t,n);else{let n=eC(e);r={x:t.x-n.x,y:t.y-n.y,width:t.width,height:t.height}}return Z(r)}function eO(e){return"static"===eb(e).position}function eL(e,t){if(!ec(e)||"fixed"===eb(e).position)return null;if(t)return t(e);let n=e.offsetParent;return ea(e)===n&&(n=n.ownerDocument.body),n}function eA(e,t){let n=el(e);if(ep(e))return n;if(!ec(e)){let t=ey(e);for(;t&&!eg(t);){if(eu(t)&&!eO(t))return t;t=ey(t)}return n}let r=eL(e,t);for(;r&&["table","td","th"].includes(ei(r))&&eO(r);)r=eL(r,t);return r&&eg(r)&&eO(r)&&!eh(r)?n:r||function(e){let t=ey(e);for(;ec(t)&&!eg(t);){if(eh(t))return t;if(ep(t))break;t=ey(t)}return null}(e)||n}let e_=async function(e){let t=this.getOffsetParent||eA,n=this.getDimensions,r=await n(e.floating);return{reference:function(e,t,n){let r=ec(t),o=ea(t),i="fixed"===n,l=eP(e,!0,i,t),a={scrollLeft:0,scrollTop:0},s=D(0);if(r||!r&&!i){if(("body"!==ei(t)||ef(o))&&(a=ev(t)),r){let e=eP(t,!0,i,t);s.x=e.x+t.clientLeft,s.y=e.y+t.clientTop}else o&&(s.x=eM(o))}let u=!o||r||i?D(0):eR(o,a);return{x:l.left+a.scrollLeft-s.x-u.x,y:l.top+a.scrollTop-s.y-u.y,width:l.width,height:l.height}}(e.reference,await t(e.floating),e.strategy),floating:{x:0,y:0,width:r.width,height:r.height}}},eN={convertOffsetParentRelativeRectToViewportRelativeRect:function(e){let{elements:t,rect:n,offsetParent:r,strategy:o}=e,i="fixed"===o,l=ea(r),a=!!t&&ep(t.floating);if(r===l||a&&i)return n;let s={scrollLeft:0,scrollTop:0},u=D(1),c=D(0),d=ec(r);if((d||!d&&!i)&&(("body"!==ei(r)||ef(l))&&(s=ev(r)),ec(r))){let e=eP(r);u=ek(r),c.x=e.x+r.clientLeft,c.y=e.y+r.clientTop}let f=!l||d||i?D(0):eR(l,s,!0);return{width:n.width*u.x,height:n.height*u.y,x:n.x*u.x-s.scrollLeft*u.x+c.x+f.x,y:n.y*u.y-s.scrollTop*u.y+c.y+f.y}},getDocumentElement:ea,getClippingRect:function(e){let{element:t,boundary:n,rootBoundary:r,strategy:o}=e,i=[..."clippingAncestors"===n?ep(t)?[]:function(e,t){let n=t.get(e);if(n)return n;let r=ew(e,[],!1).filter(e=>eu(e)&&"body"!==ei(e)),o=null,i="fixed"===eb(e).position,l=i?ey(e):e;for(;eu(l)&&!eg(l);){let t=eb(l),n=eh(l);n||"fixed"!==t.position||(o=null),(i?!n&&!o:!n&&"static"===t.position&&!!o&&["absolute","fixed"].includes(o.position)||ef(l)&&!n&&function e(t,n){let r=ey(t);return!(r===n||!eu(r)||eg(r))&&("fixed"===eb(r).position||e(r,n))}(e,l))?r=r.filter(e=>e!==l):o=t,l=ey(l)}return t.set(e,r),r}(t,this._c):[].concat(n),r],l=i[0],a=i.reduce((e,n)=>{let r=ej(t,n,o);return e.top=N(r.top,e.top),e.right=_(r.right,e.right),e.bottom=_(r.bottom,e.bottom),e.left=N(r.left,e.left),e},ej(t,l,o));return{width:a.right-a.left,height:a.bottom-a.top,x:a.left,y:a.top}},getOffsetParent:eA,getElementRects:e_,getClientRects:function(e){return Array.from(e.getClientRects())},getDimensions:function(e){let{width:t,height:n}=eE(e);return{width:t,height:n}},getScale:ek,isElement:eu,isRTL:function(e){return"rtl"===eb(e).direction}},e$=e=>({name:"arrow",options:e,async fn(t){let{x:n,y:r,placement:o,rects:i,platform:l,elements:a,middlewareData:s}=t,{element:u,padding:c=0}=F(e,t)||{};if(null==u)return{};let d=K(c),f={x:n,y:r},p=q(U(o)),h=V(p),g=await l.getDimensions(u),b="y"===p,v=b?"clientHeight":"clientWidth",y=i.reference[h]+i.reference[p]-f[p]-i.floating[h],w=f[p]-i.reference[p],x=await (null==l.getOffsetParent?void 0:l.getOffsetParent(u)),E=x?x[v]:0;E&&await (null==l.isElement?void 0:l.isElement(x))||(E=a.floating[v]||i.floating[h]);let S=E/2-g[h]/2-1,k=_(d[b?"top":"left"],S),T=_(d[b?"bottom":"right"],S),C=E-g[h]-T,P=E/2-g[h]/2+(y/2-w/2),M=N(k,_(P,C)),R=!s.arrow&&null!=B(o)&&P!==M&&i.reference[h]/2-(P<k?k:T)-g[h]/2<0,j=R?P<k?P-k:P-C:0;return{[p]:f[p]+j,data:m({[p]:M,centerOffset:P-M-j},R&&{alignmentOffset:j}),reset:R}}}),eI=(e,t,n)=>{let r=new Map,o=m({platform:eN},n),i=m({},o.platform,{_c:r});return en(e,t,m({},o,{platform:i}))};function eD(){}function eW(e){return e()}function eH(){return Object.create(null)}function eF(e){e.forEach(eW)}function ez(e){return"function"==typeof e}function eB(e,t){return e!=e?t==t:e!==t||e&&"object"==typeof e||"function"==typeof e}function eq(e,t){e.appendChild(t)}function eV(e,t,n){e.insertBefore(t,n||null)}function eU(e){e.parentNode&&e.parentNode.removeChild(e)}function eY(e){return document.createElement(e)}function eG(e){return document.createElementNS("http://www.w3.org/2000/svg",e)}function eX(e){return document.createTextNode(e)}function eK(){return eX(" ")}function eZ(e,t,n,r){return e.addEventListener(t,n,r),()=>e.removeEventListener(t,n,r)}function eJ(e,t,n){null==n?e.removeAttribute(t):e.getAttribute(t)!==n&&e.setAttribute(t,n)}let eQ=["width","height"];function e0(e,t){let n=Object.getOwnPropertyDescriptors(e.__proto__);for(let r in t)null==t[r]?e.removeAttribute(r):"style"===r?e.style.cssText=t[r]:"__value"===r?e.value=e[r]=t[r]:n[r]&&n[r].set&&-1===eQ.indexOf(r)?e[r]=t[r]:eJ(e,r,t[r])}function e1(e,t,n){e.classList.toggle(t,!!n)}function e2(){if(!r)throw Error("Function called outside component initialization");return r}function e4(e){e2().$$.after_update.push(e)}let e5=[],e6=[],e3=[],e9=[],e8=Promise.resolve(),e7=!1;function te(e){e3.push(e)}let tt=new Set,tn=0;function tr(){if(0!==tn)return;let e=r;do{try{for(;tn<e5.length;){let e=e5[tn];tn++,r=e,function(e){if(null!==e.fragment){e.update(),eF(e.before_update);let t=e.dirty;e.dirty=[-1],e.fragment&&e.fragment.p(e.ctx,t),e.after_update.forEach(te)}}(e.$$)}}catch(e){throw e5.length=0,tn=0,e}for(r=null,e5.length=0,tn=0;e6.length;)e6.pop()();for(let e=0;e<e3.length;e+=1){let t=e3[e];tt.has(t)||(tt.add(t),t())}e3.length=0}while(e5.length);for(;e9.length;)e9.pop()();e7=!1,tt.clear(),r=e}let to=new Set;function ti(){o={r:0,c:[],p:o}}function tl(){o.r||eF(o.c),o=o.p}function ta(e,t){e&&e.i&&(to.delete(e),e.i(t))}function ts(e,t,n,r){e&&e.o?to.has(e)||(to.add(e),o.c.push(()=>{to.delete(e),r&&(n&&e.d(1),r())}),e.o(t)):r&&r()}function tu(e){return(null==e?void 0:e.length)!==void 0?e:Array.from(e)}function tc(e){e&&e.c()}function td(e,t,n){let{fragment:r,after_update:o}=e.$$;r&&r.m(t,n),te(()=>{let t=e.$$.on_mount.map(eW).filter(ez);e.$$.on_destroy?e.$$.on_destroy.push(...t):eF(t),e.$$.on_mount=[]}),o.forEach(te)}function tf(e,t){let n=e.$$;null!==n.fragment&&(!function(e){let t=[],n=[];e3.forEach(r=>-1===e.indexOf(r)?t.push(r):n.push(r)),n.forEach(e=>e()),e3=t}(n.after_update),eF(n.on_destroy),n.fragment&&n.fragment.d(t),n.on_destroy=n.fragment=null,n.ctx=[])}function tp(e,t,n,o,i,l,a=null,s=[-1]){let u=r;r=e;let c=e.$$={fragment:null,ctx:[],props:l,update:eD,not_equal:i,bound:eH(),on_mount:[],on_destroy:[],on_disconnect:[],before_update:[],after_update:[],context:new Map(t.context||(u?u.$$.context:[])),callbacks:eH(),dirty:s,skip_bound:!1,root:t.target||u.$$.root};a&&a(c.root);let d=!1;if(c.ctx=n?n(e,t.props||{},(t,n,...r)=>{let o=r.length?r[0]:n;return c.ctx&&i(c.ctx[t],c.ctx[t]=o)&&(!c.skip_bound&&c.bound[t]&&c.bound[t](o),d)&&(-1===e.$$.dirty[0]&&(e5.push(e),e7||(e7=!0,e8.then(tr)),e.$$.dirty.fill(0)),e.$$.dirty[t/31|0]|=1<<t%31),n}):[],c.update(),d=!0,eF(c.before_update),c.fragment=!!o&&o(c.ctx),t.target){if(t.hydrate){let e=Array.from(t.target.childNodes);c.fragment&&c.fragment.l(e),e.forEach(eU)}else c.fragment&&c.fragment.c();t.intro&&ta(e.$$.fragment),td(e,t.target,t.anchor),tr()}r=u}class th{constructor(){this.$$=void 0,this.$$set=void 0}$destroy(){tf(this,1),this.$destroy=eD}$on(e,t){if(!ez(t))return eD;let n=this.$$.callbacks[e]||(this.$$.callbacks[e]=[]);return n.push(t),()=>{let e=n.indexOf(t);-1!==e&&n.splice(e,1)}}$set(e){this.$$set&&0!==Object.keys(e).length&&(this.$$.skip_bound=!0,this.$$set(e),this.$$.skip_bound=!1)}}function tm(e){let t,n,r,o,i;return{c(){eJ(t=eY("button"),"aria-label",n=e[3]?e[3]:null),eJ(t,"class",r=`${e[1]||""} shepherd-button ${e[4]?"shepherd-button-secondary":""}`),t.disabled=e[2],eJ(t,"tabindex","0"),eJ(t,"type","button")},m(n,r){eV(n,t,r),t.innerHTML=e[5],o||(i=eZ(t,"click",function(){ez(e[0])&&e[0].apply(this,arguments)}),o=!0)},p(o,[i]){e=o,32&i&&(t.innerHTML=e[5]),8&i&&n!==(n=e[3]?e[3]:null)&&eJ(t,"aria-label",n),18&i&&r!==(r=`${e[1]||""} shepherd-button ${e[4]?"shepherd-button-secondary":""}`)&&eJ(t,"class",r),4&i&&(t.disabled=e[2])},i:eD,o:eD,d(e){e&&eU(t),o=!1,i()}}}function tg(e,t,n){let r,o,i,l,a,s,{config:u,step:c}=t;function f(e){return d(e)?e=e.call(c):e}return e.$$set=e=>{"config"in e&&n(6,u=e.config),"step"in e&&n(7,c=e.step)},e.$$.update=()=>{192&e.$$.dirty&&(n(0,r=u.action?u.action.bind(c.tour):null),n(1,o=u.classes),n(2,i=!!u.disabled&&f(u.disabled)),n(3,l=u.label?f(u.label):null),n(4,a=u.secondary),n(5,s=u.text?f(u.text):null))},[r,o,i,l,a,s,u,c]}"undefined"!=typeof window&&(window.__svelte||(window.__svelte={v:new Set})).v.add("4");class tb extends th{constructor(e){super(),tp(this,e,tg,tm,eB,{config:6,step:7})}}function tv(e,t,n){let r=e.slice();return r[2]=t[n],r}function ty(e){let t,n;let r=tu(e[1]),o=[];for(let t=0;t<r.length;t+=1)o[t]=tw(tv(e,r,t));let i=e=>ts(o[e],1,1,()=>{o[e]=null});return{c(){for(let e=0;e<o.length;e+=1)o[e].c();t=eX("")},m(e,r){for(let t=0;t<o.length;t+=1)o[t]&&o[t].m(e,r);eV(e,t,r),n=!0},p(e,n){if(3&n){let l;for(l=0,r=tu(e[1]);l<r.length;l+=1){let i=tv(e,r,l);o[l]?(o[l].p(i,n),ta(o[l],1)):(o[l]=tw(i),o[l].c(),ta(o[l],1),o[l].m(t.parentNode,t))}for(ti(),l=r.length;l<o.length;l+=1)i(l);tl()}},i(e){if(!n){for(let e=0;e<r.length;e+=1)ta(o[e]);n=!0}},o(e){o=o.filter(Boolean);for(let e=0;e<o.length;e+=1)ts(o[e]);n=!1},d(e){e&&eU(t),function(e,t){for(let n=0;n<e.length;n+=1)e[n]&&e[n].d(t)}(o,e)}}}function tw(e){let t,n;return t=new tb({props:{config:e[2],step:e[0]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};2&n&&(r.config=e[2]),1&n&&(r.step=e[0]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tx(e){let t,n;let r=e[1]&&ty(e);return{c(){t=eY("footer"),r&&r.c(),eJ(t,"class","shepherd-footer")},m(e,o){eV(e,t,o),r&&r.m(t,null),n=!0},p(e,[n]){e[1]?r?(r.p(e,n),2&n&&ta(r,1)):((r=ty(e)).c(),ta(r,1),r.m(t,null)):r&&(ti(),ts(r,1,1,()=>{r=null}),tl())},i(e){n||(ta(r),n=!0)},o(e){ts(r),n=!1},d(e){e&&eU(t),r&&r.d()}}}function tE(e,t,n){let r;let{step:o}=t;return e.$$set=e=>{"step"in e&&n(0,o=e.step)},e.$$.update=()=>{1&e.$$.dirty&&n(1,r=o.options.buttons)},[o,r]}class tS extends th{constructor(e){super(),tp(this,e,tE,tx,eB,{step:0})}}function tk(e){let t,n,r,o,i;return{c(){t=eY("button"),(n=eY("span")).textContent="\xd7",eJ(n,"aria-hidden","true"),eJ(t,"aria-label",r=e[0].label?e[0].label:"Close Tour"),eJ(t,"class","shepherd-cancel-icon"),eJ(t,"type","button")},m(r,l){eV(r,t,l),eq(t,n),o||(i=eZ(t,"click",e[1]),o=!0)},p(e,[n]){1&n&&r!==(r=e[0].label?e[0].label:"Close Tour")&&eJ(t,"aria-label",r)},i:eD,o:eD,d(e){e&&eU(t),o=!1,i()}}}function tT(e,t,n){let{cancelIcon:r,step:o}=t;return e.$$set=e=>{"cancelIcon"in e&&n(0,r=e.cancelIcon),"step"in e&&n(2,o=e.step)},[r,e=>{e.preventDefault(),o.cancel()},o]}class tC extends th{constructor(e){super(),tp(this,e,tT,tk,eB,{cancelIcon:0,step:2})}}function tP(e){let t;return{c(){eJ(t=eY("h3"),"id",e[1]),eJ(t,"class","shepherd-title")},m(n,r){eV(n,t,r),e[3](t)},p(e,[n]){2&n&&eJ(t,"id",e[1])},i:eD,o:eD,d(n){n&&eU(t),e[3](null)}}}function tM(e,t,n){let{labelId:r,element:o,title:i}=t;return e4(()=>{d(i)&&n(2,i=i()),n(0,o.innerHTML=i,o)}),e.$$set=e=>{"labelId"in e&&n(1,r=e.labelId),"element"in e&&n(0,o=e.element),"title"in e&&n(2,i=e.title)},[o,r,i,function(e){e6[e?"unshift":"push"](()=>{n(0,o=e)})}]}class tR extends th{constructor(e){super(),tp(this,e,tM,tP,eB,{labelId:1,element:0,title:2})}}function tj(e){let t,n;return t=new tR({props:{labelId:e[0],title:e[2]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};1&n&&(r.labelId=e[0]),4&n&&(r.title=e[2]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tO(e){let t,n;return t=new tC({props:{cancelIcon:e[3],step:e[1]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};8&n&&(r.cancelIcon=e[3]),2&n&&(r.step=e[1]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tL(e){let t,n,r;let o=e[2]&&tj(e),i=e[3]&&e[3].enabled&&tO(e);return{c(){t=eY("header"),o&&o.c(),n=eK(),i&&i.c(),eJ(t,"class","shepherd-header")},m(e,l){eV(e,t,l),o&&o.m(t,null),eq(t,n),i&&i.m(t,null),r=!0},p(e,[r]){e[2]?o?(o.p(e,r),4&r&&ta(o,1)):((o=tj(e)).c(),ta(o,1),o.m(t,n)):o&&(ti(),ts(o,1,1,()=>{o=null}),tl()),e[3]&&e[3].enabled?i?(i.p(e,r),8&r&&ta(i,1)):((i=tO(e)).c(),ta(i,1),i.m(t,null)):i&&(ti(),ts(i,1,1,()=>{i=null}),tl())},i(e){r||(ta(o),ta(i),r=!0)},o(e){ts(o),ts(i),r=!1},d(e){e&&eU(t),o&&o.d(),i&&i.d()}}}function tA(e,t,n){let r,o,{labelId:i,step:l}=t;return e.$$set=e=>{"labelId"in e&&n(0,i=e.labelId),"step"in e&&n(1,l=e.step)},e.$$.update=()=>{2&e.$$.dirty&&(n(2,r=l.options.title),n(3,o=l.options.cancelIcon))},[i,l,r,o]}class t_ extends th{constructor(e){super(),tp(this,e,tA,tL,eB,{labelId:0,step:1})}}function tN(e){let t;return{c(){eJ(t=eY("div"),"class","shepherd-text"),eJ(t,"id",e[1])},m(n,r){eV(n,t,r),e[3](t)},p(e,[n]){2&n&&eJ(t,"id",e[1])},i:eD,o:eD,d(n){n&&eU(t),e[3](null)}}}function t$(e,t,n){let{descriptionId:r,element:o,step:i}=t;return e4(()=>{let{text:e}=i.options;d(e)&&(e=e.call(i)),c(e)?o.appendChild(e):n(0,o.innerHTML=e,o)}),e.$$set=e=>{"descriptionId"in e&&n(1,r=e.descriptionId),"element"in e&&n(0,o=e.element),"step"in e&&n(2,i=e.step)},[o,r,i,function(e){e6[e?"unshift":"push"](()=>{n(0,o=e)})}]}class tI extends th{constructor(e){super(),tp(this,e,t$,tN,eB,{descriptionId:1,element:0,step:2})}}function tD(e){let t,n;return t=new t_({props:{labelId:e[1],step:e[2]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};2&n&&(r.labelId=e[1]),4&n&&(r.step=e[2]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tW(e){let t,n;return t=new tI({props:{descriptionId:e[0],step:e[2]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};1&n&&(r.descriptionId=e[0]),4&n&&(r.step=e[2]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tH(e){let t,n;return t=new tS({props:{step:e[2]}}),{c(){tc(t.$$.fragment)},m(e,r){td(t,e,r),n=!0},p(e,n){let r={};4&n&&(r.step=e[2]),t.$set(r)},i(e){n||(ta(t.$$.fragment,e),n=!0)},o(e){ts(t.$$.fragment,e),n=!1},d(e){tf(t,e)}}}function tF(e){let t,n,r,o;let i=!p(e[2].options.title)||e[2].options.cancelIcon&&e[2].options.cancelIcon.enabled,l=!p(e[2].options.text),a=Array.isArray(e[2].options.buttons)&&e[2].options.buttons.length,s=i&&tD(e),u=l&&tW(e),c=a&&tH(e);return{c(){t=eY("div"),s&&s.c(),n=eK(),u&&u.c(),r=eK(),c&&c.c(),eJ(t,"class","shepherd-content")},m(e,i){eV(e,t,i),s&&s.m(t,null),eq(t,n),u&&u.m(t,null),eq(t,r),c&&c.m(t,null),o=!0},p(e,[o]){4&o&&(i=!p(e[2].options.title)||e[2].options.cancelIcon&&e[2].options.cancelIcon.enabled),i?s?(s.p(e,o),4&o&&ta(s,1)):((s=tD(e)).c(),ta(s,1),s.m(t,n)):s&&(ti(),ts(s,1,1,()=>{s=null}),tl()),4&o&&(l=!p(e[2].options.text)),l?u?(u.p(e,o),4&o&&ta(u,1)):((u=tW(e)).c(),ta(u,1),u.m(t,r)):u&&(ti(),ts(u,1,1,()=>{u=null}),tl()),4&o&&(a=Array.isArray(e[2].options.buttons)&&e[2].options.buttons.length),a?c?(c.p(e,o),4&o&&ta(c,1)):((c=tH(e)).c(),ta(c,1),c.m(t,null)):c&&(ti(),ts(c,1,1,()=>{c=null}),tl())},i(e){o||(ta(s),ta(u),ta(c),o=!0)},o(e){ts(s),ts(u),ts(c),o=!1},d(e){e&&eU(t),s&&s.d(),u&&u.d(),c&&c.d()}}}function tz(e,t,n){let{descriptionId:r,labelId:o,step:i}=t;return e.$$set=e=>{"descriptionId"in e&&n(0,r=e.descriptionId),"labelId"in e&&n(1,o=e.labelId),"step"in e&&n(2,i=e.step)},[r,o,i]}class tB extends th{constructor(e){super(),tp(this,e,tz,tF,eB,{descriptionId:0,labelId:1,step:2})}}function tq(e){let t;return{c(){eJ(t=eY("div"),"class","shepherd-arrow"),eJ(t,"data-popper-arrow","")},m(e,n){eV(e,t,n)},d(e){e&&eU(t)}}}function tV(e){let t,n,r,o,i,l,a,s;let u=e[4].options.arrow&&e[4].options.attachTo&&e[4].options.attachTo.element&&e[4].options.attachTo.on&&tq();r=new tB({props:{descriptionId:e[2],labelId:e[3],step:e[4]}});let c=[{"aria-describedby":o=p(e[4].options.text)?null:e[2]},{"aria-labelledby":i=e[4].options.title?e[3]:null},e[1],{open:"true"}],d={};for(let e=0;e<c.length;e+=1)d=function(e,t){for(let n in t)e[n]=t[n];return e}(d,c[e]);return{c(){t=eY("dialog"),u&&u.c(),n=eK(),tc(r.$$.fragment),e0(t,d),e1(t,"shepherd-has-cancel-icon",e[5]),e1(t,"shepherd-has-title",e[6]),e1(t,"shepherd-element",!0)},m(o,i){eV(o,t,i),u&&u.m(t,null),eq(t,n),td(r,t,null),e[13](t),l=!0,a||(s=eZ(t,"keydown",e[7]),a=!0)},p(e,[a]){e[4].options.arrow&&e[4].options.attachTo&&e[4].options.attachTo.element&&e[4].options.attachTo.on?u||((u=tq()).c(),u.m(t,n)):u&&(u.d(1),u=null);let s={};4&a&&(s.descriptionId=e[2]),8&a&&(s.labelId=e[3]),16&a&&(s.step=e[4]),r.$set(s),e0(t,d=function(e,t){let n={},r={},o={$$scope:1},i=e.length;for(;i--;){let l=e[i],a=t[i];if(a){for(let e in l)e in a||(r[e]=1);for(let e in a)o[e]||(n[e]=a[e],o[e]=1);e[i]=a}else for(let e in l)o[e]=1}for(let e in r)e in n||(n[e]=void 0);return n}(c,[(!l||20&a&&o!==(o=p(e[4].options.text)?null:e[2]))&&{"aria-describedby":o},(!l||24&a&&i!==(i=e[4].options.title?e[3]:null))&&{"aria-labelledby":i},2&a&&e[1],{open:"true"}])),e1(t,"shepherd-has-cancel-icon",e[5]),e1(t,"shepherd-has-title",e[6]),e1(t,"shepherd-element",!0)},i(e){l||(ta(r.$$.fragment,e),l=!0)},o(e){ts(r.$$.fragment,e),l=!1},d(n){n&&eU(t),u&&u.d(),tf(r),e[13](null),a=!1,s()}}}function tU(e){return e.split(" ").filter(e=>!!e.length)}function tY(e,t,n){var r;let o,i,l,{classPrefix:a,element:s,descriptionId:u,firstFocusableElement:c,focusableElements:d,labelId:p,lastFocusableElement:h,step:m,dataStepId:g}=t;return r=()=>{n(1,g={[`data-${a}shepherd-step-id`]:m.id}),n(9,d=s.querySelectorAll('a[href], area[href], input:not([disabled]), select:not([disabled]), textarea:not([disabled]), button:not([disabled]), [tabindex="0"]')),n(8,c=d[0]),n(10,h=d[d.length-1])},e2().$$.on_mount.push(r),e4(()=>{l!==m.options.classes&&(function(e){if(f(e)){let t=tU(e);t.length&&s.classList.remove(...t)}}(l),function(e){if(f(e)){let t=tU(e);t.length&&s.classList.add(...t)}}(l=m.options.classes))}),e.$$set=e=>{"classPrefix"in e&&n(11,a=e.classPrefix),"element"in e&&n(0,s=e.element),"descriptionId"in e&&n(2,u=e.descriptionId),"firstFocusableElement"in e&&n(8,c=e.firstFocusableElement),"focusableElements"in e&&n(9,d=e.focusableElements),"labelId"in e&&n(3,p=e.labelId),"lastFocusableElement"in e&&n(10,h=e.lastFocusableElement),"step"in e&&n(4,m=e.step),"dataStepId"in e&&n(1,g=e.dataStepId)},e.$$.update=()=>{16&e.$$.dirty&&(n(5,o=m.options&&m.options.cancelIcon&&m.options.cancelIcon.enabled),n(6,i=m.options&&m.options.title))},[s,g,u,p,m,o,i,e=>{let{tour:t}=m;switch(e.keyCode){case 9:if(0===d.length){e.preventDefault();break}e.shiftKey?(document.activeElement===c||document.activeElement.classList.contains("shepherd-element"))&&(e.preventDefault(),h.focus()):document.activeElement===h&&(e.preventDefault(),c.focus());break;case 27:t.options.exitOnEsc&&(e.preventDefault(),e.stopPropagation(),m.cancel());break;case 37:t.options.keyboardNavigation&&(e.preventDefault(),e.stopPropagation(),t.back());break;case 39:t.options.keyboardNavigation&&(e.preventDefault(),e.stopPropagation(),t.next())}},c,d,h,a,()=>s,function(e){e6[e?"unshift":"push"](()=>{n(0,s=e)})}]}class tG extends th{constructor(e){super(),tp(this,e,tY,tV,eB,{classPrefix:11,element:0,descriptionId:2,firstFocusableElement:8,focusableElements:9,labelId:3,lastFocusableElement:10,step:4,dataStepId:1,getElement:12})}get getElement(){return this.$$.ctx[12]}}class tX extends h{constructor(e,t={}){return super(),this._resolvedAttachTo=void 0,this._resolvedExtraHighlightElements=void 0,this.classPrefix=void 0,this.el=void 0,this.target=void 0,this.tour=void 0,this.tour=e,this.classPrefix=this.tour.options?j(this.tour.options.classPrefix):"",this.styles=e.styles,this._resolvedAttachTo=null,P(this),this._setOptions(t),this}cancel(){this.tour.cancel(),this.trigger("cancel")}complete(){this.tour.complete(),this.trigger("complete")}destroy(){this.cleanup&&this.cleanup(),this.cleanup=null,c(this.el)&&(this.el.remove(),this.el=null),this._updateStepTargetOnHide(),this.trigger("destroy")}getTour(){return this.tour}hide(){var e;null==(e=this.tour.modal)||e.hide(),this.trigger("before-hide"),this.el&&(this.el.hidden=!0),this._updateStepTargetOnHide(),this.trigger("hide")}_resolveExtraHiglightElements(){return this._resolvedExtraHighlightElements=this.options.extraHighlights?this.options.extraHighlights.flatMap(e=>Array.from(document.querySelectorAll(e))):[],this._resolvedExtraHighlightElements}_resolveAttachToOptions(){return this._resolvedAttachTo=function(e){let t=e.options.attachTo||{},n=Object.assign({},t);if(d(n.element)&&(n.element=n.element.call(e)),f(n.element)){try{n.element=document.querySelector(n.element)}catch(e){}n.element||console.error(`The element for this Shepherd step was not found ${t.element}`)}return n}(this),this._resolvedAttachTo}_getResolvedAttachToOptions(){return null===this._resolvedAttachTo?this._resolveAttachToOptions():this._resolvedAttachTo}isOpen(){return!!(this.el&&!this.el.hidden)}show(){return d(this.options.beforeShowPromise)?Promise.resolve(this.options.beforeShowPromise()).then(()=>this._show()):Promise.resolve(this._show())}updateStepOptions(e){Object.assign(this.options,e),this.shepherdElementComponent&&this.shepherdElementComponent.$set({step:this})}getElement(){return this.el}getTarget(){return this.target}_createTooltipContent(){let e=`${this.id}-description`,t=`${this.id}-label`;return this.shepherdElementComponent=new tG({target:this.tour.options.stepsContainer||document.body,props:{classPrefix:this.classPrefix,descriptionId:e,labelId:t,step:this,styles:this.styles}}),this.shepherdElementComponent.getElement()}_scrollTo(e){let{element:t}=this._getResolvedAttachToOptions();d(this.options.scrollToHandler)?this.options.scrollToHandler(t):t instanceof Element&&"function"==typeof t.scrollIntoView&&t.scrollIntoView(e)}_getClassOptions(e){let t=this.tour&&this.tour.options&&this.tour.options.defaultStepOptions,n=e.classes?e.classes:"",r=t&&t.classes?t.classes:"";return Array.from(new Set([...n.split(" "),...r.split(" ")])).join(" ").trim()}_setOptions(e={}){let t=this.tour&&this.tour.options&&this.tour.options.defaultStepOptions;t=k({},t||{}),this.options=Object.assign({arrow:!0},t,e,{floatingUIOptions:k(t.floatingUIOptions||{},e.floatingUIOptions||{})});let{when:n}=this.options;this.options.classes=this._getClassOptions(e),this.destroy(),this.id=this.options.id||`step-${L()}`,n&&Object.keys(n).forEach(e=>{this.on(e,n[e],this)})}_setupElements(){p(this.el)||this.destroy(),this.el=this._createTooltipContent(),this.options.advanceOn&&function(e){let{event:t,selector:n}=e.options.advanceOn||{};if(!t)return console.error("advanceOn was defined, but no event name was passed.");{let r=t=>{if(e.isOpen()){let r=e.el&&t.currentTarget===e.el;(!p(n)&&t.currentTarget.matches(n)||r)&&e.tour.next()}},o=null;if(!p(n)&&!(o=document.querySelector(n)))return console.error(`No element was found for the selector supplied to advanceOn: ${n}`);o?(o.addEventListener(t,r),e.on("destroy",()=>o.removeEventListener(t,r))):(document.body.addEventListener(t,r,!0),e.on("destroy",()=>document.body.removeEventListener(t,r,!0)))}}(this),function(e){e.cleanup&&e.cleanup();let t=e._getResolvedAttachToOptions(),n=t.element,r=function(e,t){var n,r,o,i,l,a,s,u;let c={strategy:"absolute"};c.middleware=[];let d=!!t.options.arrow&&!!t.el&&t.el.querySelector(".shepherd-arrow"),f=O(e),p=null==(n=e.on)?void 0:n.includes("auto"),h=(null==e||null==(r=e.on)?void 0:r.includes("-start"))||(null==e||null==(o=e.on)?void 0:o.includes("-end"));if(!f){if(p?c.middleware.push({name:"autoPlacement",options:l={crossAxis:!0,alignment:h?null==e||null==(i=e.on)?void 0:i.split("-").pop():null},async fn(e){var t,n,r,o;let{rects:i,middlewareData:a,placement:s,platform:u,elements:c}=e,d=F(l,e),{crossAxis:f=!1,alignment:p,allowedPlacements:h=A,autoAlignment:m=!0}=d,b=g(d,J),v=void 0!==p||h===A?((o=p||null)?[...h.filter(e=>B(e)===o),...h.filter(e=>B(e)!==o)]:h.filter(e=>z(e)===e)).filter(e=>!o||B(e)===o||!!m&&G(e)!==e):h,y=await er(e,b),w=(null==(t=a.autoPlacement)?void 0:t.index)||0,x=v[w];if(null==x)return{};let E=Y(x,i,await (null==u.isRTL?void 0:u.isRTL(c.floating)));if(s!==x)return{reset:{placement:v[0]}};let S=[y[z(x)],y[E[0]],y[E[1]]],k=[...(null==(n=a.autoPlacement)?void 0:n.overflows)||[],{placement:x,overflows:S}],T=v[w+1];if(T)return{data:{index:w+1,overflows:k},reset:{placement:T}};let C=k.map(e=>{let t=B(e.placement);return[e.placement,t&&f?e.overflows.slice(0,2).reduce((e,t)=>e+t,0):e.overflows[0],e.overflows]}).sort((e,t)=>e[1]-t[1]),P=(null==(r=C.filter(e=>e[2].slice(0,B(e[0])?2:3).every(e=>e<=0))[0])?void 0:r[0])||C[0][0];return P!==s?{data:{index:w+1,overflows:k},reset:{placement:P}}:{}}}):c.middleware.push((void 0===a&&(a={}),{name:"flip",options:a,async fn(e){var t,n,r,o,i;let{placement:l,middlewareData:s,rects:u,initialPlacement:c,platform:d,elements:f}=e,p=F(a,e),{mainAxis:h=!0,crossAxis:m=!0,fallbackPlacements:b,fallbackStrategy:v="bestFit",fallbackAxisSideDirection:y="none",flipAlignment:w=!0}=p,x=g(p,Q);if(null!=(t=s.arrow)&&t.alignmentOffset)return{};let E=z(l),S=U(c),k=z(c)===c,T=await (null==d.isRTL?void 0:d.isRTL(f.floating)),C=b||(k||!w?[X(c)]:function(e){let t=X(e);return[G(e),t,G(t)]}(c)),P="none"!==y;!b&&P&&C.push(...function(e,t,n,r){let o=B(e),i=function(e,t,n){let r=["left","right"],o=["right","left"];switch(e){case"top":case"bottom":if(n)return t?o:r;return t?r:o;case"left":case"right":return t?["top","bottom"]:["bottom","top"];default:return[]}}(z(e),"start"===n,r);return o&&(i=i.map(e=>e+"-"+o),t&&(i=i.concat(i.map(G)))),i}(c,w,y,T));let M=[c,...C],R=await er(e,x),j=[],O=(null==(n=s.flip)?void 0:n.overflows)||[];if(h&&j.push(R[E]),m){let e=Y(l,u,T);j.push(R[e[0]],R[e[1]])}if(O=[...O,{placement:l,overflows:j}],!j.every(e=>e<=0)){let e=((null==(r=s.flip)?void 0:r.index)||0)+1,t=M[e];if(t)return{data:{index:e,overflows:O},reset:{placement:t}};let n=null==(o=O.filter(e=>e.overflows[0]<=0).sort((e,t)=>e.overflows[1]-t.overflows[1])[0])?void 0:o.placement;if(!n)switch(v){case"bestFit":{let e=null==(i=O.filter(e=>{if(P){let t=U(e.placement);return t===S||"y"===t}return!0}).map(e=>[e.placement,e.overflows.filter(e=>e>0).reduce((e,t)=>e+t,0)]).sort((e,t)=>e[1]-t[1])[0])?void 0:i[0];e&&(n=e);break}case"initialPlacement":n=c}if(l!==n)return{reset:{placement:n}}}return{}}})),c.middleware.push({name:"shift",options:u={limiter:(void 0===s&&(s={}),{options:s,fn(e){let{x:t,y:n,placement:r,rects:o,middlewareData:i}=e,{offset:l=0,mainAxis:a=!0,crossAxis:u=!0}=F(s,e),c={x:t,y:n},d=U(r),f=q(d),p=c[f],h=c[d],g=F(l,e),b="number"==typeof g?{mainAxis:g,crossAxis:0}:m({mainAxis:0,crossAxis:0},g);if(a){let e="y"===f?"height":"width",t=o.reference[f]-o.floating[e]+b.mainAxis,n=o.reference[f]+o.reference[e]-b.mainAxis;p<t?p=t:p>n&&(p=n)}if(u){var v,y;let e="y"===f?"width":"height",t=["top","left"].includes(z(r)),n=o.reference[d]-o.floating[e]+(t&&(null==(v=i.offset)?void 0:v[d])||0)+(t?0:b.crossAxis),l=o.reference[d]+o.reference[e]+(t?0:(null==(y=i.offset)?void 0:y[d])||0)-(t?b.crossAxis:0);h<n?h=n:h>l&&(h=l)}return{[f]:p,[d]:h}}}),crossAxis:!0},async fn(e){let{x:t,y:n,placement:r}=e,o=F(u,e),{mainAxis:i=!0,crossAxis:l=!1,limiter:a={fn:e=>{let{x:t,y:n}=e;return{x:t,y:n}}}}=o,s=g(o,ee),c={x:t,y:n},d=await er(e,s),f=U(z(r)),p=q(f),h=c[p],b=c[f];if(i){let e="y"===p?"top":"left",t="y"===p?"bottom":"right",n=h+d[e],r=h-d[t];h=N(n,_(h,r))}if(l){let e="y"===f?"top":"left",t="y"===f?"bottom":"right",n=b+d[e],r=b-d[t];b=N(n,_(b,r))}let v=a.fn(m({},e,{[p]:h,[f]:b}));return m({},v,{data:{x:v.x-t,y:v.y-n}})}}),d){let e="object"==typeof t.options.arrow?t.options.arrow:{padding:4};c.middleware.push(e$({element:d,padding:h?e.padding:0}))}p||(c.placement=e.on)}return k(c,t.options.floatingUIOptions||{})}(t,e),o=O(t);o&&(n=document.body,e.shepherdElementComponent.getElement().classList.add("shepherd-centered")),e.cleanup=function(e,t,n,r){let o;void 0===r&&(r={});let{ancestorScroll:i=!0,ancestorResize:l=!0,elementResize:a="function"==typeof ResizeObserver,layoutShift:s="function"==typeof IntersectionObserver,animationFrame:u=!1}=r,c=eS(e),d=i||l?[...c?ew(c):[],...ew(t)]:[];d.forEach(e=>{i&&e.addEventListener("scroll",n,{passive:!0}),l&&e.addEventListener("resize",n)});let f=c&&s?function(e,t){let n,r=null,o=ea(e);function i(){var e;clearTimeout(n),null==(e=r)||e.disconnect(),r=null}return function l(a,s){void 0===a&&(a=!1),void 0===s&&(s=1),i();let{left:u,top:c,width:d,height:f}=e.getBoundingClientRect();if(a||t(),!d||!f)return;let p=I(c),h=I(o.clientWidth-(u+d)),g={rootMargin:-p+"px "+-h+"px "+-I(o.clientHeight-(c+f))+"px "+-I(u)+"px",threshold:N(0,_(1,s))||1},b=!0;function v(e){let t=e[0].intersectionRatio;if(t!==s){if(!b)return l();t?l(!1,t):n=setTimeout(()=>{l(!1,1e-7)},1e3)}b=!1}try{r=new IntersectionObserver(v,m({},g,{root:o.ownerDocument}))}catch(e){r=new IntersectionObserver(v,g)}r.observe(e)}(!0),i}(c,n):null,p=-1,h=null;a&&(h=new ResizeObserver(e=>{let[r]=e;r&&r.target===c&&h&&(h.unobserve(t),cancelAnimationFrame(p),p=requestAnimationFrame(()=>{var e;null==(e=h)||e.observe(t)})),n()}),c&&!u&&h.observe(c),h.observe(t));let g=u?eP(e):null;return u&&function t(){let r=eP(e);g&&(r.x!==g.x||r.y!==g.y||r.width!==g.width||r.height!==g.height)&&n(),g=r,o=requestAnimationFrame(t)}(),n(),()=>{var e;d.forEach(e=>{i&&e.removeEventListener("scroll",n),l&&e.removeEventListener("resize",n)}),null==f||f(),null==(e=h)||e.disconnect(),h=null,u&&cancelAnimationFrame(o)}}(n,e.el,()=>{if(!e.el){null==e.cleanup||e.cleanup();return}(function(e,t,n,r){eI(e,t.el,n).then(({x:e,y:n,placement:o,middlewareData:i})=>(t.el&&(r?Object.assign(t.el.style,{position:"fixed",left:"50%",top:"50%",transform:"translate(-50%, -50%)"}):Object.assign(t.el.style,{position:"absolute",left:`${e}px`,top:`${n}px`}),t.el.dataset.popperPlacement=o,function(e,t){let n=e.querySelector(".shepherd-arrow");if(c(n)&&t.arrow){let{x:e,y:r}=t.arrow;Object.assign(n.style,{left:null!=e?`${e}px`:"",top:null!=r?`${r}px`:""})}}(t.el,i)),t)).then(e=>new Promise(t=>{setTimeout(()=>t(e),300)})).then(e=>{null!=e&&e.el&&e.el.focus({preventScroll:!0})})})(n,e,r,o)}),e.target=t.element}(this)}_show(){var e;this.trigger("before-show"),this._resolveAttachToOptions(),this._resolveExtraHiglightElements(),this._setupElements(),this.tour.modal||this.tour.setupModal(),null==(e=this.tour.modal)||e.setupForStep(this),this._styleTargetElementForStep(this),this.el&&(this.el.hidden=!1),this.options.scrollTo&&setTimeout(()=>{this._scrollTo(this.options.scrollTo)}),this.el&&(this.el.hidden=!1);let t=this.shepherdElementComponent.getElement(),n=this.target||document.body,r=this._resolvedExtraHighlightElements;n.classList.add(`${this.classPrefix}shepherd-enabled`),n.classList.add(`${this.classPrefix}shepherd-target`),t.classList.add("shepherd-enabled"),null==r||r.forEach(e=>{e.classList.add(`${this.classPrefix}shepherd-enabled`),e.classList.add(`${this.classPrefix}shepherd-target`)}),this.trigger("show")}_styleTargetElementForStep(e){let t=e.target,n=e._resolvedExtraHighlightElements;if(!t)return;let r=e.options.highlightClass;r&&(t.classList.add(r),null==n||n.forEach(e=>e.classList.add(r))),t.classList.remove("shepherd-target-click-disabled"),null==n||n.forEach(e=>e.classList.remove("shepherd-target-click-disabled")),!1===e.options.canClickTarget&&(t.classList.add("shepherd-target-click-disabled"),null==n||n.forEach(e=>e.classList.add("shepherd-target-click-disabled")))}_updateStepTargetOnHide(){let e=this.target||document.body,t=this._resolvedExtraHighlightElements,n=this.options.highlightClass;n&&(e.classList.remove(n),null==t||t.forEach(e=>e.classList.remove(n))),e.classList.remove("shepherd-target-click-disabled",`${this.classPrefix}shepherd-enabled`,`${this.classPrefix}shepherd-target`),null==t||t.forEach(e=>{e.classList.remove("shepherd-target-click-disabled",`${this.classPrefix}shepherd-enabled`,`${this.classPrefix}shepherd-target`)})}}function tK(e){let t,n,r,o,i;return{c(){t=eG("svg"),eJ(n=eG("path"),"d",e[2]),eJ(t,"class",r=`${e[1]?"shepherd-modal-is-visible":""} shepherd-modal-overlay-container`)},m(r,l){eV(r,t,l),eq(t,n),e[11](t),o||(i=eZ(t,"touchmove",e[3]),o=!0)},p(e,[o]){4&o&&eJ(n,"d",e[2]),2&o&&r!==(r=`${e[1]?"shepherd-modal-is-visible":""} shepherd-modal-overlay-container`)&&eJ(t,"class",r)},i:eD,o:eD,d(n){n&&eU(t),e[11](null),o=!1,i()}}}function tZ(e,t,n){let r,o,{element:i,openingProperties:l}=t,a=!1;function s(){n(4,l=[{width:0,height:0,x:0,y:0,r:0}])}function u(){n(1,a=!1),p()}function c(e=0,t=0,r=0,o=0,i,a,u){if(a){let s=[a,...u||[]];for(let a of(n(4,l=[]),s)){if(!a||s.indexOf(a)!==s.lastIndexOf(a))continue;let{y:n,height:u}=function(e,t){let n=e.getBoundingClientRect(),r=n.y||n.top,o=n.bottom||r+n.height;if(t){let e=t.getBoundingClientRect(),n=e.y||e.top,i=e.bottom||n+e.height;r=Math.max(r,n),o=Math.min(o,i)}let i=Math.max(o-r,0);return{y:r,height:i}}(a,i),{x:c,width:d,left:f}=a.getBoundingClientRect();s.some(e=>{if(e===a)return!1;let t=e.getBoundingClientRect();return c>=t.left&&c+d<=t.right&&n>=t.top&&n+u<=t.bottom})||l.push({width:d+2*e,height:u+2*e,x:(c||f)+r-e,y:n+o-e,r:t})}}else s()}function d(){n(1,a=!0)}s();let f=e=>{e.preventDefault()};function p(){r&&(cancelAnimationFrame(r),r=void 0),window.removeEventListener("touchmove",f,{passive:!1})}return e.$$set=e=>{"element"in e&&n(0,i=e.element),"openingProperties"in e&&n(4,l=e.openingProperties)},e.$$.update=()=>{16&e.$$.dirty&&n(2,o=function(e){let t="",{innerWidth:n,innerHeight:r}=window;return e.forEach(e=>{let{width:n,height:r,x:o=0,y:i=0,r:l=0}=e,{topLeft:a=0,topRight:s=0,bottomRight:u=0,bottomLeft:c=0}="number"==typeof l?{topLeft:l,topRight:l,bottomRight:l,bottomLeft:l}:l;t+=`M${o+a},${i}\
      a${a},${a},0,0,0-${a},${a}\
      V${r+i-c}\
      a${c},${c},0,0,0,${c},${c}\
      H${n+o-u}\
      a${u},${u},0,0,0,${u}-${u}\
      V${i+s}\
      a${s},${s},0,0,0-${s}-${s}\
      Z`}),`M${n},${r}\
          H0\
          V0\
          H${n}\
          V${r}\
          Z\
          ${t}`.replace(/\s/g,"")}(l))},[i,a,o,e=>{e.stopPropagation()},l,()=>i,s,u,c,function(e){p(),e.tour.options.useModalOverlay?(function(e){let{modalOverlayOpeningPadding:t,modalOverlayOpeningRadius:n,modalOverlayOpeningXOffset:o=0,modalOverlayOpeningYOffset:i=0}=e.options,l=function(e){let t={top:0,left:0};if(!e)return t;let n=e.ownerDocument.defaultView;for(;n!==window.top;){var r,o,i;let e=null==(r=n)?void 0:r.frameElement;if(e){let n=e.getBoundingClientRect();t.top+=n.top+(null!=(o=n.scrollTop)?o:0),t.left+=n.left+(null!=(i=n.scrollLeft)?i:0)}n=n.parent}return t}(e.target),a=function e(t){if(!t)return null;let n=t instanceof HTMLElement&&window.getComputedStyle(t).overflowY;return"hidden"!==n&&"visible"!==n&&t.scrollHeight>=t.clientHeight?t:e(t.parentElement)}(e.target),s=()=>{r=void 0,c(t,n,o+l.left,i+l.top,a,e.target,e._resolvedExtraHighlightElements),r=requestAnimationFrame(s)};s(),window.addEventListener("touchmove",f,{passive:!1})}(e),d()):u()},d,function(e){e6[e?"unshift":"push"](()=>{n(0,i=e)})}]}class tJ extends th{constructor(e){super(),tp(this,e,tZ,tK,eB,{element:0,openingProperties:4,getElement:5,closeModalOpening:6,hide:7,positionModal:8,setupForStep:9,show:10})}get getElement(){return this.$$.ctx[5]}get closeModalOpening(){return this.$$.ctx[6]}get hide(){return this.$$.ctx[7]}get positionModal(){return this.$$.ctx[8]}get setupForStep(){return this.$$.ctx[9]}get show(){return this.$$.ctx[10]}}class tQ extends h{constructor(){super(),this.activeTour=void 0,P(this)}}class t0 extends h{constructor(e={}){return super(),this.trackedEvents=["active","cancel","complete","show"],this.classPrefix=void 0,this.currentStep=void 0,this.focusedElBeforeOpen=void 0,this.id=void 0,this.modal=void 0,this.options=void 0,this.steps=void 0,P(this),this.options=Object.assign({},{exitOnEsc:!0,keyboardNavigation:!0},e),this.classPrefix=j(this.options.classPrefix),this.steps=[],this.addSteps(this.options.steps),["active","cancel","complete","inactive","show","start"].map(e=>{(e=>{this.on(e,t=>{(t=t||{}).tour=this,t1.trigger(e,t)})})(e)}),this._setTourID(e.id),this}addStep(e,t){let n=e;return n instanceof tX?n.tour=this:n=new tX(this,n),p(t)?this.steps.push(n):this.steps.splice(t,0,n),n}addSteps(e){return Array.isArray(e)&&e.forEach(e=>{this.addStep(e)}),this}back(){let e=this.steps.indexOf(this.currentStep);this.show(e-1,!1)}async cancel(){if(this.options.confirmCancel){let e=this.options.confirmCancelMessage||"Are you sure you want to stop the tour?";(d(this.options.confirmCancel)?await this.options.confirmCancel():window.confirm(e))&&this._done("cancel")}else this._done("cancel")}complete(){this._done("complete")}getById(e){return this.steps.find(t=>t.id===e)}getCurrentStep(){return this.currentStep}hide(){let e=this.getCurrentStep();if(e)return e.hide()}isActive(){return t1.activeTour===this}next(){let e=this.steps.indexOf(this.currentStep);e===this.steps.length-1?this.complete():this.show(e+1,!0)}removeStep(e){let t=this.getCurrentStep();this.steps.some((t,n)=>{if(t.id===e)return t.isOpen()&&t.hide(),t.destroy(),this.steps.splice(n,1),!0}),t&&t.id===e&&(this.currentStep=void 0,this.steps.length?this.show(0):this.cancel())}show(e=0,t=!0){let n=f(e)?this.getById(e):this.steps[e];n&&(this._updateStateBeforeShow(),d(n.options.showOn)&&!n.options.showOn()?this._skipStep(n,t):(this.currentStep=n,this.trigger("show",{step:n,previous:this.currentStep}),n.show()))}async start(){this.trigger("start"),this.focusedElBeforeOpen=document.activeElement,this.currentStep=null,this.setupModal(),this._setupActiveTour(),this.next()}_done(e){let t=this.steps.indexOf(this.currentStep);if(Array.isArray(this.steps)&&this.steps.forEach(e=>e.destroy()),!function(e){if(e){let{steps:t}=e;t.forEach(e=>{e.options&&!1===e.options.canClickTarget&&e.options.attachTo&&(c(e.target)&&e.target.classList.remove("shepherd-target-click-disabled"),e._resolvedExtraHighlightElements&&e._resolvedExtraHighlightElements.forEach(e=>{c(e)&&e.classList.remove("shepherd-target-click-disabled")}))})}}(this),this.trigger(e,{index:t}),t1.activeTour=null,this.trigger("inactive",{tour:this}),this.modal&&this.modal.hide(),("cancel"===e||"complete"===e)&&this.modal){let e=document.querySelector(".shepherd-modal-overlay-container");e&&(e.remove(),this.modal=null)}c(this.focusedElBeforeOpen)&&this.focusedElBeforeOpen.focus()}_setupActiveTour(){this.trigger("active",{tour:this}),t1.activeTour=this}setupModal(){this.modal=new tJ({target:this.options.modalContainer||document.body,props:{styles:this.styles}})}_skipStep(e,t){let n=this.steps.indexOf(e);n===this.steps.length-1?this.complete():this.show(t?n+1:n-1,t)}_updateStateBeforeShow(){this.currentStep&&this.currentStep.hide(),this.isActive()||this._setupActiveTour()}_setTourID(e){let t=this.options.tourName||"tour",n=e||L();this.id=`${t}--${n}`}}let t1=new tQ,t2="undefined"==typeof window;t1.Step=t2?M:tX,t1.Tour=t2?R:t0;var t4=n(6556);let t5=(0,s.createContext)(null);function t6(){let e=(0,s.useContext)(t5);if(!e)throw Error("useTour must be used within a TourProvider");return e}function t3(e){let{children:t}=e,n=(0,s.useRef)(null),r=(0,u.useRouter)(),{isFirstVisit:o,markTourCompleted:i}=(0,t4.n)(),[l,c]=(0,s.useState)(!1),[d,f]=(0,s.useState)(!1),[p,h]=(0,s.useState)(!1),m=(0,s.useRef)(!1),g=()=>{n.current&&(f(!0),h(!0),n.current.start(),setTimeout(()=>{h(!1)},1e3))};(0,s.useEffect)(()=>{if(!n.current){n.current=new t1.Tour({useModalOverlay:!1,defaultStepOptions:{cancelIcon:{enabled:!0},scrollTo:{behavior:"smooth",block:"center"},arrow:!1,highlightClass:"shepherd-highlight",when:{show(){var e,t,n;let r=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),o=null==r?void 0:r.getElement(),i=null==o?void 0:o.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(r))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),o){let e=o.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=o.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}let a=null==r?void 0:r.getTarget();a&&a instanceof HTMLElement&&(a.style.outline="3px solid #3b82f6",a.style.outlineOffset="2px",a.style.borderRadius="8px",a.style.position="relative",a.style.zIndex="9999",a.setAttribute("data-shepherd-highlighted","true"))},hide(){let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted")),document.documentElement.style.removeProperty("--shepherd-dialog-height");let t=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');t&&t instanceof HTMLElement&&(t.style.removeProperty("height"),t.style.removeProperty("max-height"))}}}});let e=document.createElement("style");e.id="shepherd-global-custom-style",e.textContent='\n          .shepherd-element {\n            /* Uniform 1px border using inner box-shadow so corners stay consistent */\n            border: none !important;\n            border-radius: 10px !important;\n            z-index: 30000 !important;\n            box-shadow: 0 0 0 1px #d1d5db inset, 0 10px 15px -3px rgba(0, 0, 0, 0.1), 0 4px 6px -2px rgba(0, 0, 0, 0.05) !important;\n            overflow: visible !important;\n            background-clip: padding-box !important;\n          }\n\n          .shepherd-title {\n              font-weight: bold;\n              color: #111827;\n              margin: 0;\n          }\n\n          .shepherd-element .shepherd-header {\n              padding: 1rem 1rem 0.5rem 1rem;\n          }\n\n          .shepherd-element .shepherd-text {\n              padding: 0.5rem 1rem 1rem 1rem;\n          }\n\n          /* Fix mobile menu gap when tour dialog is present */\n          @media (max-width: 767px) {\n            /* Very specific selector to override Tailwind\'s h-[calc(100vh-56px)] class */\n            div.fixed.top-14.left-0.w-64.bg-white.border-r.shadow-lg.z-50.transform,\n            .fixed.top-14.left-0.w-64.bg-white.shadow-lg.z-50,\n            div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"][class*="bg-white"][class*="shadow-lg"][class*="z-50"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n              max-height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Target the mobile menu by its exact class combination from the HTML */\n            .fixed.w-64.bg-white.border-r.border-gray-200.shadow-lg.z-50.transform,\n            .fixed[class*="w-64"][class*="bg-white"][class*="border-r"][class*="shadow-lg"][class*="z-50"][class*="transform"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Even more specific - target by multiple class combinations */\n            .fixed.top-14.left-0[class*="w-64"],\n            div.fixed[class*="top-14"][class*="left-0"][class*="w-64"] {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Super aggressive approach - use high specificity to override Tailwind */\n            body div.fixed.w-64:not(.shepherd-element),\n            html body div.fixed.w-64:not(.shepherd-element) {\n              height: calc(100vh - 56px - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Fallback selectors for other mobile menu patterns */\n            nav[data-state="open"],\n            .mobile-menu.open,\n            [data-mobile-menu="true"] {\n              height: calc(100vh - var(--shepherd-dialog-height, 200px)) !important;\n              max-height: calc(100vh - var(--shepherd-dialog-height, 200px)) !important;\n            }\n\n            /* Ensure mobile menu content flows properly */\n            .fixed.w-64 nav,\n            .fixed[class*="w-64"] nav {\n              height: 100% !important;\n              overflow-y: auto !important;\n            }\n          }\n\n        ',document.getElementById("shepherd-global-custom-style")||document.head.appendChild(e),n.current.on("complete",()=>{let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted"));let t=document.getElementById("shepherd-column-overlay");t&&t.remove();let n=document.getElementById("shepherd-column-anchor");n&&n.remove();let r=document.getElementById("shepherd-user-column-overlay");r&&r.remove();let o=document.getElementById("shepherd-user-column-anchor");o&&o.remove();let l=document.getElementById("shepherd-global-custom-style");l&&l.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let a=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');a&&a instanceof HTMLElement&&(a.style.removeProperty("height"),a.style.removeProperty("max-height")),f(!1),h(!1),i()}),n.current.on("cancel",()=>{let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted"));let t=document.getElementById("shepherd-column-overlay");t&&t.remove();let n=document.getElementById("shepherd-column-anchor");n&&n.remove();let r=document.getElementById("shepherd-user-column-overlay");r&&r.remove();let o=document.getElementById("shepherd-user-column-anchor");o&&o.remove();let l=document.getElementById("shepherd-global-custom-style");l&&l.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let a=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');a&&a instanceof HTMLElement&&(a.style.removeProperty("height"),a.style.removeProperty("max-height")),f(!1),h(!1),i()}),[{title:"\uD83D\uDC4B Welcome to SkyPilot!",text:"\n              <p>SkyPilot is a system for managing AI workloads on any cluster and cloud infrastructure.</p>\n            ",buttons:[{text:"Skip Tour",action(){this.cancel()},classes:"shepherd-button-secondary"},{text:"Start Tour",action(){this.next()}}]},{title:"Clusters",text:"\n              <p>Spin up <strong>Sky Clusters</strong> on any infrastructure you have access to.</p>\n              <p>Easily SSH into any node, connect an IDE, or queue development jobs.</p>\n            ",attachTo:{element:'a[href="/dashboard/clusters"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()};"/dashboard/clusters"!==r.pathname?(m.current=!0,r.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let r=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),o=null==r?void 0:r.getElement(),i=null==o?void 0:o.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(r))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),o){let e=o.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=o.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}let a=document.querySelector('a[href="/dashboard/clusters"]');a&&a instanceof HTMLElement&&(a.style.outline="3px solid #3b82f6",a.style.outlineOffset="2px",a.style.borderRadius="8px",a.style.position="relative",a.style.zIndex="9999",a.setAttribute("data-shepherd-highlighted","true"))},hide(){let e=document.querySelector('[data-shepherd-highlighted="true"]');e&&e instanceof HTMLElement&&(e.style.outline="",e.style.outlineOffset="",e.style.borderRadius="",e.style.boxShadow="",e.style.position="",e.style.zIndex="",e.removeAttribute("data-shepherd-highlighted")),document.documentElement.style.removeProperty("--shepherd-dialog-height");let t=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');t&&t instanceof HTMLElement&&(t.style.removeProperty("height"),t.style.removeProperty("max-height"))}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"SkyPilot is infra-agnostic",text:"\n              <p>Run compute on any hyperscaler, neocloud, or Kubernetes cluster — all within a unified system.</p>\n            ",attachTo:{element:function(){let e=document.getElementById("shepherd-column-anchor");if(e)return e;let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(t){let e=t.closest("table"),n=Array.from(t.parentElement.children).indexOf(t);if(e){let t=e.querySelectorAll("tbody tr"),r=null;for(let e=t.length-1;e>=0;e--){let o=t[e].children[n];if(o){r=o;break}}if(r)return r}return t}return document.querySelector("table")||"body"},on:"bottom",offset:{skidding:0,distance:15}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(t&&t instanceof HTMLElement){let r=t.closest("table");if(r){let o=t.getBoundingClientRect(),i=window.innerWidth,l=r.closest('.overflow-x-auto, .overflow-auto, [style*="overflow"]')||r.parentElement;if((o.right>i||o.left<0)&&l&&l instanceof HTMLElement){let r=l.getBoundingClientRect(),i=t.offsetLeft-r.width/2+o.width/2;l.scrollTo({left:Math.max(0,i),behavior:"smooth"}),setTimeout(()=>{n(),e()},300);return}n()}}e()},n=()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),r=e.getBoundingClientRect(),o=t.querySelectorAll("tbody tr"),i=r;o.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=i.bottom+4,a=document.createElement("div");a.id="shepherd-column-anchor",a.style.position="fixed",a.style.left="".concat(r.left+r.width/2,"px"),a.style.top="".concat(l,"px"),a.style.width="1px",a.style.height="1px",a.style.zIndex="9999",a.style.pointerEvents="none",a.style.backgroundColor="transparent",a.style.transform="translate(-50%, -50%)",document.body.appendChild(a)}}};"/dashboard/clusters"!==window.location.pathname?(m.current=!0,r.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let r=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),o=null==r?void 0:r.getElement(),i=null==o?void 0:o.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(r))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),o){let e=o.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=o.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}setTimeout(()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"Infra"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),r=e.getBoundingClientRect(),o=t.querySelectorAll("tbody tr"),i=r;o.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=document.createElement("div");l.id="shepherd-column-overlay",l.style.position="fixed",l.style.left="".concat(r.left-4,"px"),l.style.top="".concat(r.top-4,"px"),l.style.width="".concat(r.width+8,"px"),l.style.height="".concat(i.bottom-r.top+8,"px"),l.style.outline="3px solid #3b82f6",l.style.outlineOffset="2px",l.style.borderRadius="8px",l.style.zIndex="9998",l.style.pointerEvents="none",l.style.backgroundColor="transparent",document.body.appendChild(l)}}},100)},hide(){let e=document.getElementById("shepherd-column-overlay");e&&e.remove();let t=document.getElementById("shepherd-column-anchor");t&&t.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let n=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');n&&n instanceof HTMLElement&&(n.style.removeProperty("height"),n.style.removeProperty("max-height"));let r=document.getElementById("shepherd-global-custom-style");r&&r.remove()}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Multi-user support",text:"\n              <p>SkyPilot supports multiple users in an organization.</p>\n              <p>Each user can have their own clusters and jobs, with proper access controls.</p>\n            ",attachTo:{element:function(){let e=document.getElementById("shepherd-user-column-anchor");if(e)return e;let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(t){let e=t.closest("table"),n=Array.from(t.parentElement.children).indexOf(t);if(e){let t=e.querySelectorAll("tbody tr"),r=null;for(let e=t.length-1;e>=0;e--){let o=t[e].children[n];if(o){r=o;break}}if(r)return r}return t}return document.querySelector("table")||"body"},on:"bottom",offset:{skidding:0,distance:15}},beforeShowPromise:function(){return new Promise(e=>{let t=()=>{let t=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(t&&t instanceof HTMLElement){let r=t.closest("table");if(r){let o=t.getBoundingClientRect(),i=window.innerWidth,l=r.closest('.overflow-x-auto, .overflow-auto, [style*="overflow"]')||r.parentElement;if((o.right>i||o.left<0)&&l&&l instanceof HTMLElement){let r=l.getBoundingClientRect(),i=t.offsetLeft-r.width/2+o.width/2;l.scrollTo({left:Math.max(0,i),behavior:"smooth"}),setTimeout(()=>{n(),e()},300);return}n()}}e()},n=()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),r=e.getBoundingClientRect(),o=t.querySelectorAll("tbody tr"),i=r;o.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=i.bottom+4,a=document.createElement("div");a.id="shepherd-user-column-anchor",a.style.position="fixed",a.style.left="".concat(r.left+r.width/2,"px"),a.style.top="".concat(l,"px"),a.style.width="1px",a.style.height="1px",a.style.zIndex="9999",a.style.pointerEvents="none",a.style.backgroundColor="transparent",a.style.transform="translate(-50%, -50%)",document.body.appendChild(a)}}};"/dashboard/clusters"!==window.location.pathname?(m.current=!0,r.push("/clusters").then(()=>{m.current=!1,setTimeout(t,200)})):t()})},when:{show(){var e,t,n;let r=null===(e=t1.activeTour)||void 0===e?void 0:e.getCurrentStep(),o=null==r?void 0:r.getElement(),i=null==o?void 0:o.querySelector(".shepherd-footer"),l=document.createElement("span");if(l.className="shepherd-progress",l.innerText="".concat((null===(t=t1.activeTour)||void 0===t?void 0:t.steps.indexOf(r))+1," of ").concat(null===(n=t1.activeTour)||void 0===n?void 0:n.steps.length),null==i||i.insertBefore(l,i.firstChild),o){let e=o.offsetHeight;if(document.documentElement.style.setProperty("--shepherd-dialog-height","".concat(e+20,"px")),window.innerWidth<768){let e=null;for(let t of[".fixed.top-14.left-0.w-64","div.fixed.w-64.bg-white.border-r",".fixed.w-64.transform",'[class*="fixed"][class*="w-64"][class*="bg-white"]','div[class*="fixed"][class*="top-14"][class*="left-0"][class*="w-64"]'])if(e=document.querySelector(t))break;if(!e)for(let t of document.querySelectorAll("div.fixed")){let n=t.getBoundingClientRect();if(256===n.width&&0===n.left&&n.top>=50){e=t;break}}if(e&&e instanceof HTMLElement){let t=o.getBoundingClientRect().top-56;e.style.setProperty("height","".concat(t,"px"),"important"),e.style.setProperty("max-height","".concat(t,"px"),"important")}}}setTimeout(()=>{let e=Array.from(document.querySelectorAll("thead th")).find(e=>e.textContent&&"User"===e.textContent.trim());if(e&&e instanceof HTMLElement){let t=e.closest("table");if(t){let n=Array.from(e.parentElement.children).indexOf(e),r=e.getBoundingClientRect(),o=t.querySelectorAll("tbody tr"),i=r;o.forEach(e=>{let t=e.children[n];t&&(i=t.getBoundingClientRect())});let l=document.createElement("div");l.id="shepherd-user-column-overlay",l.style.position="fixed",l.style.left="".concat(r.left-4,"px"),l.style.top="".concat(r.top-4,"px"),l.style.width="".concat(r.width+8,"px"),l.style.height="".concat(i.bottom-r.top+8,"px"),l.style.outline="3px solid #3b82f6",l.style.outlineOffset="2px",l.style.borderRadius="8px",l.style.zIndex="9998",l.style.pointerEvents="none",l.style.backgroundColor="transparent",document.body.appendChild(l)}}},100)},hide(){let e=document.getElementById("shepherd-user-column-overlay");e&&e.remove();let t=document.getElementById("shepherd-user-column-anchor");t&&t.remove(),document.documentElement.style.removeProperty("--shepherd-dialog-height");let n=document.querySelector(".fixed.top-14.left-0.w-64")||document.querySelector("div.fixed.w-64.bg-white.border-r")||document.querySelector(".fixed.w-64.transform")||document.querySelector('[class*="fixed"][class*="w-64"][class*="bg-white"]');n&&n instanceof HTMLElement&&(n.style.removeProperty("height"),n.style.removeProperty("max-height"));let r=document.getElementById("shepherd-global-custom-style");r&&r.remove()}},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Spin up compute in seconds",text:'\n            <p>Spin up clusters using the Python SDK or the CLI.</p>\n            <div class="space-y-2">\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch --gpus L4:8</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n              <div class="rounded-lg border text-card-foreground shadow-sm p-3 bg-gray-50">\n                <div class="flex items-center justify-between">\n                  <pre class="text-sm w-full whitespace-pre-wrap">\n                    <code class="block">sky launch --num-nodes 10 --cpus 32+</code>\n                  </pre>\n                  <button class="inline-flex items-center justify-center whitespace-nowrap text-sm font-medium ring-offset-background transition-colors focus-visible:outline-none focus-visible:ring-2 focus-visible:ring-ring focus-visible:ring-offset-2 disabled:pointer-events-none disabled:opacity-50 hover:bg-accent hover:text-accent-foreground h-8 w-8 rounded-full" onclick="copyDashboardCodeBlock(this)" title="Copy to clipboard">\n                    <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-copy h-4 w-4">\n                      <rect width="14" height="14" x="8" y="8" rx="2" ry="2"></rect>\n                      <path d="M4 16c-1.1 0-2-.9-2-2V4c0-1.1.9-2 2-2h10c1.1 0 2 .9 2 2"></path>\n                    </svg>\n                  </button>\n                </div>\n              </div>\n            </div>\n            ',buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Jobs",text:"\n            <p>Use <strong>Managed Jobs</strong> for long-running workloads.</p>\n            <p>They provide automatic recovery against failures, such as recovering from preemptions or transient errors.</p>\n          ",attachTo:{element:function(){return document.querySelector('a[href="/dashboard/jobs"]')||document.querySelector('a[href="/dashboard/jobs"]')||'a[href="/dashboard/jobs"]'},on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Bring one or many infrastructure",text:"\n            <p>SkyPilot combines your Kubernetes clusters, cloud VMs, or on-premise nodes into a unified compute pool.</p>\n            <p>You can monitor them in this page.</p>\n          ",attachTo:{element:function(){return document.querySelector('a[href="/dashboard/infra"]')||document.querySelector('a[href="/dashboard/infra"]')||'a[href="/dashboard/infra"]'},on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Workspaces",text:"\n            <p>For team deployments, admins can use Workspaces to manage teams or projects.</p>\n          ",attachTo:{element:'a[href="/dashboard/workspaces"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"Users",text:"\n            <p>SkyPilot provides user management with RBAC and SSO support. Admins can manage users in this page.</p>\n          ",attachTo:{element:'a[href="/dashboard/users"]',on:"bottom",offset:{skidding:0,distance:10}},beforeShowPromise:function(){return new Promise(e=>{(()=>{let t=window.innerWidth<768,n=document.querySelector('[data-testid="mobile-menu-button"], button[aria-label*="menu"], button[aria-label*="Menu"], .mobile-menu-button, [role="button"][aria-expanded]');if(t&&n&&!("true"===n.getAttribute("aria-expanded")||n.classList.contains("open")||document.querySelector('nav[data-state="open"], .mobile-menu.open, [data-mobile-menu="true"]'))){n instanceof HTMLElement&&n.click(),setTimeout(()=>{e()},300);return}e()})()})},buttons:[{text:"Back",action(){this.back()},classes:"shepherd-button-secondary"},{text:"Next",action(){this.next()}}]},{title:"\uD83C\uDF89 Tour complete!",text:'\n              <p>We invite you to to explore the rest of the dashboard.</p>\n              <p>To get started with SkyPilot, refer to <a href="https://docs.skypilot.co/en/latest/getting-started/quickstart.html">Quickstart</a>. You can restart this tour by clicking the <svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="inline-block h-4 w-4 align-middle"><circle cx="12" cy="12" r="10"></circle><path d="M9.09 9a3 3 0 0 1 5.83 1c0 2-3 3-3 3"></path><path d="M12 17h.01"></path></svg> icon in the bottom right corner.</p>\n              <p>Have questions? Join Slack to directly chat with the SkyPilot team.</p>\n            ',buttons:[{text:"Finish",action(){this.complete()},classes:"shepherd-button-secondary"},{text:"Join Slack",action(){window.open("https://slack.skypilot.co/","_blank"),this.complete()}}]}].forEach(e=>{n.current.addStep(e)})}return o&&!l&&(g(),c(!0)),()=>{n.current&&n.current.complete()}},[o]),(0,s.useEffect)(()=>{if(!d||p)return;let e=e=>{if(!m.current){if(window.confirm("The tour is currently in progress. Do you want to abort the tour and navigate away?\n\nYou can resume the tour by clicking the question mark on the bottom right."))n.current&&n.current.cancel();else throw r.events.emit("routeChangeError"),"Route change aborted by user during tour."}},t=e=>(e.preventDefault(),e.returnValue="The tour is currently in progress. Are you sure you want to leave?",e.returnValue);return r.events.on("routeChangeStart",e),window.addEventListener("beforeunload",t),()=>{r.events.off("routeChangeStart",e),window.removeEventListener("beforeunload",t)}},[d,p,r]);let b={startTour:g,completeTour:()=>{n.current&&n.current.complete(),h(!1)},tour:n.current};return(0,a.jsx)(t5.Provider,{value:b,children:t})}window.copyDashboardCodeBlock=function(e){let t=e.closest(".bg-gray-50").querySelector("pre").querySelector("code.block").textContent;navigator.clipboard.writeText(t).then(()=>{let t=e.innerHTML;e.innerHTML='<svg xmlns="http://www.w3.org/2000/svg" width="24" height="24" viewBox="0 0 24 24" fill="none" stroke="currentColor" stroke-width="2" stroke-linecap="round" stroke-linejoin="round" class="lucide lucide-check h-5 w-5 text-green-600"><path d="m9 12 2 2 4-4"/></svg>',setTimeout(()=>{e.innerHTML=t},2e3)})}},6378:function(e,t,n){"use strict";n.d(t,{kq:function(){return l}});let r=n(51214).ej.DEFAULT_TTL;function o(e){let t=5381;for(let n=0;n<e.length;n++)t=(t<<5)+t+e.charCodeAt(n);return t>>>0}class i{setPreloader(e){this.preloader=e}async get(e){let t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:[],n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{},o=n.ttl||r,i=!1!==n.refreshOnAccess,l=this._generateKey(e,t),a=e.name||"anonymous",s=this.cache.get(l),u=Date.now();if(s&&u-s.lastUpdated<o){let n=Math.round((u-s.lastUpdated)/1e3);if(this._debug("Cache HIT for ".concat(a," (age: ").concat(n,"s, TTL: ").concat(Math.round(o/1e3),"s)")),i&&(this.cache.set(l,{data:s.data,lastUpdated:u}),this._debug("Cache TTL refreshed for ".concat(a))),!this.backgroundJobs.has(l)){var c;(null===(c=this.preloader)||void 0===c?void 0:c.wasRecentlyPreloaded(e,t))?this._debug("Skipping background refresh for ".concat(a," - recently preloaded")):this._refreshInBackground(e,t,l)}return s.data}if(this.pendingRequests.has(l))return this._debug("Request deduplication: Waiting for pending request for ".concat(a)),this.pendingRequests.get(l);let d=(async()=>{try{let n=await e(...t);if(n&&n.__skipCache){if(this._debug("Skip caching for ".concat(a," due to __skipCache flag on result")),s)return s.data;return n}return this.cache.set(l,{data:n,lastUpdated:Date.now()}),n}catch(e){if(s)return console.warn("Failed to fetch fresh data for ".concat(l,"/").concat(a,", returning stale data:"),e),s.data;throw e}finally{this.pendingRequests.delete(l)}})();return this.pendingRequests.set(l,d),d}invalidate(e){let t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:[],n=this._generateKey(e,t);this.cache.delete(n),this.backgroundJobs.delete(n),this.pendingRequests.delete(n)}invalidateFunction(e){let t=o(e.toString()),n=[];for(let e of this.cache.keys())e.startsWith("".concat(t,"_"))&&n.push(e);n.forEach(e=>{this.cache.delete(e),this.backgroundJobs.delete(e),this.pendingRequests.delete(e)})}clear(){this.cache.clear(),this.backgroundJobs.clear(),this.pendingRequests.clear()}getStats(){return{cacheSize:this.cache.size,backgroundJobs:this.backgroundJobs.size,pendingRequests:this.pendingRequests.size,keys:Array.from(this.cache.keys())}}getDetailedStats(){let e=Date.now(),t=[];for(let[n,r]of this.cache.entries()){let o=e-r.lastUpdated;t.push({key:n,age:Math.round(o/1e3),lastUpdated:new Date(r.lastUpdated).toISOString(),hasBackgroundJob:this.backgroundJobs.has(n),hasPendingRequest:this.pendingRequests.has(n)})}return{cacheSize:this.cache.size,backgroundJobs:this.backgroundJobs.size,pendingRequests:this.pendingRequests.size,entries:t.sort((e,t)=>e.age-t.age)}}setDebugMode(e){this.debugMode=e}_debug(e){for(var t=arguments.length,n=Array(t>1?t-1:0),r=1;r<t;r++)n[r-1]=arguments[r];this.debugMode&&console.log("[DashboardCache] ".concat(e),...n)}_refreshInBackground(e,t,n){this.backgroundJobs.set(n,!0),e(...t).then(e=>{e&&e.__skipCache||this.cache.set(n,{data:e,lastUpdated:Date.now()})}).catch(e=>{console.warn("Background refresh failed for ".concat(n,":"),e)}).finally(()=>{this.backgroundJobs.delete(n)})}_generateKey(e,t){let n=o(e.toString()),r=t.length>0?JSON.stringify(t):"";return"".concat(n,"_").concat(r)}constructor(){this.cache=new Map,this.backgroundJobs=new Map,this.pendingRequests=new Map,this.debugMode=!1,this.preloader=null}}let l=new i;t.ZP=l},51214:function(e,t,n){"use strict";n.d(t,{MO:function(){return i},ej:function(){return r},nb:function(){return o}});let r={DEFAULT_TTL:12e4},o={REFRESH_INTERVAL:3e4,GPU_REFRESH_INTERVAL:3e4},i={NAME_TRUNCATE_LENGTH:20}},46990:function(e,t,n){"use strict";n.d(t,{cn:function(){return Z}});var r=n(90512);let o=e=>{let t=s(e),{conflictingClassGroups:n,conflictingClassGroupModifiers:r}=e;return{getClassGroupId:e=>{let n=e.split("-");return""===n[0]&&1!==n.length&&n.shift(),i(n,t)||a(e)},getConflictingClassGroupIds:(e,t)=>{let o=n[e]||[];return t&&r[e]?[...o,...r[e]]:o}}},i=(e,t)=>{if(0===e.length)return t.classGroupId;let n=e[0],r=t.nextPart.get(n),o=r?i(e.slice(1),r):void 0;if(o)return o;if(0===t.validators.length)return;let l=e.join("-");return t.validators.find(({validator:e})=>e(l))?.classGroupId},l=/^\[(.+)\]$/,a=e=>{if(l.test(e)){let t=l.exec(e)[1],n=t?.substring(0,t.indexOf(":"));if(n)return"arbitrary.."+n}},s=e=>{let{theme:t,prefix:n}=e,r={nextPart:new Map,validators:[]};return f(Object.entries(e.classGroups),n).forEach(([e,n])=>{u(n,r,e,t)}),r},u=(e,t,n,r)=>{e.forEach(e=>{if("string"==typeof e){(""===e?t:c(t,e)).classGroupId=n;return}if("function"==typeof e){if(d(e)){u(e(r),t,n,r);return}t.validators.push({validator:e,classGroupId:n});return}Object.entries(e).forEach(([e,o])=>{u(o,c(t,e),n,r)})})},c=(e,t)=>{let n=e;return t.split("-").forEach(e=>{n.nextPart.has(e)||n.nextPart.set(e,{nextPart:new Map,validators:[]}),n=n.nextPart.get(e)}),n},d=e=>e.isThemeGetter,f=(e,t)=>t?e.map(([e,n])=>[e,n.map(e=>"string"==typeof e?t+e:"object"==typeof e?Object.fromEntries(Object.entries(e).map(([e,n])=>[t+e,n])):e)]):e,p=e=>{if(e<1)return{get:()=>void 0,set:()=>{}};let t=0,n=new Map,r=new Map,o=(o,i)=>{n.set(o,i),++t>e&&(t=0,r=n,n=new Map)};return{get(e){let t=n.get(e);return void 0!==t?t:void 0!==(t=r.get(e))?(o(e,t),t):void 0},set(e,t){n.has(e)?n.set(e,t):o(e,t)}}},h=e=>{let{separator:t,experimentalParseClassName:n}=e,r=1===t.length,o=t[0],i=t.length,l=e=>{let n;let l=[],a=0,s=0;for(let u=0;u<e.length;u++){let c=e[u];if(0===a){if(c===o&&(r||e.slice(u,u+i)===t)){l.push(e.slice(s,u)),s=u+i;continue}if("/"===c){n=u;continue}}"["===c?a++:"]"===c&&a--}let u=0===l.length?e:e.substring(s),c=u.startsWith("!"),d=c?u.substring(1):u;return{modifiers:l,hasImportantModifier:c,baseClassName:d,maybePostfixModifierPosition:n&&n>s?n-s:void 0}};return n?e=>n({className:e,parseClassName:l}):l},m=e=>{if(e.length<=1)return e;let t=[],n=[];return e.forEach(e=>{"["===e[0]?(t.push(...n.sort(),e),n=[]):n.push(e)}),t.push(...n.sort()),t},g=e=>({cache:p(e.cacheSize),parseClassName:h(e),...o(e)}),b=/\s+/,v=(e,t)=>{let{parseClassName:n,getClassGroupId:r,getConflictingClassGroupIds:o}=t,i=[],l=e.trim().split(b),a="";for(let e=l.length-1;e>=0;e-=1){let t=l[e],{modifiers:s,hasImportantModifier:u,baseClassName:c,maybePostfixModifierPosition:d}=n(t),f=!!d,p=r(f?c.substring(0,d):c);if(!p){if(!f||!(p=r(c))){a=t+(a.length>0?" "+a:a);continue}f=!1}let h=m(s).join(":"),g=u?h+"!":h,b=g+p;if(i.includes(b))continue;i.push(b);let v=o(p,f);for(let e=0;e<v.length;++e){let t=v[e];i.push(g+t)}a=t+(a.length>0?" "+a:a)}return a};function y(){let e,t,n=0,r="";for(;n<arguments.length;)(e=arguments[n++])&&(t=w(e))&&(r&&(r+=" "),r+=t);return r}let w=e=>{let t;if("string"==typeof e)return e;let n="";for(let r=0;r<e.length;r++)e[r]&&(t=w(e[r]))&&(n&&(n+=" "),n+=t);return n},x=e=>{let t=t=>t[e]||[];return t.isThemeGetter=!0,t},E=/^\[(?:([a-z-]+):)?(.+)\]$/i,S=/^\d+\/\d+$/,k=new Set(["px","full","screen"]),T=/^(\d+(\.\d+)?)?(xs|sm|md|lg|xl)$/,C=/\d+(%|px|r?em|[sdl]?v([hwib]|min|max)|pt|pc|in|cm|mm|cap|ch|ex|r?lh|cq(w|h|i|b|min|max))|\b(calc|min|max|clamp)\(.+\)|^0$/,P=/^(rgba?|hsla?|hwb|(ok)?(lab|lch))\(.+\)$/,M=/^(inset_)?-?((\d+)?\.?(\d+)[a-z]+|0)_-?((\d+)?\.?(\d+)[a-z]+|0)/,R=/^(url|image|image-set|cross-fade|element|(repeating-)?(linear|radial|conic)-gradient)\(.+\)$/,j=e=>L(e)||k.has(e)||S.test(e),O=e=>V(e,"length",U),L=e=>!!e&&!Number.isNaN(Number(e)),A=e=>V(e,"number",L),_=e=>!!e&&Number.isInteger(Number(e)),N=e=>e.endsWith("%")&&L(e.slice(0,-1)),$=e=>E.test(e),I=e=>T.test(e),D=new Set(["length","size","percentage"]),W=e=>V(e,D,Y),H=e=>V(e,"position",Y),F=new Set(["image","url"]),z=e=>V(e,F,X),B=e=>V(e,"",G),q=()=>!0,V=(e,t,n)=>{let r=E.exec(e);return!!r&&(r[1]?"string"==typeof t?r[1]===t:t.has(r[1]):n(r[2]))},U=e=>C.test(e)&&!P.test(e),Y=()=>!1,G=e=>M.test(e),X=e=>R.test(e),K=function(e,...t){let n,r,o;let i=function(a){return r=(n=g(t.reduce((e,t)=>t(e),e()))).cache.get,o=n.cache.set,i=l,l(a)};function l(e){let t=r(e);if(t)return t;let i=v(e,n);return o(e,i),i}return function(){return i(y.apply(null,arguments))}}(()=>{let e=x("colors"),t=x("spacing"),n=x("blur"),r=x("brightness"),o=x("borderColor"),i=x("borderRadius"),l=x("borderSpacing"),a=x("borderWidth"),s=x("contrast"),u=x("grayscale"),c=x("hueRotate"),d=x("invert"),f=x("gap"),p=x("gradientColorStops"),h=x("gradientColorStopPositions"),m=x("inset"),g=x("margin"),b=x("opacity"),v=x("padding"),y=x("saturate"),w=x("scale"),E=x("sepia"),S=x("skew"),k=x("space"),T=x("translate"),C=()=>["auto","contain","none"],P=()=>["auto","hidden","clip","visible","scroll"],M=()=>["auto",$,t],R=()=>[$,t],D=()=>["",j,O],F=()=>["auto",L,$],V=()=>["bottom","center","left","left-bottom","left-top","right","right-bottom","right-top","top"],U=()=>["solid","dashed","dotted","double","none"],Y=()=>["normal","multiply","screen","overlay","darken","lighten","color-dodge","color-burn","hard-light","soft-light","difference","exclusion","hue","saturation","color","luminosity"],G=()=>["start","end","center","between","around","evenly","stretch"],X=()=>["","0",$],K=()=>["auto","avoid","all","avoid-page","page","left","right","column"],Z=()=>[L,$];return{cacheSize:500,separator:":",theme:{colors:[q],spacing:[j,O],blur:["none","",I,$],brightness:Z(),borderColor:[e],borderRadius:["none","","full",I,$],borderSpacing:R(),borderWidth:D(),contrast:Z(),grayscale:X(),hueRotate:Z(),invert:X(),gap:R(),gradientColorStops:[e],gradientColorStopPositions:[N,O],inset:M(),margin:M(),opacity:Z(),padding:R(),saturate:Z(),scale:Z(),sepia:X(),skew:Z(),space:R(),translate:R()},classGroups:{aspect:[{aspect:["auto","square","video",$]}],container:["container"],columns:[{columns:[I]}],"break-after":[{"break-after":K()}],"break-before":[{"break-before":K()}],"break-inside":[{"break-inside":["auto","avoid","avoid-page","avoid-column"]}],"box-decoration":[{"box-decoration":["slice","clone"]}],box:[{box:["border","content"]}],display:["block","inline-block","inline","flex","inline-flex","table","inline-table","table-caption","table-cell","table-column","table-column-group","table-footer-group","table-header-group","table-row-group","table-row","flow-root","grid","inline-grid","contents","list-item","hidden"],float:[{float:["right","left","none","start","end"]}],clear:[{clear:["left","right","both","none","start","end"]}],isolation:["isolate","isolation-auto"],"object-fit":[{object:["contain","cover","fill","none","scale-down"]}],"object-position":[{object:[...V(),$]}],overflow:[{overflow:P()}],"overflow-x":[{"overflow-x":P()}],"overflow-y":[{"overflow-y":P()}],overscroll:[{overscroll:C()}],"overscroll-x":[{"overscroll-x":C()}],"overscroll-y":[{"overscroll-y":C()}],position:["static","fixed","absolute","relative","sticky"],inset:[{inset:[m]}],"inset-x":[{"inset-x":[m]}],"inset-y":[{"inset-y":[m]}],start:[{start:[m]}],end:[{end:[m]}],top:[{top:[m]}],right:[{right:[m]}],bottom:[{bottom:[m]}],left:[{left:[m]}],visibility:["visible","invisible","collapse"],z:[{z:["auto",_,$]}],basis:[{basis:M()}],"flex-direction":[{flex:["row","row-reverse","col","col-reverse"]}],"flex-wrap":[{flex:["wrap","wrap-reverse","nowrap"]}],flex:[{flex:["1","auto","initial","none",$]}],grow:[{grow:X()}],shrink:[{shrink:X()}],order:[{order:["first","last","none",_,$]}],"grid-cols":[{"grid-cols":[q]}],"col-start-end":[{col:["auto",{span:["full",_,$]},$]}],"col-start":[{"col-start":F()}],"col-end":[{"col-end":F()}],"grid-rows":[{"grid-rows":[q]}],"row-start-end":[{row:["auto",{span:[_,$]},$]}],"row-start":[{"row-start":F()}],"row-end":[{"row-end":F()}],"grid-flow":[{"grid-flow":["row","col","dense","row-dense","col-dense"]}],"auto-cols":[{"auto-cols":["auto","min","max","fr",$]}],"auto-rows":[{"auto-rows":["auto","min","max","fr",$]}],gap:[{gap:[f]}],"gap-x":[{"gap-x":[f]}],"gap-y":[{"gap-y":[f]}],"justify-content":[{justify:["normal",...G()]}],"justify-items":[{"justify-items":["start","end","center","stretch"]}],"justify-self":[{"justify-self":["auto","start","end","center","stretch"]}],"align-content":[{content:["normal",...G(),"baseline"]}],"align-items":[{items:["start","end","center","baseline","stretch"]}],"align-self":[{self:["auto","start","end","center","stretch","baseline"]}],"place-content":[{"place-content":[...G(),"baseline"]}],"place-items":[{"place-items":["start","end","center","baseline","stretch"]}],"place-self":[{"place-self":["auto","start","end","center","stretch"]}],p:[{p:[v]}],px:[{px:[v]}],py:[{py:[v]}],ps:[{ps:[v]}],pe:[{pe:[v]}],pt:[{pt:[v]}],pr:[{pr:[v]}],pb:[{pb:[v]}],pl:[{pl:[v]}],m:[{m:[g]}],mx:[{mx:[g]}],my:[{my:[g]}],ms:[{ms:[g]}],me:[{me:[g]}],mt:[{mt:[g]}],mr:[{mr:[g]}],mb:[{mb:[g]}],ml:[{ml:[g]}],"space-x":[{"space-x":[k]}],"space-x-reverse":["space-x-reverse"],"space-y":[{"space-y":[k]}],"space-y-reverse":["space-y-reverse"],w:[{w:["auto","min","max","fit","svw","lvw","dvw",$,t]}],"min-w":[{"min-w":[$,t,"min","max","fit"]}],"max-w":[{"max-w":[$,t,"none","full","min","max","fit","prose",{screen:[I]},I]}],h:[{h:[$,t,"auto","min","max","fit","svh","lvh","dvh"]}],"min-h":[{"min-h":[$,t,"min","max","fit","svh","lvh","dvh"]}],"max-h":[{"max-h":[$,t,"min","max","fit","svh","lvh","dvh"]}],size:[{size:[$,t,"auto","min","max","fit"]}],"font-size":[{text:["base",I,O]}],"font-smoothing":["antialiased","subpixel-antialiased"],"font-style":["italic","not-italic"],"font-weight":[{font:["thin","extralight","light","normal","medium","semibold","bold","extrabold","black",A]}],"font-family":[{font:[q]}],"fvn-normal":["normal-nums"],"fvn-ordinal":["ordinal"],"fvn-slashed-zero":["slashed-zero"],"fvn-figure":["lining-nums","oldstyle-nums"],"fvn-spacing":["proportional-nums","tabular-nums"],"fvn-fraction":["diagonal-fractions","stacked-fractions"],tracking:[{tracking:["tighter","tight","normal","wide","wider","widest",$]}],"line-clamp":[{"line-clamp":["none",L,A]}],leading:[{leading:["none","tight","snug","normal","relaxed","loose",j,$]}],"list-image":[{"list-image":["none",$]}],"list-style-type":[{list:["none","disc","decimal",$]}],"list-style-position":[{list:["inside","outside"]}],"placeholder-color":[{placeholder:[e]}],"placeholder-opacity":[{"placeholder-opacity":[b]}],"text-alignment":[{text:["left","center","right","justify","start","end"]}],"text-color":[{text:[e]}],"text-opacity":[{"text-opacity":[b]}],"text-decoration":["underline","overline","line-through","no-underline"],"text-decoration-style":[{decoration:[...U(),"wavy"]}],"text-decoration-thickness":[{decoration:["auto","from-font",j,O]}],"underline-offset":[{"underline-offset":["auto",j,$]}],"text-decoration-color":[{decoration:[e]}],"text-transform":["uppercase","lowercase","capitalize","normal-case"],"text-overflow":["truncate","text-ellipsis","text-clip"],"text-wrap":[{text:["wrap","nowrap","balance","pretty"]}],indent:[{indent:R()}],"vertical-align":[{align:["baseline","top","middle","bottom","text-top","text-bottom","sub","super",$]}],whitespace:[{whitespace:["normal","nowrap","pre","pre-line","pre-wrap","break-spaces"]}],break:[{break:["normal","words","all","keep"]}],hyphens:[{hyphens:["none","manual","auto"]}],content:[{content:["none",$]}],"bg-attachment":[{bg:["fixed","local","scroll"]}],"bg-clip":[{"bg-clip":["border","padding","content","text"]}],"bg-opacity":[{"bg-opacity":[b]}],"bg-origin":[{"bg-origin":["border","padding","content"]}],"bg-position":[{bg:[...V(),H]}],"bg-repeat":[{bg:["no-repeat",{repeat:["","x","y","round","space"]}]}],"bg-size":[{bg:["auto","cover","contain",W]}],"bg-image":[{bg:["none",{"gradient-to":["t","tr","r","br","b","bl","l","tl"]},z]}],"bg-color":[{bg:[e]}],"gradient-from-pos":[{from:[h]}],"gradient-via-pos":[{via:[h]}],"gradient-to-pos":[{to:[h]}],"gradient-from":[{from:[p]}],"gradient-via":[{via:[p]}],"gradient-to":[{to:[p]}],rounded:[{rounded:[i]}],"rounded-s":[{"rounded-s":[i]}],"rounded-e":[{"rounded-e":[i]}],"rounded-t":[{"rounded-t":[i]}],"rounded-r":[{"rounded-r":[i]}],"rounded-b":[{"rounded-b":[i]}],"rounded-l":[{"rounded-l":[i]}],"rounded-ss":[{"rounded-ss":[i]}],"rounded-se":[{"rounded-se":[i]}],"rounded-ee":[{"rounded-ee":[i]}],"rounded-es":[{"rounded-es":[i]}],"rounded-tl":[{"rounded-tl":[i]}],"rounded-tr":[{"rounded-tr":[i]}],"rounded-br":[{"rounded-br":[i]}],"rounded-bl":[{"rounded-bl":[i]}],"border-w":[{border:[a]}],"border-w-x":[{"border-x":[a]}],"border-w-y":[{"border-y":[a]}],"border-w-s":[{"border-s":[a]}],"border-w-e":[{"border-e":[a]}],"border-w-t":[{"border-t":[a]}],"border-w-r":[{"border-r":[a]}],"border-w-b":[{"border-b":[a]}],"border-w-l":[{"border-l":[a]}],"border-opacity":[{"border-opacity":[b]}],"border-style":[{border:[...U(),"hidden"]}],"divide-x":[{"divide-x":[a]}],"divide-x-reverse":["divide-x-reverse"],"divide-y":[{"divide-y":[a]}],"divide-y-reverse":["divide-y-reverse"],"divide-opacity":[{"divide-opacity":[b]}],"divide-style":[{divide:U()}],"border-color":[{border:[o]}],"border-color-x":[{"border-x":[o]}],"border-color-y":[{"border-y":[o]}],"border-color-s":[{"border-s":[o]}],"border-color-e":[{"border-e":[o]}],"border-color-t":[{"border-t":[o]}],"border-color-r":[{"border-r":[o]}],"border-color-b":[{"border-b":[o]}],"border-color-l":[{"border-l":[o]}],"divide-color":[{divide:[o]}],"outline-style":[{outline:["",...U()]}],"outline-offset":[{"outline-offset":[j,$]}],"outline-w":[{outline:[j,O]}],"outline-color":[{outline:[e]}],"ring-w":[{ring:D()}],"ring-w-inset":["ring-inset"],"ring-color":[{ring:[e]}],"ring-opacity":[{"ring-opacity":[b]}],"ring-offset-w":[{"ring-offset":[j,O]}],"ring-offset-color":[{"ring-offset":[e]}],shadow:[{shadow:["","inner","none",I,B]}],"shadow-color":[{shadow:[q]}],opacity:[{opacity:[b]}],"mix-blend":[{"mix-blend":[...Y(),"plus-lighter","plus-darker"]}],"bg-blend":[{"bg-blend":Y()}],filter:[{filter:["","none"]}],blur:[{blur:[n]}],brightness:[{brightness:[r]}],contrast:[{contrast:[s]}],"drop-shadow":[{"drop-shadow":["","none",I,$]}],grayscale:[{grayscale:[u]}],"hue-rotate":[{"hue-rotate":[c]}],invert:[{invert:[d]}],saturate:[{saturate:[y]}],sepia:[{sepia:[E]}],"backdrop-filter":[{"backdrop-filter":["","none"]}],"backdrop-blur":[{"backdrop-blur":[n]}],"backdrop-brightness":[{"backdrop-brightness":[r]}],"backdrop-contrast":[{"backdrop-contrast":[s]}],"backdrop-grayscale":[{"backdrop-grayscale":[u]}],"backdrop-hue-rotate":[{"backdrop-hue-rotate":[c]}],"backdrop-invert":[{"backdrop-invert":[d]}],"backdrop-opacity":[{"backdrop-opacity":[b]}],"backdrop-saturate":[{"backdrop-saturate":[y]}],"backdrop-sepia":[{"backdrop-sepia":[E]}],"border-collapse":[{border:["collapse","separate"]}],"border-spacing":[{"border-spacing":[l]}],"border-spacing-x":[{"border-spacing-x":[l]}],"border-spacing-y":[{"border-spacing-y":[l]}],"table-layout":[{table:["auto","fixed"]}],caption:[{caption:["top","bottom"]}],transition:[{transition:["none","all","","colors","opacity","shadow","transform",$]}],duration:[{duration:Z()}],ease:[{ease:["linear","in","out","in-out",$]}],delay:[{delay:Z()}],animate:[{animate:["none","spin","ping","pulse","bounce",$]}],transform:[{transform:["","gpu","none"]}],scale:[{scale:[w]}],"scale-x":[{"scale-x":[w]}],"scale-y":[{"scale-y":[w]}],rotate:[{rotate:[_,$]}],"translate-x":[{"translate-x":[T]}],"translate-y":[{"translate-y":[T]}],"skew-x":[{"skew-x":[S]}],"skew-y":[{"skew-y":[S]}],"transform-origin":[{origin:["center","top","top-right","right","bottom-right","bottom","bottom-left","left","top-left",$]}],accent:[{accent:["auto",e]}],appearance:[{appearance:["none","auto"]}],cursor:[{cursor:["auto","default","pointer","wait","text","move","help","not-allowed","none","context-menu","progress","cell","crosshair","vertical-text","alias","copy","no-drop","grab","grabbing","all-scroll","col-resize","row-resize","n-resize","e-resize","s-resize","w-resize","ne-resize","nw-resize","se-resize","sw-resize","ew-resize","ns-resize","nesw-resize","nwse-resize","zoom-in","zoom-out",$]}],"caret-color":[{caret:[e]}],"pointer-events":[{"pointer-events":["none","auto"]}],resize:[{resize:["none","y","x",""]}],"scroll-behavior":[{scroll:["auto","smooth"]}],"scroll-m":[{"scroll-m":R()}],"scroll-mx":[{"scroll-mx":R()}],"scroll-my":[{"scroll-my":R()}],"scroll-ms":[{"scroll-ms":R()}],"scroll-me":[{"scroll-me":R()}],"scroll-mt":[{"scroll-mt":R()}],"scroll-mr":[{"scroll-mr":R()}],"scroll-mb":[{"scroll-mb":R()}],"scroll-ml":[{"scroll-ml":R()}],"scroll-p":[{"scroll-p":R()}],"scroll-px":[{"scroll-px":R()}],"scroll-py":[{"scroll-py":R()}],"scroll-ps":[{"scroll-ps":R()}],"scroll-pe":[{"scroll-pe":R()}],"scroll-pt":[{"scroll-pt":R()}],"scroll-pr":[{"scroll-pr":R()}],"scroll-pb":[{"scroll-pb":R()}],"scroll-pl":[{"scroll-pl":R()}],"snap-align":[{snap:["start","end","center","align-none"]}],"snap-stop":[{snap:["normal","always"]}],"snap-type":[{snap:["none","x","y","both"]}],"snap-strictness":[{snap:["mandatory","proximity"]}],touch:[{touch:["auto","none","manipulation"]}],"touch-x":[{"touch-pan":["x","left","right"]}],"touch-y":[{"touch-pan":["y","up","down"]}],"touch-pz":["touch-pinch-zoom"],select:[{select:["none","text","all","auto"]}],"will-change":[{"will-change":["auto","scroll","contents","transform",$]}],fill:[{fill:[e,"none"]}],"stroke-w":[{stroke:[j,O,A]}],stroke:[{stroke:[e,"none"]}],sr:["sr-only","not-sr-only"],"forced-color-adjust":[{"forced-color-adjust":["auto","none"]}]},conflictingClassGroups:{overflow:["overflow-x","overflow-y"],overscroll:["overscroll-x","overscroll-y"],inset:["inset-x","inset-y","start","end","top","right","bottom","left"],"inset-x":["right","left"],"inset-y":["top","bottom"],flex:["basis","grow","shrink"],gap:["gap-x","gap-y"],p:["px","py","ps","pe","pt","pr","pb","pl"],px:["pr","pl"],py:["pt","pb"],m:["mx","my","ms","me","mt","mr","mb","ml"],mx:["mr","ml"],my:["mt","mb"],size:["w","h"],"font-size":["leading"],"fvn-normal":["fvn-ordinal","fvn-slashed-zero","fvn-figure","fvn-spacing","fvn-fraction"],"fvn-ordinal":["fvn-normal"],"fvn-slashed-zero":["fvn-normal"],"fvn-figure":["fvn-normal"],"fvn-spacing":["fvn-normal"],"fvn-fraction":["fvn-normal"],"line-clamp":["display","overflow"],rounded:["rounded-s","rounded-e","rounded-t","rounded-r","rounded-b","rounded-l","rounded-ss","rounded-se","rounded-ee","rounded-es","rounded-tl","rounded-tr","rounded-br","rounded-bl"],"rounded-s":["rounded-ss","rounded-es"],"rounded-e":["rounded-se","rounded-ee"],"rounded-t":["rounded-tl","rounded-tr"],"rounded-r":["rounded-tr","rounded-br"],"rounded-b":["rounded-br","rounded-bl"],"rounded-l":["rounded-tl","rounded-bl"],"border-spacing":["border-spacing-x","border-spacing-y"],"border-w":["border-w-s","border-w-e","border-w-t","border-w-r","border-w-b","border-w-l"],"border-w-x":["border-w-r","border-w-l"],"border-w-y":["border-w-t","border-w-b"],"border-color":["border-color-s","border-color-e","border-color-t","border-color-r","border-color-b","border-color-l"],"border-color-x":["border-color-r","border-color-l"],"border-color-y":["border-color-t","border-color-b"],"scroll-m":["scroll-mx","scroll-my","scroll-ms","scroll-me","scroll-mt","scroll-mr","scroll-mb","scroll-ml"],"scroll-mx":["scroll-mr","scroll-ml"],"scroll-my":["scroll-mt","scroll-mb"],"scroll-p":["scroll-px","scroll-py","scroll-ps","scroll-pe","scroll-pt","scroll-pr","scroll-pb","scroll-pl"],"scroll-px":["scroll-pr","scroll-pl"],"scroll-py":["scroll-pt","scroll-pb"],touch:["touch-x","touch-y","touch-pz"],"touch-x":["touch"],"touch-y":["touch"],"touch-pz":["touch"]},conflictingClassGroupModifiers:{"font-size":["leading"]}}});function Z(){for(var e=arguments.length,t=Array(e),n=0;n<e;n++)t[n]=arguments[n];return K((0,r.W)(t))}},32586:function(e,t,n){"use strict";n.r(t);var r=n(85893),o=n(67294),i=n(20745),l=n(5152),a=n.n(l),s=n(45697),u=n.n(s);n(36501);var c=n(93225),d=n(94849),f=n(23800),p=n(91623);let h=a()(()=>Promise.all([n.e(835),n.e(850),n.e(353),n.e(938)]).then(n.bind(n,9353)).then(e=>e.Layout),{loadableGenerated:{webpack:()=>[9353]},ssr:!1});function m(e){let{Component:t,pageProps:n}=e;return(0,o.useEffect)(()=>{let e=document.createElement("link");e.rel="icon",e.href="".concat(c.GW,"/favicon.ico"),document.head.appendChild(e)},[]),(0,r.jsx)(f.ol,{children:(0,r.jsx)(p.pU,{children:(0,r.jsx)(d.C,{children:(0,r.jsx)(h,{highlighted:n.highlighted,children:(0,r.jsx)(t,{...n})})})})})}window.React=o,window.ReactDOM=i,m.propTypes={Component:u().elementType.isRequired,pageProps:u().object.isRequired},t.default=m},23800:function(e,t,n){"use strict";n.d(t,{d7:function(){return v},dL:function(){return x},ol:function(){return g},t6:function(){return w},x1:function(){return y}});var r=n(85893),o=n(67294),i=n(93225),l=n(47145),a=n(6378),s=n(17853);let u=(0,o.createContext)({topNavLinks:[],routes:[],components:{}}),c={topNavLinks:[],routes:[],components:{}},d={REGISTER_TOP_NAV_LINK:"REGISTER_TOP_NAV_LINK",REGISTER_ROUTE:"REGISTER_ROUTE",REGISTER_COMPONENT:"REGISTER_COMPONENT"};function f(e,t){switch(t.type){case d.REGISTER_TOP_NAV_LINK:return{...e,topNavLinks:p(e.topNavLinks,t.payload)};case d.REGISTER_ROUTE:return{...e,routes:p(e.routes,t.payload)};case d.REGISTER_COMPONENT:{let{slot:n}=t.payload,r=p(e.components[n]||[],t.payload);return r.sort((e,t)=>(e.order||100)-(t.order||100)),{...e,components:{...e.components,[n]:r}}}default:return e}}function p(e,t){let n=e.findIndex(e=>e.id===t.id);if(-1===n)return[...e,t];let r=[...e];return r[n]=t,r}let h=new Map;async function m(){try{let e=await l.x.get("/api/plugins");if(!e.ok)return console.warn("[SkyDashboardPlugin] Failed to fetch plugin manifest:",e.status,e.statusText),[];let t=await e.json();if(!t||!Array.isArray(t.plugins))return[];return console.log("Plugin manifest:",t.plugins),t.plugins}catch(e){return console.warn("[SkyDashboardPlugin] Error fetching plugin manifest:",e),[]}}function g(e){let{children:t}=e,[n,l]=(0,o.useReducer)(f,c);(0,o.useEffect)(()=>{let e=!1,t={registerTopNavLink(e){var t,n;let r=e&&e.id&&e.label&&e.href?{id:String(e.id),label:String(e.label),href:String(e.href),order:Number.isFinite(e.order)?e.order:0,group:e.group?String(e.group):null,target:"_blank"===e.target?"_blank":"_self",rel:null!==(t=e.rel)&&void 0!==t?t:"_blank"===e.target||/^https?:\/\//.test(String(e.href))?"noopener noreferrer":void 0,external:null!==(n=e.external)&&void 0!==n?n:/^(https?:)?\/\//.test(String(e.href))||"_blank"===e.target,badge:"string"==typeof e.badge?e.badge:null,icon:"string"==typeof e.icon?e.icon:null,description:"string"==typeof e.description?e.description:void 0}:(console.warn("[SkyDashboardPlugin] Invalid top nav link registration:",e),null);return r?(l({type:d.REGISTER_TOP_NAV_LINK,payload:r}),r.id):null},registerRoute(e){let t=function(e){if(!e||"object"!=typeof e||!e.id||!e.path||"function"!=typeof e.mount)return console.warn("[SkyDashboardPlugin] Invalid route registration:",e),null;let t=String(e.path),n=t.startsWith("/")?t:"/".concat(t);return{id:String(e.id),path:n,title:"string"==typeof e.title?e.title:void 0,description:"string"==typeof e.description?e.description:void 0,mount:e.mount,unmount:"function"==typeof e.unmount?e.unmount:void 0,context:e.context&&"object"==typeof e.context?e.context:void 0}}(e);return t?(l({type:d.REGISTER_ROUTE,payload:t}),t.id):null},registerComponent(e){let t=e&&"object"==typeof e&&e.id&&e.slot&&"function"==typeof e.component?{id:String(e.id),slot:String(e.slot),component:e.component,order:Number.isFinite(e.order)?e.order:100,conditions:e.conditions&&"object"==typeof e.conditions?{pages:Array.isArray(e.conditions.pages)?e.conditions.pages.map(String):void 0}:void 0}:(console.warn("[SkyDashboardPlugin] Invalid component registration:",e),null);return t?(l({type:d.REGISTER_COMPONENT,payload:t}),t.id):null},getContext:()=>({basePath:i.GW,apiEndpoint:i.f4,dashboardCache:a.ZP,grafanaUtils:{checkGrafanaAvailability:s.TO,getGrafanaUrl:s.ki}})};return window.SkyDashboardPluginAPI=t,window.dispatchEvent(new CustomEvent("skydashboard:plugins-ready",{detail:t})),(async()=>{let t=await m();e||t.map(e=>e&&"object"==typeof e&&e.js_extension_path?(console.log("Extracting JS extension path:",e.js_extension_path),e.js_extension_path):null).filter(Boolean).forEach(t=>{e||function(e){let t=function(e){if(!e||"string"!=typeof e)return null;if(/^https?:\/\//.test(e))return e;try{return new URL(e,window.location.origin).toString()}catch(t){return console.warn("[SkyDashboardPlugin] Failed to resolve plugin script path:",e,t),null}}(e);if(!t)return;if(h.has(t))return h.get(t);console.log("Loading plugin script:",t);let n=new Promise(e=>{let n=document.createElement("script");n.type="text/javascript",n.async=!0,n.src=t,n.onload=()=>e(),n.onerror=n=>{console.warn("[SkyDashboardPlugin] Failed to load plugin script:",t,n),e()},document.head.appendChild(n)});h.set(t,n)}(t)})})(),()=>{e=!0,window.SkyDashboardPluginAPI===t&&delete window.SkyDashboardPluginAPI}},[]);let p=(0,o.useMemo)(()=>n,[n]);return(0,r.jsx)(u.Provider,{value:p,children:t})}function b(){return(0,o.useContext)(u)}function v(){let{topNavLinks:e}=b();return(0,o.useMemo)(()=>{let t=[...e].sort((e,t)=>e.order-t.order);return{ungrouped:t.filter(e=>!e.group),groups:t.filter(e=>e.group).reduce((e,t)=>{let n=t.group;return e[n]||(e[n]=[]),e[n].push(t),e},{})}},[e])}function y(){let{routes:e}=b();return e}function w(e){let t=y();return(0,o.useMemo)(()=>e&&t.find(t=>t.path===e)||null,[e,t])}function x(e){let{components:t}=b();return(0,o.useMemo)(()=>{if(!e)return[];let n=t[e]||[],r=window.location.pathname;return n.filter(e=>{var t;return null===(t=e.conditions)||void 0===t||!t.pages||e.conditions.pages.some(e=>r.includes(e))})},[e,t])}},17853:function(e,t,n){"use strict";n.d(t,{TO:function(){return i},ki:function(){return l}});let r=null,o=null,i=async()=>null!==r?r:o||(o=(async()=>{try{let e="".concat(window.location.origin,"/grafana"),t=await fetch("".concat(e,"/api/health"),{method:"GET",credentials:"include",headers:{Accept:"application/json"},signal:AbortSignal.timeout(5e3)});return 200==t.status}catch(e){return console.debug("Grafana availability check failed:",e),r=!1,!1}finally{o=null}})()),l=()=>"".concat(window.location.origin,"/grafana")},36501:function(){},77663:function(e){!function(){var t={229:function(e){var t,n,r,o=e.exports={};function i(){throw Error("setTimeout has not been defined")}function l(){throw Error("clearTimeout has not been defined")}function a(e){if(t===setTimeout)return setTimeout(e,0);if((t===i||!t)&&setTimeout)return t=setTimeout,setTimeout(e,0);try{return t(e,0)}catch(n){try{return t.call(null,e,0)}catch(n){return t.call(this,e,0)}}}!function(){try{t="function"==typeof setTimeout?setTimeout:i}catch(e){t=i}try{n="function"==typeof clearTimeout?clearTimeout:l}catch(e){n=l}}();var s=[],u=!1,c=-1;function d(){u&&r&&(u=!1,r.length?s=r.concat(s):c=-1,s.length&&f())}function f(){if(!u){var e=a(d);u=!0;for(var t=s.length;t;){for(r=s,s=[];++c<t;)r&&r[c].run();c=-1,t=s.length}r=null,u=!1,function(e){if(n===clearTimeout)return clearTimeout(e);if((n===l||!n)&&clearTimeout)return n=clearTimeout,clearTimeout(e);try{n(e)}catch(t){try{return n.call(null,e)}catch(t){return n.call(this,e)}}}(e)}}function p(e,t){this.fun=e,this.array=t}function h(){}o.nextTick=function(e){var t=Array(arguments.length-1);if(arguments.length>1)for(var n=1;n<arguments.length;n++)t[n-1]=arguments[n];s.push(new p(e,t)),1!==s.length||u||a(f)},p.prototype.run=function(){this.fun.apply(null,this.array)},o.title="browser",o.browser=!0,o.env={},o.argv=[],o.version="",o.versions={},o.on=h,o.addListener=h,o.once=h,o.off=h,o.removeListener=h,o.removeAllListeners=h,o.emit=h,o.prependListener=h,o.prependOnceListener=h,o.listeners=function(e){return[]},o.binding=function(e){throw Error("process.binding is not supported")},o.cwd=function(){return"/"},o.chdir=function(e){throw Error("process.chdir is not supported")},o.umask=function(){return 0}}},n={};function r(e){var o=n[e];if(void 0!==o)return o.exports;var i=n[e]={exports:{}},l=!0;try{t[e](i,i.exports,r),l=!1}finally{l&&delete n[e]}return i.exports}r.ab="//";var o=r(229);e.exports=o}()},5152:function(e,t,n){e.exports=n(19606)},41664:function(e,t,n){e.exports=n(24116)},11163:function(e,t,n){e.exports=n(26036)},92703:function(e,t,n){"use strict";var r=n(50414);function o(){}function i(){}i.resetWarningCache=o,e.exports=function(){function e(e,t,n,o,i,l){if(l!==r){var a=Error("Calling PropTypes validators directly is not supported by the `prop-types` package. Use PropTypes.checkPropTypes() to call them. Read more at http://fb.me/use-check-prop-types");throw a.name="Invariant Violation",a}}function t(){return e}e.isRequired=e;var n={array:e,bigint:e,bool:e,func:e,number:e,object:e,string:e,symbol:e,any:e,arrayOf:t,element:e,elementType:e,instanceOf:t,node:e,objectOf:t,oneOf:t,oneOfType:t,shape:t,exact:t,checkPropTypes:i,resetWarningCache:o};return n.PropTypes=n,n}},45697:function(e,t,n){e.exports=n(92703)()},50414:function(e){"use strict";e.exports="SECRET_DO_NOT_PASS_THIS_OR_YOU_WILL_BE_FIRED"},6223:function(e,t,n){"use strict";n.d(t,{Z:function(){return K}});var r,o,i,l,a,s,u,c=function(){return(c=Object.assign||function(e){for(var t,n=1,r=arguments.length;n<r;n++)for(var o in t=arguments[n])Object.prototype.hasOwnProperty.call(t,o)&&(e[o]=t[o]);return e}).apply(this,arguments)};function d(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&0>t.indexOf(r)&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols)for(var o=0,r=Object.getOwnPropertySymbols(e);o<r.length;o++)0>t.indexOf(r[o])&&Object.prototype.propertyIsEnumerable.call(e,r[o])&&(n[r[o]]=e[r[o]]);return n}"function"==typeof SuppressedError&&SuppressedError;var f=n(67294),p="right-scroll-bar-position",h="width-before-scroll-bar";function m(e,t){return"function"==typeof e?e(t):e&&(e.current=t),e}var g="undefined"!=typeof window?f.useLayoutEffect:f.useEffect,b=new WeakMap,v=(void 0===o&&(o={}),(void 0===i&&(i=function(e){return e}),l=[],a=!1,s={read:function(){if(a)throw Error("Sidecar: could not `read` from an `assigned` medium. `read` could be used only with `useMedium`.");return l.length?l[l.length-1]:null},useMedium:function(e){var t=i(e,a);return l.push(t),function(){l=l.filter(function(e){return e!==t})}},assignSyncMedium:function(e){for(a=!0;l.length;){var t=l;l=[],t.forEach(e)}l={push:function(t){return e(t)},filter:function(){return l}}},assignMedium:function(e){a=!0;var t=[];if(l.length){var n=l;l=[],n.forEach(e),t=l}var r=function(){var n=t;t=[],n.forEach(e)},o=function(){return Promise.resolve().then(r)};o(),l={push:function(e){t.push(e),o()},filter:function(e){return t=t.filter(e),l}}}}).options=c({async:!0,ssr:!1},o),s),y=function(){},w=f.forwardRef(function(e,t){var n,r,o,i,l=f.useRef(null),a=f.useState({onScrollCapture:y,onWheelCapture:y,onTouchMoveCapture:y}),s=a[0],u=a[1],p=e.forwardProps,h=e.children,w=e.className,x=e.removeScrollBar,E=e.enabled,S=e.shards,k=e.sideCar,T=e.noRelative,C=e.noIsolation,P=e.inert,M=e.allowPinchZoom,R=e.as,j=e.gapMode,O=d(e,["forwardProps","children","className","removeScrollBar","enabled","shards","sideCar","noRelative","noIsolation","inert","allowPinchZoom","as","gapMode"]),L=(n=[l,t],r=function(e){return n.forEach(function(t){return m(t,e)})},(o=(0,f.useState)(function(){return{value:null,callback:r,facade:{get current(){return o.value},set current(value){var e=o.value;e!==value&&(o.value=value,o.callback(value,e))}}}})[0]).callback=r,i=o.facade,g(function(){var e=b.get(i);if(e){var t=new Set(e),r=new Set(n),o=i.current;t.forEach(function(e){r.has(e)||m(e,null)}),r.forEach(function(e){t.has(e)||m(e,o)})}b.set(i,n)},[n]),i),A=c(c({},O),s);return f.createElement(f.Fragment,null,E&&f.createElement(k,{sideCar:v,removeScrollBar:x,shards:S,noRelative:T,noIsolation:C,inert:P,setCallbacks:u,allowPinchZoom:!!M,lockRef:l,gapMode:j}),p?f.cloneElement(f.Children.only(h),c(c({},A),{ref:L})):f.createElement(void 0===R?"div":R,c({},A,{className:w,ref:L}),h))});w.defaultProps={enabled:!0,removeScrollBar:!0,inert:!1},w.classNames={fullWidth:h,zeroRight:p};var x=function(e){var t=e.sideCar,n=d(e,["sideCar"]);if(!t)throw Error("Sidecar: please provide `sideCar` property to import the right car");var r=t.read();if(!r)throw Error("Sidecar medium not found");return f.createElement(r,c({},n))};x.isSideCarExport=!0;var E=function(){var e=0,t=null;return{add:function(o){if(0==e&&(t=function(){if(!document)return null;var e=document.createElement("style");e.type="text/css";var t=r||n.nc;return t&&e.setAttribute("nonce",t),e}())){var i,l;(i=t).styleSheet?i.styleSheet.cssText=o:i.appendChild(document.createTextNode(o)),l=t,(document.head||document.getElementsByTagName("head")[0]).appendChild(l)}e++},remove:function(){--e||!t||(t.parentNode&&t.parentNode.removeChild(t),t=null)}}},S=function(){var e=E();return function(t,n){f.useEffect(function(){return e.add(t),function(){e.remove()}},[t&&n])}},k=function(){var e=S();return function(t){return e(t.styles,t.dynamic),null}},T={left:0,top:0,right:0,gap:0},C=function(e){return parseInt(e||"",10)||0},P=function(e){var t=window.getComputedStyle(document.body),n=t["padding"===e?"paddingLeft":"marginLeft"],r=t["padding"===e?"paddingTop":"marginTop"],o=t["padding"===e?"paddingRight":"marginRight"];return[C(n),C(r),C(o)]},M=function(e){if(void 0===e&&(e="margin"),"undefined"==typeof window)return T;var t=P(e),n=document.documentElement.clientWidth,r=window.innerWidth;return{left:t[0],top:t[1],right:t[2],gap:Math.max(0,r-n+t[2]-t[0])}},R=k(),j="data-scroll-locked",O=function(e,t,n,r){var o=e.left,i=e.top,l=e.right,a=e.gap;return void 0===n&&(n="margin"),"\n  .".concat("with-scroll-bars-hidden"," {\n   overflow: hidden ").concat(r,";\n   padding-right: ").concat(a,"px ").concat(r,";\n  }\n  body[").concat(j,"] {\n    overflow: hidden ").concat(r,";\n    overscroll-behavior: contain;\n    ").concat([t&&"position: relative ".concat(r,";"),"margin"===n&&"\n    padding-left: ".concat(o,"px;\n    padding-top: ").concat(i,"px;\n    padding-right: ").concat(l,"px;\n    margin-left:0;\n    margin-top:0;\n    margin-right: ").concat(a,"px ").concat(r,";\n    "),"padding"===n&&"padding-right: ".concat(a,"px ").concat(r,";")].filter(Boolean).join(""),"\n  }\n  \n  .").concat(p," {\n    right: ").concat(a,"px ").concat(r,";\n  }\n  \n  .").concat(h," {\n    margin-right: ").concat(a,"px ").concat(r,";\n  }\n  \n  .").concat(p," .").concat(p," {\n    right: 0 ").concat(r,";\n  }\n  \n  .").concat(h," .").concat(h," {\n    margin-right: 0 ").concat(r,";\n  }\n  \n  body[").concat(j,"] {\n    ").concat("--removed-body-scroll-bar-size",": ").concat(a,"px;\n  }\n")},L=function(){var e=parseInt(document.body.getAttribute(j)||"0",10);return isFinite(e)?e:0},A=function(){f.useEffect(function(){return document.body.setAttribute(j,(L()+1).toString()),function(){var e=L()-1;e<=0?document.body.removeAttribute(j):document.body.setAttribute(j,e.toString())}},[])},_=function(e){var t=e.noRelative,n=e.noImportant,r=e.gapMode,o=void 0===r?"margin":r;A();var i=f.useMemo(function(){return M(o)},[o]);return f.createElement(R,{styles:O(i,!t,o,n?"":"!important")})},N=!1;if("undefined"!=typeof window)try{var $=Object.defineProperty({},"passive",{get:function(){return N=!0,!0}});window.addEventListener("test",$,$),window.removeEventListener("test",$,$)}catch(e){N=!1}var I=!!N&&{passive:!1},D=function(e,t){if(!(e instanceof Element))return!1;var n=window.getComputedStyle(e);return"hidden"!==n[t]&&!(n.overflowY===n.overflowX&&"TEXTAREA"!==e.tagName&&"visible"===n[t])},W=function(e,t){var n=t.ownerDocument,r=t;do{if("undefined"!=typeof ShadowRoot&&r instanceof ShadowRoot&&(r=r.host),H(e,r)){var o=F(e,r);if(o[1]>o[2])return!0}r=r.parentNode}while(r&&r!==n.body);return!1},H=function(e,t){return"v"===e?D(t,"overflowY"):D(t,"overflowX")},F=function(e,t){return"v"===e?[t.scrollTop,t.scrollHeight,t.clientHeight]:[t.scrollLeft,t.scrollWidth,t.clientWidth]},z=function(e,t,n,r,o){var i,l=(i=window.getComputedStyle(t).direction,"h"===e&&"rtl"===i?-1:1),a=l*r,s=n.target,u=t.contains(s),c=!1,d=a>0,f=0,p=0;do{if(!s)break;var h=F(e,s),m=h[0],g=h[1]-h[2]-l*m;(m||g)&&H(e,s)&&(f+=g,p+=m);var b=s.parentNode;s=b&&b.nodeType===Node.DOCUMENT_FRAGMENT_NODE?b.host:b}while(!u&&s!==document.body||u&&(t.contains(s)||t===s));return d&&(o&&1>Math.abs(f)||!o&&a>f)?c=!0:!d&&(o&&1>Math.abs(p)||!o&&-a>p)&&(c=!0),c},B=function(e){return"changedTouches"in e?[e.changedTouches[0].clientX,e.changedTouches[0].clientY]:[0,0]},q=function(e){return[e.deltaX,e.deltaY]},V=function(e){return e&&"current"in e?e.current:e},U=0,Y=[],G=(u=function(e){var t=f.useRef([]),n=f.useRef([0,0]),r=f.useRef(),o=f.useState(U++)[0],i=f.useState(k)[0],l=f.useRef(e);f.useEffect(function(){l.current=e},[e]),f.useEffect(function(){if(e.inert){document.body.classList.add("block-interactivity-".concat(o));var t=(function(e,t,n){if(n||2==arguments.length)for(var r,o=0,i=t.length;o<i;o++)!r&&o in t||(r||(r=Array.prototype.slice.call(t,0,o)),r[o]=t[o]);return e.concat(r||Array.prototype.slice.call(t))})([e.lockRef.current],(e.shards||[]).map(V),!0).filter(Boolean);return t.forEach(function(e){return e.classList.add("allow-interactivity-".concat(o))}),function(){document.body.classList.remove("block-interactivity-".concat(o)),t.forEach(function(e){return e.classList.remove("allow-interactivity-".concat(o))})}}},[e.inert,e.lockRef.current,e.shards]);var a=f.useCallback(function(e,t){if("touches"in e&&2===e.touches.length||"wheel"===e.type&&e.ctrlKey)return!l.current.allowPinchZoom;var o,i=B(e),a=n.current,s="deltaX"in e?e.deltaX:a[0]-i[0],u="deltaY"in e?e.deltaY:a[1]-i[1],c=e.target,d=Math.abs(s)>Math.abs(u)?"h":"v";if("touches"in e&&"h"===d&&"range"===c.type)return!1;var f=W(d,c);if(!f)return!0;if(f?o=d:(o="v"===d?"h":"v",f=W(d,c)),!f)return!1;if(!r.current&&"changedTouches"in e&&(s||u)&&(r.current=o),!o)return!0;var p=r.current||o;return z(p,t,e,"h"===p?s:u,!0)},[]),s=f.useCallback(function(e){if(Y.length&&Y[Y.length-1]===i){var n="deltaY"in e?q(e):B(e),r=t.current.filter(function(t){var r;return t.name===e.type&&(t.target===e.target||e.target===t.shadowParent)&&(r=t.delta)[0]===n[0]&&r[1]===n[1]})[0];if(r&&r.should){e.cancelable&&e.preventDefault();return}if(!r){var o=(l.current.shards||[]).map(V).filter(Boolean).filter(function(t){return t.contains(e.target)});(o.length>0?a(e,o[0]):!l.current.noIsolation)&&e.cancelable&&e.preventDefault()}}},[]),u=f.useCallback(function(e,n,r,o){var i={name:e,delta:n,target:r,should:o,shadowParent:function(e){for(var t=null;null!==e;)e instanceof ShadowRoot&&(t=e.host,e=e.host),e=e.parentNode;return t}(r)};t.current.push(i),setTimeout(function(){t.current=t.current.filter(function(e){return e!==i})},1)},[]),c=f.useCallback(function(e){n.current=B(e),r.current=void 0},[]),d=f.useCallback(function(t){u(t.type,q(t),t.target,a(t,e.lockRef.current))},[]),p=f.useCallback(function(t){u(t.type,B(t),t.target,a(t,e.lockRef.current))},[]);f.useEffect(function(){return Y.push(i),e.setCallbacks({onScrollCapture:d,onWheelCapture:d,onTouchMoveCapture:p}),document.addEventListener("wheel",s,I),document.addEventListener("touchmove",s,I),document.addEventListener("touchstart",c,I),function(){Y=Y.filter(function(e){return e!==i}),document.removeEventListener("wheel",s,I),document.removeEventListener("touchmove",s,I),document.removeEventListener("touchstart",c,I)}},[]);var h=e.removeScrollBar,m=e.inert;return f.createElement(f.Fragment,null,m?f.createElement(i,{styles:"\n  .block-interactivity-".concat(o," {pointer-events: none;}\n  .allow-interactivity-").concat(o," {pointer-events: all;}\n")}):null,h?f.createElement(_,{noRelative:e.noRelative,gapMode:e.gapMode}):null)},v.useMedium(u),x),X=f.forwardRef(function(e,t){return f.createElement(w,c({},e,{ref:t,sideCar:G}))});X.classNames=w.classNames;var K=X},36206:function(e,t,n){"use strict";function r(e,t,{checkForDefaultPrevented:n=!0}={}){return function(r){if(e?.(r),!1===n||!r.defaultPrevented)return t?.(r)}}n.d(t,{M:function(){return r}})},28771:function(e,t,n){"use strict";n.d(t,{F:function(){return i},e:function(){return l}});var r=n(67294);function o(e,t){if("function"==typeof e)return e(t);null!=e&&(e.current=t)}function i(...e){return t=>{let n=!1,r=e.map(e=>{let r=o(e,t);return n||"function"!=typeof r||(n=!0),r});if(n)return()=>{for(let t=0;t<r.length;t++){let n=r[t];"function"==typeof n?n():o(e[t],null)}}}}function l(...e){return r.useCallback(i(...e),e)}},25360:function(e,t,n){"use strict";n.d(t,{b:function(){return l},k:function(){return i}});var r=n(67294),o=n(85893);function i(e,t){let n=r.createContext(t),i=e=>{let{children:t,...i}=e,l=r.useMemo(()=>i,Object.values(i));return(0,o.jsx)(n.Provider,{value:l,children:t})};return i.displayName=e+"Provider",[i,function(o){let i=r.useContext(n);if(i)return i;if(void 0!==t)return t;throw Error(`\`${o}\` must be used within \`${e}\``)}]}function l(e,t=[]){let n=[],i=()=>{let t=n.map(e=>r.createContext(e));return function(n){let o=n?.[e]||t;return r.useMemo(()=>({[`__scope${e}`]:{...n,[e]:o}}),[n,o])}};return i.scopeName=e,[function(t,i){let l=r.createContext(i),a=n.length;n=[...n,i];let s=t=>{let{scope:n,children:i,...s}=t,u=n?.[e]?.[a]||l,c=r.useMemo(()=>s,Object.values(s));return(0,o.jsx)(u.Provider,{value:c,children:i})};return s.displayName=t+"Provider",[s,function(n,o){let s=o?.[e]?.[a]||l,u=r.useContext(s);if(u)return u;if(void 0!==i)return i;throw Error(`\`${n}\` must be used within \`${t}\``)}]},function(...e){let t=e[0];if(1===e.length)return t;let n=()=>{let n=e.map(e=>({useScope:e(),scopeName:e.scopeName}));return function(e){let o=n.reduce((t,{useScope:n,scopeName:r})=>{let o=n(e)[`__scope${r}`];return{...t,...o}},{});return r.useMemo(()=>({[`__scope${t.scopeName}`]:o}),[o])}};return n.scopeName=t.scopeName,n}(i,...t)]}},46063:function(e,t,n){"use strict";n.d(t,{XB:function(){return f}});var r,o=n(67294),i=n(36206),l=n(75320),a=n(28771),s=n(79698),u=n(85893),c="dismissableLayer.update",d=o.createContext({layers:new Set,layersWithOutsidePointerEventsDisabled:new Set,branches:new Set}),f=o.forwardRef((e,t)=>{let{disableOutsidePointerEvents:n=!1,onEscapeKeyDown:f,onPointerDownOutside:m,onFocusOutside:g,onInteractOutside:b,onDismiss:v,...y}=e,w=o.useContext(d),[x,E]=o.useState(null),S=x?.ownerDocument??globalThis?.document,[,k]=o.useState({}),T=(0,a.e)(t,e=>E(e)),C=Array.from(w.layers),[P]=[...w.layersWithOutsidePointerEventsDisabled].slice(-1),M=C.indexOf(P),R=x?C.indexOf(x):-1,j=w.layersWithOutsidePointerEventsDisabled.size>0,O=R>=M,L=function(e,t=globalThis?.document){let n=(0,s.W)(e),r=o.useRef(!1),i=o.useRef(()=>{});return o.useEffect(()=>{let e=e=>{if(e.target&&!r.current){let r=function(){h("dismissableLayer.pointerDownOutside",n,o,{discrete:!0})},o={originalEvent:e};"touch"===e.pointerType?(t.removeEventListener("click",i.current),i.current=r,t.addEventListener("click",i.current,{once:!0})):r()}else t.removeEventListener("click",i.current);r.current=!1},o=window.setTimeout(()=>{t.addEventListener("pointerdown",e)},0);return()=>{window.clearTimeout(o),t.removeEventListener("pointerdown",e),t.removeEventListener("click",i.current)}},[t,n]),{onPointerDownCapture:()=>r.current=!0}}(e=>{let t=e.target,n=[...w.branches].some(e=>e.contains(t));!O||n||(m?.(e),b?.(e),e.defaultPrevented||v?.())},S),A=function(e,t=globalThis?.document){let n=(0,s.W)(e),r=o.useRef(!1);return o.useEffect(()=>{let e=e=>{e.target&&!r.current&&h("dismissableLayer.focusOutside",n,{originalEvent:e},{discrete:!1})};return t.addEventListener("focusin",e),()=>t.removeEventListener("focusin",e)},[t,n]),{onFocusCapture:()=>r.current=!0,onBlurCapture:()=>r.current=!1}}(e=>{let t=e.target;[...w.branches].some(e=>e.contains(t))||(g?.(e),b?.(e),e.defaultPrevented||v?.())},S);return!function(e,t=globalThis?.document){let n=(0,s.W)(e);o.useEffect(()=>{let e=e=>{"Escape"===e.key&&n(e)};return t.addEventListener("keydown",e,{capture:!0}),()=>t.removeEventListener("keydown",e,{capture:!0})},[n,t])}(e=>{R!==w.layers.size-1||(f?.(e),!e.defaultPrevented&&v&&(e.preventDefault(),v()))},S),o.useEffect(()=>{if(x)return n&&(0===w.layersWithOutsidePointerEventsDisabled.size&&(r=S.body.style.pointerEvents,S.body.style.pointerEvents="none"),w.layersWithOutsidePointerEventsDisabled.add(x)),w.layers.add(x),p(),()=>{n&&1===w.layersWithOutsidePointerEventsDisabled.size&&(S.body.style.pointerEvents=r)}},[x,S,n,w]),o.useEffect(()=>()=>{x&&(w.layers.delete(x),w.layersWithOutsidePointerEventsDisabled.delete(x),p())},[x,w]),o.useEffect(()=>{let e=()=>k({});return document.addEventListener(c,e),()=>document.removeEventListener(c,e)},[]),(0,u.jsx)(l.WV.div,{...y,ref:T,style:{pointerEvents:j?O?"auto":"none":void 0,...e.style},onFocusCapture:(0,i.M)(e.onFocusCapture,A.onFocusCapture),onBlurCapture:(0,i.M)(e.onBlurCapture,A.onBlurCapture),onPointerDownCapture:(0,i.M)(e.onPointerDownCapture,L.onPointerDownCapture)})});function p(){let e=new CustomEvent(c);document.dispatchEvent(e)}function h(e,t,n,{discrete:r}){let o=n.originalEvent.target,i=new CustomEvent(e,{bubbles:!1,cancelable:!0,detail:n});t&&o.addEventListener(e,t,{once:!0}),r?(0,l.jH)(o,i):o.dispatchEvent(i)}f.displayName="DismissableLayer",o.forwardRef((e,t)=>{let n=o.useContext(d),r=o.useRef(null),i=(0,a.e)(t,r);return o.useEffect(()=>{let e=r.current;if(e)return n.branches.add(e),()=>{n.branches.delete(e)}},[n.branches]),(0,u.jsx)(l.WV.div,{...e,ref:i})}).displayName="DismissableLayerBranch"},27552:function(e,t,n){"use strict";n.d(t,{EW:function(){return i}});var r=n(67294),o=0;function i(){r.useEffect(()=>{let e=document.querySelectorAll("[data-radix-focus-guard]");return document.body.insertAdjacentElement("afterbegin",e[0]??l()),document.body.insertAdjacentElement("beforeend",e[1]??l()),o++,()=>{1===o&&document.querySelectorAll("[data-radix-focus-guard]").forEach(e=>e.remove()),o--}},[])}function l(){let e=document.createElement("span");return e.setAttribute("data-radix-focus-guard",""),e.tabIndex=0,e.style.outline="none",e.style.opacity="0",e.style.position="fixed",e.style.pointerEvents="none",e}},95420:function(e,t,n){"use strict";let r;n.d(t,{M:function(){return f}});var o=n(67294),i=n(28771),l=n(75320),a=n(79698),s=n(85893),u="focusScope.autoFocusOnMount",c="focusScope.autoFocusOnUnmount",d={bubbles:!1,cancelable:!0},f=o.forwardRef((e,t)=>{let{loop:n=!1,trapped:r=!1,onMountAutoFocus:f,onUnmountAutoFocus:b,...v}=e,[y,w]=o.useState(null),x=(0,a.W)(f),E=(0,a.W)(b),S=o.useRef(null),k=(0,i.e)(t,e=>w(e)),T=o.useRef({paused:!1,pause(){this.paused=!0},resume(){this.paused=!1}}).current;o.useEffect(()=>{if(r){let e=function(e){if(T.paused||!y)return;let t=e.target;y.contains(t)?S.current=t:m(S.current,{select:!0})},t=function(e){if(T.paused||!y)return;let t=e.relatedTarget;null===t||y.contains(t)||m(S.current,{select:!0})};document.addEventListener("focusin",e),document.addEventListener("focusout",t);let n=new MutationObserver(function(e){if(document.activeElement===document.body)for(let t of e)t.removedNodes.length>0&&m(y)});return y&&n.observe(y,{childList:!0,subtree:!0}),()=>{document.removeEventListener("focusin",e),document.removeEventListener("focusout",t),n.disconnect()}}},[r,y,T.paused]),o.useEffect(()=>{if(y){g.add(T);let e=document.activeElement;if(!y.contains(e)){let t=new CustomEvent(u,d);y.addEventListener(u,x),y.dispatchEvent(t),t.defaultPrevented||(function(e,{select:t=!1}={}){let n=document.activeElement;for(let r of e)if(m(r,{select:t}),document.activeElement!==n)return}(p(y).filter(e=>"A"!==e.tagName),{select:!0}),document.activeElement===e&&m(y))}return()=>{y.removeEventListener(u,x),setTimeout(()=>{let t=new CustomEvent(c,d);y.addEventListener(c,E),y.dispatchEvent(t),t.defaultPrevented||m(e??document.body,{select:!0}),y.removeEventListener(c,E),g.remove(T)},0)}}},[y,x,E,T]);let C=o.useCallback(e=>{if(!n&&!r||T.paused)return;let t="Tab"===e.key&&!e.altKey&&!e.ctrlKey&&!e.metaKey,o=document.activeElement;if(t&&o){let t=e.currentTarget,[r,i]=function(e){let t=p(e);return[h(t,e),h(t.reverse(),e)]}(t);r&&i?e.shiftKey||o!==i?e.shiftKey&&o===r&&(e.preventDefault(),n&&m(i,{select:!0})):(e.preventDefault(),n&&m(r,{select:!0})):o===t&&e.preventDefault()}},[n,r,T.paused]);return(0,s.jsx)(l.WV.div,{tabIndex:-1,...v,ref:k,onKeyDown:C})});function p(e){let t=[],n=document.createTreeWalker(e,NodeFilter.SHOW_ELEMENT,{acceptNode:e=>{let t="INPUT"===e.tagName&&"hidden"===e.type;return e.disabled||e.hidden||t?NodeFilter.FILTER_SKIP:e.tabIndex>=0?NodeFilter.FILTER_ACCEPT:NodeFilter.FILTER_SKIP}});for(;n.nextNode();)t.push(n.currentNode);return t}function h(e,t){for(let n of e)if(!function(e,{upTo:t}){if("hidden"===getComputedStyle(e).visibility)return!0;for(;e&&(void 0===t||e!==t);){if("none"===getComputedStyle(e).display)return!0;e=e.parentElement}return!1}(n,{upTo:t}))return n}function m(e,{select:t=!1}={}){if(e&&e.focus){var n;let r=document.activeElement;e.focus({preventScroll:!0}),e!==r&&(n=e)instanceof HTMLInputElement&&"select"in n&&t&&e.select()}}f.displayName="FocusScope";var g=(r=[],{add(e){let t=r[0];e!==t&&t?.pause(),(r=b(r,e)).unshift(e)},remove(e){r=b(r,e),r[0]?.resume()}});function b(e,t){let n=[...e],r=n.indexOf(t);return -1!==r&&n.splice(r,1),n}},91276:function(e,t,n){"use strict";n.d(t,{M:function(){return s}});var r,o=n(67294),i=n(9981),l=(r||(r=n.t(o,2)))[" useId ".trim().toString()]||(()=>void 0),a=0;function s(e){let[t,n]=o.useState(l());return(0,i.b)(()=>{e||n(e=>e??String(a++))},[e]),e||(t?`radix-${t}`:"")}},42651:function(e,t,n){"use strict";n.d(t,{h:function(){return s}});var r=n(67294),o=n(73935),i=n(75320),l=n(9981),a=n(85893),s=r.forwardRef((e,t)=>{let{container:n,...s}=e,[u,c]=r.useState(!1);(0,l.b)(()=>c(!0),[]);let d=n||u&&globalThis?.document?.body;return d?o.createPortal((0,a.jsx)(i.WV.div,{...s,ref:t}),d):null});s.displayName="Portal"},75320:function(e,t,n){"use strict";n.d(t,{WV:function(){return a},jH:function(){return s}});var r=n(67294),o=n(73935),i=n(88426),l=n(85893),a=["a","button","div","form","h2","h3","img","input","label","li","nav","ol","p","select","span","svg","ul"].reduce((e,t)=>{let n=(0,i.Z8)(`Primitive.${t}`),o=r.forwardRef((e,r)=>{let{asChild:o,...i}=e,a=o?n:t;return"undefined"!=typeof window&&(window[Symbol.for("radix-ui")]=!0),(0,l.jsx)(a,{...i,ref:r})});return o.displayName=`Primitive.${t}`,{...e,[t]:o}},{});function s(e,t){e&&o.flushSync(()=>e.dispatchEvent(t))}},88426:function(e,t,n){"use strict";n.d(t,{Z8:function(){return l},g7:function(){return a}});var r=n(67294),o=n(28771),i=n(85893);function l(e){let t=function(e){let t=r.forwardRef((e,t)=>{let{children:n,...i}=e;if(r.isValidElement(n)){let e,l;let a=(e=Object.getOwnPropertyDescriptor(n.props,"ref")?.get)&&"isReactWarning"in e&&e.isReactWarning?n.ref:(e=Object.getOwnPropertyDescriptor(n,"ref")?.get)&&"isReactWarning"in e&&e.isReactWarning?n.props.ref:n.props.ref||n.ref,s=function(e,t){let n={...t};for(let r in t){let o=e[r],i=t[r];/^on[A-Z]/.test(r)?o&&i?n[r]=(...e)=>{let t=i(...e);return o(...e),t}:o&&(n[r]=o):"style"===r?n[r]={...o,...i}:"className"===r&&(n[r]=[o,i].filter(Boolean).join(" "))}return{...e,...n}}(i,n.props);return n.type!==r.Fragment&&(s.ref=t?(0,o.F)(t,a):a),r.cloneElement(n,s)}return r.Children.count(n)>1?r.Children.only(null):null});return t.displayName=`${e}.SlotClone`,t}(e),n=r.forwardRef((e,n)=>{let{children:o,...l}=e,a=r.Children.toArray(o),s=a.find(u);if(s){let e=s.props.children,o=a.map(t=>t!==s?t:r.Children.count(e)>1?r.Children.only(null):r.isValidElement(e)?e.props.children:null);return(0,i.jsx)(t,{...l,ref:n,children:r.isValidElement(e)?r.cloneElement(e,void 0,o):null})}return(0,i.jsx)(t,{...l,ref:n,children:o})});return n.displayName=`${e}.Slot`,n}var a=l("Slot"),s=Symbol("radix.slottable");function u(e){return r.isValidElement(e)&&"function"==typeof e.type&&"__radixId"in e.type&&e.type.__radixId===s}},79698:function(e,t,n){"use strict";n.d(t,{W:function(){return o}});var r=n(67294);function o(e){let t=r.useRef(e);return r.useEffect(()=>{t.current=e}),r.useMemo(()=>(...e)=>t.current?.(...e),[])}},77342:function(e,t,n){"use strict";n.d(t,{T:function(){return a}});var r,o=n(67294),i=n(9981),l=(r||(r=n.t(o,2)))[" useInsertionEffect ".trim().toString()]||i.b;function a({prop:e,defaultProp:t,onChange:n=()=>{},caller:r}){let[i,a,s]=function({defaultProp:e,onChange:t}){let[n,r]=o.useState(e),i=o.useRef(n),a=o.useRef(t);return l(()=>{a.current=t},[t]),o.useEffect(()=>{i.current!==n&&(a.current?.(n),i.current=n)},[n,i]),[n,r,a]}({defaultProp:t,onChange:n}),u=void 0!==e,c=u?e:i;{let t=o.useRef(void 0!==e);o.useEffect(()=>{let e=t.current;if(e!==u){let t=u?"controlled":"uncontrolled";console.warn(`${r} is changing from ${e?"controlled":"uncontrolled"} to ${t}. Components should not switch from controlled to uncontrolled (or vice versa). Decide between using a controlled or uncontrolled value for the lifetime of the component.`)}t.current=u},[u,r])}return[c,o.useCallback(t=>{if(u){let n="function"==typeof t?t(e):t;n!==e&&s.current?.(n)}else a(t)},[u,e,a,s])]}Symbol("RADIX:SYNC_STATE")},9981:function(e,t,n){"use strict";n.d(t,{b:function(){return o}});var r=n(67294),o=globalThis?.document?r.useLayoutEffect:()=>{}},90512:function(e,t,n){"use strict";function r(){for(var e,t,n=0,r="",o=arguments.length;n<o;n++)(e=arguments[n])&&(t=function e(t){var n,r,o="";if("string"==typeof t||"number"==typeof t)o+=t;else if("object"==typeof t){if(Array.isArray(t)){var i=t.length;for(n=0;n<i;n++)t[n]&&(r=e(t[n]))&&(o&&(o+=" "),o+=r)}else for(r in t)t[r]&&(o&&(o+=" "),o+=r)}return o}(e))&&(r&&(r+=" "),r+=t);return r}n.d(t,{W:function(){return r}}),t.Z=r},68588:function(e,t,n){"use strict";n.d(t,{M:function(){return r}});let r="data-"+(0,n(93193).D)("framerAppearId")},2445:function(e,t,n){"use strict";function r(e){return null!==e&&"object"==typeof e&&"function"==typeof e.start}n.d(t,{H:function(){return r}})},9442:function(e,t,n){"use strict";n.d(t,{featureDefinitions:function(){return o}});let r={animation:["animate","variants","whileHover","whileTap","exit","whileInView","whileFocus","whileDrag"],exit:["exit"],drag:["drag","dragControls"],focus:["whileFocus"],hover:["whileHover","onHoverStart","onHoverEnd"],tap:["whileTap","onTap","onTapStart","onTapCancel"],pan:["onPan","onPanStart","onPanSessionStart","onPanEnd"],inView:["whileInView","onViewportEnter","onViewportLeave"],layout:["layout","layoutId"]},o={};for(let e in r)o[e]={isEnabled:t=>r[e].some(e=>!!t[e])}},96728:function(e,t,n){"use strict";n.d(t,{j:function(){return i}});var r=n(59354);let o={};function i(e,{layout:t,layoutId:n}){return r.G.has(e)||e.startsWith("origin")||(t||void 0!==n)&&(!!o[e]||"opacity"===e)}},93193:function(e,t,n){"use strict";n.d(t,{D:function(){return r}});let r=e=>e.replace(/([a-z])([A-Z])/gu,"$1-$2").toLowerCase()},62627:function(e,t,n){"use strict";n.d(t,{q:function(){return o}});let r=["animate","circle","defs","desc","ellipse","g","image","line","filter","marker","mask","metadata","path","pattern","polygon","polyline","rect","stop","switch","symbol","svg","text","tspan","use","view"];function o(e){if("string"!=typeof e||e.includes("-"));else if(r.indexOf(e)>-1||/[A-Z]/u.test(e))return!0;return!1}},41069:function(e,t,n){"use strict";n.d(t,{r:function(){return u}});var r=n(59354),o=n(21219);let i=(e,t)=>t&&"number"==typeof e?t.transform(e):e;var l=n(20354);let a={x:"translateX",y:"translateY",z:"translateZ",transformPerspective:"perspective"},s=r._.length;function u(e,t,n){let{style:u,vars:c,transformOrigin:d}=e,f=!1,p=!1;for(let e in t){let n=t[e];if(r.G.has(e)){f=!0;continue}if((0,o.f)(e)){c[e]=n;continue}{let t=i(n,l.j[e]);e.startsWith("origin")?(p=!0,d[e]=t):u[e]=t}}if(!t.transform&&(f||n?u.transform=function(e,t,n){let o="",u=!0;for(let c=0;c<s;c++){let s=r._[c],d=e[s];if(void 0===d)continue;let f=!0;if(!(f="number"==typeof d?d===(s.startsWith("scale")?1:0):0===parseFloat(d))||n){let e=i(d,l.j[s]);if(!f){u=!1;let t=a[s]||s;o+=`${t}(${e}) `}n&&(t[s]=e)}}return o=o.trim(),n?o=n(t,u?"":o):u&&(o="none"),o}(t,e.transform,n):u.transform&&(u.transform="none")),p){let{originX:e="50%",originY:t="50%",originZ:n=0}=d;u.transformOrigin=`${e} ${t} ${n}`}}},50189:function(e,t,n){"use strict";n.d(t,{U:function(){return i}});var r=n(32558),o=n(96728);function i(e,t,n){let{style:i}=e,l={};for(let a in i)((0,r.i)(i[a])||t.style&&(0,r.i)(t.style[a])||(0,o.j)(a,e)||n?.getValue(a)?.liveStyle!==void 0)&&(l[a]=i[a]);return l}},42617:function(e,t,n){"use strict";n.d(t,{i:function(){return a}});var r=n(41069),o=n(37259);let i={offset:"stroke-dashoffset",array:"stroke-dasharray"},l={offset:"strokeDashoffset",array:"strokeDasharray"};function a(e,{attrX:t,attrY:n,attrScale:a,pathLength:s,pathSpacing:u=1,pathOffset:c=0,...d},f,p,h){if((0,r.r)(e,d,p),f){e.style.viewBox&&(e.attrs.viewBox=e.style.viewBox);return}e.attrs=e.style,e.style={};let{attrs:m,style:g}=e;m.transform&&(g.transform=m.transform,delete m.transform),(g.transform||m.transformOrigin)&&(g.transformOrigin=m.transformOrigin??"50% 50%",delete m.transformOrigin),g.transform&&(g.transformBox=h?.transformBox??"fill-box",delete m.transformBox),void 0!==t&&(m.x=t),void 0!==n&&(m.y=n),void 0!==a&&(m.scale=a),void 0!==s&&function(e,t,n=1,r=0,a=!0){e.pathLength=1;let s=a?i:l;e[s.offset]=o.px.transform(-r);let u=o.px.transform(t),c=o.px.transform(n);e[s.array]=`${u} ${c}`}(m,s,u,c,!1)}},79854:function(e,t,n){"use strict";n.d(t,{a:function(){return r}});let r=e=>"string"==typeof e&&"svg"===e.toLowerCase()},16832:function(e,t,n){"use strict";n.d(t,{U:function(){return l}});var r=n(32558),o=n(59354),i=n(50189);function l(e,t,n){let l=(0,i.U)(e,t,n);for(let n in e)((0,r.i)(e[n])||(0,r.i)(t[n]))&&(l[-1!==o._.indexOf(n)?"attr"+n.charAt(0).toUpperCase()+n.substring(1):n]=e[n]);return l}},7504:function(e,t,n){"use strict";n.d(t,{G:function(){return l},M:function(){return a}});var r=n(2445),o=n(97732),i=n(22963);function l(e){return(0,r.H)(e.animate)||i.V.some(t=>(0,o.$)(e[t]))}function a(e){return!!(l(e)||e.variants)}},97732:function(e,t,n){"use strict";function r(e){return"string"==typeof e||Array.isArray(e)}n.d(t,{$:function(){return r}})},79432:function(e,t,n){"use strict";function r(e){let t=[{},{}];return e?.values.forEach((e,n)=>{t[0][n]=e.get(),t[1][n]=e.getVelocity()}),t}function o(e,t,n,o){if("function"==typeof t){let[i,l]=r(o);t=t(void 0!==n?n:e.custom,i,l)}if("string"==typeof t&&(t=e.variants&&e.variants[t]),"function"==typeof t){let[i,l]=r(o);t=t(void 0!==n?n:e.custom,i,l)}return t}n.d(t,{o:function(){return o}})},22963:function(e,t,n){"use strict";n.d(t,{V:function(){return o},e:function(){return r}});let r=["animate","whileInView","whileFocus","whileHover","whileTap","whileDrag","exit"],o=["initial",...r]},11741:function(e,t,n){"use strict";n.d(t,{j:function(){return r}});let r="undefined"!=typeof window},21219:function(e,t,n){"use strict";n.d(t,{f:function(){return o},t:function(){return l}});let r=e=>t=>"string"==typeof t&&t.startsWith(e),o=r("--"),i=r("var(--"),l=e=>!!i(e)&&a.test(e.split("/*")[0].trim()),a=/var\(--(?:[\w-]+\s*|[\w-]+\s*,(?:\s*[^)(\s]|\s*\((?:[^)(]|\([^)(]*\))*\))+\s*)\)$/iu},93674:function(e,t,n){"use strict";n.d(t,{Z:function(){return l}});var r=n(42218);let o=["setup","read","resolveKeyframes","preUpdate","update","preRender","render","postRender"];var i=n(57275);function l(e,t){let n=!1,l=!0,a={delta:0,timestamp:0,isProcessing:!1},s=()=>n=!0,u=o.reduce((e,n)=>(e[n]=function(e,t){let n=new Set,r=new Set,o=!1,l=!1,a=new WeakSet,s={delta:0,timestamp:0,isProcessing:!1},u=0;function c(t){a.has(t)&&(d.schedule(t),e()),u++,t(s)}let d={schedule:(e,t=!1,i=!1)=>{let l=i&&o?n:r;return t&&a.add(e),l.has(e)||l.add(e),e},cancel:e=>{r.delete(e),a.delete(e)},process:e=>{if(s=e,o){l=!0;return}o=!0,[n,r]=[r,n],n.forEach(c),t&&i.f.value&&i.f.value.frameloop[t].push(u),u=0,n.clear(),o=!1,l&&(l=!1,d.process(e))}};return d}(s,t?n:void 0),e),{}),{setup:c,read:d,resolveKeyframes:f,preUpdate:p,update:h,preRender:m,render:g,postRender:b}=u,v=()=>{let o=r.c.useManualTiming?a.timestamp:performance.now();n=!1,r.c.useManualTiming||(a.delta=l?1e3/60:Math.max(Math.min(o-a.timestamp,40),1)),a.timestamp=o,a.isProcessing=!0,c.process(a),d.process(a),f.process(a),p.process(a),h.process(a),m.process(a),g.process(a),b.process(a),a.isProcessing=!1,n&&t&&(l=!1,e(v))},y=()=>{n=!0,l=!0,a.isProcessing||e(v)};return{schedule:o.reduce((e,t)=>{let r=u[t];return e[t]=(e,t=!1,o=!1)=>(n||y(),r.schedule(e,t,o)),e},{}),cancel:e=>{for(let t=0;t<o.length;t++)u[o[t]].cancel(e)},state:a,steps:u}}},59354:function(e,t,n){"use strict";n.d(t,{G:function(){return o},_:function(){return r}});let r=["transformPerspective","x","y","z","translateX","translateY","translateZ","scale","scaleX","scaleY","rotate","rotateX","rotateY","rotateZ","skew","skewX","skewY"],o=new Set(r)},57275:function(e,t,n){"use strict";n.d(t,{f:function(){return r}});let r={value:null,addProjectionMetrics:null}},77596:function(e,t,n){"use strict";function r(e){return"object"==typeof e&&null!==e&&"offsetHeight"in e}n.d(t,{R:function(){return r}})},20354:function(e,t,n){"use strict";n.d(t,{j:function(){return a}});var r=n(57958);let o={...r.Rx,transform:Math.round};var i=n(37259);let l={rotate:i.RW,rotateX:i.RW,rotateY:i.RW,rotateZ:i.RW,scale:r.bA,scaleX:r.bA,scaleY:r.bA,scaleZ:r.bA,skew:i.RW,skewX:i.RW,skewY:i.RW,distance:i.px,translateX:i.px,translateY:i.px,translateZ:i.px,x:i.px,y:i.px,z:i.px,perspective:i.px,transformPerspective:i.px,opacity:r.Fq,originX:i.$C,originY:i.$C,originZ:i.px},a={borderWidth:i.px,borderTopWidth:i.px,borderRightWidth:i.px,borderBottomWidth:i.px,borderLeftWidth:i.px,borderRadius:i.px,radius:i.px,borderTopLeftRadius:i.px,borderTopRightRadius:i.px,borderBottomRightRadius:i.px,borderBottomLeftRadius:i.px,width:i.px,maxWidth:i.px,height:i.px,maxHeight:i.px,top:i.px,right:i.px,bottom:i.px,left:i.px,padding:i.px,paddingTop:i.px,paddingRight:i.px,paddingBottom:i.px,paddingLeft:i.px,margin:i.px,marginTop:i.px,marginRight:i.px,marginBottom:i.px,marginLeft:i.px,backgroundPositionX:i.px,backgroundPositionY:i.px,...l,zIndex:o,fillOpacity:r.Fq,strokeOpacity:r.Fq,numOctaves:o}},57958:function(e,t,n){"use strict";n.d(t,{Fq:function(){return i},Rx:function(){return o},bA:function(){return l}});var r=n(488);let o={test:e=>"number"==typeof e,parse:parseFloat,transform:e=>e},i={...o,transform:e=>(0,r.u)(0,1,e)},l={...o,default:1}},37259:function(e,t,n){"use strict";n.d(t,{$C:function(){return u},RW:function(){return o},aQ:function(){return i},px:function(){return l},vh:function(){return a},vw:function(){return s}});let r=e=>({test:t=>"string"==typeof t&&t.endsWith(e)&&1===t.split(" ").length,parse:parseFloat,transform:t=>`${t}${e}`}),o=r("deg"),i=r("%"),l=r("px"),a=r("vh"),s=r("vw"),u={...i,parse:e=>i.parse(e)/100,transform:e=>i.transform(100*e)}},32558:function(e,t,n){"use strict";n.d(t,{i:function(){return r}});let r=e=>!!(e&&e.getVelocity)},488:function(e,t,n){"use strict";n.d(t,{u:function(){return r}});let r=(e,t,n)=>n>t?t:n<e?e:n},42218:function(e,t,n){"use strict";n.d(t,{c:function(){return r}});let r={}}},function(e){var t=function(t){return e(e.s=t)};e.O(0,[774,179],function(){return t(91118),t(26036)}),_N_E=e.O()}]);